/***************************************************************************/
/* J Drive (tm) Engine                                                     */
/* Copyright 2003-2005, Xilinx, Inc.                                       */
/* The use, reproduction, and distribution of this code is subject         */
/* to the J Drive license requirements and restrictions.                   */
/***************************************************************************/
#include <windows.h>
#include <stdio.h>
#include <stdlib.h>

#include "windrvr.h"
#include "bjdrvif.h"
#include "wdrv_lic.h"

jdriver::jdriver() {
  hWD = (HANDLE) NULL;
  io_increment = 0;
  jcab_base_port = (DWORD) -1;
  max_write_size = ( getenv("XDBGJTAGWRTRANSSIZE") == (char *) NULL ? 4096 : atoi( getenv("XDBGJTAGWRTRANSSIZE") ) );
  m_dwBufferSize = max_write_size * 2;
  m_transfer = (WD_TRANSFER *) NULL;
}

jdriver::~jdriver() {
  hWD = (HANDLE) NULL;
  io_increment = 0;
  jcab_base_port = (DWORD) -1;
  m_dwBufferSize = 0;
  max_write_size = 0;
  if (m_transfer) {
    delete m_transfer;
    m_transfer = (WD_TRANSFER *) NULL;
  }
}


int
jdriver::GetParallelPorts( DWORD port_index ) {

  WD_CARD_REGISTER cardReg;
  BOOLEAN pc98 = FALSE;

  if (open() == -1) {
    return( -1 );
  }

  // get the pc type, PC98 or AT Compatible

  // The best way to distinguish between AT compatibles and PC-9800 
  // is to check the data following the ROM BIOS entry point (FFFF:0000). 
  // This area*s data format is shown below. The date is written 
  // in ASCII code, so when the separator "/" (2Fh in hexadecimal) 
  // is not present, it is PC-9800. 
  // One should check FFFF:0007 and FFFF:000A. This detection method 
  // assumes that machines other than AT compatibles are PC-9800.

  // Address            Length          AT Compatibles          PC-9800
  // FFFF:0000  5 Bytes         JMP far                         JMP far
  // FFFF:0005  8 Bytes         Date 'mm/dd/yy'         Not used
  // FFFF:000D  1 Byte          Not used                        Not used
  // FFFF:000E  1 Byte          Model ID                        ROM checksum
  // FFFF:000F  1 Byte          Reserved                        ROM checksum

  cardReg.Card.dwItems = 1;
  cardReg.fCheckLockOnly = FALSE;
  cardReg.Card.Item[0].item = ITEM_MEMORY;
  cardReg.Card.Item[0].fNotSharable = FALSE;
  cardReg.Card.Item[0].I.Mem.dwBytes = 1;
  cardReg.Card.Item[0].I.Mem.dwPhysicalAddr = 0xFFFF7;

  WD_CardRegister (hWD, &cardReg);

  if(cardReg.hCard) {
    BYTE *ptr;

    ptr = (BYTE *) cardReg.Card.Item[0].I.Mem.dwUserDirectAddr;
    
    BYTE slash = *ptr;               // at location 7

    if(slash == '/') {
      slash = *(ptr+=3);
      if(slash == '/') {
        pc98 = FALSE;
      } else {
        pc98 = TRUE;
      }
    } else {
      pc98 = FALSE;
    }
  } else {
    close();
    return( -1 );
  }
  WD_CardUnregister(hWD, &cardReg);

  if (pc98) {
    jcab_base_port = PC98_BASE_IO_ADDR;
    io_increment = 2;
  } else {
    io_increment = 1;
    
    // setup WinDrv to read physical memory
    cardReg.Card.dwItems = 1;
    cardReg.Card.Item[0].item = ITEM_MEMORY;
    cardReg.Card.Item[0].fNotSharable = FALSE;
    cardReg.Card.Item[0].I.Mem.dwBytes = 6;
    cardReg.fCheckLockOnly = FALSE;
    
    // check for LPT1, LPT2 and LPT3
    cardReg.Card.Item[0].I.Mem.dwPhysicalAddr = LPT_BASE_MEM_ADDR; // 0x00000408
    WD_CardRegister( hWD, &cardReg );
    if (cardReg.hCard) {
      WORD *ptr;

      ptr = (WORD *) cardReg.Card.Item[0].I.Mem.dwUserDirectAddr;      
         jcab_base_port = ptr[port_index];
    } else {
         close();
         return( -1 );
    }
    WD_CardUnregister(hWD, &cardReg);
  }   
  
  return( jcab_base_port );
}


BOOLEAN 
jdriver::WRTransfer( LPBYTE pOutBuf, LPBYTE pInBuf, DWORD dwSize )
{

  DWORD increment = 0;
  DWORD m_dwTransferSize = 0;

  if (m_transfer == (WD_TRANSFER *) NULL) {
    m_transfer = new WD_TRANSFER[m_dwBufferSize];
  }

  m_dwTransferSize = max_write_size;
  if (pInBuf) {

    // both input and output bytes are counted

    increment = 2;
    dwSize *= 2; 
  } else {
    increment = 1;
  }

  if (dwSize < m_dwTransferSize) {

    // send the whole thing

    m_dwTransferSize = dwSize;
  }

  DWORD j = 0;
  DWORD k = 0;
  while (dwSize > 0) {
      for (DWORD i = 0; i < m_dwTransferSize; i+=increment, j++ ) {
        m_transfer[i].cmdTrans = WP_BYTE;
        m_transfer[i].dwPort = GetDataPort();
        m_transfer[i].Data.Byte = pOutBuf[j];
        
        if (pInBuf) {
          m_transfer[i+1].cmdTrans = RP_BYTE;
          m_transfer[i+1].dwPort = GetStatusPort();
          m_transfer[i+1].Data.Byte = 0x0;
        }
      }

      WD_MultiTransfer( hWD, m_transfer, m_dwTransferSize );
      
      if (pInBuf) {
        for (DWORD i = 0; i < m_dwTransferSize; i+=increment, k++ ) {
          pInBuf[k] = m_transfer[i+1].Data.Byte;
        }
      }

      dwSize -= m_dwTransferSize;
      if (dwSize < m_dwTransferSize) {

        // send the whole thing

        m_dwTransferSize = dwSize;
      }
  }

  return( TRUE );
}


BYTE
jdriver::ReadByte( DWORD dwIOAddr )
{
  WD_TRANSFER trns;

  trns.cmdTrans = RP_BYTE;
  trns.dwPort = dwIOAddr;
  WD_Transfer( hWD, &trns );
  return( trns.Data.Byte );
}

BYTE
jdriver::WriteByte( DWORD dwIOAddr, BYTE bData )
{
  WD_TRANSFER trns;

  trns.cmdTrans = WP_BYTE;
  trns.dwPort = dwIOAddr;
  trns.Data.Byte = bData;
  WD_Transfer( hWD, &trns );
  return( trns.Data.Byte );
}

int 
jdriver::open() {
  if (hWD == (HANDLE) NULL) {
    hWD = WD_Open();
    if (hWD == INVALID_HANDLE_VALUE) {
         hWD = (HANDLE) NULL;
         return( -1 );
    }
    WD_LICENSE lic;

    strcpy(lic.cLicense, WDRV_LICENSE);
    WD_License(hWD, &lic);
  }
  return(0);
}

void 
jdriver::close() {
  WD_Close( hWD );
  hWD = (HANDLE) NULL;
}

