/***************************************************************************/
/* J Drive (tm) Engine                                                     */
/* Copyright 2003-2005, Xilinx, Inc.                                       */
/* The use, reproduction, and distribution of this code is subject         */
/* to the J Drive license requirements and restrictions.                   */
/***************************************************************************/
/*******************************************************/
/* file: ports.h                                       */
/* abstract:  This file contains extern declarations   */
/*            for providing stimulus to the JTAG ports.*/
/*******************************************************/
#ifndef ports_dot_h
#define ports_dot_h

#if 0
#define DEBUG_MODE  /* this line and output the                   */
		    /* TMS and TDI values on the rising edge of                   */
		    /* the clock                                                  */
#endif

#ifdef WIN95PP
#define DATA_OFFSET    (unsigned short) 0
#define STATUS_OFFSET  (unsigned short) 1
#define CONTROL_OFFSET (unsigned short) 2
#endif

//==!========================================================================
//==! 2003/01/28: From detectCable() code
//==!========================================================================
#define BUSY_PIN_MASK   0x80
#define ACK_PIN_MASK    0x40
#define PE_PIN_MASK     0x20
//==!========================================================================

#if defined( WIN95PP ) || defined( WINDRVR )
#pragma warning( disable: 4214 )
typedef union outPortUnion {
	unsigned char value;
	struct opBitsStr {
		unsigned char tdi:1;
		unsigned char tck:1;
		unsigned char tms:1;
		unsigned char zero:1;
		unsigned char one:1;
		unsigned char bit5:1;
		unsigned char bit6:1;
		unsigned char bit7:1;
	} bits;
} outPortType;

typedef union inPortUnion {
	unsigned char value;
	struct ipBitsStr {
		unsigned char bit0:1;
		unsigned char bit1:1;
		unsigned char bit2:1;
		unsigned char bit3:1;
		unsigned char tdo:1;
		unsigned char bit5:1;
		unsigned char bit6:1;
		unsigned char bit7:1;
	} bits;
} inPortType;
#pragma warning( default: 4214 )
#endif

/* these constants are used to send the appropriate ports to setPort */
/* they should be enumerated types, but some of the microcontroller  */
/* compilers don't like enumerated types */
#define TCK (short) 0
#define TMS (short) 1
#define TDI (short) 2

#if defined(__cplusplus)
/* set the port "p" (TCK, TMS, or TDI) to val (0 or 1) */
extern "C" {
#if 0
	void setPort(short p, short val);
    void closePort();
#else
    int setPort(short p, short val);
    void closePort();
    void xportsInitialize();
    int  burstData(int iMode,int iByteBoundary,int iReadTDO,
              long *plBurstDataIndex,int iTms,int iTdi,long *plReadCnt,
              char *pcArrayRead);

#endif

	/* read the TDO bit and store it in val */
	unsigned char readTDOBit();

	/* make clock go down->up->down*/
	unsigned char pulseClock( int getTdo );

	/* read the next byte of data from the xsvf file */
	void readByte(unsigned char *data);

	void waitTime(long microsec);
}
#else
/* set the port "p" (TCK, TMS, or TDI) to val (0 or 1) */
#if 0
extern void setPort(short p, short val);
#else
extern int setPort(short p, short val);
extern void closePort();
#endif

/* read the TDO bit and store it in val */
extern unsigned char readTDOBit();

/* make clock go down->up->down*/
extern unsigned char pulseClock( int getTdo );

/* read the next byte of data from the xsvf file */
extern void readByte(unsigned char *data);

extern void waitTime(long microsec);
#endif
#endif

//==!========================================================================
//==! 2002/03/28: Max bit burst shift size used by burst_pTransfer data
//==! structure (64K was arbitrarily selected). burst_pTransfer[] will hold
//==! hold the data and the command to be used by the port driver on that command.
//==!
//==! To shift out 64K bytes (64*1024)*(8 bits/byte) requires at most
//==! 4 entries:
//==! - 1 to write the TDI and TMS data
//==! - 1 to read the TDO data from the port
//==! - 1 to write/drive TCK high
//==! - 1 to write/drive TCK back low
//==!========================================================================
#define MAX_BYTE_BURST_SIZE   64*1024
#define MAX_BIT_BURST_SIZE    MAX_BYTE_BURST_SIZE*8*4
//==!========================================================================

//==!========================================================================

