/***************************************************************************/
/* J Drive (tm) Engine                                                     */
/* Copyright 2003-2005, Xilinx, Inc.                                       */
/* The use, reproduction, and distribution of this code is subject         */
/* to the J Drive license requirements and restrictions.                   */
/***************************************************************************/
//==!=====================================================================!==
//==! See XPROJECT.H for project file list and project version history.
//==!=====================================================================!==
//==! Revision history: XCHKFUNC.CPP
//==! 2003/04/11: Contains functions for checking JTAG features
//==!=====================================================================!==
#include "xcommon.h"
#include "x_bsdl.h"
/*==!===================================================================!==*/
/*==! Drive TMS and TDI signals on chain, clock once, and return sampled   */
/*==! TDO value.                                                           */
/*==! 2002/03/29: Added iMode to support burst mode. If iMode = 0, then    */
/*==! update the iTapState variable which keeps track if the TAP state     */
/*==! and call setPort(). If 1, then just update iTapState but leave the   */
/*==! actual port communication to dataBurst().                            */
/*==!                                                                      */
/*==! Function Parameters : sTmsVal = TMS value                            */
/*==!                     : sTdiVal = TDI value                            */
/*==!                     : iMode = see function description               */
/*==! Return Value        : sampled TDO value                              */
/*==!===================================================================!==*/
int callPort(short sTmsVal,short sTdiVal,int iMode)
{ unsigned char uChr;

#if PC_DEBUG == 1
  switch(iTapState)
  { case TAP_RESET:
      if (sTmsVal == 1)
      { iTapState = TAP_RESET;        printf("RTI     -");
      } /* IF */
      else
      { iTapState = TAP_IDLE;         printf("IDLE    -");
      } /* ELSE */
      break;
    case TAP_IDLE:
      if (sTmsVal == 1)
      { iTapState = TAP_DRSELECT;     printf("DRSELECT-");
      } /* IF */
      else
      { iTapState = TAP_IDLE;         printf("IDLE    -");
      } /* ELSE */
      break;
    case TAP_DRSELECT:
      if (sTmsVal == 1)
      { iTapState = TAP_IRSELECT;     printf("IRSELECT -");
      } /* IF */
      else
      { iTapState = TAP_DRCAPTURE;    printf("DRCAPTURE-");
      } /* ELSE */
      break;
    case TAP_DRCAPTURE:
      if (sTmsVal == 1)
      { iTapState = TAP_DREXIT1;      printf("DREXIT1  -");
      } /* IF */
      else
      { iTapState = TAP_DRSHIFT;      printf("DRSHIFT  -");
      } /* ELSE */
      break;
    case TAP_DRSHIFT:
      if (sTmsVal == 1)
      { iTapState = TAP_DREXIT1;      printf("DREXIT1  -");
      } /* IF */
      else
      { iTapState = TAP_DRSHIFT;      printf("DRSHIFT  -");
      } /* ELSE */
      break;
    case TAP_DREXIT1:
      if (sTmsVal == 1)
      { iTapState = TAP_DRUPDATE;     printf("DRUPDATE -");
      } /* IF */
      else
      { iTapState = TAP_DRPAUSE;      printf("DRPAUSE  -");
      } /* ELSE */
      break;
    case TAP_DRPAUSE:
      if (sTmsVal == 1)
      { iTapState = TAP_DREXIT2;      printf("DREXIT2  -");
      } /* IF */
      else
      { iTapState = TAP_DRPAUSE;      printf("DRPAUSE  -");
      } /* ELSE */
      break;
    case TAP_DREXIT2:
      if (sTmsVal == 1)
      { iTapState = TAP_DRUPDATE;     printf("DRUPDATE -");
      } /* IF */
      else
      { iTapState = TAP_DRSHIFT;      printf("DRSHIFT  -");
      } /* ELSE */
      break;
    case TAP_DRUPDATE:
      if (sTmsVal == 1)
      { iTapState = TAP_DRSELECT;     printf("DRSELECT -");
      } /* IF */
      else
      { iTapState = TAP_IDLE;         printf("IDLE     -");
      } /* ELSE */
      break;
    case TAP_IRSELECT:
      if (sTmsVal == 1)
      { iTapState = TAP_RESET;        printf("RTI      -");
      } /* IF */
      else
      { iTapState = TAP_IRCAPTURE;    printf("IRCAPTURE-");
      } /* ELSE */
      break;
    case TAP_IRCAPTURE:
      if (sTmsVal == 1)
      { iTapState = TAP_IREXIT1;      printf("IREXIT1  -");
      } /* IF */
      else
      { iTapState = TAP_IRSHIFT;      printf("IRSHIFT  -");
      } /* ELSE */
      break;
    case TAP_IRSHIFT:
      if (sTmsVal == 1)
      { iTapState = TAP_IREXIT1;      printf("IREXIT1  -");
      } /* IF */
      else
      { iTapState = TAP_IRSHIFT;      printf("IRSHIFT  -");
      } /* ELSE */
      break;
    case TAP_IREXIT1:
      if (sTmsVal == 1)
      { iTapState = TAP_IRUPDATE;     printf("IRUPDATE -");
      } /* IF */
      else
      { iTapState = TAP_IRPAUSE;      printf("IRPAUSE  -");
      } /* ELSE */
      break;
    case TAP_IRPAUSE:
      if (sTmsVal == 1)
      { iTapState = TAP_IREXIT2;      printf("IREXIT2  -");
      } /* IF */
      else
      { iTapState = TAP_IRPAUSE;      printf("IRPAUSE  -");
      } /* ELSE */
      break;
    case TAP_IREXIT2:
      if (sTmsVal == 1)
      { iTapState = TAP_IRUPDATE;     printf("IRUPDATE -");
      } /* IF */
      else
      { iTapState = TAP_IRSHIFT;      printf("IRSHIFT  -");
      } /* ELSE */
      break;
    case TAP_IRUPDATE:
      if (sTmsVal == 1)
      { iTapState = TAP_DRSELECT;     printf("DRSELECT -");
      } /* IF */
      else
      { iTapState = TAP_IDLE;         printf("IDLE     -");
      } /* ELSE */
      break;
    default:
      break;
  } // SWITCH

  if (++iTapPrint > 5)
  { printf("\n");
    iTapPrint = 0;
  } // IF
#endif

  if (iMode == 1)
  { return 1;   //==! Always return 0 (or 1 in burst mode)
  } // IF

  iPortStatus = setPort(TMS,sTmsVal);
  setPort(TDI,sTdiVal);
  uChr = pulseClock(1);
  return (uChr == 0 ? 0 : 1);

} /* callPort() */
//==!=====================================================================!==
//==! doSyncJtagReset(): do synchronous JTAG reset. If iMode = 1, then move
//==! to RUN-TEST-IDLE state after reset.
//==!
//==! Function Parameters: iMode = see description above
//==! Return Value       : none
//==!=====================================================================!==
void doSyncJtagReset(int iMode)
{
#if USE_BURST == 0
//==! Non-burst version of function
  long lVar;

  for (lVar = 0; lVar < 5; ++lVar)
  { callPort(1,0,0);
  } // FOR lVar < 5
  if (iMode == 1)
  {
    callPort(0,0,0);
  } // IF

#else
  long lVar,lBurstCnt,lReadCnt;

  for (lVar = 0,lBurstCnt = 0; lVar < 5; ++lVar)
  { burstData(0,0,0,&lBurstCnt,1,0,&lReadCnt,cArrayBurstReadTdo);
  } // FOR lVar < 5
  if (iMode == 1)
  { burstData(0,0,0,&lBurstCnt,0,0,&lReadCnt,cArrayBurstReadTdo);
  } // IF
  burstData(1,0,0,&lBurstCnt,0,0,&lReadCnt,cArrayBurstReadTdo);
#endif
} // doSyncJtagReset()
//==!=====================================================================!==
//==! doHeaderTrailer(): enMode specifies header (eHeader) or trailer
//==! (eTrailer) while enReg specifies instruction (eIreg) or data (eDreg)
//==! register.
//==!
//==! Function Parameters: enMode = see function description
//==!                      enReg = see function description
//==! Return Value       : none
//==!=====================================================================!==
void doHeaderTrailer(EnumTapType enMode,EnumTapType enReg)
{
#if USE_BURST == 0
//==! Non-burst version of function
  long lVar,lRegLength;

  switch(enMode)
  { case eHeader:
      lRegLength = lHir;
      if (enReg == eDreg)
      { lRegLength = lHdr;
      } // IF
      if (lRegLength <= 0)
      { return;
      } // IF
      for (lVar = 0; lVar < lRegLength; ++lVar)
      { callPort(0,1,0);
      } // FOR lVar < lRegLength
      break;
    case eTrailer:
    default:
      lRegLength = lTir;
      if (enReg == eDreg)
      { lRegLength = lTdr;
      } // IF
      if (lRegLength <= 0)
      { return;
      } // IF
      for (lVar = 0; lVar < lRegLength; ++lVar)
      { if (lVar == lRegLength-1)
        { callPort(1,1,0);
        } // IF
        else
        { callPort(0,1,0);
        } // ELSE
      } // FOR lVar < lRegLength
      break;
  } // SWITCH (enMode)
#else
  long lVar,lRegLength,lBurstCnt,lReadCnt;

  switch(enMode)
  { case eHeader:
      lRegLength = lHir;
      if (enReg == eDreg)
      { lRegLength = lHdr;
      } // IF
      if (lRegLength <= 0)
      { return;
      } // IF
      for (lVar = 0,lBurstCnt = 0; lVar < lRegLength; ++lVar)
      { burstData(0,0,0,&lBurstCnt,0,1,&lReadCnt,cArrayBurstReadTdo);
      } // FOR lVar < lRegLength
      break;
    case eTrailer:
    default:
      lRegLength = lTir;
      if (enReg == eDreg)
      { lRegLength = lTdr;
      } // IF
      if (lRegLength <= 0)
      { return;
      } // IF
      for (lVar = 0,lBurstCnt = 0; lVar < lRegLength; ++lVar)
      { if (lVar == lRegLength-1)
        { burstData(0,0,0,&lBurstCnt,1,1,&lReadCnt,cArrayBurstReadTdo);
        } // IF
        else
        { burstData(0,0,0,&lBurstCnt,0,1,&lReadCnt,cArrayBurstReadTdo);
        } // ELSE
      } // FOR lVar < lRegLength
      break;
  } // SWITCH (enMode)
  burstData(1,0,0,&lBurstCnt,0,0,&lReadCnt,cArrayBurstReadTdo);
#endif
} // doHeaderTrailer()
//==!=====================================================================!==
//==! readBurstTdoArray(): read bits from the bytes in the iaBurstRead[]
//==! array and put it into the pcReadArray[] starting at array location
//==! plReadIndex. Max number of bits should not exceed lMaxReadCnt.
//==!
//==! Function Parameters: lCnt = number of bytes readable from iaBurstRead[]
//==!                      caBurstRead = char array holding data read from
//==!                              burst operation
//==!                      pcReadArray = pointer to data array to hold read
//==!                              data extracted from iaBurstRead[]
//==!                      plReadIndex = pointer to read array index variable
//==!                      lMaxReadCnt = max number of bits to read back
//==! Return Value       : none
//==!=====================================================================!==
void readBurstTdoArray(long lCnt,char *caBurstRead,char *pcReadArray,
                       long *plReadIndex,long lMaxReadCnt)

{ long lVar0,lVar1,lVar2;

  if (lCnt == 0)
  { jMessage(DISP_AND_EXIT,"rBTA-0",0,NULL);
    return;
  } // IF
  for (lVar0 = *plReadIndex,lVar1 = 0,lVar2 = 0;
       lVar0 < lMaxReadCnt; ++lVar0)
  { if (lVar1 >= 8)
    { ++lVar2;
      lVar1 = 0;
      if (lVar2 >= lCnt)
      { return;
      } // IF lVar2 >= lReadCnt
    } // IF lVar1 >= 8
    pcReadArray[lMaxReadCnt-lVar0-1] =
       ((caBurstRead[lVar2] >> lVar1) & 0x01) == 1 ? '1' : '0';
    ++lVar1;
  } // FOR lVar0 < lMaxReadCnt

} // readBurstTdoArray()
//==!=====================================================================!==
//==! jtagShift(): do JTAG shift. Note that the LSB of the data to be shifted
//==! in is in the highest array location - 1. For example: The char array
//==! 0110011 is arranged from A[0 to 6]. Shifting will start from index 6
//==! and moves down to index 0. The TDO buffer will be filled the same way;
//==! i.e., first bit of data will be stored in A[6] and the last TDO read
//==! will be in A[0].
//==!
//==! Function Parameters: iMode = TAP_IRSHIFT or TAP_DRSHIFT
//==!                      lShiftLen = shift reg length
//==!                      pcDataIn = pointer to data to shift
//==!                      pcDataOut = pointer to array to hold TDO data
//==!                          shifted out
//==!                      iTdoCheck = if 1 then read TDO, else just shift TDI
//==!                      iWaitInUpdatexR = if 1 then do not go to RTI after
//==!                        shifting the IR or DR data
//==! Return Value       : none
//==!=====================================================================!==
void jtagShift(int iMode,long lShiftLen,char *pcDataIn,char *pcDataOut,
               int iTdoCheck,int iWaitInUpdatexR)
{
#if USE_BURST == 0
  long lVar;
  int iDataVal;

  if (iMode == TAP_IRSHIFT)
  { callPort(1,0,0);
  } // IF
  callPort(1,0,0);
  callPort(0,0,0);
  callPort(0,0,0);
  doHeaderTrailer(eHeader,iMode == TAP_IRSHIFT ? eIreg : eDreg);

  for (lVar = lShiftLen-1; lVar >= 0; --lVar)
  { iDataVal = 1;
    if (pcDataIn != NULL)
    { iDataVal = pcDataIn[lVar] == '1' ? 1 : 0;
    } // IF

    if (lVar == 0 &&
           ((iMode == TAP_IRSHIFT && lTir == 0) ||
            (iMode == TAP_DRSHIFT && lTdr == 0)
           )
       )
    { pcDataOut[lVar] = callPort(1,iDataVal,0) + '0';
    } // IF
    else
    { pcDataOut[lVar] = callPort(0,iDataVal,0) + '0';
    } // ELSE
  } // FOR lVar >= 0
#else
  long lVar,lBurstCnt,lReadCnt,lVar1,lVar2;
  int iDataVal;

  if (iMode == TAP_IRSHIFT)
  { callPort(1,0,0);
  } // IF
  callPort(1,0,0);
  callPort(0,0,0);
  callPort(0,0,0);
  doHeaderTrailer(eHeader,iMode == TAP_IRSHIFT ? eIreg : eDreg);

  for (lVar = lShiftLen-1,lBurstCnt = 0,lVar1 = 0,lVar2 = 0;
       lVar >= 0; --lVar,++lVar1)
  { iDataVal = 1;
    if (pcDataIn != NULL)
    { iDataVal = pcDataIn[lVar] == '1' ? 1 : 0;
    } // IF

    if (lVar == 0 &&
           ((iMode == TAP_IRSHIFT && lTir == 0) ||
            (iMode == TAP_DRSHIFT && lTdr == 0)
           )
       )
    { if (iTdoCheck == 1)
      { burstData(0,1,1,&lBurstCnt,1,iDataVal,&lReadCnt,cArrayBurstReadTdo);
      } // IF
      else
      { burstData(0,0,0,&lBurstCnt,1,iDataVal,&lReadCnt,cArrayBurstReadTdo);
      } // ELSE
    } // IF
    else
    { if (iTdoCheck == 1)
      { burstData(0,1,1,&lBurstCnt,0,iDataVal,&lReadCnt,cArrayBurstReadTdo);
      } // IF
      else
      { burstData(0,0,0,&lBurstCnt,0,iDataVal,&lReadCnt,cArrayBurstReadTdo);
      } // ELSE
    } // ELSE

//==! If number of bits to burst reaches limit, then burst these bits out
//==! first before continuing the burst array load operation. lVar2 keeps
//==! track of number of bits read back.
    if (iTdoCheck == 1 &&
         (lVar1%(MAX_BYTE_BURST_SIZE-2) == MAX_BYTE_BURST_SIZE-3)
       )
    { burstData(1,1,1,&lBurstCnt,0,0,&lReadCnt,cArrayBurstReadTdo);
      readBurstTdoArray(lReadCnt,cArrayBurstReadTdo,pcDataOut,&lVar2,
                        lShiftLen);
    } // IF

  } // FOR lVar >= 0

  if (iTdoCheck == 1)
  { burstData(1,0,1,&lBurstCnt,0,0,&lReadCnt,cArrayBurstReadTdo);
    readBurstTdoArray(lReadCnt,cArrayBurstReadTdo,pcDataOut,&lVar2,lShiftLen);
  } // IF
  else
  { burstData(1,0,0,&lBurstCnt,0,0,&lReadCnt,cArrayBurstReadTdo);
  } // ELSE

#endif

  pcDataOut[lShiftLen] = '\0';  // Terminate char array

  doHeaderTrailer(eTrailer,iMode == TAP_IRSHIFT ? eIreg : eDreg);
  callPort(1,0,0);
  if (iWaitInUpdatexR == 0)
  { callPort(0,0,0);
  } // IF iWaitInUpdatexR == 0

} // jtagShift()
//==!=====================================================================!==
//==! Format data according to iMode:
//==!   - if iMode == 0, then format as 4 bit units separated by spaces
//==!   - if iMode == 1, then format as hex
//==!
//==! pcData[] contains either 0/d or 1/u.
//==!
//==! Note that binary data is stored as a bit per character, and that the
//==! MSB of the binary sequence is stored starting at location 0. This is
//==! why data is read from the lDataLen-1 position when converting the data
//==! to hex format (i.e., iMode = 1). When formatting in 4 bit units, the
//==! 4 bit units start at the LSB position (i.e., at array position
//==! lDataLen-1).
//==!
//==! Function Parameters: iMode = see function description
//==!                      lDataLen = data length
//==!                      pcData = pointer to data to format
//==!                      pcDataBuf = pointer to buffer for formatted-data
//==!                      bufTdo = TDO buffer
//==! Return Value       : none
//==!=====================================================================!==
void formatData(int iMode,long lDataLen,char *pcData,char *pcDataBuf)
{ long lVar0,lVar1,lVar2;

  switch(iMode)
  { case 0:
    { for (lVar0 = lDataLen-1,lVar1 = lDataLen + ((lDataLen-1)/4) - 1,
                              lVar2 = 0;
           lVar0 >= 0; --lVar0,--lVar1,++lVar2)
      { if (lVar2 != 0 && lVar2%4 == 0)
        { pcDataBuf[lVar1] = ' ';
          --lVar1;
        } // IF
        pcDataBuf[lVar1] = pcData[lVar0];
      } // FOR lVar0 < lDataLen
      pcDataBuf[lDataLen + ((lDataLen-1)/4)] = '\0';
      return;
    } // CASE 0
    case 1:
    default:
    { for (lVar0 = ((lDataLen+3)/4)-1,lVar1 = lDataLen-1,lVar2 = 0;
           lVar1 >= 0; --lVar1)
      { switch(pcData[lVar1])
        { case '1':
          case 'u':
          case 'H':
            lVar2 += 1 << ((lDataLen-1)-lVar1)%4;
            break;
          case '0':
          case 'd':
          case 'L':
          case 'x':
            break;
          default:
            sprintf(sMsgOnly,"fData[%s/%ld]",pcData,lVar1);
            jMessage(DISP_AND_EXIT,sMsgOnly,0,NULL);
            break;
        } // SWITCH(pcData[lVar1])

//==! The lVar1 == 0 test is just in case lDataLen is not an multiple of 4.
        if (((lDataLen-1)-lVar1)%4 == 3 || lVar1 == 0)
        { pcDataBuf[lVar0] = hextochr(lVar2);
          if (isalpha(pcDataBuf[lVar0]))
          { pcDataBuf[lVar0] = toupper(pcDataBuf[lVar0]);
          } // IF
          --lVar0;
          lVar2 = 0;
        } // IF
      } // FOR lVar1 >= 0
      pcDataBuf[(lDataLen+3)/4] = '\0';
      return;
    } // CASE 1

  } // SWITCH (iMode)

} // formatData()
//==!=====================================================================!==
//==! getBinaryVal() returns 0 or 1 if cChr is '0' or '1', respectively.
//==!
//==! Function Parameters: cChr = binary character
//==! Return Value       : 0 or 1 as described above
//==!=====================================================================!==
int getBinaryVal(char cChr)
{

  switch(cChr)
  { case '0':
    case '1':
    { return cChr - '0';
      break;
    } // CASE '0' or '1'
    default:
    { sprintf(sMsgBuf00,"getBval[%c]",cChr);
      jMessage(DISP_AND_EXIT,sMsgBuf00,0,NULL);
    } // DEFAULT
  } // SWITCH
  return 0;

} // getBinaryVal
//==!=====================================================================!==
//==! long2bit(): converts long value into a bit vector of size lSize. The
//==! iMode parameter controls which end of the array holds the LSB:
//==! iMode = 0: LSB of bit/binary pattern will start from index 0
//==!         1: LSB is at highest index specified by lSize
//==!
//==! Function Parameters: cBuf = buffer to hold bit pattern
//==!                      lVal = long value to convert to binary
//==!                      lSize = number of bits to convert lVal from
//==!                      iMode = see function description
//==! Return Value       : none
//==!=====================================================================!==
void long2bit(char *cBuf,long lVal,long lSize,int iMode)
{ long lCnt;

  for (lCnt = 0; lCnt < lSize; ++lCnt)
  { if (iMode == 0)
    { cBuf[lCnt] = (((lVal >> lCnt) & 0x01) == 1) ? '1' : '0';
    } // IF
    else
    { cBuf[lSize-lCnt-1] = (((lVal >> lCnt) & 0x01) == 1) ? '1' : '0';
    } // ELSE
  } // FOR lCnt < lVal
  cBuf[lSize] = '\0';

} // long2bit()
//==!=====================================================================!==
//==! portVal2Bin(): convert port values to binary format; port values and
//==! binary values are stored as ASCII characters in pcBuf[] which is NULL
//==! terminated.
//==!
//==! Ex 1: dduu will be converted to 0011
//==!    2: HLLL will be converted to 1000
//==!
//==! Function Parameters: pcBuf = char buffer containing port values
//==! Return Value       : none
//==!=====================================================================!==
void portVal2Bin(char *pcBuf)
{ long lVar;

  for (lVar = 0; pcBuf[lVar] != '\0'; ++lVar)
  { switch(pcBuf[lVar])
    { case 'd':
      case 'D':
      case 'l':
      case 'L':
      { pcBuf[lVar] = '0';
        break;
      } // CASE 'D' or 'L'
      case 'u':
      case 'U':
      case 'h':
      case 'H':
      { pcBuf[lVar] = '1';
        break;
      } // CASE 'U' or 'H'
      default:
      { sprintf(sMsgOnly,"pv2b[%c]",pcBuf[lVar]);
        jMessage(DISP_AND_EXIT,sMsgOnly,0,NULL);
        break;
      } // DEFAULT
    } // SWITCH (pcBuf[lVar])
  } // FOR pcBuf[lVar] != '\0'

} // portVal2Bin()
//==!=====================================================================!==
//==! bin2long(): convert binary value stored in pcBuf into a long value.
//==! pcBuf[] must be NULL-terminated.
//==!
//==! Function Parameters: pcBuf = char buffer containing bin value in ASCII
//==! Return Value       : long value
//==!=====================================================================!==
long bin2long(char *pcBuf)
{ long lVal,lBinVal;

  for (lVal = 0,lBinVal = 0; pcBuf[lVal] != '\0'; ++lVal)
  { switch(pcBuf[lVal])
    { case '0':
      case '1':
      { lBinVal = (lBinVal << 1) + (pcBuf[lVal] == '1' ? 1 : 0);
        break;
      } // CASE '0' or '1'
      default:
      { sprintf(sMsgOnly,"bin2long(%s)",pcBuf);
        jMessage(DISP_AND_EXIT,sMsgOnly,0,NULL);
        break;
      } // DEFAULT
    } // SWITCH (pcBuf[lVal])
  } // FOR pcBuf[lVal] != '\0'
  return lBinVal;

} // bin2long()
//==!=====================================================================!==
//==! hex2long(): convert hex value stored in pcBuf into a long value.
//==! pcBuf[] must be NULL-terminated.
//==!
//==! Function Parameters: pcBuf = char buffer containing hex value in ASCII
//==! Return Value       : unsigned long hex value
//==!=====================================================================!==
unsigned long hex2ulong(char *pcBuf)
{ long lVal;
  unsigned long ulHexVal;

  for (lVal = 0,ulHexVal = 0; pcBuf[lVal] != '\0'; ++lVal)
  { if (!isxdigit(pcBuf[lVal]))
    { sprintf(sMsgOnly,"hex2ulong(%s)",pcBuf);
      jMessage(DISP_AND_EXIT,sMsgOnly,0,NULL);
    } // IF
    ulHexVal = (ulHexVal*16) + chrtohex(pcBuf[lVal]);
  } // FOR pcBuf[lVal] != '\0'
  return ulHexVal;

} // hex2ulong()
//==!=====================================================================!==
//==! doByspass(): load bypass instruction
//==!
//==! Function Parameters: pcBufTdi = TDI buffer
//==!                      pcBufTdo = TDO buffer
//==! Return Value       : none
//==!=====================================================================!==
void doBypass(char *pcBufTdi,char *pcBufTdo)
{

  xInitCharArray(sMsgBuf03,pBsdlData->lInstrLength,'1',1);
  jtagShift(TAP_IRSHIFT,pBsdlData->lInstrLength,sMsgBuf03,pcBufTdo,1,0);

} // doBypass()
//==!=====================================================================!==
//==! setBscSafeValues(): use the safe values for all the BSCs in the device
//==! as specified in the BSDL file.
//==!
//==! Function Parameters: pcBufTdi = TDI buffer
//==! Return Value       : none
//==!=====================================================================!==
void setBscSafeValues(char *pcBufTdi)
{ long lVar;

  xInitCharArray(pcBufTdi,pBsdlData->lBscLength,'0',1);
  for (lVar = 0; lVar < pBsdlData->lBscLength; ++lVar)
  { if (pBsdlData->pBscData[lVar]->lBscIndex != lVar)
    { jMessage(DISP_AND_EXIT,"Set BSC safe err",0,NULL);
    } // IF
    if (pBsdlData->pBscData[lVar]->iSafeVal != -1)
    { pcBufTdi[pBsdlData->lBscLength - lVar - 1] =
                   (char)(pBsdlData->pBscData[lVar]->iSafeVal + '0');
    } // IF
  } // FOR lVar < pBsdlData->lBscLength

} // setBscSafeValues()
//==!=====================================================================!==
//==! getULWord(): get an unsigned long word from the pMCS file pointer.
//==! Function will use an MCS byte reader program which will read and
//==! combine 2 bytes to get a word: first byte is lower byte, second is
//==! upper byte. If the last read results in only half a word, then the
//==! upper byte is set to 0.
//==!
//==! Function Parameters: pMCS = MCS file pointer
//==!                      pulVar = pointer to return unsigned long value
//==!                      plAddr = pointer to return word address of word
//==!                             to be read
//==! Return Value       : 1 if successful, 0 otherwise
//==!=====================================================================!==
int getULWord(CMCSfile *pMCS,unsigned long *pulVar,long *plAddr)
{ int iX;

  *pulVar = 0;
  iX = getMCSByte(pMCS);
  *plAddr = (pMCS->pCMCSinfo->lCurAddr)/2;

  if (iX == 1)
  {
//==! If the byte address somehow started on an odd byte, then this byte will
//==! be the upper byte of the word and the lower byte will be set to 0. All
//==! bytes read after this byte will now be aligned on a word boundary.
    if (pMCS->pCMCSinfo->lCurAddr%2 == 1)
    { *pulVar = pMCS->ucByte << 8;
      return 1;
    } // IF

    *pulVar = pMCS->ucByte;
    iX = getMCSByte(pMCS);
    if (iX == 1)
    { *pulVar = pMCS->ucByte << 8 | *pulVar;
    } // IF
  } // IF
  else
  { return 0;
  } // ELSE
  return 1;

} // getULWord()
//==!=====================================================================!==
//==! end of module XCHKFUNC.CPP
//==!=====================================================================!==

