/***************************************************************************/
/* J Drive (tm) Engine                                                     */
/* Copyright 2003-2005, Xilinx, Inc.                                       */
/* The use, reproduction, and distribution of this code is subject         */
/* to the J Drive license requirements and restrictions.                   */
/***************************************************************************/
//==!=====================================================================!==
//==! See XPROJECT.H for project file list and project version history.
//==!=====================================================================!==
//==! Revision history: XCOMMON.H
//==! 2003/03/30: Multiple C/C++ files in a project can include this header
//==!             file, but only one file should define COMPILE_DEFINE_VARS
//==!             to define the variables. All other C/C++ files that include
//==!             this header file will have these variables declared as
//==!             EXTERNal.
//==! 2003/05/02: _CRTDBG_MAP_ALLOC define to use memory leak detection
//==!             capability in Visual C++. As outlined in the Microsoft
//==!             Technical Article "Detecting and Isolating Memory Leaks
//==!             using Microsft Visual C++", the use of this define and the
//==!             crtdbg.h must be in the order as used in this file. The
//==!             function _CrtSetDbgFlag() at the beginning of main() will
//==!             cause a leak check to be done when the program exits.
//==! 2003/12/31: Added m_iEndOfDataStream and m_iBufDataReadEnd to cDataPort
//==!             to support buffered read operations.
//==!=====================================================================!==
//==! Development platform selector
//==!=====================================================================!==
#define SYS_UNIX           0   // Unix system  (not tested as of 2003/05/31)
#define SYS_MICROSOFT_OS   1   // Microsoft OS
//==!=====================================================================!==
#include <stdio.h>

#if SYS_MICROSOFT_OS == 1
  #define _CRTDBG_MAP_ALLOC
  #include <stdlib.h>
  #include <crtdbg.h>
  #include <conio.h>
#else
  #include <stdlib.h>
#endif

#include <ctype.h>
#include <string.h>
#include <time.h>
//==!=====================================================================!==
//==! COMPILE_DEFINE_VARS controls global variable definition/declaration.
//==!=====================================================================!==
#ifdef COMPILE_DEFINE_VARS
  #define DEF_DEC
#else
  #define DEF_DEC extern
#endif
//==!=====================================================================!==
//==! Data mask to set debug variables.
//==!=====================================================================!==
#define OPT_STATUS_OFF    0x01  //==! Bit 0 = turn off status display
#define OPT_PROGDEBUG     0x02  //==! Bit 1 = print program debug info
#define OPT_PRT_BSDL      0x04  //==! Bit 2 = print BSDL info to output file
#define OPT_SKIP_IDCHECK  0x08  //==! Bit 3 = skip IDCODE check
#define OPT_FLOWLOOPDBG   0x10  //==! Bit 4 = turn on flow loop debug code
#define OPT_RPTCNTDISPOFF 0x20  //==! Bit 5 = turn off action repeat count
                                //==!         display
//==!=====================================================================!==
//==! Parser constants
//==!=====================================================================!==
#define MAX_XDATABUF_SIZE     300  // max parser data buffer length
#define MAX_XDATA_SIZE        200  // max parser data length
//==!=====================================================================!==
//==! Supported XDATA comment styles
//==!=====================================================================!==
#define XDATA_COMMENT_STYLE_BSDL  '-'  //==! BSDL/VHDL comment (double '-')
#define XDATA_COMMENT_STYLE_CPP   '/'  //==! C++ comment (double '/')
#define XDATA_COMMENT_STYLE_SVF   '!'  //==! SVF comment (single '!')
//==!=====================================================================!==
#define EXIT_SUCCESS            0  // used by exit()
#define EXIT_ERROR              1

#define XDATA_READ              2
#define XDATA_READ_RAW          3
//==!=====================================================================!==
//==! Parsed data attributes
//==!=====================================================================!==
enum EnumPARSE_ATTR
{ eATTR_NUM, eATTR_VAR, eATTR_PUNCT, eATTR_WSPACE, eATTR_OTHERS
};
//==!=====================================================================!==
//==! Message display instructions for dispMsg()
//==!=====================================================================!==
#define DISP_USE_CUR_POS     0x00 // print from current position
#define DISP_CLEAR           0x01 // clear terminal/monitor position
#define DISP_PRE_CR          0x02 // do carriage-return before message
#define DISP_PRE_LF          0x04 // do line-feed before message
#define DISP_POST_CR         0x08 // do carraige return after message
#define DISP_POST_LF         0x10 // do line-feed after message
//==!=====================================================================!==
//==! Used by jMessage()
//==!=====================================================================!==
#define DISP_SHOW_MSG         100
#define DISP_AND_EXIT         101
//==!=====================================================================!==
//==! Constants used by cDataPort()
//==!=====================================================================!==
#define dREADBUFMAX     256*1024 // max size of dataport read buffer
#define dWRITEBUFMAX    256*1024 // max size of dataport write buffer
//==!=====================================================================!==
//==! Constants used by getFloat()
//==!=====================================================================!==
#define GF_REALFORMAT \
               "Use the ff number syntax: digits[.digits][e[+|-]digits]"
#define GF_POSNUMBER    "Positive number expected"
#define GF_FLOAT              110
#define GF_LONG               111
#define GF_UNKNOWNTYPE        112
//==!=====================================================================!==
//==! cCharStr class: multiple instances are combined into a linked list to
//==! hold strings longer than MAX_XDATA_SIZE.
//==!=====================================================================!==
class cCharStr
{
public:
  long m_lSize;
  char *m_csStr;

  cCharStr *m_pNextcCharStr;

  cCharStr();
  ~cCharStr();

}; // class cCharStr
//==!=====================================================================!==
//==! cDataPort class: keeps track of a data port's read/write context.
//==!=====================================================================!==
class cDataPort
{
public:
  FILE *pFILE;                     // pointer to data port (usually a file)
  char *psPortName,                // data port name

//==! sOrigData and sData are initially set to MAX_XDATA_SIZE, but will
//==! change if the parsed data is larger than this size. These 2 variables
//==! can be changed by a call to xDataBsdl() which can read very long
//==! char or hex strings in a BSDL file separated by the & character. The
//==! long char or hex string is first read char by char until a delimiter
//==! is found. These chars are stored in temporary arrays that are
//==! MAX_XDATA_SIZE long and combined into a linked list pointed to by
//==! m_pcCharList. Once the char string is terminated, its size is
//==! calculated by getting the size of all arrays in m_pcCharStr. sData and
//==! sOrigData are then resized, and the chars from the arrays in the linked
//==! list are copied to them. The size of sData/sOrigData is stored in the
//==! member variable m_lDpDataSize.
       *sOrigData,                 // current dataport data
       *sData,                     // sOrigData converted to lowercase

       *m_csReadBuf,               // array pointer for buffered read ops
       *m_csWriteBuf;              // pointer to fwrite buffer

  cCharStr *m_pcCharStr;           // used by xDataBsdl() to read long char
                                   // strings representing hex values;
                                   // once size is determined, will then
                                   // change sData/sOrigData size (which will
                                   // also be recorded in m_lDpDataSize)

  long m_lDpDataSize,              // size of sData and sOrigData
       lLineNum,                   // current line number
       lCharCount,                 // current number of characters read
       m_lReadBufIndex,            // index into m_csReadBuf
       m_lReadBufMax,              // max m_csReadBuf index set by fread()
       m_lWriteBufCnt,             // number of items/chars in m_csWriteBuf[]
       m_lWriteTmpBufCnt;          // number of items/chars in
                                   // m_csTmpWriteBuf[]
  EnumPARSE_ATTR eDataAttr;        // enumerated parsed data attribute
  int  iCurChar,                   // current data port char
       iCommentStyle,              // comment style for this file
       iBracketLevel,              // keeps track of bracket level/depth
       iBsdlEntityEnd,             // 1 if BSDL entity ended with a ';'
       iParserGetData,             // flag to get next data from data port
       iParserGetChar,             // flag to get next char from data port
       iParserWarnings,            // parser warnings while processing file

//==! 2003/12/31: The EOF condition can be encountered when performing
//==! buffered reads from a file source. The data read routine should then
//==! read data from the read buffer.
//==! - m_iBufDataReadEnd is set to 1 to prevent further buffered read
//==!   attempts
//==! - m_iEndOfDataStream is set to 1 after all data in read buffer from
//==!   last read attempt has been read
       m_iEndOfDataStream,         // 1 if at end of data stream
       m_iBufDataReadEnd,          // 1 if dataport buffered read encountered
                                   // end condition
       m_iDataFormat,              // 1 for writing data in ISC data format
       m_iFirstData;               // 1 if first data not yet written, else 0
                                   // (used during formatting to put commas
                                   // between data units/records after the
                                   // first data unit/record)
  cDataPort *pNextDataPort;        // pointer to next cDataPort

  cDataPort(char *psStr);
  ~cDataPort();
  void resetDataPort();
  void xPortReopen(char *sMode);
  void xCloseDataPort();
  int writeToCDP(int iMode,char *csData);    // write to cDataPort

}; // class cDataPort
//==!=====================================================================!==
//==! Global variables
//==!=====================================================================!==
DEF_DEC cDataPort
  *pDP_List,          //==! Pointer to linked list of all data ports
  *pDP_Input,         //==! Pointer to input data port(s)
  *pDP_Output,        //==! Pointer to output data port
  *pDP_Log,           //==! Pointer to log data port
  *pNewDataPort;      //==! Global temporary data port pointer

DEF_DEC char
  sMsgOnly[MAX_XDATABUF_SIZE],sGlobMessageBuf[MAX_XDATABUF_SIZE],
  sMsgBuf00[MAX_XDATABUF_SIZE],sMsgBuf01[MAX_XDATABUF_SIZE],
  sMsgBuf02[MAX_XDATABUF_SIZE],sMsgBuf03[MAX_XDATABUF_SIZE],
  sHoldData[MAX_XDATA_SIZE],   // Buffer to hold cDataPort data
  csTmpWriteBuf[MAX_XDATA_SIZE]; // Buffer to hold formatted data before
                                 // sending to cDataPort->writeToCDP()

DEF_DEC unsigned long lSpecOptions;

DEF_DEC time_t t_ProgStart,t_ProgFinish;

DEF_DEC int
  iOptShowStatus,
  iOptOverWritePrompt,     //==! If 1, then prompt the user before
                           //==! deleting/overwriting a file.
  iGlobOptBufferedReadOp,  //==! If 1 then read operations from dataport are
                           //==! buffered; else read data one by one. Default
                           //==! is 1, set to 0 using hidden command line
                           //==! option "-nobufread"
  iGlobOptBufferedWriteOp; //==! If 1 then use fwrite() to write data to
                           //==! destination, else use fprintf() (set to 0
                           //==! using -nobufwrite option)
DEF_DEC long
  lGlobOptBufferedReadSize, //==! Specifies size of array for buffered read
                           //==! operations using fread(). Default to
                           //==! dREADBUFMAX, but can be changed via hidden
                           //==! command line option "-readbufsize N"
  lGlobOptBufferedWriteSize; //==! Specified size of array for buffered write
                           //==! operations using fwrite(). Default to
                           //==! dWRITEBUFMAX, but can be changed via hidden
                           //==! command line option "-writebufsize N"
//==!=====================================================================!==
//==! Function prototypes
//==!=====================================================================!==
void isPtrNull(void *pvPtr,char *sMsg);
int  isInRange(int iMode,long lVal,long lMinV,long lMaxV,char *sMsg,
        int iStat);
void resetGlobalVars();
void initGlobalVars();
void initApplicationVars();
void freeGlobalMem();
void freeApplicationMem();
void waitUsec(long lMicroSec);
char *readCmdLineData(int *argc,char ***argv,int iMode);
void readCmdLine(int argc,char **argv);
void getExtFilename(cDataPort *pChn,char *pBuf,int iSize);
void getDpChar(cDataPort *pDP,int iMode);
void xParseNumData(cDataPort *pChn,char cChr);
int isComment(cDataPort *pDP);
void xParseData(cDataPort *pChn,int iMode);
char xQuery(char *pQueryPrompt,char *pValidResp);
void dispMsg(char *sMsg,unsigned int uiDispInstr,int iMode,int iLogMode);
void jMessage(int iMsgType,char *sMsg,int iMode,cDataPort *pChn);
void skipDpLine(cDataPort *pDP);
void copyStr(char **pStr,char *pSrcStr);
int  strCaseConvert(char *pStr,long lMax,int iMode);
void xFileMsg(char *sFileName,int iFileStatus,cDataPort *pChn,int iMode);
void readCombinedOption(unsigned long lOpts);
long lReadDecimal(cDataPort *pChn);
void xFclose(FILE **pFP);
FILE *xFileOpen(char *pFileName,char *pChkMode);
int findStr(cDataPort *pDP,char *sStr,int iMode);
int findChar(cDataPort *pDP,char sChr,int iMode);
int getFloat(cDataPort *pDP,float *fNum,long *lNum);
int  xstrcmp(char *sA,char *sB);
void xstrcat(char **sDest,char *sSrc);
void xInitCharArray(char *cBuf,int iCnt,char cFill,int iMode);
//==!=====================================================================!==
//==! end of XCOMMON.H
//==!=====================================================================!==

