/***************************************************************************/
/* J Drive (tm) Engine                                                     */
/* Copyright 2003-2005, Xilinx, Inc.                                       */
/* The use, reproduction, and distribution of this code is subject         */
/* to the J Drive license requirements and restrictions.                   */
/***************************************************************************/
//==!=====================================================================!==
//==! See XPROJECT.H for project file list and project version history.
//==!=====================================================================!==
//==! Revision history: XDRIVE.CPP
//==! 2003/04/14: setup and main() calls for 1532 XDRIVE program
//==!=====================================================================!==
#define COMPILE_DEFINE_VARS 1
#include "xcommon.h"
#include "x_bsdl.h"
/*==!===================================================================!==*/
/*==! Display info or messages to the system screen. Modify or remove if   */
/*==! code is to be used in an embedded system.                            */
/*==!                                                                      */
/*==! Function Parameters : iMode = exit if 1, else continue               */
/*==!                       sMsg = string to display                       */
/*==! Return Value        : none                                           */
/*==!===================================================================!==*/
void dispInfo(int iMode,char *sMsg)
{ int iTemp;

  iTemp = iOptShowStatus;
  iOptShowStatus = 1;

  dispMsg(sMsg,DISP_USE_CUR_POS,1,1);
  if (iMode == 1)
  { freeGlobalMem();
    freeApplicationMem();
    exit(EXIT_ERROR);
  } // IF

  iOptShowStatus = iTemp;

} /* dispInfo() */
//==!=====================================================================!==
CMCSrecord::CMCSrecord()
{

  pUCarray = NULL;
  resetMCSrec();

} // CMCSrecord::CMCSrecord()
//==!=====================================================================!==
//==! Reset the CMCSrecord object
//==!
//==! Function Parameters: none
//==! Return Value       : none
//==!=====================================================================!==
void CMCSrecord::resetMCSrec()
{
  lByteCount = 0;
  lByteAddr = -1;
  lOffset = 0;
  lBufIndex = 0;
  lCurAddr = -1;
  lCurLinearOffset = -1;
  iRecType = -1;
  iDataAvail = 1;
  freeArray(pUCarray);

} // CMCSrecord::CMCSrecord()
//==!=====================================================================!==
//==! CMCSrecord destructor.
//==!
//==! Function Parameters: none
//==! Return Value       : none
//==!=====================================================================!==
CMCSrecord::~CMCSrecord()
{

  freeArray(pUCarray);

} // CMCSrecord::~CMCSrecord()
//==!=====================================================================!==
//==! CMCSfile constructor.
//==!
//==! Function Parameters: none
//==! Return Value       : none
//==!=====================================================================!==
CMCSfile::CMCSfile()
{
  lCMCSfileCnt = 0;
  iBitPos = -1;
  ucByte = 0;
  pDP = NULL;
  pCMCSinfo = NULL;
  pNextCMCSfile = NULL;

} // CMCSfile::CMCSfile()
//==!=====================================================================!==
//==! CMCSfile member function to reopen an MCS file.
//==!
//==! Function Parameters: sMode = file open mode
//==! Return Value       : none
//==!=====================================================================!==
void CMCSfile::reopenMCS(char *sMode)
{

  ucByte = 0;
  if (pDP != NULL)
  { xFclose(&(pDP->pFILE));
    pDP->pFILE = xFileOpen(pDP->psPortName,sMode);
    pDP->resetDataPort();
  } // IF
  if (pCMCSinfo != NULL)
  { pCMCSinfo->resetMCSrec();
  } // IF

} // CMCSfile::reopenMCS()
//==!=====================================================================!==
//==! CMCSfile destructor.
//==!
//==! Function Parameters: none
//==! Return Value       : none
//==!=====================================================================!==
CMCSfile::~CMCSfile()
{

  freeData(pCMCSinfo);

} // CMCSfile::~CMCSfile()
//==!=====================================================================!==
//==! Setup the MCS file object given a filename.
//==!
//==! Function Parameters: sMcsName = MCS filename
//==!                      pMcsObj = pointer to MCS object pointer
//==! Return Value       : none
//==!=====================================================================!==
void setupMcsFile(char *sMcsName,CMCSfile **pMcsObj)
{ FILE *pFptr;
  cDataPort *pDP_Tmp;

  pFptr = xFileOpen(sMcsName,"r");
  xFclose(&pFptr);
  pDP_Tmp = new cDataPort(sMcsName);
  if (*pMcsObj != NULL)
  { if ((*pMcsObj)->pDP != NULL)
    { freeData((*pMcsObj)->pDP);
    } // IF
    freeData(*pMcsObj);
  } // IF
  *pMcsObj = new CMCSfile;
  (*pMcsObj)->lCMCSfileCnt = 1;
  (*pMcsObj)->pDP = pDP_Tmp;
  pDP_Tmp = NULL;

} // setupMcsObj()
//==!=====================================================================!==
//==! Process program-specific options. Other generic options like input,
//==! output, or log filenames are done in readCmdLine().
//==!
//==! Function Parameters: pointers to argc and argv (from command line)
//==! Return Value       : 1 if a valid option was specified, 0 otherwise
//==!=====================================================================!==
int processProgOptions(int *argc,char ***argv)
{ char *pChr;
  int iFlag;
  FILE *pFP_tmp;

  pChr = (*argv[0]) + 1;

  if (!xstrcmp(pChr,"checkbsdl"))
  { iGlobOptCheckBsdl = 1;
    iOpt_CheckInputFiles = 1;   // Force program to skip processing of
                                // 1532 operations in CMD file
    pChr = readCmdLineData(&(*argc),&(*argv),1);
    if (pChr == NULL)
    { jMessage(DISP_AND_EXIT,"Specify BSDL file to check",0,NULL);
    }
    else
    { pFP_tmp = xFileOpen(pChr,"t");
      if (pFP_tmp != NULL)
      { xFclose(&pFP_tmp);
        copyStr(&psGlobBsdlStr,pChr);
      } // IF
      else
      { sprintf(sMsgOnly,"Cannot open/read BSDL file [%s]",pChr);
        jMessage(DISP_AND_EXIT,sMsgOnly,0,NULL);
        return 0;
      } // ELSE
    } // ELSE
    return 1;
  } // IF option = checkbsdl

  if (!xstrcmp(pChr,"checkinpfiles"))
  { iOpt_CheckInputFiles = 1;
    return 1;
  } // IF option = checkinpfiles

  if (!xstrcmp(pChr,"continue_on_fail"))
  { iGlobOptContinueOnFail = 1;
    return 1;
  } // IF option = continue_on_fail

  if (!xstrcmp(pChr,"bsdlwarnings"))
  { iGlobOptDisplayBsdlWarnings = 1;
    return 1;
  } // IF option = bsdlwarnings

  if (!xstrcmp(pChr,"maxbiterr"))
  { pChr = readCmdLineData(&(*argc),&(*argv),1);
    if (pChr == NULL)
    { jMessage(DISP_AND_EXIT,"Specify max # of bit errors to display",0,NULL);
    }
    else
    { lOptMaxBitErr = atol(pChr);
    } // ELSE
    return 1;
  } // IF option = maxbiterr

  if (!xstrcmp(pChr,"m"))  //==! Processing mode
  { iFlag = 0;
    pChr = readCmdLineData(&(*argc),&(*argv),1);
    if (pChr == NULL)
    { iFlag = 1;
    } // IF
    else
    { if (!xstrcmp(pChr,"concurrent"))
      { iGlobProcMode = 1;
      } // IF
      else
      { if (!xstrcmp(pChr,"sequential"))
        { iGlobProcMode = 0;
        } // IF
        else
        { if (!xstrcmp(pChr,"max_concurrent"))
          { iGlobProcMode = 1;
            iGlobOptMaxConcurrent = 1;
          } // IF
          else
          { iFlag = 1;
          } // ELSE
        } // ELSE
      } // ELSE
    } // ELSE

    if (iFlag == 1)
    { jMessage(DISP_AND_EXIT,"Specify -m [sequential | concurrent]",0,NULL);
    } // IF
    return 1;
  } // IF option = m

  if (!xstrcmp(pChr,"min0clk"))
  { iOpt_MinOneClk = 0;
    return 1;
  } // IF option = min1clk

  if (!xstrcmp(pChr,"nomemiscdata"))
  { iGlobOptUseMemIscData = 0;
    return 1;
  } // IF option = nomemiscdata

  if (!xstrcmp(pChr,"skip_error_proc"))
  { iGlobOptUseErrorProc = 0;
    return 1;
  } // IF option = skip_error_proc

  if (!xstrcmp(pChr,"skip_exit_on_error"))
  { iGlobOptSkipExitOnError = 1;
    return 1;
  } // IF option = skip_exit_on_error

  if (!xstrcmp(pChr,"xshowallopts"))
  { iOpt_ShowAllOpts = 1;
    return 1;
  } // IF option = xshowallopts

  if (!xstrcmp(pChr,"xshowcmdfileopts"))
  { iOpt_ShowCmdFileOpts = 1;
    return 1;
  } // IF option = xshowcmdfileopts

  return 0;

} // processProgOptions()
//==!=====================================================================!==
//==! Check if command line option combinations are valid.
//==!
//==! Function Parameters: none
//==! Return Value       : 1 if successful, 0 otherwise
//==!=====================================================================!==
int checkOptCombo()
{

  return 1;

} // checkOptCombo()
//==!=====================================================================!==
//==! Display program-specific command line options.
//==!
//==! Function Parameters: none
//==! Return Value       : none
//==!=====================================================================!==
void displayProgOptions()
{
  dispMsg("Options:",DISP_POST_LF,1,1);
  dispMsg(
"  -bsdlwarnings = display BSDL DESIGN_WARNINGs and ISC_DESIGN_WARNINGs",
     DISP_POST_LF,1,1);
  dispMsg("  -checkinpfiles = check files specified in command file",
     DISP_POST_LF,1,1);
  if (iOpt_ShowAllOpts == 1)
  { dispMsg("  -checkbsdl <BSDL file> = check specified BSDL file then exit",
       DISP_POST_LF,1,1);
    dispMsg("  -continue_on_fail = continue device processing even on failure",
       DISP_POST_LF,1,1);
    dispMsg("  -min0clk = turn off single 0-1-0 TCK pulse when waiting X secs",
       DISP_POST_LF,1,1);
    dispMsg("  -nobufread = turn off buffered read operations",
       DISP_POST_LF,1,1);
    dispMsg("  -no_fwrite = use fprintf() instead of fwrite() to write data",
       DISP_POST_LF,1,1);
    dispMsg("  -nomemiscdata = do not hold ISC data in memory",
       DISP_POST_LF,1,1);
    dispMsg("  -readbufsize N = specify fread buffer size",
       DISP_POST_LF,1,1);
    dispMsg("  -writebufsize N = specify fwrite buffer size",
       DISP_POST_LF,1,1);
    xQuery("-==< Press ENTER to continue >==-",NULL);
    dispMsg("  -xopt N = debugging options (N = sum of the following)",
       DISP_POST_LF,1,1);
    dispMsg("        =>   1: turn off status display",
       DISP_POST_LF,1,1);
    dispMsg("        =>   2: print program debug information",
       DISP_POST_LF,1,1);
    dispMsg("        =>   4: print BSDL info to output file",
       DISP_POST_LF,1,1);
    dispMsg("        =>   8: skip IDCODE check",
       DISP_POST_LF,1,1);
    dispMsg("        =>  16: turn on flow loop debug code",
       DISP_POST_LF,1,1);
    dispMsg("        =>  32: display ACTION repeat count",
       DISP_POST_LF,1,1);
    dispMsg("  -xshowallopts = show all options",
       DISP_POST_LF,1,1);
    dispMsg("  -xshowcmdfileopts = show ISC command file format and options",
       DISP_POST_LF,1,1);
    xQuery("-==< Press ENTER to continue >==-",NULL);
  } // IF iOpt_ShowAllOpts == 1

  if (iOpt_ShowCmdFileOpts == 1)
  { dispMsg("--- ISC Command File Format and Options ----------------",
       DISP_PRE_LF | DISP_POST_LF,1,1);
    dispMsg("  = devices are specified in TDO to TDI order",
       DISP_POST_LF,1,1);
    dispMsg("  = format: device_name [options]; [device_name [options];]",
       DISP_POST_LF,1,1);
    dispMsg("  = options:",
       DISP_POST_LF,1,1);
    dispMsg("     -a\"<action name>\"",
       DISP_POST_LF,1,1);
    dispMsg("     -b\"<BSDL filename>\"",
       DISP_POST_LF,1,1);
    dispMsg(
    "     -d\"<ISC data filename>\"  <-- specify file if required by action",
       DISP_POST_LF,1,1);
    dispMsg(
    "     -iN   <-- N = JTAG instr reg length (to BYPASS part if no BSDL)",
       DISP_POST_LF,1,1);
    dispMsg("     -o\"<device output filename>\"",
       DISP_POST_LF,1,1);
    dispMsg("     -data_isc  <-- use ISC data format for -o\"filename\"",
       DISP_POST_LF,1,1);
    dispMsg(
    "     -skip_dev  <-- skip device (device is physically not in JTAG chain)",
       DISP_POST_LF,1,1);
    xQuery("-==< Press ENTER to continue >==-",NULL);
  } // IF iOpt_ShowCmdFileOpts == 1

  dispMsg("  -l <log file> = write diagnostic messages to log file",
     DISP_POST_LF,1,1);
  dispMsg("  -m [sequential (default) | concurrent] = processing mode",
     DISP_POST_LF,1,1);
  dispMsg("  -maxbiterr N = N -> max number of bit errors to display",
     DISP_POST_LF,1,1);
  dispMsg("  -o <output file> = write debug messages to output file",
     DISP_POST_LF,1,1);
  dispMsg(
"  -skip_error_proc = use BYPASS instead of PROC_ERROR_EXIT when errors occur",
     DISP_POST_LF,1,1);
  dispMsg(
"  -skip_exit_on_error = process other devices even on flow exit_on_error",
     DISP_POST_LF,1,1);

} // displayProgOptions()
//==!=====================================================================!==
//==! Initialize variables specific to this application. Different from
//==! initGlobalVars() which initializes the variables used by the common
//==! libraries, DLLs, or object files used by the program (e.g., the
//==! parser, file manager, etc.)
//==!
//==! Function Parameters: none
//==! Return Value       : none
//==!=====================================================================!==
void initApplicationVars()
{

  iOpt_PrintBsdl = 0;
  iOpt_ProgDebug = 0;
  iOpt_RptCntDispOff = 0;
  iOpt_MinOneClk = 1;
  iOpt_SkipIdCheck = 0;
  iOpt_ShowAllOpts = 0;
  iOpt_ShowCmdFileOpts = 0;
  iOpt_CheckInputFiles = 0;

  iGlobOptUseMemIscData = 1; //==! Set to 0 with "-nomemiscdata" command line
                             //==! option
  iGlobOptMaxConcurrent = 0;
  iGlobOptCheckBsdl = 0;     //==! Set to 1 using "-checkbsdl <BSDL file>"
  psGlobBsdlStr = NULL;

  iGlobOptDisplayBsdlWarnings = 0;  //==! Set to 1 with "-bsdlwarnings"
                             //==! command line option
  iGlobOptUseErrorProc = 1;  //==! Set to 0 with "-skip_error_proc" command
                             //==! line option
  iGlobOptSkipExitOnError = 0;
                              //==! Set to 1 using "-skip_exit_on_error" to
                              //==! force program to continue processing even
                              //==! if a flow with the EXIT_ON_ERROR flag has
                              //==! failed
  iGlobExitOnError = 0;       //==! Global variable set to 1 if a flow with
                              //==! EXIT_ON_ERROR flag has failed and the
                              //==! option iGlobOptSkipExitOnError = 0

  iGlobOptContinueOnFail = 0; //==! Set to 1 using "-continue_on_fail"
                              //==! command line option

  pBsdlData = NULL;
  pgCDD_list = NULL;      //==! Global device data list (TDO to TDI)
  pgCDD_FromTdi = NULL;   //==! Global device data list (TDI to TDO)
  pgCBD_list = NULL;      //==! global BSDL file list
  pgCMIDS_list = NULL;    //==! global CMemIscDataSrc list
  pTempCPortData = NULL;
  iGlobProcMode = 0;

  pCMCSobject = new CMCSrecord;
  pCMCSobject->lByteCount = 32;
  pCMCSobject->pUCarray = new unsigned char [pCMCSobject->lByteCount];
  pCMCSobject->lBufIndex = 0;
  pCMCSobject->lCurAddr = 0;
  pCMCSobject->lCurLinearOffset = 0;

#if PC_DEBUG == 1
  iTapState = TAP_RESET;
  iTapPrint = 0;
#endif

  iStatus = 0;
  iMatchAll = 0;
  lCntBytesRead = 0;
  iGlobReadByte = 0;
  iGlobReadBitCnt = 0;
  iReadUserArea = 0;
  iReadUserAreaDone = 0;
  iReadDevice = 0;
  lMismatchCnt = 0;
  lTdiTdoInstrCnt = 0;

  g_verbose = 0;
  iGlobParportPin17Setting = 0;

  lHdr = lHir = lTdr = lTir = 0;
  lOptMaxBitErr = 50;
  lMaxIrChainLength = 400;
  lMaxDrChainLength = 100;

  pMcsFile = NULL;

  cTdiData = NULL;
  cTdoData = NULL;
  cTdoCopy = NULL;
  cGlobTapDataReg = NULL;

  iOptCalibrateTimer = 0;
  lMicroSecCount = 40;

  iOptByteOrder = 0;       // set to 1 using -byteorder to display flash
                           // read data in byte order; default is word order

} // initApplicationVars()
//==!=====================================================================!==
//==! Free/deallocate any memory used by application-specific variables.
//==!
//==! Function Parameters: none
//==! Return Value       : none
//==!=====================================================================!==
void freeApplicationMem()
{ CDeviceData *pCDD_var;
  CBsdlData *pCBD_var;
  CMemIscDataSrc *pCMIDS_var;

  iOpt_PrintBsdl = 0;
  iOpt_ProgDebug = 0;
  iOpt_RptCntDispOff = 0;
  iOpt_MinOneClk = 1;
  iOpt_SkipIdCheck = 0;
  iOpt_ShowAllOpts = 0;
  iOpt_ShowCmdFileOpts = 0;
  iOpt_CheckInputFiles = 0;

  iGlobOptCheckBsdl = 0;
  freeArray(psGlobBsdlStr);

  freeData(pBsdlData);

  while (pgCDD_list != NULL)
  { pCDD_var = pgCDD_list->m_pNextCDeviceData;
    pgCDD_list->m_pNextCDeviceData = NULL;
    freeData(pgCDD_list);
    pgCDD_list = pCDD_var;
  } // WHILE pgCDD_list != NULL
  pgCDD_FromTdi = NULL;

  while (pgCBD_list != NULL)
  { pCBD_var = pgCBD_list->m_pNextCBsdlData;
    pgCBD_list->m_pNextCBsdlData = NULL;
    freeData(pgCBD_list);
    pgCBD_list = pCBD_var;
  } // WHILE pgCBD_list != NULL

  while (pgCMIDS_list != NULL)
  { pCMIDS_var = pgCMIDS_list->m_pNextCMemIscDataSrc;
    pgCMIDS_list->m_pNextCMemIscDataSrc = NULL;
    freeData(pgCMIDS_list);
    pgCMIDS_list = pCMIDS_var;
  } // WHILE pgCMIDS_list != NULL

  freeData(pTempCPortData);

  iGlobProcMode = 0;

  closePort();

  if (pCMCSobject != NULL)
  { if (pCMCSobject->pUCarray != NULL)
    { freeArray(pCMCSobject->pUCarray);
    } // IF
    freeData(pCMCSobject);
  } // IF

  if (pMcsFile != NULL)
  { if (pMcsFile->pDP != NULL)
    { freeData(pMcsFile->pDP);
    } // IF
    freeData(pMcsFile);
  } // IF

  freeArray(cTdiData);
  freeArray(cTdoData);
  freeArray(cTdoCopy);
  freeArray(cGlobTapDataReg);

} // freeApplicationMem()
//==!=====================================================================!==
//==! waitUsec(): wait specified number of microseconds by using the
//==! calibrated value lMicroSecCount. Note that timing using this function
//==! is dependent on other applications running in the system. The print
//==! code in the FOR loop that does the timing was put there to force the
//==! compiler to generate code (instead of optimizing the FOR loop away if
//==! the loop body was empty).
//==!
//==! Function Parameters: lMicroSec = number of microseconds to wait
//==! Return Value       : none
//==!=====================================================================!==
void waitUsec(long lMicroSec)
{ long lUsecTime,lVar;

  lUsecTime = lMicroSec * lMicroSecCount;
  for (lVar = 0; lVar < lUsecTime; ++lVar)
  { if (iOpt_ProgDebug == 1 && lVar%1000000 == 999999)
    { sprintf(sMsgOnly,"  = wait usec [%ld]",lVar);
      dispMsg(sMsgOnly,DISP_PRE_CR,1,0);
    } // IF
  } // FOR lVar < lUsecTime

} // waitUsec()
//==!=====================================================================!==
//==! calibrateTimer(): sets lMicroSecCount variable to a value which will be
//==! used by the FOR loop in waitUsec() to wait the specified number of
//==! microseconds. The functions Sleep() and waitTime() provide millisecond
//==! resolution only.
//==!
//==! Function Parameters: none (use all global variables)
//==! Return Value       : none
//==!=====================================================================!==
void calibrateTimer()
{ time_t t_Stop;

  lMicroSecCount = 0;
  t_Stop = time(NULL) + 30;
  while (time(NULL) != t_Stop)
  { ++lMicroSecCount;
  } // WHILE

  lMicroSecCount = lMicroSecCount/1000000 + 1;

} // calibrateTimer()
//==!=====================================================================!==
//==! Prepare files used by the application:
//==! - open input file(s) or open them when needed
//==! - open output file pointer (can be opened in binary mode)
//==! - open the log file
//==! - open any other files
//==!
//==! Function Parameters: sysArgc = command line arg count
//==!                      sysArgv = command line arg data
//==! Return Value       : none
//==!=====================================================================!==
void prepareFiles(int sysArgc,char **sysArgv)
{ int iX;
  long lPrintPos;

  if (pDP_Output != NULL)
  { if (pDP_Output->psPortName != NULL)
    { xFclose(&(pDP_Output->pFILE));
      pDP_Output->pFILE = xFileOpen(pDP_Output->psPortName,"w");
      fprintf(pDP_Output->pFILE,
              "%s %s\n%s\n",PROG_NAME_TM,VERSION_NUM,PROG_DESC);
      lPrintPos = fprintf(pDP_Output->pFILE,"- command line:");
      for (iX = 0; iX < sysArgc; ++iX)
      { lPrintPos += fprintf(pDP_Output->pFILE," %s",sysArgv[iX]);
        if (lPrintPos > 70)
        { lPrintPos = fprintf(pDP_Output->pFILE,"\n\t");
        } // IF
      } // FOR iX < sysArgc
    } // IF
  } // IF
  if (pDP_Log != NULL)
  { if (pDP_Log->psPortName != NULL)
    { xFclose(&(pDP_Log->pFILE));
      pDP_Log->pFILE = xFileOpen(pDP_Log->psPortName,"w");
    } // IF
  } // IF

  if (pDP_Input != NULL)
  { xFclose(&(pDP_Input->pFILE));
    pDP_Input->pFILE = xFileOpen(pDP_Input->psPortName,"r");
  } // IF

} // prepareFiles()
//==!=====================================================================!==
//==! Main() call. Returns program status that can be used in batch
//==! processing mode.
//==!
//==! Function Parameters: argc and argv (from command line)
//==! Return Value       : program EXIT value
//==!=====================================================================!==
int main(int argc,char **argv)
{ int iProgStat;          //==! Program status
  long lOperRepeatCount = 0,lOperPass = 0,lTotProgRunTime;
  time_t t_OperStart;
  cDataPort *pCDP_Bsdl;
  CBsdlData *pCBD_var;

#if SYS_UNIX == 0
  _CrtSetDbgFlag(_CRTDBG_ALLOC_MEM_DF | _CRTDBG_LEAK_CHECK_DF);
#endif

  iGlobProgramState = dPROGSTATE_INITIALIZE;

  initGlobalVars();
  initApplicationVars();
  readCmdLine(argc,argv);

  if (iOptCalibrateTimer == 1)
  { calibrateTimer();
  } // IF

//==! Initialize output or log files before processing input file(s)
  prepareFiles(argc,argv);

  sprintf(sMsgOnly,"%s %s\n%s",PROG_NAME_TM,VERSION_NUM,PROG_DESC);
  dispMsg(sMsgOnly,DISP_POST_LF,1,1);

  t_ProgStart = time(NULL);
  t_OperStart = t_ProgStart;  // Start of operation
  strcpy(sMsgBuf00,ctime(&t_ProgStart));
  dispMsg(sMsgBuf00,DISP_PRE_LF,1,1);

  if (iOpt_CheckInputFiles == 0)
  { xportsInitialize();
  } // IF

  if (iGlobOptCheckBsdl == 1)  // Check BSDL file
  { pCDP_Bsdl = new cDataPort(psGlobBsdlStr);
    sprintf(sMsgBuf00,"--> Checking BSDL file [%s]",pCDP_Bsdl->psPortName);
    dispMsg(sMsgBuf00,DISP_POST_LF,1,1);
    pCDP_Bsdl->pFILE = xFileOpen(pCDP_Bsdl->psPortName,"r");
    pCBD_var = new CBsdlData;      // create new BSDL object
    iProgStat = pCBD_var->readBsdl(pCDP_Bsdl); // read BSDL info
    if (iProgStat == 1)
    { dispMsg("    = File OK",DISP_POST_LF,1,1);
    } // IF
    freeData(pCDP_Bsdl);
    freeData(pCBD_var);
    freeArray(psGlobBsdlStr);
  } // IF
  else
  { iProgStat = readCmdFile(pDP_Input);
    freeData(pDP_Input);   //==! Free command file
    iProgStat = checkCmdOptions(pgCDD_list);
  } // ELSE

  if (iOpt_CheckInputFiles == 0 && iProgStat == 1)
  { prepareDataBuffers(pgCDD_list,iGlobProcMode);
    for (lOperRepeatCount = 0,lOperPass = 0;;++lOperRepeatCount)
    { if (processDeviceList(pgCDD_list,lOperRepeatCount) == 1)
      { ++lOperPass;   // keep track of successful operations
      } // IF
      t_ProgFinish = time(NULL);

      sprintf(sMsgOnly,
          "---< Elapsed clock time for iteration [%ld] = %ld seconds >---",
          lOperRepeatCount+1,(long)(difftime(t_ProgFinish,t_ProgStart)));
      dispMsg(sMsgOnly,DISP_PRE_LF | DISP_POST_LF,1,1);

      if (xQuery("==> Repeat operation (Y/N) ? ","YN") == 'N')
      { break;
      } // IF

      t_ProgStart  = time(NULL);
      strcpy(sMsgBuf00,ctime(&t_ProgStart));
      dispMsg(sMsgBuf00,DISP_PRE_LF,1,1);
    } // FOR (;;)
    freeDataBuffers(pgCDD_list);

    iProgStat = (lOperPass < (lOperRepeatCount+1) ? 0 : 1);
  } // IF

  t_ProgFinish = time(NULL);
  strcpy(sMsgBuf00,ctime(&t_ProgFinish));
  dispMsg(sMsgBuf00,DISP_PRE_LF,1,1);

  lTotProgRunTime = (long)(difftime(t_ProgFinish,t_OperStart));
  sprintf(sMsgBuf00,"---[ Total program runtime: %ld sec%s ]---",
          lTotProgRunTime,(lTotProgRunTime > 1 ? "s" : " "));
  if (iOpt_CheckInputFiles == 0 && iGlobOptCheckBsdl == 0)
  { dispMsg(sMsgBuf00,DISP_PRE_LF,1,1);
    sprintf(sMsgBuf00,"   => passed %ld of %ld iteration%s",
            lOperPass,lOperRepeatCount+1,lOperRepeatCount > 0 ? "s" : " ");
  } // IF
  dispMsg(sMsgBuf00,DISP_PRE_LF | DISP_POST_LF,1,1);

//==! Free memory used by common/global variables followed by memory used
//==! by application-specific variables before terminating program.
  freeGlobalMem();
  freeApplicationMem();

  exit(iProgStat == 0 ? EXIT_ERROR : EXIT_SUCCESS);
  
  return(EXIT_SUCCESS);  // Added to turn off compiler warning

} // main()
//==!=====================================================================!==
//==! end of module XDRIVE.CPP
//==!=====================================================================!==

