/***************************************************************************/
/* J Drive (tm) Engine                                                     */
/* Copyright 2003-2005, Xilinx, Inc.                                       */
/* The use, reproduction, and distribution of this code is subject         */
/* to the J Drive license requirements and restrictions.                   */
/***************************************************************************/
//==!=====================================================================!==
//==! Revision history: XPROJECT.H
//==! - created 2003/06/30
//==! - project file list and project version history
//==! - program name and version number of project
//==! - template functions
//==!=====================================================================!==
//==! Project [XDRIVE] file list
//==!=====================================================================!==
//==! - XDRIVE reads BSDL data and 1532 instructions from a BSDL file. Reads
//==!   data from an ISC file when required by a 1532 operation/action.
//==!
//==!   XCOMMON.H      = common includes and variables for parser
//==!   XPARSER.CPP    = parser
//==!                                                
//==!   XPROJECT.H     = program name and version number
//==!   X_BSDL.H       = defines for reader functions
//==!   PORTS.H        = defines for parallel port functions
//==!   BJDRVIF.H      = header for JTAG driver
//==!   WINDRVR.H      = header for Windows parallel port driver
//==!
//==!   XDRIVE.CPP     = main()
//==!   XCHKFUNC.CPP   = higher-level JTAG shift functions
//==!   XRD_BSDL.CPP   = BSDL reader
//==!   XRD_DATA.CPP   = ISC data reader
//==!   XROUTINE.CPP   = MCS data reader
//==!   PORTS.CPP      = parallel port functions
//==!   BASJDRVI.CPP   = Windows parallel port drive functions
//==!=====================================================================!==
//==! Project version history
//==!=====================================================================!==
//==! 2003/06/30 ver 01.00: Initial version gets boundary scan, process,
//==!                       flow, and action information from an input BSDL
//==!                       file.
//==! 2003/07/30 ver 04.00: Jumped to 4.00 to signify new code version from
//==!                       current 3.00.
//==! 2004/03/31 ver 04.01: Added the following features:
//==!                       - concurrent flow processing
//==!                         = default is safe concurrent processing where
//==!                           non-identical flow types are processed
//==!                           separately and where the other flow is placed
//==!                           in bypass mode
//==!                         = max_concurrent flow processing where flows
//==!                           that do not specify an input or output data
//==!                           shift will have the last TDI data shifted in
//==!                           and/or its TDO data shifted out
//==!                       - BSDL and ISC data files are read into memory
//==!                         for faster processing on embedded systems
//==!                         = files read into memory are also shared by
//==!                           devices that reference the same filename
//==!                       - data CRC and file transmission CRCs are
//==!                         calculated and compared against the ones
//==!                         specified in the files (if specified)
//==!                       - output operations (using flow output operator)
//==!                         can be done using buffered writes
//==!                         = data can be written using ISC data format
//==!                         = data generated from a flow specification must
//==!                           match any flow that will use this data
//==!                       - devices specified in the JTAG chain can be
//==!                         skipped without removing them from the chain
//==!                         description (skip_dev command file option)
//==! 2004/05/26 ver 04.02: Added code to CBsdlData::readBsdlAttributes() to:
//==!                       - print out DESIGN_WARNINGs and
//==!                         ISC_DESIGN_WARNINGs if "-bsdlwarnings" command
//==!                         line option specified
//==!                       - check specified ISC_PIN_BEHAVIOR opcode
//==!                       - check opcode(s) for ISC_ILLEGAL_EXIT
//==!                       Added code to processAction() and
//==!                       processActionsConcurrently() to call the
//==!                       proc_error_exit procedure for a device/devices if
//==!                       an error occurred while processing the action for
//==!                       the device.
//==! 2004/09/22 ver 04.03: Added enumerated port type ePortBuffer for buffer
//==!                       port type used in new function checkPortType().
//==!                       Assigned enPortType to all elements in a port
//==!                       bit_vector (see NOTE_0011). Added lBscCellNum and
//==!                       changed use of lBscCellType in CBscData class to
//==!                       handle cell type AC_n and BC_n (lBscCellType = 1
//==!                       and 0; respectively).
//==! 2004/10/04 ver 04.04: Added check in CBsdlData::readIscFlows() to check
//==!                       if a flow has a data name if it uses the '?' or
//==!                       '?!' operator to get input (see NOTE_0012).
//==!                       Added new command line option
//==!                       "-checkbsdl <BSDL file>"
//==!                       to check the BSDL file without having to first
//==!                       create a command file. This option will override
//==!                       all other options.
//==! 2004/12/08 ver 04.05: Added the following CBsdlData member functions:
//==!                       - CBsdlData::checkRegOpcodeAssociation() to
//==!                         check if the mandatory/optional JTAG opcodes
//==!                         are associated with the correct data registers
//==!                         (see NOTE_0013)
//==!                       - CBsdlData::check1149StdLevel() to record JTAG
//==!                         standard level; used to determine if PRELOAD is
//==!                         a mandatory instruction (only for std 2001 and
//==!                         later)
//==! 2005/02/10 ver 04.06: Used new Windows Parallel Port driver license
//==!                       string (see WDRV_LIC.H for details).
//==! 2005/06/06 ver 04.07: Added CMultFlow to allow multiple repeat sections
//==!                       in a flow and in the data file.
//==! 2005/10/10 ver 04.08: Added code to CBsdlData::readBsdlAttributes() to
//==!                       check INSTRUCTION_CAPTURE attribute length.
//==!=====================================================================!==
#define PROG_NAME_TM     "jdrive(tm)"  // program name with trademark
#define PROGRAM_NAME     "jdrive"      // program name
#define PROG_DESC        "Xilinx jDrive"
#define VERSION_NUM \
     "Version 4.08 Copyright (c) 2003-2005 Xilinx, Inc. All Rights Reserved"
#define PROG_VERSION     "4.08"
//==!=====================================================================!==
//==! Template function to free/deallocate memory used by a pointer.
//==!
//==! Function Parameters: a = pointer to an object
//==! Return Value       : none
//==!=====================================================================!==
template <class T> void freeData(T& a)
{

  if (a != NULL)
  { delete a;
    a = NULL;
  } // IF

} // freeData()
//==!=====================================================================!==
//==! Template function to free/deallocate memory used by a pointer.
//==!
//==! Function Parameters: a = pointer to an array of objects
//==! Return Value       : none
//==!=====================================================================!==
template <class T> void freeArray(T& a)
{

  if (a != NULL)
  { delete [] a;
    a = NULL;
  } // IF

} // freeArray()
//==!=====================================================================!==
//==! end of XPROJECT.H
//==!=====================================================================!==

