/***************************************************************************/
/* J Drive (tm) Engine                                                     */
/* Copyright 2003-2005, Xilinx, Inc.                                       */
/* The use, reproduction, and distribution of this code is subject         */
/* to the J Drive license requirements and restrictions.                   */
/***************************************************************************/
//==!=====================================================================!==
//==! See XPROJECT.H for project file list and project version history.
//==!=====================================================================!==
//==! Revision history: XRD_DATA.CPP
//==! 2003/09/02: 1532 ISC command and data file reader.
//==!=====================================================================!==
#include "xcommon.h"
#include "x_bsdl.h"
//==!=====================================================================!==
//==! Output binary data in string as hex chars.
//==!
//==! Function Parameters: sBinData = data to stderr and or output file
//==!                      lLen = length of binary data in sBinData
//==!                      iMode = 0: send to stderr only
//==!                              1: send to destination file also
//==!                              2: send to destination file only
//==!                      iDest = 0: output file
//==!                              1: log file
//==! Return Value       : none
//==!=====================================================================!==
void debugStrHex(char *sBinData,long lLen,int iMode,int iDest)
{ long lX,lY;
  unsigned short usVal;

  for (usVal = 0,lX = (lLen-1)%4,lY = 0; lY < lLen; ++lY)
  { if (sBinData[lY] != '0' && sBinData[lY] != '1')
    { sprintf(sMsgOnly,"debugStrHex [%s]",sBinData);
      jMessage(DISP_AND_EXIT,sMsgOnly,0,NULL);
    } // IF
    if (sBinData[lY] == '1')
    { usVal = usVal | (1 << lX);
    } // IF
    --lX;
    if (lX < 0)
    { if (iMode != 0)
      { if (iDest == 0)
        { if (pDP_Output != NULL)
          { if (pDP_Output->pFILE != NULL)
            { fprintf(pDP_Output->pFILE,"%X",usVal);
            } // IF
          } // IF
        } // IF
        else
        { if (pDP_Log != NULL)
          { if (pDP_Log->pFILE != NULL)
            { fprintf(pDP_Log->pFILE,"%X",usVal);
            } // IF
          } // IF
        } // ELSE
      } // IF iMode != 0
      if (iMode != 2)
      { fprintf(stderr,"%X",usVal);
      } // IF
      lX = 3;
      usVal = 0;
    } // IF lX < 0

  } // FOR lY < lLen

} // debugStrHex()
//==!=====================================================================!==
//==! Output debug messages.
//==!
//==! Function Parameters: sMsg = debug message to stderr and or output file
//==!                      iMode = 0: send to stderr only
//==!                              1: send to destination file also
//==!                              2: send to destination file only
//==!                      iDest = 0: output file
//==!                              1: log file
//==! Return Value       : none
//==!=====================================================================!==
void debugMsg(char *sMsg,int iMode,int iDest)
{

  if (iMode != 0)
  { if (iDest == 0)
    { if (pDP_Output != NULL)
      { if (pDP_Output->pFILE != NULL)
        { fprintf(pDP_Output->pFILE,"%s",sMsg);
        } // IF
      } // IF
    } // IF
    else
    { if (pDP_Log != NULL)
      { if (pDP_Log->pFILE != NULL)
        { fprintf(pDP_Log->pFILE,"%s",sMsg);
        } // IF
      } // IF
    } // ELSE
  } // IF
  if (iMode != 2)
  { fprintf(stderr,"%s",sMsg);
  } // IF

} // debugMsg()
//==!=====================================================================!==
//==! Date record and data file transmission CRC calculation functions as
//==! defined in IEEE STD 1532-2002:
//==!
//==! Data file transmission CRC functions:
//==! - init_transmission_crc()
//==! - compute_transmission_crc()
//==!
//==! Data record CRC functions:
//==! - init_crc()
//==! - crc_result()
//==! - compute_crc()
//==!=====================================================================!==
void init_crc()
{
  uiCRC_32register = 0;        // Clear CRC value
} // init_crc()

unsigned int crc_result()
{
  return uiCRC_32register;     // Return the CRC value
} // crc_result()

void compute_crc(unsigned int Bit)  // LSB has bit to accumulate
{
  Bit &= Low_Bit;                    // Assure only LSB present
  if (uiCRC_32register & High_Bit)   // if MSB set
  { uiCRC_32register = (uiCRC_32register << 1)^Polynomial^Bit;
  } // IF
  else
  { uiCRC_32register = (uiCRC_32register << 1)^Bit;
  } // ELSE
} // compute_crc()

void init_transmission_crc()
{
  usCRC_16register = 0xFFFF;    // start with all ones in shift register
} // init_transmission_crc()

void compute_transmission_crc(unsigned char in_byte)
{ int bit,feedback;

//==! 2003/11/04: Must skip LF (\n) char also. The 1532-2002 standard only
//==! skips the CR (\r) char; this is supposed to be corrected in the next
//==! revision.
  if (in_byte != '\r' && in_byte != '\n') // Skip carriage return and LF char
  { for (bit = 0; bit < CHAR_BIT; ++bit)  // Compute for each bit in byte
    { feedback = (in_byte ^ usCRC_16register) & 0x01;  // XOR LSB
      usCRC_16register >>= 1;      // shift the shift register
      if (feedback)
      { usCRC_16register ^= CCITT_CRC;  // invert selected bits
      } // IF
      in_byte >>= 1;              // get the next bit of in_byte
    } // FOR bit < CHAR_BIT
  } // IF in_byte != '\r'

} // compute_transmission_crc()
//==!=====================================================================!==
//==! CxCSetting constructor
//==!
//==! Function Parameters: none
//==! Return Value       : none
//==!=====================================================================!==
CxCSetting::CxCSetting()
{
  m_csProcName = NULL;
  m_csProcData = NULL;
  m_enProcOption = ePROC_NO_SETTING;
  m_pNextCxCSetting = NULL;

} // CxCSetting::CxCSetting()
//==!=====================================================================!==
//==! CxCSetting destructor
//==!
//==! Function Parameters: none
//==! Return Value       : none
//==!=====================================================================!==
CxCSetting::~CxCSetting()
{
  freeArray(m_csProcName);
  freeArray(m_csProcData);
  m_enProcOption = ePROC_NO_SETTING;
  m_pNextCxCSetting = NULL;

} // CxCSetting::~CxCSetting()
//==!=====================================================================!==
//==! CxCVarData constructor
//==!
//==! Function Parameters: none
//==! Return Value       : none
//==!=====================================================================!==
CxCVarData::CxCVarData()
{
  m_csActionName = NULL;
  m_csActionDataName = NULL;
  m_pCxCS_ProcSetting = NULL;
  m_pNextCxCVarData = NULL;

} // CxCVarData::CxCVarData()
//==!=====================================================================!==
//==! CxCVarData destructor
//==!
//==! Function Parameters: none
//==! Return Value       : none
//==!=====================================================================!==
CxCVarData::~CxCVarData()
{ CxCSetting *pCxCSetting;

  freeArray(m_csActionName);
  freeArray(m_csActionDataName);

  while (m_pCxCS_ProcSetting != NULL)
  { pCxCSetting = m_pCxCS_ProcSetting->m_pNextCxCSetting;
    m_pCxCS_ProcSetting->m_pNextCxCSetting = NULL;
    freeData(m_pCxCS_ProcSetting);
    m_pCxCS_ProcSetting = pCxCSetting;
  } // WHILE m_pCxCS_ProcSetting != NULL

  m_pNextCxCVarData = NULL;

} // CxCVarData::~CxCVarData()
//==!=====================================================================!==
//==! CIscData constructor
//==!
//==! Function Parameters: cIscDataFileName = ISC data filename
//==! Return Value       : none
//==!=====================================================================!==
CIscData::CIscData(char *cIscDataFileName)
{
  m_csStdVersion = NULL;
  m_csCreationDate = NULL;
  m_csCreatorString = NULL;
  m_lBracketLevel = 0;

  m_pCxCStr_EntityList = NULL;
  m_pCxCVD_OverrideList = NULL;
  m_pCDP_IscData = new cDataPort(cIscDataFileName);

} // CIscData::CIscData()
//==!=====================================================================!==
//==! CIscData destructor
//==!
//==! Function Parameters: none
//==! Return Value       : none
//==!=====================================================================!==
CIscData::~CIscData()
{ CxCStr *pCxCStr;
  CxCVarData *pCxCVD;

  freeArray(m_csStdVersion);
  freeArray(m_csCreationDate);
  freeArray(m_csCreatorString);
  m_lBracketLevel = 0;

  while (m_pCxCStr_EntityList != NULL)
  { pCxCStr = m_pCxCStr_EntityList->pNextCxCStr;
    m_pCxCStr_EntityList->pNextCxCStr = NULL;
    freeData(m_pCxCStr_EntityList);
    m_pCxCStr_EntityList = pCxCStr;
  } // WHILE m_pCxCStr_EntityList != NULL;

  while (m_pCxCVD_OverrideList != NULL)
  { pCxCVD = m_pCxCVD_OverrideList->m_pNextCxCVarData;
    m_pCxCVD_OverrideList->m_pNextCxCVarData = NULL;
    freeData(m_pCxCVD_OverrideList);
    m_pCxCVD_OverrideList = pCxCVD;
  } // WHILE m_pCxCVD_OverrideList != NULL;

  freeData(m_pCDP_IscData);

} // CIscData::~CIscData()
//==!=====================================================================!==
//==! Test if cDateString follows the yyyymmdd.hhmm creation date format.
//==!
//==! Function Parameters: cDateString = creation date char string
//==! Return Value       : none
//==!=====================================================================!==
void CIscData::testCreationDate(char *cDateString)
{ char cTmp[20];
  long lVar0,lVar1,lMaxDay,lYear;

  if (strlen(cDateString) != 13)
  { sprintf(sMsgOnly,"[%s]: invalid date (must use YYYYMMDD.HHMM format)",
            cDateString);
    jMessage(DISP_SHOW_MSG,sMsgOnly,0,NULL);
    return;
  } // IF

  strcpy(cTmp,cDateString);
  cTmp[4] = '\0';
  lYear = atol(cTmp);

  strcpy(cTmp,&(cDateString[4]));
  cTmp[2] = '\0';
  lVar0 = atol(cTmp);
  sprintf(sMsgBuf01,"yyyyMMdd.hhmm = %s: invalid month %ld",
          cDateString,lVar0);
  isInRange(1,lVar0,1,12,sMsgBuf01,DISP_SHOW_MSG);

  strcpy(cTmp,&(cDateString[6]));
  cTmp[2] = '\0';
  lVar1 = atol(cTmp);
  switch(lVar0)
  { case 2:         // if February
      lMaxDay = 28;
      if (lYear%4 == 0 || lYear%100 == 0)  // Check for leap year
      { lMaxDay = 29;
      } // IF
      break;
    case 4:
    case 6:
    case 9:
    case 11:
      lMaxDay = 30;
      break;
    default:
      lMaxDay = 31;
      break;
  } // SWITCH (lVar0)
  lMaxDay = 31;
  sprintf(sMsgBuf01,"yyyymmDD.hhmm = %s: invalid day for month %ld",
          cDateString,lVar0);
  isInRange(1,lVar1,1,lMaxDay,sMsgBuf01,DISP_SHOW_MSG);

  strcpy(cTmp,&(cDateString[9]));
  cTmp[2] = '\0';
  lVar0 = atol(cTmp);
  sprintf(sMsgBuf01,"yyyymmdd.HHmm = %s: invalid hour value %ld",
            cDateString,lVar0);
  isInRange(1,lVar0,0,23,sMsgBuf01,DISP_SHOW_MSG);

  strcpy(cTmp,&(cDateString[11]));
  cTmp[2] = '\0';
  lVar0 = atol(cTmp);
  sprintf(sMsgBuf01,"yyyymmdd.hhMM = %s: invalid minute value %ld",
          cDateString,lVar0);
  isInRange(1,lVar0,0,59,sMsgBuf01,DISP_SHOW_MSG);

} // CIscData::testCreationDate()
//==!=====================================================================!==
//==! Convert string to one of the enumerated proc override settings.
//==!
//==! Function Parameters: sSet = proc setting
//==! Return Value       : enumerated proc override setting
//==!=====================================================================!==
EnumProcSetting CIscData::cnvProcSet(char *sSetting)
{

  if (!strcmp(m_pCDP_IscData->sData,"on"))
  { return ePROC_ON;
  } // IF
  if (!strcmp(m_pCDP_IscData->sData,"off"))
  { return ePROC_OFF;
  } // IF
  if (!strcmp(m_pCDP_IscData->sData,"optional"))
  { return ePROC_OPTIONAL;
  } // IF
  if (!strcmp(m_pCDP_IscData->sData,"recommended"))
  { return ePROC_RECOMMENDED;
  } // IF
  jMessage(DISP_AND_EXIT,
          "use ON, OFF, OPTIONAL, or RECOMMENDED",1,m_pCDP_IscData);
  return ePROC_NO_SETTING;

} // CIscData::cnvProcSet()
//==!=====================================================================!==
//==! Check if array with name csDataArrayName is in the m_pCMIDA_list of the
//==! CMemIscDataSrc object. Return pointer to the array object.
//==!
//==! Function Parameters: pCMIDS_param = pointer to data src object
//==!                      csDataArrayName_param = array name to look for
//==! Return Value       : data array pointer
//==!=====================================================================!==
CMemIscDataArray *findCMemIscDataArray(CMemIscDataSrc *pCMIDS_param,
        char *csDataArrayName_param)
{ CMemIscDataArray *pCMIDAvar;

  if (pCMIDS_param == NULL)   // this should never happen
  { return NULL;
  } // IF
  for (pCMIDAvar = pCMIDS_param->m_pCMIDA_list; pCMIDAvar != NULL;
       pCMIDAvar = pCMIDAvar->m_pNextCMemIscDataArray)
  { if (!strcmp(pCMIDAvar->m_csDataArrayName,csDataArrayName_param))
    { break;
    } // IF
  } // FOR pCMIDAvar != NULL
  return pCMIDAvar;

} // findCMemIscDataArray()
//==!=====================================================================!==
//==! Find pointer to CMemIscDataSrc object with same filename as pIscData
//==! in the global list pgCMIDS_list. Can return NULL if not found.
//==!
//==! Function Parameters: pCDP_IscData = pointer to ISC data port
//==! Return Value       : CMemIscDataSrc pointer if NULL, else NULL
//==!=====================================================================!==
CMemIscDataSrc *findCMemIscDataSrc(cDataPort *pIscData)
{ CMemIscDataSrc *pCMIDSvar;

  for (pCMIDSvar = pgCMIDS_list; pCMIDSvar != NULL;
       pCMIDSvar = pCMIDSvar->m_pNextCMemIscDataSrc)
  { if (!strcmp(pCMIDSvar->m_csDataSrcName,pIscData->psPortName))
    { break;
    } // IF
  } // FOR pCMIDSvar != NULL
  return pCMIDSvar;

} // findCMemIscDataSrc()
//==!=====================================================================!==
//==! Calculate the CRC for this row. This result can be either the final CRC
//==! or an intermediate CRC which can then be added to when this function is
//==! called again.
//==!
//==! Function Parameters: pCDP_IscData = pointer to ISC data port
//==! Return Value       : none
//==! - side effect      : will update the global uiCRC_32register variable
//==!=====================================================================!==
void calcDataCrc(CMemIscDataRow *pDataRow)
{ CMemIscDataRowSection *pCMIDRSvar;
  unsigned int uiHexVal;
  long lCnt;
  int iX;

  if (pDataRow == NULL)
  { return;
  } // IF

  for (pCMIDRSvar = pDataRow->m_pCMIDRS_Head; pCMIDRSvar != NULL;
       pCMIDRSvar = pCMIDRSvar->m_pNextCMIDRS_MSB)
  { for (lCnt = pCMIDRSvar->m_lDataCnt - 1; lCnt >= 0; --lCnt)
    { uiHexVal = chrtohex(pCMIDRSvar->m_csHexData[lCnt]);

//==! Set to 1 to display data being sent to CRC computation routine
#if 0
      if (lGlob_IscCounter%32 == 0)
      { fprintf(pDP_Output->pFILE,"\n");
        lGlob_IscCounter = 0;
      } // IF
      fprintf(pDP_Output->pFILE,"%c",pCMIDRSvar->m_csHexData[lCnt]);
      ++lGlob_IscCounter;
#endif

      for (iX = 0; iX < 4; ++iX)
      { compute_crc(uiHexVal >> iX);
      } // FOR iX < 4
    } // FOR lCnt >= 0
  } // FOR pCMIDRSvar != NULL

} // calcDataCrc()
//==!=====================================================================!==
//==! Check format of ISC data record. If iMode = 0, then read the data array
//==! name and confirm that the DATA_CRC of the data record(s) is correct,
//==! else just check to see that the data record contains sections for
//==! INITIALIZE, REPEAT, TERMINATE, and DATA_CRC, and then look for the
//==! closing brace.
//==!
//==! Function Parameters: iMode = see function description
//==!                      csDataArray = data array name pointer
//==!                      p_pCMIDS = pointer to CMemIscDataSrc pointer to
//==!                              return ISC data information (useful
//==!                              only when iGlobOptUseMemIscData = 1)
//==! Return Value       : 1 if successful 0 otherwise
//==!=====================================================================!==
int CIscData::checkDataRecord(int iMode,char *csDataArray,
                 CMemIscDataSrc **p_pCMIDS)
{ int iX,iRptSect;
  CMultDataSection *pCMDSvar;
  CMemIscDataSrc *pCMIDSvar;
  CMemIscDataArray *pCMIDAvar;
  CMemIscDataRow **p_IscDataRow,*p_IscDataRowTail;

  if (iGlobOptUseMemIscData == 1)
  { pCMIDSvar = findCMemIscDataSrc(m_pCDP_IscData);
    if (pCMIDSvar == NULL)
    { pCMIDSvar = new CMemIscDataSrc;
      copyStr(&(pCMIDSvar->m_csDataSrcName),m_pCDP_IscData->psPortName);
      pCMIDSvar->m_pNextCMemIscDataSrc = pgCMIDS_list;
      pgCMIDS_list = pCMIDSvar;
      *p_pCMIDS = pCMIDSvar;
    } // IF
    else
    { *p_pCMIDS = pCMIDSvar;
    } // ELSE
  } // IF iGlobOptUseMemIscData == 1

  if (iMode == 0)
  { xGetData(m_pCDP_IscData,0,eATTR_VAR,"data array name expected");
    strcpy(sMsgBuf00,m_pCDP_IscData->sData);
    init_crc();

    if (iGlobOptUseMemIscData == 1)
    { pCMIDAvar = findCMemIscDataArray(pCMIDSvar,sMsgBuf00);
      if (pCMIDAvar != NULL)
      { sprintf(sMsgOnly,
                "Data array (%s) already defined in ISC file [%s]",
                sMsgBuf00,pCMIDSvar->m_csDataSrcName);
        jMessage(DISP_AND_EXIT,sMsgOnly,1,m_pCDP_IscData);
        return 0;
      } // IF
      pCMIDAvar = new CMemIscDataArray;
      copyStr(&(pCMIDAvar->m_csDataArrayName),sMsgBuf00);
      pCMIDAvar->m_pNextCMemIscDataArray = pCMIDSvar->m_pCMIDA_list;
      pCMIDSvar->m_pCMIDA_list = pCMIDAvar;
    } // IF iGlobOptUseMemIscData == 1
  } // IF
  else
  { strcpy(sMsgBuf00,csDataArray);
  } // ELSE

  findStr(m_pCDP_IscData,"(",0);
  ++m_lBracketLevel;
  while (1)
  { xParseData(m_pCDP_IscData,XDATA_READ);
    if (m_pCDP_IscData->sData[0] == ')')
    { --m_lBracketLevel;
      return 1;
    } // IF
    iX = 0;  // Set iX to 1 if DATA_CRC keyword found
    iRptSect = 0;
    p_IscDataRow = NULL;

    if (strcmp(m_pCDP_IscData->sData,"initialize") &&
        strcmp(m_pCDP_IscData->sData,"repeat"    ) &&
        strcmp(m_pCDP_IscData->sData,"terminate" ) &&
        strcmp(m_pCDP_IscData->sData,"data_crc"  ))
    { sprintf(sMsgOnly,
              "INITIALIZE/REPEAT/TERMINATE/DATA_CRC for data(%s) expected",
              sMsgBuf00);
      jMessage(DISP_AND_EXIT,sMsgOnly,1,m_pCDP_IscData);
      return 0;
    } // IF
    else
    { if (iMode == 0 && iGlobOptUseMemIscData == 1)
      { if (!strcmp(m_pCDP_IscData->sData,"initialize"))
        { p_IscDataRow = &(pCMIDAvar->m_pCMIDR_Init);
        } // IF
        else
        { if (!strcmp(m_pCDP_IscData->sData,"repeat"))
          { iRptSect = 1;
          } // IF
          else
          { if (!strcmp(m_pCDP_IscData->sData,"terminate"))
            { p_IscDataRow = &(pCMIDAvar->m_pCMIDR_Term);
            } //IF
          } // ELSE
        } // ELSE
        if (strcmp(m_pCDP_IscData->sData,"data_crc") && (iRptSect == 0))
        { if (*p_IscDataRow != NULL)
          { sprintf(sMsgOnly,"Data array (%s) %s section already defined",
                    pCMIDAvar->m_csDataArrayName,m_pCDP_IscData->sData);
            jMessage(DISP_AND_EXIT,sMsgOnly,1,m_pCDP_IscData);
          } // IF
        } // IF strcmp(m_pCDP_IscData->sData,"data_crc")
      } // IF iMode == 0 && iGlobOptUseMemIscData == 1
    } // ELSE

    if (!strcmp(m_pCDP_IscData->sData,"data_crc"))
    { iX = 1;
    } // IF

    findStr(m_pCDP_IscData,"(",0);

    if (iMode == 0)
    { if (iX == 1)
      { xParseData(m_pCDP_IscData,XDATA_READ);
        sprintf(sMsgBuf01,"%s",m_pCDP_IscData->sData);
        sprintf(sMsgBuf00,"%08x",crc_result());

//==! Display only if ISC data was loaded into memory because this lets
//==! program read ISC data in the proper order (i.e., last hex char to
//==! first). If reading from the file, we would have to store everything
//==! into a temporary memory array and then calculate the data CRC.
        if (iGlobOptUseMemIscData == 1)
        { sprintf(sMsgOnly,"    = Data CRC [%s]: ",sMsgBuf01);
          if (strcmp(sMsgBuf00,sMsgBuf01))
          { sprintf(sMsgBuf02,"mismatch (computed [%s])",sMsgBuf00);
            strcat(sMsgOnly,sMsgBuf02);
          } // IF
          else
          { strcat(sMsgOnly,"OK");
          } // ELSE
          dispMsg(sMsgOnly,DISP_USE_CUR_POS | DISP_POST_LF,1,1);
        } // IF iGlobOptUseMemIscData == 1
        findStr(m_pCDP_IscData,")",0);
      } // IF iX == 1
      else
      { if (iGlobOptUseMemIscData == 1)
        { if (iRptSect == 1)
          { if (pCMIDAvar->m_pCMultDataSect_Rpt == NULL)
            { pCMIDAvar->m_pCMultDataSect_Rpt = new CMultDataSection;
              pCMDSvar = pCMIDAvar->m_pCMultDataSect_Rpt;
            } // IF
            else
            { for (pCMDSvar = pCMIDAvar->m_pCMultDataSect_Rpt;
                   pCMDSvar->m_pNextCMultDataSection != NULL;
                   pCMDSvar = pCMDSvar->m_pNextCMultDataSection)
              { ;
              } // FOR pCMDSvar->m_pNextCMultDataSection != NULL
              pCMDSvar->m_pNextCMultDataSection = new CMultDataSection;
              pCMDSvar = pCMDSvar->m_pNextCMultDataSection;
            } // ELSE
            p_IscDataRow = &(pCMDSvar->m_pCMIDR_DataRow);
          } // IF iRptSect == 1
          *p_IscDataRow = new CMemIscDataRow;
          p_IscDataRowTail = *p_IscDataRow;
        } // IF
        do
        { getDpChar(m_pCDP_IscData,1);
          if (m_pCDP_IscData->iCurChar == ')')
          { if (iGlobOptUseMemIscData == 1)
            { p_IscDataRowTail->optimizeSpace();
              calcDataCrc(p_IscDataRowTail);
            } // IF
            break;
          } // IF
          if (!isxdigit(m_pCDP_IscData->iCurChar))
          { if (m_pCDP_IscData->iCurChar == ',')
            { if (iGlobOptUseMemIscData == 1)
              { p_IscDataRowTail->optimizeSpace();
                calcDataCrc(p_IscDataRowTail);
                p_IscDataRowTail->m_pNextCMemIscDataRow = new CMemIscDataRow;
                p_IscDataRowTail = p_IscDataRowTail->m_pNextCMemIscDataRow;
              } // IF
              continue;
            } // IF
            sprintf(sMsgOnly,"[%c]: data record must have hex digits only",
                    m_pCDP_IscData->iCurChar);
            jMessage(DISP_AND_EXIT,sMsgOnly,1,m_pCDP_IscData);
          } // IF

          if (iGlobOptUseMemIscData == 1)
          { p_IscDataRowTail->addHexChar(m_pCDP_IscData->iCurChar);
          } // IF iGlobOptUseMemIscData == 1
        } while (1);
      } // ELSE
    } // IF iMode == 0
    else
    { findChar(m_pCDP_IscData,')',1);  //==! Find matching brace/bracket
    } // ELSE

  } // WHILE (1)

  return 0;

} // CIscData::checkDataRecord()
//==!=====================================================================!==
//==! Convert hex string to an unsigned long value.
//==!
//==! Function Parameters: sHexStr = hex value in ASCII text format
//==!                      pDP = data port pointer
//==! Return Value       : hex value as unsigned long
//==!=====================================================================!==
unsigned long cnvHexToULong(char *sHexStr,cDataPort *pDP)
{ long lCnt;
  unsigned long ulVal;

  for (lCnt = 0,ulVal = 0; lCnt < (long)strlen(sHexStr); ++lCnt)
  { if (!isxdigit(sHexStr[lCnt]))
    { jMessage(DISP_AND_EXIT,"CRC must have hex chars only",1,pDP);
    } // IF
    ulVal = (ulVal << 4) + chrtohex(sHexStr[lCnt]);
  } // FOR lCnt < strlen(sHexStr)
  return ulVal;

} // cnvHexToULong()
//==!=====================================================================!==
//==! Calculate CRC for ISC data port. Compute CRC for lMaxByteCount bytes.
//==!
//==! Function Parameters: pDP = data port pointer
//==!                      lMaxByteCount = max number of bytes for CRC
//==!                          computation
//==! Return Value       : 16-bit unsigned short
//==!=====================================================================!==
unsigned short calcIscCRC(cDataPort *pDP,long lMaxByteCount)
{ long lCnt;

  pDP->xPortReopen("r");

  init_transmission_crc();

  for (lCnt = 0; lCnt < lMaxByteCount; ++lCnt)
  { getDpChar(pDP,0);
    compute_transmission_crc((unsigned char)pDP->iCurChar);
  } // lCnt < lMaxByteCount

  return ~usCRC_16register;

} // calcIscCRC()
//==!=====================================================================!==
//==! Overloaded function: calculate CRC for ISC data port.
//==!
//==! Function Parameters: pDP = data port pointer
//==! Return Value       : 16-bit unsigned short
//==!=====================================================================!==
unsigned short calcIscCRC(cDataPort *pDP)
{

  pDP->xPortReopen("r");

  init_transmission_crc();

  do
  { getDpChar(pDP,0);
    if (pDP->m_iEndOfDataStream == 0)
    { compute_transmission_crc((unsigned char)pDP->iCurChar);
    } // IF
  } while (pDP->m_iEndOfDataStream == 0);

  return ~usCRC_16register;

} // calcIscCRC()
//==!=====================================================================!==
//==! Process the ISC override list. If iMode = 1, then the override info is
//==! compared with the action+procedure list in the pCBDvar object.
//==!
//==! Function Parameters: iMode = 1: store override info into override
//==!                                 setting list
//==!                      pCBDvar = pointer to BSDL object
//==!                              0: syntax check only
//==! Return Value       : none
//==!=====================================================================!==
void CIscData::processOverrideList(int iMode,CBsdlData *pCBDvar)
{

  xParseData(m_pCDP_IscData,XDATA_READ);
  if (strcmp(m_pCDP_IscData->sData,"override"))
  { m_pCDP_IscData->iParserGetData = 0;
    return;
  } // IF

  while (!strcmp(m_pCDP_IscData->sData,"override"))
  { xGetData(m_pCDP_IscData,0,eATTR_VAR,"procedure name expected");
    strcpy(sMsgBuf00,m_pCDP_IscData->sData);
    xParseData(m_pCDP_IscData,XDATA_READ);

    sMsgBuf01[0] = '\0';
//==! Check if procedure has a data name
    if (m_pCDP_IscData->sData[0] == '(')
    { xGetData(m_pCDP_IscData,0,eATTR_VAR,"procedure data name expected");
      strcpy(sMsgBuf01,m_pCDP_IscData->sData);
      findStr(m_pCDP_IscData,")",0);
      xParseData(m_pCDP_IscData,XDATA_READ);
    } // IF sData = '('

    if (m_pCDP_IscData->eDataAttr != eATTR_VAR)
    { jMessage(DISP_AND_EXIT,"action name expected",1,m_pCDP_IscData);
    } // IF
    strcpy(sMsgBuf02,m_pCDP_IscData->sData);

    sMsgBuf03[0] = '\0';
//==! Check if action has a data name
    xParseData(m_pCDP_IscData,XDATA_READ);
    if (m_pCDP_IscData->sData[0] == '(')
    { xGetData(m_pCDP_IscData,0,eATTR_VAR,"action data name expected");
      strcpy(sMsgBuf03,m_pCDP_IscData->sData);
      findStr(m_pCDP_IscData,")",0);
      xParseData(m_pCDP_IscData,XDATA_READ);
    } // IF sData = '('

    if (iMode == 1)
    { addSetting(sMsgBuf02,sMsgBuf03,sMsgBuf00,sMsgBuf01,
                 this->cnvProcSet(m_pCDP_IscData->sData),m_pCDP_IscData,
                 pCBDvar);
    } // IF

    xParseData(m_pCDP_IscData,XDATA_READ);
    if (m_pCDP_IscData->sData[0] == ',')
    { findStr(m_pCDP_IscData,"override",0);
    } // IF
    else
    { if (!strcmp(m_pCDP_IscData->sData,"override"))
      { jMessage(DISP_AND_EXIT,
                 "Separate multiple overrides with ','",1,m_pCDP_IscData);
      } // IF
      m_pCDP_IscData->iParserGetData = 0;
      break;
    } // ELSE
  } // WHILE sData = override

} // CIscData::processOverrideList()
//==!=====================================================================!==
//==! Find data subheader string (e.g., initialize, repeat, terminate) of
//==! data array csDataArray in ISC data file.
//==!
//==! 2005/06/21: Added lMultDataHeaderIndex so function can pick which
//==! data array to point to if there are multiple data arrays of the same
//==! type. Specifically added to support multiple REPEAT arrays which are
//==! allowed in the 1532 standard specification.
//==!
//==! Function Parameters: csDataArray = data array string
//==!                      csSubHeader = data sub header string
//==!                      lMultDataHeaderIndex = specifies which data header
//==!                         in a multi-data header section to point to (see
//==!                         function description)
//==!                      pCMemIscDataContext_obj = pointer to
//==!                           CMemIscDataContext object
//==! Return Value       : none
//==!=====================================================================!==
void CIscData::readIscDataSubHeader(char *csDataArray,char *csSubHeader,
                 long lMultDataHeaderIndex,
                 CMemIscDataContext *pCMemIscDataContext_obj)
{ long lX;
  CMultDataSection *pCMDSvar;

//==! Prepare error message
  sprintf(sMsgBuf00,"[%s] section of data array [%s] not found",
          csSubHeader,csDataArray);

  if (iGlobOptUseMemIscData == 1)
  { if (pCMemIscDataContext_obj->m_pCMIDArray == NULL)
    { jMessage(DISP_AND_EXIT,sMsgBuf00,1,m_pCDP_IscData);
    } // IF
    if (!xstrcmp(csSubHeader,"initialize"))
    { pCMemIscDataContext_obj->m_pCMIDR_current =
                 pCMemIscDataContext_obj->m_pCMIDArray->m_pCMIDR_Init;
      return;
    } // IF
    if (!xstrcmp(csSubHeader,"repeat"))
    { pCMDSvar = pCMemIscDataContext_obj->m_pCMIDArray->m_pCMultDataSect_Rpt;
      if (pCMDSvar == NULL)
      { pCMemIscDataContext_obj->m_pCMIDR_current = NULL;
      } // IF
      for (lX = 0; lX < lMultDataHeaderIndex; ++lX)
      { if (pCMDSvar == NULL)
        { pCMemIscDataContext_obj->m_pCMIDR_current = NULL;
          break;
        } // IF
        pCMemIscDataContext_obj->m_pCMIDR_current =
                                            pCMDSvar->m_pCMIDR_DataRow;
        pCMDSvar = pCMDSvar->m_pNextCMultDataSection;
      } // FOR lX < lMultDataHeaderIndex
      return;
    } // IF
    if (!xstrcmp(csSubHeader,"terminate"))
    { pCMemIscDataContext_obj->m_pCMIDR_current =
                 pCMemIscDataContext_obj->m_pCMIDArray->m_pCMIDR_Term;
      return;
    } // IF
    jMessage(DISP_AND_EXIT,"rdIDSH:InitRptTerm",1,m_pCDP_IscData);
  } // IF iGlobOptUseMemIscData == 1

  while (1)
  { xParseData(m_pCDP_IscData,XDATA_READ);
    if (m_pCDP_IscData->sData[0] == ')')
    { jMessage(DISP_AND_EXIT,sMsgBuf00,1,m_pCDP_IscData);
      --m_lBracketLevel;
      return;
    } // IF
    if (!strcmp(csSubHeader,m_pCDP_IscData->sData))
    { return;
    } // IF
    if (strcmp(m_pCDP_IscData->sData,"initialize") &&
        strcmp(m_pCDP_IscData->sData,"repeat"    ) &&
        strcmp(m_pCDP_IscData->sData,"terminate" ) &&
        strcmp(m_pCDP_IscData->sData,"data_crc"  ))
    { sprintf(sMsgOnly,
              "INITIALIZE/REPEAT/TERMINATE/DATA_CRC for data(%s) expected",
              csDataArray);
      jMessage(DISP_AND_EXIT,sMsgOnly,1,m_pCDP_IscData);
      return;
    } // IF
    findStr(m_pCDP_IscData,"(",0);
    findChar(m_pCDP_IscData,')',1);  //==! Find matching brace/bracket
  } // WHILE (1)

} // CIscData::readIscDataSubHeader()
//==!=====================================================================!==
//==! Copy the ISC data info from a device in the global CDeviceData list
//==! that shares the same m_pCDP_IscData object. Copy the CIscData info of
//==! a device that uses the same m_pCDP_IscData object with non-NULL data
//==! for its CIscData member variables. Note that it is possible for a
//==! device to have all NULL CIscData member variables (except
//==! m_pCDP_IscData), in which case there is nothing to copy.
//==!
//==! Function Parameters: pCDD = pointer to global CDeviceData list
//==! Return Value       : none
//==!=====================================================================!==
void CIscData::copyCIscDataInfo(CDeviceData *pCDD)
{ CDeviceData *pCDDvar;
  CxCStr *pCxCStr_var,*pCxCStr_obj,*pCxCStr_tail;
  CxCVarData *pCxCVD_var,*pCxCVD_obj,*pCxCVD_tail;
  CxCSetting *pCxCSetting_var,*pCxCSetting_obj,*pCxCSetting_tail;

  for (pCDDvar = pCDD; pCDDvar != NULL; pCDDvar = pCDDvar->m_pNextCDeviceData)
  { if (pCDDvar->m_pCIscData == NULL)
    { continue;
    } // IF
    if (pCDDvar->m_pCIscData->m_pCDP_IscData == NULL)
    { continue;
    } // IF
    if (!strcmp(pCDDvar->m_pCIscData->m_pCDP_IscData->psPortName,
                m_pCDP_IscData->psPortName))
    { if (pCDDvar->m_pCIscData->m_csStdVersion        != NULL ||
          pCDDvar->m_pCIscData->m_csCreationDate      != NULL ||
          pCDDvar->m_pCIscData->m_csCreatorString     != NULL ||
          pCDDvar->m_pCIscData->m_pCxCStr_EntityList  != NULL ||
          pCDDvar->m_pCIscData->m_pCxCVD_OverrideList != NULL)
      { break;
      } // IF
    } // IF
  } // FOR pCDDvar != NULL

  if (pCDDvar == NULL)  // ISC data file should have been read previously
  { sprintf(sMsgBuf00,"copyCIscData[%s]",m_pCDP_IscData->psPortName);
    jMessage(DISP_AND_EXIT,sMsgBuf00,1,NULL);
  } // IF

  if (pCDDvar->m_pCIscData->m_csStdVersion != NULL)
  { copyStr(&(m_csStdVersion),pCDDvar->m_pCIscData->m_csStdVersion);
  } // IF
  if (pCDDvar->m_pCIscData->m_csCreationDate != NULL)
  { copyStr(&(m_csCreationDate),pCDDvar->m_pCIscData->m_csCreationDate);
  } // IF
  if (pCDDvar->m_pCIscData->m_csCreatorString != NULL)
  { copyStr(&(m_csCreatorString),pCDDvar->m_pCIscData->m_csCreatorString);
  } // IF

  for (pCxCStr_var = pCDDvar->m_pCIscData->m_pCxCStr_EntityList;
       pCxCStr_var != NULL;
       pCxCStr_var = pCxCStr_var->pNextCxCStr)
  { pCxCStr_obj = new CxCStr;
    copyStr(&(pCxCStr_obj->pStr),pCxCStr_var->pStr);
    if (m_pCxCStr_EntityList == NULL)
    { m_pCxCStr_EntityList = pCxCStr_obj;
    } // IF
    else
    { pCxCStr_tail->pNextCxCStr = pCxCStr_obj;
    } // IF
    pCxCStr_tail = pCxCStr_obj;
  } // FOR pCxCStr_var != NULL

  for (pCxCVD_var = pCDDvar->m_pCIscData->m_pCxCVD_OverrideList;
       pCxCVD_var != NULL;
       pCxCVD_var = pCxCVD_var->m_pNextCxCVarData)
  { pCxCVD_obj = new CxCVarData;
    copyStr(&(pCxCVD_obj->m_csActionName),pCxCVD_var->m_csActionName);
    copyStr(&(pCxCVD_obj->m_csActionDataName),pCxCVD_var->m_csActionDataName);

    for (pCxCSetting_var = pCxCVD_var->m_pCxCS_ProcSetting;
         pCxCSetting_var != NULL;
         pCxCSetting_var = pCxCSetting_var->m_pNextCxCSetting)
    { pCxCSetting_obj = new CxCSetting;
      copyStr(&(pCxCSetting_obj->m_csProcName),pCxCSetting_var->m_csProcName);
      copyStr(&(pCxCSetting_obj->m_csProcData),pCxCSetting_var->m_csProcData);
      pCxCSetting_obj->m_enProcOption = pCxCSetting_var->m_enProcOption;
      if (pCxCVD_obj->m_pCxCS_ProcSetting == NULL)
      { pCxCVD_obj->m_pCxCS_ProcSetting = pCxCSetting_obj;
      } // IF
      else
      { pCxCSetting_tail->m_pNextCxCSetting = pCxCSetting_obj;
      } // ELSE
      pCxCSetting_tail = pCxCSetting_obj;
    } // FOR pCxCSetting_var != NULL

    if (m_pCxCVD_OverrideList == NULL)
    { m_pCxCVD_OverrideList = pCxCVD_obj;
    } // IF
    else
    { pCxCVD_tail->m_pNextCxCVarData = pCxCVD_obj;
    } // IF
    pCxCVD_tail = pCxCVD_obj;
  } // FOR pCxCVD_var != NULL;

} // CIscData::copyCIscDataInfo()
//==!=====================================================================!==
//==! Read ISC data file header and store information into the appropriate
//==! member variables. Function assumes that data file has the following
//==! sections in this order: header + override (optional) + data
//==!
//==! Function Parameters: iMode = 0: read and store header and override info
//==!                                 only; check the data sections for
//==!                                 correct matching braces and CRC if
//==!                                 specified
//==!                              1: skip header info and search for
//==!                                 csDataArray data section
//==!                      pCDD = pointer to CDeviceData list
//==!                      pCBDvar = pointer to BSDL object
//==!                      csDataArray = data array name (i.e., name of data
//==!                              array to be used by a procedure)
//==!                      pCMemIscDataContext_obj = pointer to
//==!                              CMemIscDataContext object to hold
//==!                              information regarding ISC data
//==!                              information held in memory
//==!                         - when iMode = 0, a CMemIscDataSrc object is
//==!                           assigned to a pointer in the context object
//==!                         - when iMode = 1, check that p_pCMIDS already
//==!                           has a CMemIscDataSrc object assigned to a
//==!                           pointer in the context object
//==! Return Value       : 1 if successful 0 otherwise
//==!=====================================================================!==
int CIscData::readIscDataHeader(int iMode,CDeviceData *pCDD,
                  CBsdlData *pCBDvar,char *csDataArray,
                  CMemIscDataContext *pCMemIscDataContext_obj)
{ CxCStr *pCxCStr;
  unsigned short usCrcVal,usCalcCrc;
  CMemIscDataSrc *pCMIDSvar;
  long lDataPortPos;
  int iFlag;

//==! If there is no BSDL file, then the ISC data file should not be read
//==! because there will be no action+procedure+flow information to use the
//==! ISC data.
  if (pCBDvar == NULL)
  { return 1;
  } // IF

  iGlobProgramState = dPROGSTATE_READISCDATA;

//==! The following combinations should never occur:
//==! - if iMode == 0, then the pointer must be NULL (i.e., no
//==!   memory ISC data object assigned yet)
//==! - if iMode == 1, then the pointer must not be NULL (i.e.,
//==!   a memory ISC data object should already have been assigned)
  if (iGlobOptUseMemIscData == 1)
  { if (iMode == 0)
    { if (pCMemIscDataContext_obj->m_pCMIDS != NULL)
      { jMessage(DISP_AND_EXIT,"rdIDH:*p!=NULL",0,NULL);
      } // IF
      pCMIDSvar = findCMemIscDataSrc(m_pCDP_IscData);
      if (pCMIDSvar != NULL)
      { copyCIscDataInfo(pCDD);
        pCMemIscDataContext_obj->m_pCMIDS = pCMIDSvar;
        return 1;
      } // IF
    } // IF iMode == 0
    else
    { if (pCMemIscDataContext_obj->m_pCMIDS == NULL)
      { jMessage(DISP_AND_EXIT,"rdIDH:*p==NULL",0,NULL);
        return 0;
      } // IF
      else
      { pCMemIscDataContext_obj->m_pCMIDArray =
               findCMemIscDataArray(pCMemIscDataContext_obj->m_pCMIDS,
                                    csDataArray);
        if (pCMemIscDataContext_obj->m_pCMIDArray == NULL)
        { isPtrNull(pCMemIscDataContext_obj->m_pCMIDS,"cid_0:ridh");
          sprintf(sMsgBuf00,
                  "\t! => Warning: no data array [%s] in ISC file [%s]",
                  csDataArray,
                  pCMemIscDataContext_obj->m_pCMIDS->m_csDataSrcName);
          dispMsg(sMsgBuf00,DISP_PRE_LF,1,1);
          return 0;
        } // IF
        return 1;
      } // ELSE
    } // ELSE
  } // IF iGlobOptUseMemIscData == 1

  isPtrNull(m_pCDP_IscData,"rIDH");
  m_pCDP_IscData->xPortReopen("r");
  findStr(m_pCDP_IscData,"ieee_1532_data",0);
  findStr(m_pCDP_IscData,"(",0);
  ++m_lBracketLevel;
  findStr(m_pCDP_IscData,"header",0);
  findStr(m_pCDP_IscData,"(",0);
  ++m_lBracketLevel;

  if (iMode == 0)
  { sprintf(sMsgBuf00,"--> Reading ISC data file [%s]",
            m_pCDP_IscData->psPortName);
    jMessage(DISP_SHOW_MSG,sMsgBuf00,1,NULL);

    findStr(m_pCDP_IscData,"std_version",0);
    xGetData(m_pCDP_IscData,0,eATTR_VAR,"version expected");
    copyStr(&(m_csStdVersion),m_pCDP_IscData->sData);
    if (strcmp(m_pCDP_IscData->sData,"std_1532_2000") &&
        strcmp(m_pCDP_IscData->sData,"std_1532_2001"))
    { sprintf(sMsgBuf00,"Unsupported version [%s]",m_pCDP_IscData->sData);
      jMessage(DISP_SHOW_MSG,sMsgBuf00,1,NULL);
    } // IF

    for (xParseData(m_pCDP_IscData,XDATA_READ);
         m_pCDP_IscData->sData[0] != ')';
         xParseData(m_pCDP_IscData,XDATA_READ))
    { if (!strcmp(m_pCDP_IscData->sData,"creation_date"))
      { xGetData(m_pCDP_IscData,0,eATTR_NUM,"creation date expected");
        strcpy(sMsgBuf00,m_pCDP_IscData->sData);
        findStr(m_pCDP_IscData,".",0);
        xGetData(m_pCDP_IscData,0,eATTR_NUM,"hours:mins part expected");
        strcat(sMsgBuf00,".");
        strcat(sMsgBuf00,m_pCDP_IscData->sData);
        testCreationDate(sMsgBuf00);
        copyStr(&(m_csCreationDate),sMsgBuf00);
        continue;
      } // IF sData = creation_date

      if (!strcmp(m_pCDP_IscData->sData,"creator"))
      { findStr(m_pCDP_IscData,"\"",0);
        for (xParseData(m_pCDP_IscData,XDATA_READ_RAW);
             m_pCDP_IscData->sData[0] != '"';
             xParseData(m_pCDP_IscData,XDATA_READ_RAW))
        { xstrcat(&m_csCreatorString,m_pCDP_IscData->sData);
        } // FOR m_pCDP_IscData->sData[0] != '"'
        continue;
      } // IF sData = creator

      if (!strcmp(m_pCDP_IscData->sData,"entity"))
      { iFlag = 0;
        do
        { xGetData(m_pCDP_IscData,0,eATTR_VAR,"device entity name expected");
          pCxCStr = new CxCStr;
          copyStr(&(pCxCStr->pStr),m_pCDP_IscData->sData);
          if (!strcmp(m_pCDP_IscData->sData,pCBDvar->csEntityName))
          { iFlag = 1;
          } // IF
          pCxCStr->pNextCxCStr = m_pCxCStr_EntityList;
          m_pCxCStr_EntityList = pCxCStr;
          xParseData(m_pCDP_IscData,XDATA_READ);
        } while (m_pCDP_IscData->sData[0] == ',');
        m_pCDP_IscData->iParserGetData = 0;

        if (iFlag == 0)
        { sprintf(sMsgOnly,
                  "  = ISC datafile [%s] being used with BSDL entity [%s]",
                  m_pCDP_IscData->psPortName,pCBDvar->csEntityName);
          jMessage(DISP_SHOW_MSG,sMsgOnly,1,NULL);
        } // IF

        continue;
      } // IF sData = entity
    } // FOR m_pCDP_IscData->sData[0] != ')'
    --m_lBracketLevel;

    processOverrideList(1,pCBDvar);

//==! Check data record(s)
    for (iFlag = 0;;)
    { xParseData(m_pCDP_IscData,XDATA_READ);
      if (m_pCDP_IscData->sData[0] == ')')
      { --m_lBracketLevel;
        break;
      } // IF

      if (iFlag == 1)
      { if (m_pCDP_IscData->sData[0] != ',')
        { jMessage(DISP_AND_EXIT,"Use ',' to separate data records",
                   1,m_pCDP_IscData);
        } // IF
        else
        { xParseData(m_pCDP_IscData,XDATA_READ);
        } // ELSE
      } // IF
      iFlag = 1;

      if (!strcmp(m_pCDP_IscData->sData,"unprocessed"))
      { xParseData(m_pCDP_IscData,XDATA_READ);
      } // IF
      if (strcmp(m_pCDP_IscData->sData,"data"))
      { jMessage(DISP_AND_EXIT,"UNPROCESSED or DATA expected",1,m_pCDP_IscData);
      } // IF

      checkDataRecord(0,NULL,&(pCMemIscDataContext_obj->m_pCMIDS));
    } // FOR (iFlag = 0;;)

//==! Check for transmission CRC. Can use ftell() to find current position if
//==! data port is a file.
    lDataPortPos = m_pCDP_IscData->lCharCount;
    xParseData(m_pCDP_IscData,XDATA_READ);
    if (m_pCDP_IscData->m_iEndOfDataStream == 0 ||
        m_pCDP_IscData->eDataAttr != eATTR_WSPACE)
    { if (strlen(m_pCDP_IscData->sData) != 4)
      { jMessage(DISP_SHOW_MSG,
                 "CRC must have 4 hex digits",1,m_pCDP_IscData);
      } // IF
      usCrcVal = (unsigned short)cnvHexToULong(m_pCDP_IscData->sData,
                                               m_pCDP_IscData);
      usCalcCrc = calcIscCRC(m_pCDP_IscData,lDataPortPos);
      sprintf(sMsgOnly,"    = ISC transmission CRC [%04X]: ",usCrcVal);
      if (usCrcVal != usCalcCrc)
      { sprintf(sMsgBuf00," mismatch (computed [%04X])",usCalcCrc);
        strcat(sMsgOnly,sMsgBuf00);
      } // IF
      else
      { strcat(sMsgOnly,"OK");
      } // ELSE
      dispMsg(sMsgOnly,DISP_USE_CUR_POS | DISP_POST_LF,1,1);
    } // IF not end of file or data != white space

    xFclose(&(m_pCDP_IscData->pFILE));
    return 1;
  } // IF iMode == 0
  else  //==! Skip header section by looking for matching ')'
  { findStr(m_pCDP_IscData,")",1);
    --m_lBracketLevel;
    isPtrNull(csDataArray,"data array name required");
    processOverrideList(0,pCBDvar);
  } // ELSE

  do
  { xParseData(m_pCDP_IscData,XDATA_READ);
    if (!strcmp(m_pCDP_IscData->sData,"unprocessed"))
    { xParseData(m_pCDP_IscData,XDATA_READ);
    } // IF
    if (strcmp(m_pCDP_IscData->sData,"data"))
    { sprintf(sMsgOnly,"UNPROCESSED/DATA: array [%s] not found",csDataArray);
      jMessage(DISP_AND_EXIT,sMsgOnly,1,m_pCDP_IscData);
    } // IF
    xGetData(m_pCDP_IscData,0,eATTR_VAR,"data array name expected");

    if (!strcmp(m_pCDP_IscData->sData,csDataArray))
    { findStr(m_pCDP_IscData,"(",0);
      ++m_lBracketLevel;
      return 1;
    } // IF

    checkDataRecord(1,csDataArray,&(pCMemIscDataContext_obj->m_pCMIDS));

    xParseData(m_pCDP_IscData,XDATA_READ);
    if (m_pCDP_IscData->sData[0] != ',')
    { if (!strcmp(m_pCDP_IscData->sData,"unprocessed") ||
          !strcmp(m_pCDP_IscData->sData,"data"))
      { jMessage(DISP_AND_EXIT,"Use ',' to separate data records",
                 1,m_pCDP_IscData);
      } // IF
    } // IF
  } while (m_pCDP_IscData->sData[0] != ')');

  sprintf(sMsgOnly,"\t! => Warning: No data record [%s] in ISC file [%s]",
          csDataArray,m_pCDP_IscData->psPortName);
  dispMsg(sMsgOnly,DISP_PRE_LF,1,1);
  return 0;

} // CIscData::readIscDataHeader()
//==!=====================================================================!==
//==! Add the action-proc setting to the override list. First check if the
//==! action+procedure(proc data) name already exists, and if it has a
//==! different setting. If a new one, then check if the procedure setting
//==! can be changed. For example, something declared as optional or
//==! recommended can be changed to recommended, optional, on, or off, but
//==! if this procedure is mandatory (i.e., no specification), then this is
//==! an error.
//==!
//==! Function Parameters: csAction = action
//==!                      csActionData = optional action data
//==!                      csProc = procedure
//==!                      csProcData = optional procedure data
//==!                      eProcSetting = procedure setting
//==!                      pCBDvar = pointer to BSDL object
//==! Return Value       : none
//==!=====================================================================!==
void CIscData::addSetting(char *csAction,char *csActionData,
    char *csProc,char *csProcData,EnumProcSetting eProcSetting,
    cDataPort *pDP,CBsdlData *pCBDvar)
{ CxCVarData *pCxCVD;
  CxCSetting *pCxCSet;

//==! Check if this action+proc+(optional)procdata exists and whether the
//==! specified procsetting is valid (see function description).
  if (pCBDvar->verifyOverride(csAction,csActionData,csProc,csProcData,
                              eProcSetting,pDP) == 0)
  { return;
  } // IF

  for (pCxCVD = m_pCxCVD_OverrideList; pCxCVD != NULL;
       pCxCVD = pCxCVD->m_pNextCxCVarData)
  { if (strcmp(pCxCVD->m_csActionName,csAction))
    { continue;
    } // IF
    if (pCxCVD->m_csActionDataName != NULL && csActionData != NULL)
    { if (strcmp(pCxCVD->m_csActionDataName,csActionData))
      { continue;
      } // IF
    } // IF
    for (pCxCSet = pCxCVD->m_pCxCS_ProcSetting; pCxCSet != NULL;
         pCxCSet = pCxCSet->m_pNextCxCSetting)
    { if (!strcmp(pCxCSet->m_csProcName,csProc))
      { if ((csProcData == NULL && pCxCSet->m_csProcData == NULL) ||
            (csProcData != NULL && pCxCSet->m_csProcData != NULL))
        { if (csProcData != NULL)
          { if (strcmp(csProcData,pCxCSet->m_csProcData))
            { continue;
            } // IF
          } // IF proc data name exists
          if (pCxCSet->m_enProcOption != eProcSetting)
          { sprintf(sMsgOnly,"Action [%s] Proc [%s",csAction,csProc);
            if (csProcData != NULL)
            { sprintf(sMsgBuf00,"(%s)",csProcData);
              strcat(sMsgOnly,sMsgBuf00);
            } // IF
            strcat(sMsgOnly,"] override setting already defined");
            jMessage(DISP_SHOW_MSG,sMsgOnly,1,NULL);
          } // IF previous setting != current setting
          return;
        } // IF no proc data name or proc data name exists
      } // IF procname = csProcName
    } // FOR pCxCSet != NULL

//==! Add another procedure override object to this action
    pCxCSet = new CxCSetting;
    copyStr(&(pCxCSet->m_csProcName),csProc);
    if (csProcData != NULL)
    { copyStr(&(pCxCSet->m_csProcData),csProcData);
    } // IF
    pCxCSet->m_enProcOption = eProcSetting;
    pCxCSet->m_pNextCxCSetting = pCxCVD->m_pCxCS_ProcSetting;
    pCxCVD->m_pCxCS_ProcSetting = pCxCSet;
    return;
  } // FOR pCxCvd != NULL

//==! Add a new action+actiondata+proc+procdata object
  pCxCSet = new CxCSetting;
  copyStr(&(pCxCSet->m_csProcName),csProc);
  if (csProcData != NULL)
  { copyStr(&(pCxCSet->m_csProcData),csProcData);
  } // IF
  pCxCSet->m_enProcOption = eProcSetting;
  pCxCVD = new CxCVarData;
  copyStr(&(pCxCVD->m_csActionName),csAction);
  if (csActionData != NULL)
  { copyStr(&(pCxCVD->m_csActionDataName),csActionData);
  } // IF
  pCxCVD->m_pCxCS_ProcSetting = pCxCSet;
  pCxCVD->m_pNextCxCVarData = m_pCxCVD_OverrideList;
  m_pCxCVD_OverrideList = pCxCVD;
  return;

} // CIscData::addSetting()
//==!=====================================================================!==
//==! CMemIscDataRowSection constructor
//==!
//==! Function Parameters: iMode = 0: initialize pointer array to NULL
//==!                              1: allocate dCMIDRS_SIZE bytes to pointer
//==!                                 and initialize array elements to NULL
//==! Return Value       : none
//==!=====================================================================!==
CMemIscDataRowSection::CMemIscDataRowSection(int iMode)
{

  if (iMode == 0)
  { m_csHexData = NULL;
  } // IF
  else
  { m_csHexData = new char [dCMIDRS_SIZE];
    isPtrNull(m_csHexData,"Insufficient RAM to hold data file info in mem");
//==! No need to initialize because function will fill with data from data
//==! source and will keep track of max data in m_lDataCnt.
  } // ELSE

  m_lDataCnt = 0;
  m_pNextCMIDRS_LSB = NULL;
  m_pNextCMIDRS_MSB = NULL;

} // CMemIscDataRowSection::CMemIscDataRowSection()
//==!=====================================================================!==
//==! CMemIscDataRowSection destructor
//==!
//==! Function Parameters: none
//==! Return Value       : none
//==!=====================================================================!==
CMemIscDataRowSection::~CMemIscDataRowSection()
{
  freeArray(m_csHexData);
  m_lDataCnt = 0;
  m_pNextCMIDRS_LSB = NULL;
  m_pNextCMIDRS_MSB = NULL;

} // CMemIscDataRowSection::~CMemIscDataRowSection()
//==!=====================================================================!==
//==! CMemIscDataRow constructor
//==!
//==! Function Parameters: none
//==! Return Value       : none
//==!=====================================================================!==
CMemIscDataRow::CMemIscDataRow()
{

  m_lTotDataCnt = 0;
  m_pCMIDRS_Head = NULL;
  m_pCMIDRS_Tail = NULL;
  m_pNextCMemIscDataRow = NULL;

} // CMemIscDataRow::CMemIscDataRow()
//==!=====================================================================!==
//==! CMemIscDataRow destructor
//==!
//==! Function Parameters: none
//==! Return Value       : none
//==!=====================================================================!==
CMemIscDataRow::~CMemIscDataRow()
{ CMemIscDataRowSection *pCMIDRSvar;

  m_lTotDataCnt = 0;
  while (m_pCMIDRS_Head != NULL)
  { pCMIDRSvar = m_pCMIDRS_Head->m_pNextCMIDRS_MSB;
    freeData(m_pCMIDRS_Head);
    m_pCMIDRS_Head = pCMIDRSvar;
  } // WHILE m_pCMIDRS_Head != NULL

  m_pCMIDRS_Tail = NULL;
  m_pNextCMemIscDataRow = NULL;

} // CMemIscDataRow::~CMemIscDataRow()
//==!=====================================================================!==
//==! CMemIscDataRow::addHexChar(): add hex char to row. Add a row section if
//==! current row section buffer is full.
//==!
//==! Function Parameters: iChar = char to add
//==! Return Value       : none
//==!=====================================================================!==
void CMemIscDataRow::addHexChar(int iChar)
{ CMemIscDataRowSection *pCMIDRSvar;

  if (m_pCMIDRS_Head == NULL)
  { m_pCMIDRS_Head = new CMemIscDataRowSection(1);
    m_pCMIDRS_Tail = m_pCMIDRS_Head;
  } // IF m_pCMIDRS_Head == NULL

  if (m_pCMIDRS_Head->m_lDataCnt >= dCMIDRS_SIZE)
  { pCMIDRSvar = new CMemIscDataRowSection(1);
//==! Connect to head of doubly-linked list
    pCMIDRSvar->m_pNextCMIDRS_MSB = m_pCMIDRS_Head;
    m_pCMIDRS_Head->m_pNextCMIDRS_LSB = pCMIDRSvar;
    m_pCMIDRS_Head = pCMIDRSvar;
  } // IF m_pCMIDRS_Head->m_lDataCnt >= dCMIDRS_SIZE

  m_pCMIDRS_Head->m_csHexData[m_pCMIDRS_Head->m_lDataCnt] = iChar;

  ++m_lTotDataCnt;
  ++(m_pCMIDRS_Head->m_lDataCnt);

} // CMemIscDataRow::addHexChar()
//==!=====================================================================!==
//==! CMemIscDataRow::optimizeSpace(): optimize the space used by the row.
//==! Hex data for a row is stored in groups (called row sections) of size
//==! dCMIDRS_SIZE in a doubly linked list. For example, if dCMIDRS_SIZE = 20,
//==! then a 30 char row is stored in 2 row sections, with the last row
//==! section containing only 10 chars. This routine will resize the last row
//==! section's char array to exactly 10 chars (i.e., char array is not
//==! NULL-terminated).
//==!
//==! Function Parameters: none
//==! Return Value       : none
//==!=====================================================================!==
void CMemIscDataRow::optimizeSpace()
{ char *cBuf;
  long lCnt;

  if (m_pCMIDRS_Head == NULL)
  { return;
  } // IF m_pCMIDRS_Head == NULL

  if (m_pCMIDRS_Head->m_lDataCnt > 0 &&
      m_pCMIDRS_Head->m_lDataCnt < dCMIDRS_SIZE)
  { cBuf = new char [m_pCMIDRS_Head->m_lDataCnt];
    for (lCnt = 0; lCnt < m_pCMIDRS_Head->m_lDataCnt; ++lCnt)
    { cBuf[lCnt] = m_pCMIDRS_Head->m_csHexData[lCnt];
    } // FOR lCnt < m_pCMIDRS_Head->m_lDataCnt
    freeArray(m_pCMIDRS_Head->m_csHexData);
    m_pCMIDRS_Head->m_csHexData = cBuf;
  } // IF m_pCMIDRS_Head->m_lDataCnt > 0 && < dCMIDRS_SIZE

} // CMemIscDataRow::optimizeSpace()
//==!=====================================================================!==
//==! CMultDataSection constructor
//==!
//==! Function Parameters: none
//==! Return Value       : none
//==!=====================================================================!==
CMultDataSection::CMultDataSection()
{

  m_pCMIDR_DataRow = NULL;
  m_pNextCMultDataSection = NULL;

} // CMultDataSection::CMultDataSection()
//==!=====================================================================!==
//==! CMultDataSection desstructor
//==!
//==! Function Parameters: none
//==! Return Value       : none
//==!=====================================================================!==
CMultDataSection::~CMultDataSection()
{ CMemIscDataRow *pCMIDRvar;

  while (m_pCMIDR_DataRow != NULL)
  { pCMIDRvar = m_pCMIDR_DataRow->m_pNextCMemIscDataRow;
    freeData(m_pCMIDR_DataRow);
    m_pCMIDR_DataRow = pCMIDRvar;
  } // WHILE m_pCMIDR_DataRow != NULL

  m_pNextCMultDataSection = NULL;

} // CMultDataSection::~CMultDataSection()
//==!=====================================================================!==
//==! CMemIscDataArray constructor
//==!
//==! Function Parameters: none
//==! Return Value       : none
//==!=====================================================================!==
CMemIscDataArray::CMemIscDataArray()
{

  m_csDataArrayName = NULL;
  m_pCMIDR_Init = NULL;
  m_pCMIDR_Term = NULL;
  m_pCMultDataSect_Rpt = NULL;
  m_pNextCMemIscDataArray = NULL;

} // CMemIscDataArray::CMemIscDataArray()
//==!=====================================================================!==
//==! CMemIscDataArray destructor
//==!
//==! Function Parameters: none
//==! Return Value       : none
//==!=====================================================================!==
CMemIscDataArray::~CMemIscDataArray()
{ CMemIscDataRow *pCMIDRvar;
  CMultDataSection *pCMDSvar;

  freeArray(m_csDataArrayName);

  while (m_pCMIDR_Init != NULL)
  { pCMIDRvar = m_pCMIDR_Init->m_pNextCMemIscDataRow;
    freeData(m_pCMIDR_Init);
    m_pCMIDR_Init = pCMIDRvar;
  } // WHILE m_pCMIDR_Init != NULL

  while (m_pCMultDataSect_Rpt != NULL)
  { pCMDSvar = m_pCMultDataSect_Rpt->m_pNextCMultDataSection;
    freeData(m_pCMultDataSect_Rpt);
    m_pCMultDataSect_Rpt = pCMDSvar;
  } // WHILE m_pCMultDataSect_Rpt != NULL

  while (m_pCMIDR_Term != NULL)
  { pCMIDRvar = m_pCMIDR_Term->m_pNextCMemIscDataRow;
    freeData(m_pCMIDR_Term);
    m_pCMIDR_Term = pCMIDRvar;
  } // WHILE m_pCMIDR_Term != NULL

  m_pNextCMemIscDataArray = NULL;

} // CMemIscDataArray::~CMemIscDataArray()
//==!=====================================================================!==
//==! CMemIscDataSrc constructor
//==!
//==! Function Parameters: none
//==! Return Value       : none
//==!=====================================================================!==
CMemIscDataSrc::CMemIscDataSrc()
{
  m_csDataSrcName = NULL;
  m_pCMIDA_list = NULL;
  m_pNextCMemIscDataSrc = NULL;

} // CMemIscDataSrc::CMemIscDataSrc()
//==!=====================================================================!==
//==! CMemIscDataSrc destructor
//==!
//==! Function Parameters: none
//==! Return Value       : none
//==!=====================================================================!==
CMemIscDataSrc::~CMemIscDataSrc()
{ CMemIscDataArray *pCMIDAvar;

  freeArray(m_csDataSrcName);
  while (m_pCMIDA_list != NULL)
  { pCMIDAvar = m_pCMIDA_list->m_pNextCMemIscDataArray;
    freeData(m_pCMIDA_list);
    m_pCMIDA_list = pCMIDAvar;
  } // WHILE m_pCMIDA_list != NULL
  m_pNextCMemIscDataSrc = NULL;

} // CMemIscDataSrc::CMemIscDataSrc()
//==!=====================================================================!==
//==! CMemIscDataContext constructor
//==!
//==! Function Parameters: none
//==! Return Value       : none
//==!=====================================================================!==
CMemIscDataContext::CMemIscDataContext()
{

  m_pCMIDS = NULL;
  m_pCMIDArray = NULL;
  m_pCMIDR_current = NULL;
  m_lCMIDR_index = 0;

} // CMemIscDataContext::CMemIscDataContext()
//==!=====================================================================!==
//==! CMemIscDataContext destructor
//==!
//==! Function Parameters: none
//==! Return Value       : none
//==!=====================================================================!==
CMemIscDataContext::~CMemIscDataContext()
{

//==! Memory data source objects will be cleared through the global pointer
//==! pgCMIDS_list
  m_pCMIDS = NULL;
  m_pCMIDArray = NULL;
  m_pCMIDR_current = NULL;
  m_lCMIDR_index = 0;

} // CMemIscDataContext::~CMemIscDataContext()
//==!=====================================================================!==
//==! CProcContext constructor
//==!
//==! Function Parameters: none
//==! Return Value       : none
//==!=====================================================================!==
CProcContext::CProcContext()
{
  m_iBypass = 0;
  m_iPhase = 0;
  m_iSkipFlag = 0;
  m_iFlowActStat = 0;
  m_iDataAvail = 0;
  m_iFlag0 = 0;
  m_iProcessingLoop = 0;

  m_lX = 0;
  m_lY = 0;
  m_lZ = 0;
  m_lRptMax = 0;
  m_lGetData = 0;

  m_lLoopMin = 0;
  m_lLoopMax = 0;
  m_lLoopVar = 0;
  m_l_OST = 0;

  m_pCIAvar = NULL;
  m_pProcObj = NULL;
  m_pFlowObjVar = NULL;
  m_pCFAptr = NULL;
  m_pCFAvar = NULL;
  m_pCFAloopVar = NULL;
  m_pCFA_LoopActions = NULL;

} // CProcContext::CProcContext()
//==!=====================================================================!==
//==! CProcContext destructor (same as constructor)
//==!
//==! Function Parameters: none
//==! Return Value       : none
//==!=====================================================================!==
CProcContext::~CProcContext()
{
  m_iBypass = 0;
  m_iPhase = 0;
  m_iSkipFlag = 0;
  m_iFlowActStat = 0;
  m_iDataAvail = 0;
  m_iFlag0 = 0;
  m_iProcessingLoop = 0;

  m_lX = 0;
  m_lY = 0;
  m_lZ = 0;
  m_lRptMax = 0;
  m_lGetData = 0;

  m_lLoopMin = 0;
  m_lLoopMax = 0;
  m_lLoopVar = 0;
  m_l_OST = 0;

  m_pCIAvar = NULL;
  m_pProcObj = NULL;
  m_pFlowObjVar = NULL;
  m_pCFAptr = NULL;
  m_pCFAvar = NULL;
  m_pCFAloopVar = NULL;
  m_pCFA_LoopActions = NULL;

} // CProcContext::~CProcContext()
//==!=====================================================================!==
//==! CDeviceData constructor
//==!
//==! Function Parameters: none
//==! Return Value       : none
//==!=====================================================================!==
CDeviceData::CDeviceData()
{
  m_csDeviceName = NULL;
  m_csAction = NULL;
  m_csActionData = NULL;
  m_pCBsdlData = NULL;
  m_pCDP_Out = NULL;
  m_lInstrRegLength = 0;
  m_iSkipDev = 0;
  m_iIscDataFormat = 0;
  m_pCProcContext = NULL;
  m_iProcessingStatus = 1;

  m_cSIR = NULL;
  m_cSIR_bypass = NULL;
  m_cTdi = NULL;
  m_cTdo = NULL;
  m_cTdoCopy = NULL;
  m_iFlowType = -1;
  m_iHoldFlowType = -1;

  m_pCIscData = NULL;
  m_pNextCDeviceData = NULL;
  m_pNextCDDfromTdi = NULL;

} // CDeviceData::CDeviceData()
//==!=====================================================================!==
//==! CDeviceData destructor
//==!
//==! Function Parameters: none
//==! Return Value       : none
//==!=====================================================================!==
CDeviceData::~CDeviceData()
{
  freeArray(m_csDeviceName);
  freeArray(m_csAction);
  freeArray(m_csActionData);
  m_pCBsdlData = NULL;   // will be freed when deleting the global pgCBD_list
  freeData(m_pCDP_Out);
  m_iSkipDev = 0;
  m_iIscDataFormat = 0;
  freeData(m_pCProcContext);
  m_iProcessingStatus = 1;

  freeArray(m_cSIR);
  freeArray(m_cSIR_bypass);
  freeArray(m_cTdi);
  freeArray(m_cTdo);
  freeArray(m_cTdoCopy);
  m_iFlowType = -1;
  m_iHoldFlowType = -1;

  freeData(m_pCIscData);
  m_pNextCDeviceData = NULL;
  m_pNextCDDfromTdi = NULL;

} // CDeviceData::~CDeviceData()
//==!=====================================================================!==
//==! CDeviceData member function to create a register long enough to hold
//==! instruction or data for a particular action.
//==!
//==! Function Parameters: l_IR = instruction reg length
//==!                      l_DR = data reg length
//==! Return Value       : none
//==!=====================================================================!==
void CDeviceData::createTapReg(long l_IR,long l_DR)
{ long lVar;

  m_cSIR = new char [l_IR+1];
  m_cSIR_bypass = new char [l_IR+1];
  xInitCharArray(m_cSIR,l_IR,'1',1);   // initialize to all 1s
  xInitCharArray(m_cSIR_bypass,l_IR,'1',1);   // initialize to all 1s

  lVar = l_DR;
  if (l_IR > l_DR)
  { lVar = l_IR;
  } // IF
  m_cTdi = new char [lVar+1];  // initialize data regs to all 0s
  m_cTdo = new char [lVar+1];
  m_cTdoCopy = new char [lVar+1];

  xInitCharArray(m_cTdi,lVar,'0',1);
  xInitCharArray(m_cTdo,lVar,'0',1);
  xInitCharArray(m_cTdoCopy,lVar,'0',1);

} // CDeviceData::createTapReg()
//==!=====================================================================!==
//==! Get information for opcode iOpIndex.
//==!
//==! Function Parameters: lOpIndex = opcode index
//==!                      pcsOpBits = opcode bit encoding
//==! Return Value       : pointer to opcode name
//==!=====================================================================!==
char *CDeviceData::getOpcodeInfo(long lOpIndex,char *pcsOpBits)
{
  isInRange(1,lOpIndex,0,m_pCBsdlData->lNumOpcodes-1,
            "gOpInfo",DISP_AND_EXIT);
  strcpy(pcsOpBits,m_pCBsdlData->pCBsdlOpInfo[lOpIndex]
                               ->m_CxCStr_OpcodeList->pStr);
  return m_pCBsdlData->pCBsdlOpInfo[lOpIndex]->csOpcodeName;

} // CDeviceData::getOpcodeInfo()
//==!=====================================================================!==
//==! Load instruction register with BYPASS opcode. Function was written
//==! just in case the BYPASS opcode in the BSDL file (if specified) is not
//==! all 1s.
//==!
//==! Function Parameters: none
//==! Return Value       : none
//==!=====================================================================!==
void CDeviceData::loadBypassOp()
{

  if (m_pCBsdlData == NULL)
  { xInitCharArray(m_cSIR,m_lInstrRegLength,'1',1); // initialize to all 1s
  } // IF
  else
  { isInRange(1,m_pCBsdlData->lOpcodeBypass,0,m_pCBsdlData->lNumOpcodes-1,
              "No BypassOp",DISP_AND_EXIT);
    strcpy(m_cSIR,m_pCBsdlData->pCBsdlOpInfo[m_pCBsdlData->lOpcodeBypass]
                              ->m_CxCStr_OpcodeList->pStr);
  } // ELSE

} // CDeviceData::loadBypassOp()
//==!=====================================================================!==
//==! Write any output data to the output file in ISC data format if
//==! m_iIscDataFormat == 1.
//==!
//==! Function Parameters: iMode = 0: write header
//==!                              1: write trailer
//==!                              2: write DATA with cStrData as array name
//==!                              3: write INIT/REPEAT/TERMINATE data
//==!                              4: write closing ')' for DATA
//==!                      cStrData = data string for different iModes
//==!                      lParam = used if iMode = 3
//==!                         0: print INITIALIZE keyword
//==!                         1: print REPEAT keyword
//==!                         2: print TERMINATE keyword
//==! Return Value       : none
//==!=====================================================================!==
void CDeviceData::writeIscData(int iMode,char *cStrData,long lParam)
{ struct tm *tmTimeData;
  time_t tCurTime;

  if (m_pCDP_Out == NULL)
  { return;
  } // IF

  if (m_pCDP_Out->pFILE == NULL)
  { return;
  } // IF

  if (m_iIscDataFormat == 0)
  { if (iMode == 1)
    { m_pCDP_Out->writeToCDP(1,NULL);
    } // IF
    return;
  } // IF

  switch(iMode)
  { case 0:
    { m_pCDP_Out->writeToCDP(0,"IEEE_1532_Data");
      m_pCDP_Out->writeToCDP(0,"\n(Header (STD_Version STD_1532_2001");

      time(&tCurTime);
      tmTimeData = localtime(&tCurTime);
      sprintf(sMsgBuf03,
            "\n         Creation_Date %04d%02d%02d.%02d%02d",
            tmTimeData->tm_year+1900,tmTimeData->tm_mon+1,
            tmTimeData->tm_mday,tmTimeData->tm_hour,tmTimeData->tm_min);
      m_pCDP_Out->writeToCDP(0,sMsgBuf03);

      sprintf(sMsgBuf03,"\n         Creator \"%s %s\"",
              PROGRAM_NAME,PROG_VERSION);
      m_pCDP_Out->writeToCDP(0,sMsgBuf03);
      sprintf(sMsgBuf03,"\n         Entity %s\n        )",
              m_pCBsdlData->csEntityName);
      m_pCDP_Out->writeToCDP(0,sMsgBuf03);
      break;
    } // CASE 0
    case 1:
    { m_pCDP_Out->writeToCDP(0,"\n)");
      m_pCDP_Out->writeToCDP(1,NULL);
      break;
    } // CASE 1
    case 2:
    { m_pCDP_Out->writeToCDP(0,"\n  DATA ");
      if (cStrData != NULL)
      { sprintf(sMsgBuf03,"%s (",cStrData);
        m_pCDP_Out->writeToCDP(0,sMsgBuf03);
      } // IF
      break;
    } // CASE 2
    case 3:
    { switch (lParam)
      { case -2:
        { m_pCDP_Out->writeToCDP(0,"\n    )");
          break;
        } // CASE 0
        case 0:
        { m_pCDP_Out->writeToCDP(0,"\n    initialize (\n");
          m_pCDP_Out->m_iFirstData = 1;
          break;
        } // CASE 0
        case 1:
        { m_pCDP_Out->writeToCDP(0,"\n    repeat (\n");
          m_pCDP_Out->m_iFirstData = 1;
          break;
        } // CASE 1
        case 2:
        { m_pCDP_Out->writeToCDP(0,"\n    terminate (\n");
          m_pCDP_Out->m_iFirstData = 1;
          break;
        } // CASE 2
        default:
        { jMessage(DISP_AND_EXIT,"writeIscData iMode",0,NULL);
          break;
        } // DEFAULT
      } // SWITCH (lParam)
      break;
    } // CASE 3
    case 4:
    { m_pCDP_Out->writeToCDP(0,"\n  )");
      break;
    } // CASE 4
    default:   // should never get here
    { break;
    } // DEFAULT
  } // SWITCH (iMode)

} // CDeviceData::writeIscData()
//==!=====================================================================!==
//==! Check the flow object according to iMode. If iMode:
//==! = -1: check INIT/REPEAT/TERMINATE flow actions for presence of output
//==!       operator
//==! =  0: check INIT flow action for output operator
//==! =  1: check REPEAT flow action for output operator
//==! =  2: check TERMINATE flow action for output operator
//==! lFlowIndex is usually used when iMode = 1 (REPEAT flow/phase) to select
//==! which REPEAT flow to process.
//==!
//==! Function Parameters: pCFOpar = pointer to flow object
//==!                      iMode = see function description
//==!                      lFlowIndex = see function description
//==! Return Value       : 1 if successful, 0 otherwise
//==!=====================================================================!==
int checkFlowObj(CxFlowObj *pCFOpar,int iMode,long lFlowIndex)
{ CMultFlow *p_CMFlow;
  long lX;

  if (pCFOpar != NULL)
  { if (pCFOpar->m_pFlowItem != NULL)
    { switch (iMode)
      { case -1:
        { if (pCFOpar->m_pFlowItem->m_iInitHasOutOp == 1)
          { return 1;
          } // IF
          for (p_CMFlow = pCFOpar->m_pFlowItem->m_pCMF_Repeat;
               p_CMFlow != NULL;
               p_CMFlow = p_CMFlow->m_pNextCMultFlow)
          { if (p_CMFlow->m_iRptHasOutOp == 1)
            { return 1;
            } // IF
          } // FOR p_CMFLow != NULL
          if (pCFOpar->m_pFlowItem->m_iTermHasOutOp == 1)
          { return 1;
          } // IF
          break;
        } // CASE -1
        case  0:
        { if (pCFOpar->m_pFlowItem->m_iInitHasOutOp == 1)
          { return 1;
          } // IF
          break;
        } // CASE 0
        case  1:
        { for (p_CMFlow = pCFOpar->m_pFlowItem->m_pCMF_Repeat,lX = 1;
               p_CMFlow != NULL;
               p_CMFlow = p_CMFlow->m_pNextCMultFlow,++lX)
          { if (lX == lFlowIndex)
            { if (p_CMFlow->m_iRptHasOutOp == 1)
              { return 1;
              } // IF
              return 0;
            } // IF
          } // FOR p_CMFlow != NULL
          isPtrNull(p_CMFlow,"chkFlowObj:0");
          return 0;
        } // CASE 1
        case  2:
        { if (pCFOpar->m_pFlowItem->m_iTermHasOutOp == 1)
          { return 1;
          } // IF
          break;
        } // CASE 2
        default:   // should never get here
        { jMessage(DISP_AND_EXIT,"CheckFlowObj iMode",0,NULL);
        } // DEFAULT
      } // SWITCH (iMode)
    } // IF pCFOpar->m_pFlowItem != NULL
  } // IF pCFOpar != NULL

  return 0;

} // checkFlowObj()
//==!=====================================================================!==
//==! Get data from command file data pointer and return it as a character
//==! string. Data can be a single delimited token, or a string containing
//==! spaces enclosed by double quotes. Data enclosed by quotes should be on
//==! one line only (i.e., string should not span multiple lines or have
//==! carraige-returns or line-feeds).
//==!
//==! Ex: data1
//==!     "data with spaces "  <- will include last space before closing
//==!                             double quote
//==!
//==! Function Parameters: pDP = pointer to command data port
//==! Return Value       : data (see function description)
//==!=====================================================================!==
char *getCmdData(cDataPort *pDP)
{ char *pStr = NULL;

  xParseData(pDP,XDATA_READ);
  if (pDP->eDataAttr == eATTR_VAR)
  { copyStr(&pStr,pDP->sData);
    return pStr;
  } // IF
  if (strcmp(pDP->sData,"\""))
  { jMessage(DISP_AND_EXIT,"Invalid data format",1,pDP);
  } // IF
  while (1)
  { xParseData(pDP,XDATA_READ_RAW);
    if (pDP->sData[0] == '"')
    { break;
    } // IF
    if (pDP->sData[0] == '\r' || pDP->sData[0] == '\n')
    { jMessage(DISP_SHOW_MSG,"data/string spans multiple lines",1,pDP);
      xstrcat(&pStr," ");
      continue;
    } // IF
    xstrcat(&pStr,pDP->sData);
  } // WHILE (1)
  return pStr;

} // getCmdData()
//==!=====================================================================!==
//==! Find a pointer to the IscAction specified for the device.
//==!
//==! Function Parameters: pDeviceObj = pointer to device object
//==! Return Value       : pointer to CIscAct action
//==!=====================================================================!==
CIscAct *findIscAction(CDeviceData *pDeviceObj)
{ CIscAct *pCFAvar;

  if (pDeviceObj->m_csAction == NULL)
  { return NULL;
  } // IF

  for (pCFAvar = pDeviceObj->m_pCBsdlData->m_pCIscActInfo;
       pCFAvar != NULL;
       pCFAvar = pCFAvar->m_pNextCIscAct)
  { if (!strcmp(pDeviceObj->m_csAction,pCFAvar->m_csActName))
    { if (pDeviceObj->m_csActionData == NULL &&
          pCFAvar->m_csActDataName == NULL)
      { break;
      } // IF
      if (pDeviceObj->m_csActionData != NULL &&
          pCFAvar->m_csActDataName != NULL)
      { if (!strcmp(pCFAvar->m_csActDataName,pDeviceObj->m_csActionData))
        { break;
        } // IF
      } // IF
    } // IF
  } // FOR pCFAvar != NULL
  return pCFAvar;

} // findIscAction()
//==!=====================================================================!==
//==! Check command file option combinations for each device. For example,
//==! check that:
//==! - action exists in BSDL file (check that BSDL file was specified)
//==! - data exists for the action in the ISC file
//==!
//==! Function Parameters: pDeviceList = pointer to device list
//==! Return Value       : 1 if successful, 0 otherwise
//==!=====================================================================!==
int checkCmdOptions(CDeviceData *pDeviceList)
{ CDeviceData *pCDDvar;
  CIscAct *pCFAvar;
  int iFlag;

  iGlobProgramState = dPROGSTATE_CHECKCMDOPTIONS;

  for (pCDDvar = pDeviceList,iFlag = 1;
       pCDDvar != NULL;
       pCDDvar = pCDDvar->m_pNextCDeviceData)
  { //==! Check instr reg length definition
    if (pCDDvar->m_pCBsdlData != NULL)
    { if (pCDDvar->m_pCBsdlData->lInstrLength != -1)
      { if (pCDDvar->m_lInstrRegLength == 0) // Update if not defined
        { pCDDvar->m_lInstrRegLength = pCDDvar->m_pCBsdlData->lInstrLength;
        } // IF
        else // Compare against instr reg length in BSDL file
        { if (pCDDvar->m_pCBsdlData->lInstrLength != pCDDvar->m_lInstrRegLength)
          { sprintf(sMsgOnly,
               "  => [%s]: will use instr reg length [%ld] from [%s]",
               pCDDvar->m_csDeviceName,pCDDvar->m_pCBsdlData->lInstrLength,
               pCDDvar->m_pCBsdlData->csBsdlFileName);
            jMessage(DISP_SHOW_MSG,sMsgOnly,0,NULL);
          } // IF
        } // ELSE
      } // IF BSDL instr reg length != -1
    } // IF pCDDvar->m_pCBsdlData != NULL

//==! Allocate processing context structures per device if concurrent mode
//==! selected. These structures keep track of each device's context: which
//==! procedure/flow/action is being processed.
    if (iGlobProcMode == 1)
    { pCDDvar->m_pCProcContext = new CProcContext;
    } // IF iGlobProcMode == 1

    if (pCDDvar->m_csAction == NULL)  //==! No action specified
    { continue;
    } // IF

//==! The device and its data file each have a variable indicating the data
//==! format so that the DATA_ISC option and the data output filename can be
//==! specified in any order in the command file.
    if (pCDDvar->m_iIscDataFormat == 1)
    { if (pCDDvar->m_pCDP_Out != NULL)
      { pCDDvar->m_pCDP_Out->m_iDataFormat = 1;
      } // IF
    } // IF pCDDvar->m_iIscDataFormat == 1

    if (pCDDvar->m_pCBsdlData == NULL)
    { sprintf(sMsgOnly,"Device [%s] action [%s] requires 1532 BSDL file",
              pCDDvar->m_csDeviceName,pCDDvar->m_csAction);
      jMessage(DISP_AND_EXIT,sMsgOnly,0,NULL);
    } // IF

    if (findIscAction(pCDDvar) == NULL)
    { if (pCDDvar->m_csActionData == NULL)
      { sprintf(sMsgOnly,"[%s]: invalid action for [%s (%s)]",
              pCDDvar->m_csAction,pCDDvar->m_csDeviceName,
              pCDDvar->m_pCBsdlData->csBsdlFileName);
      } // IF
      else
      { sprintf(sMsgOnly,"[%s(%s)]: invalid action for device [%s (%s)]",
              pCDDvar->m_csAction,pCDDvar->m_csActionData,
              pCDDvar->m_csDeviceName,
              pCDDvar->m_pCBsdlData->csBsdlFileName);
      } // ELSE
      jMessage(DISP_SHOW_MSG,sMsgOnly,0,NULL);
      iFlag = 0;

//==! Display list of actions for this device
      sprintf(sMsgOnly,"  - Valid [%s] actions are:",
              pCDDvar->m_pCBsdlData->csBsdlFileName);
      dispMsg(sMsgOnly,DISP_POST_LF,1,1);
      for (pCFAvar = pCDDvar->m_pCBsdlData->m_pCIscActInfo;
           pCFAvar != NULL;
           pCFAvar = pCFAvar->m_pNextCIscAct)
      { sprintf(sMsgOnly,"    = %s",pCFAvar->m_csActName);
        if (pCFAvar->m_csActDataName != NULL)
        { strcat(sMsgOnly,"(");
          strcat(sMsgOnly,pCFAvar->m_csActDataName);
          strcat(sMsgOnly,")");
        } // IF
        dispMsg(sMsgOnly,DISP_POST_LF,1,1);
      } // FOR pCFAvar != NULL
    } // IF findIscAction() == NULL

    if (isIscDataRequired(pCDDvar) == 1)
    { if (pCDDvar->m_pCIscData == NULL)   // check for ISC data file
      { if (pCDDvar->m_csActionData == NULL)
        { sprintf(sMsgOnly,
                  "datafile needed for action [%s] for device [%s (%s)]",
                  pCDDvar->m_csAction,
                  pCDDvar->m_csDeviceName,
                  pCDDvar->m_pCBsdlData->csBsdlFileName);
        } // IF
        else
        { sprintf(sMsgOnly,
                  "datafile needed for action [%s(%s)] for device [%s (%s)]",
                  pCDDvar->m_csAction,
                  pCDDvar->m_csActionData,
                  pCDDvar->m_csDeviceName,
                  pCDDvar->m_pCBsdlData->csBsdlFileName);
        } // ELSE
        jMessage(DISP_SHOW_MSG,sMsgOnly,0,NULL);
        iFlag = 0;
      } // IF
      else
      { pCDDvar->m_pCIscData
               ->readIscDataHeader(0,pDeviceList,pCDDvar->m_pCBsdlData,NULL,
                                   &(pCDDvar->m_CMemIscDataContext));
        pCDDvar->m_pCIscData->m_pCDP_IscData->xCloseDataPort();
      } // ELSE
    } // IF
  } // FOR pCDDvar != NULL

  return iFlag;

} // checkCmdOptions()
//==!=====================================================================!==
//==! Read command file and create the CDeviceData list pgCDD_list. Closes
//==! command file before leaving function.
//==!
//==! Function Parameters: pDP = pointer to command data port
//==! Return Value       : 1 if successful, 0 otherwise
//==!=====================================================================!==
int readCmdFile(cDataPort *pDP)
{ CDeviceData *pCDDvar,*pCDDtail;
  CBsdlData *pCBDvar;
  char *pcPtr;
  long lVar0,lVar1;
  cDataPort *pCDP_data;
  FILE *pFP;

  pDP->iCommentStyle = XDATA_COMMENT_STYLE_CPP;

  while (1)
  { iGlobProgramState = dPROGSTATE_READCMDFILE;

    xParseData(pDP,XDATA_READ);
    if (pDP->m_iEndOfDataStream == 1)
    { xFclose(&(pDP->pFILE));
      return 1;
    } // IF

    if (pDP->eDataAttr != eATTR_VAR)
    { jMessage(DISP_AND_EXIT,"JTAG device description expected",1,pDP);
    } // IF

//==! Check if device name is unique
    for (pCDDvar = pgCDD_list; pCDDvar != NULL;
         pCDDvar = pCDDvar->m_pNextCDeviceData)
    { if (!strcmp(pDP->sData,pCDDvar->m_csDeviceName))
      { jMessage(DISP_AND_EXIT,"Device name already exists",1,pDP);
      } // IF
    } // FOR pCDDvar != NULL

    pCDDvar = new CDeviceData;
    copyStr(&(pCDDvar->m_csDeviceName),pDP->sData);
    while (1)
    { xParseData(pDP,XDATA_READ);
      if (pDP->m_iEndOfDataStream == 1)
      { jMessage(DISP_AND_EXIT,"Terminate device description with ';'",1,pDP);
      } // IF
      else
      { pDP->iParserGetData = 0;
      } // ELSE
      findStr(pDP,"-",0);
      xParseData(pDP,XDATA_READ_RAW);
      switch (pDP->eDataAttr)
      { case eATTR_VAR:
        {
//==! specify device action
          if (!strcmp(pDP->sData,"a"))
          { lVar0 = 0;
            xParseData(pDP,XDATA_READ);
            if (pDP->sData[0] == '"')
            { lVar0 = 1;
            } // IF
            else
            { pDP->iParserGetData = 0;
            } // ELSE
            xGetData(pDP,0,eATTR_VAR,"Action expected");
            copyStr(&(pCDDvar->m_csAction),pDP->sData);
            xParseData(pDP,XDATA_READ);
            if (pDP->sData[0] == '(')
            { xGetData(pDP,0,eATTR_VAR,"action data name expected");
              copyStr(&(pCDDvar->m_csActionData),pDP->sData);
              findStr(pDP,")",0);
            } // IF
            else
            { pDP->iParserGetData = 0;
            } // ELSE
            if (lVar0 == 1)
            { findStr(pDP,"\"",0);
            } // IF
            break;
          } // IF sData == 'a'

//==! specify device BSDL file
          if (!strcmp(pDP->sData,"b"))
          { pcPtr = getCmdData(pDP);
            pFP = xFileOpen(pcPtr,"t");
            if (pFP == NULL)
            { sprintf(sMsgOnly,"BSDL file [%s] does not exist",pcPtr);
              jMessage(DISP_AND_EXIT,sMsgOnly,1,pDP);
              break;
            } // IF
            xFclose(&pFP);

#if 0
//==! Check if BSDL file has already been read by checking the pgCBD_list; if
//==! present, then point to it, else add a CBsdlData object to the list
//==! and then point to it.
            for (pCBDvar = pgCBD_list;
                 pCBDvar != NULL;
                 pCBDvar = pCBDvar->m_pNextCBsdlData)
            { if (!strcmp(pcPtr,pCBDvar->csBsdlFileName))
              { break;
              } // IF
            } // FOR pCBDvar != NULL

            if (pCBDvar == NULL) // BSDL file not in pgCBD_list
            { pCDP_data = new cDataPort(pcPtr);
              sprintf(sMsgBuf00,"--> Reading BSDL file [%s]",
                      pCDP_data->psPortName);
              dispMsg(sMsgBuf00,DISP_POST_LF,1,1);
              pCDP_data->pFILE = xFileOpen(pCDP_data->psPortName,"r");

              pCBDvar = new CBsdlData;      // create new BSDL object
              pCBDvar->readBsdl(pCDP_data); // read BSDL info

              pCBDvar->m_pNextCBsdlData = pgCBD_list;  // add to pgCBD_list
              pgCBD_list = pCBDvar;

              freeData(pCDP_data);

              pCDDvar->m_pCBsdlData = pCBDvar; // assign BSDL file to device
            } // IF
            else
            { pCDDvar->m_pCBsdlData = xCopyBsdl(pCBDvar); // copy BSDL file
            } // ELSE
            freeArray(pcPtr);
#else
            pCDP_data = new cDataPort(pcPtr);
            sprintf(sMsgBuf00,"--> Reading BSDL file [%s]",
                    pCDP_data->psPortName);
            dispMsg(sMsgBuf00,DISP_POST_LF,1,1);
            pCDP_data->pFILE = xFileOpen(pCDP_data->psPortName,"r");

            pCBDvar = new CBsdlData;      // create new BSDL object
            pCBDvar->readBsdl(pCDP_data); // read BSDL info

            pCBDvar->m_pNextCBsdlData = pgCBD_list;  // add to pgCBD_list
            pgCBD_list = pCBDvar;

            freeData(pCDP_data);
            freeArray(pcPtr);
            pCDDvar->m_pCBsdlData = pCBDvar; // assign BSDL file to device
#endif

            break;
          } // IF sData == 'b'

//==! specify ISC data file
          if (!strcmp(pDP->sData,"d"))
          { pcPtr = getCmdData(pDP);
            pFP = xFileOpen(pcPtr,"t");
            if (pFP == NULL)
            { sprintf(sMsgOnly,"ISC data file [%s] does not exist",pcPtr);
              jMessage(DISP_AND_EXIT,sMsgOnly,1,pDP);
              break;
            } // IF
            xFclose(&pFP);
            pCDDvar->m_pCIscData = new CIscData (pcPtr);
            freeArray(pcPtr);
            break;
          } // IF sData[0] == 'd'

//==! Get the instruction reg length. Note that the decimal number can be
//==! right next to the -i option (e.g., -i20) or as a separate token
//==! (e.g., -i 20).
          if (pDP->sData[0] == 'i')
          { lVar0 = strlen(pDP->sData);
            if (lVar0 > 1)
            { for (lVar1 = 1; lVar1 < lVar0; ++lVar1)
              { if (!isdigit(pDP->sData[lVar1]))
                { jMessage(DISP_AND_EXIT,"use decimal numbers only",1,pDP);
                } // IF !isdigit()
              } // FOR lVar1 < lVar0
              pCDDvar->m_lInstrRegLength = atol(&(pDP->sData[1]));
            } // IF
            else
            { pCDDvar->m_lInstrRegLength = lReadDecimal(pDP);
            } // ELSE
            break;
          } // IF sData[0] == 'i'

//==! specify output file to hold data read from this device. Check if file
//==! exists before overwriting only when the action is performed (see
//==! NOTE_0007).
          if (!strcmp(pDP->sData,"o"))
          { pcPtr = getCmdData(pDP);
            pCDDvar->m_pCDP_Out = new cDataPort (pcPtr);
            freeArray(pcPtr);
            break;
          } // IF sData == 'o'

//==! write data to output file in ISC data format
          if (!strcmp(pDP->sData,"data_isc"))
          { pCDDvar->m_iIscDataFormat = 1;
            break;
          } // IF sData == 'data_isc'

//==! Skip device as if it were not in the JTAG chain (do not put into JTAG
//==! BYPASS). Used to debug the skip device code which will be used mainly
//==! in device programmer environments where one or more devices are defined
//==! in the command file to correspond to the maximum number of sockets on
//==! the programmer. The device programmer control program can then set the
//==! m_iSkipDev flag to skip sockets which do not have devices in them or
//==! which do not pass some preliminary test (i.e., device ID check, etc).
          if (!strcmp(pDP->sData,"skip_dev"))
          { pCDDvar->m_iSkipDev = 1;
            break;
          } // IF sData == 'skip_dev'

          jMessage(DISP_AND_EXIT,"Unknown option",1,pDP);
          break;
        } // CASE eATTR_VAR

        default:
        { jMessage(DISP_AND_EXIT,"Unknown command",1,pDP);
          break;
        } // DEFAULT

      } // SWITCH (pDP->eDataAttr)

      xParseData(pDP,XDATA_READ);
      if (!strcmp(pDP->sData,";"))
      { break;
      } // IF
      else
      { pDP->iParserGetData = 0;
      } // ELSE

    } // WHILE (1)

    if (pgCDD_list == NULL)
    { pgCDD_list = pCDDvar;
    } // IF
    else
    { pCDDtail->m_pNextCDeviceData = pCDDvar;
    } // ELSE
    pCDDtail = pCDDvar;

    pCDDvar->m_pNextCDDfromTdi = pgCDD_FromTdi;
    pgCDD_FromTdi = pCDDvar;

  } // WHILE (1)

  xFclose(&(pDP->pFILE));
  return 1;

} // readCmdFile()
//==!=====================================================================!==
//==! Wait the amount of time specified by the device object. Do the number
//==! of wait TCK cycles first, and then wait the specified amount of time.
//==! This routine can be optimized by overlapping the TCK wait time with
//==! the absolute wait time specified.
//==!
//==! Function Parameters: pCFWobj = flow wait object being processed
//==! Return Value       : none
//==!=====================================================================!==
void doFlowWait(CFlowWait *pCFWobj)
{ long lX;

//==! Non-overlapped version: do TCK wait time first, then absolute flow wait
//==! time in seconds.
  if (pCFWobj->m_lMinFlowWaitCycles > 0)
  { for (lX = 0; lX < pCFWobj->m_lMinFlowWaitCycles; ++lX)
    { pulseClock(0); //==! Do LOW-HIGH transition
    } // FOR lX < pCFWobj->m_lMinFlowWaitCycles
    setPort(TCK,0);  //==! Bring TCK LOW
    if (pCFWobj->m_fMinFlowWaitTimeSec <= 0)
    { return;
    } // IF
  } // IF lMinFlowWaitCycles > 0

  if (pCFWobj->m_fMinFlowWaitTimeSec > 0)
  { if (pCFWobj->m_fMinFlowWaitTimeSec > 10)
    { sprintf(sMsgOnly," -> wait %.2f secs",
              pCFWobj->m_fMinFlowWaitTimeSec);
      debugMsg(sMsgOnly,1,1);
    } // IF
    lX = (long)(pCFWobj->m_fMinFlowWaitTimeSec * 1e6);
//==! 2003/10/14: Pulse clock at least once when waiting X seconds. Seems to
//==! be required for XC18V04/X2CP.
    if (iOpt_MinOneClk == 1)
    { pulseClock(0);   //==! Do LOW-HIGH transition
      setPort(TCK,0);  //==! Bring TCK LOW
    } // IF
    waitTime(lX);
  } // IF

} // doFlowWait()
//==!=====================================================================!==
//==! Process the specified flow activity. Return 0 if activity was
//==! processed successfully, 1 otherwise.
//==!
//==! The iDataStatus variable pointer returns a value that is interpreted as
//==! follows:
//==! - 1: if there is still data available in the ISC data file
//==! - 2: if no more data in ISC data file
//==!
//==!
//==! Function Parameters: pCDDobj = device being processed
//==!                      pCFAobj = flow activity object being processed
//==!                      pCIFobj = pointer to CIscFlow object containing
//==!                         the variables used by this flow activity
//==!                      iDataStatus = variable pointer to return value
//==!                         (see function description)
//==!                         - return 1 if there is still data available in
//==!                           the ISC data file
//==!                         - return 2 if no more data in ISC data file
//==!                      lCurRptCount = current repeat count (i.e., how
//==!                         many times this flow was processed)
//==!                      iInLoop = 1 if function is part of a loop, 0
//==!                         otherwise
//==! Return Value       : 0 or 1 as described above
//==!=====================================================================!==
int processFlowActivity(CDeviceData *pCDDobj,
                        CFlowAct *pCFAobj,CIscFlow *pCIFobj,int *iDataStatus,
                        long lCurRptCount,int iInLoop)
{ long lLoopMin,lLoopMax,lX,l_OST;
  CFlowAct *pCFAvar;

  iGlobProgramState = dPROGSTATE_PROCESSFLOWACTIVITY;

  if (pCFAobj->m_pLoopActions != NULL)
  { lLoopMin = pCFAobj->m_lMinLoopCount;
    lLoopMax = pCFAobj->m_lMaxLoopCount;
    if (iOpt_FlowLoopDbg == 1)
    { sprintf(sMsgOnly," -> loop max: %ld times",lLoopMax);
      debugMsg(sMsgOnly,1,1);
    } // IF iOpt_FlowLoopDbg == 1
    for (lX = 0; lX < lLoopMax; ++lX)
    { if (iOpt_FlowLoopDbg == 1)
      { sprintf(sMsgBuf03,"\n== Flow loop iteration [%ld] --",lX);
        debugMsg(sMsgBuf03,1,1);
      } // IF
      for (pCFAvar = pCFAobj->m_pLoopActions,l_OST = 0;
           pCFAvar != NULL;
           pCFAvar = pCFAvar->m_pNextCFlowAct)
      { l_OST += processFlowActivity(pCDDobj,pCFAvar,pCIFobj,
                 &(*iDataStatus),lCurRptCount,1);
      } // FOR pCFAvar != NULL
      if (lX >= lLoopMin && l_OST == 0)
      { break;
      } // IF
    } // FOR lX < lLoopMax
    if (lX >= lLoopMax && l_OST > 0)
    { sprintf(sMsgOnly,"\n*** ==> Max flow loop count [%ld] reached",lLoopMax);
      jMessage(DISP_AND_EXIT,sMsgOnly,0,NULL);
    } // IF
    if (iOpt_FlowLoopDbg == 1)
    { sprintf(sMsgOnly,"\n\t-> iterations through loop: %ld (max %ld)",
              lX,lLoopMax);
      debugMsg(sMsgOnly,1,1);
    } // IF iOpt_FlowLoopDbg == 1
    return 0;
  } // IF pCFAobj->m_pLoopActions != NULL

  sprintf(sMsgOnly,"\n\t- opcode [%s]",
          pCDDobj->getOpcodeInfo(pCFAobj->m_lOpcodeIndex,cTdiData));
  if (iOpt_ProgDebug == 1)
  { debugMsg(sMsgOnly,2,1);
  } // IF

//==! If there is update data to shift, then load the opcode but stop in
//==! UPDATE-IR (see jtagShift() parameter description). The update data
//==! shift will go from UPDATE_IR to SELECT-DR (skipping RTI), then end in
//==! RTI after SHIFT-DR. This TAP state transition sequence is required so
//==! that the previously loaded opcode will not start executing until after
//==! the data (if required) is loaded. If there is no input data to shift,
//==! the JTAG IR shift operation will load the opcode and stop in RTI.
  jtagShift(TAP_IRSHIFT,pCDDobj->m_lInstrRegLength,cTdiData,cTdoData,0,
            (pCFAobj->m_pUF_Update != NULL ? 1 : 0));

  pCFAobj->processUpdateField(0,cTdiData,cTdoData,
                              pCDDobj,pCIFobj,&(*iDataStatus));

  doFlowWait(pCFAobj->m_pFlowWaitInfo);

  return pCFAobj->processCaptureField(0,cTdiData,cTdoData,pCDDobj,pCIFobj,
                    lCurRptCount,&(*iDataStatus),iInLoop);

} // processFlowActivity()
//==!=====================================================================!==
//==! Convert the enumerated type to a string (for debug purposes)
//==!
//==! Function Parameters: eProcValue
//==! Return Value       : char string for enumerated type
//==!=====================================================================!==
char *cnvOverrideStr(EnumProcSetting eProcValue)
{

  switch(eProcValue)
  { case ePROC_ON:
      return "ON";
    case ePROC_OFF:
      return "OFF";
    case ePROC_OPTIONAL:
      return "OPTIONAL";
    case ePROC_RECOMMENDED:
      return "RECOMMENDED";
    case ePROC_NO_SETTING:
    default:
      return "NO_SETTING";
  } // SWITCH eProcValue

} // cnvOverrideStr()
//==!=====================================================================!==
//==! Check if an ISC data file is needed for the action specified for the
//==! device.
//==!
//==! Function Parameters: pCDDobj = device being processed
//==! Return Value       : 1 if datafile required, 0 otherwise
//==!=====================================================================!==
int isIscDataRequired(CDeviceData *pCDDobj)
{ CIscAct *pCIAvar;
  CxProcObj *pProcObjVar;
  CxFlowObj *pFlowObjVar;

  pCIAvar = findIscAction(pCDDobj);

  if (pCIAvar == NULL)
  { return 0;
  } // IF

//==! Look through procedures in this action.
  for (pProcObjVar = pCIAvar->m_pActionProcs;
       pProcObjVar != NULL;
       pProcObjVar = pProcObjVar->m_pNextCxProcObj)
  { //==! Check if data file needed by flows in each procedure.
    for (pFlowObjVar = pProcObjVar->m_pProcItem->m_pProcFlows;
         pFlowObjVar != NULL;
         pFlowObjVar = pFlowObjVar->m_pNextCxFlowObj)
    { if (pFlowObjVar->m_pFlowItem->m_lUseDataName > 0)
      { return 1;
      } // IF
    } // FOR pFlowObjVar != NULL
  } // FOR pProcObjVar != NULL
  return 0;

} // isIscDataRequired()
//==!=====================================================================!==
//==! Check procedure processing status: check if procedure is optional,
//==! recommended, or always required.
//==!
//==! Function Parameters: pCDD = CDeviceData being processed
//==!                      pCxProcObjArg = procedure object
//==!                      piSkipFlag = pointer return if procedure must be
//==!                          skipped (1) or not (0)
//==! Return Value       : none
//==!=====================================================================!==
void checkProcedureProcessingStatus(CDeviceData *pCDD,
    CxProcObj *pCxProcObjArg,int *piSkipFlag)
{

  *piSkipFlag = 0;
  sprintf(sMsgBuf00,"\n  *** [%s] %s(%s)",pCDD->m_csDeviceName,
          pCxProcObjArg->m_pProcItem->m_csProcName,
          pCxProcObjArg->m_csProcParam != NULL
                ? pCxProcObjArg->m_csProcParam : " ");
  switch(pCxProcObjArg->m_enOptionOverride)
  { case ePROC_OFF:         //==! Skip this procedure
      sprintf(sMsgBuf01," (%s -> OFF)",
              cnvOverrideStr(pCxProcObjArg->m_enOptionSpec));
      strcat(sMsgBuf00,sMsgBuf01);
      *piSkipFlag = 1;
      break;
    case ePROC_ON:          //==! Do not skip this procedure
      sprintf(sMsgBuf01," (%s -> ON)",
              cnvOverrideStr(pCxProcObjArg->m_enOptionSpec));
      strcat(sMsgBuf00,sMsgBuf01);
      break;
    case ePROC_OPTIONAL:    //==! Procedure now optional
      sprintf(sMsgBuf01," (%s -> OPTIONAL)",
              cnvOverrideStr(pCxProcObjArg->m_enOptionSpec));
      strcat(sMsgBuf00,sMsgBuf01);
      *piSkipFlag = 1;
      break;
    case ePROC_RECOMMENDED: //==! Procedure now recommended; do not skip
      sprintf(sMsgBuf01," (%s -> RECOMMENDED)",
      cnvOverrideStr(pCxProcObjArg->m_enOptionSpec));
      strcat(sMsgBuf00,sMsgBuf01);
      break;
    case ePROC_NO_SETTING:
      if (pCxProcObjArg->m_enOptionSpec == ePROC_OPTIONAL)
      { *piSkipFlag = 1;
        strcat(sMsgBuf00," (OPTIONAL)");
      } // IF
      else
      { if (pCxProcObjArg->m_enOptionSpec == ePROC_RECOMMENDED)
        { strcat(sMsgBuf00," (RECOMMENDED)");
        } // IF
      } // ELSE
      break;
  } // SWITCH (pCxProbObjArg->m_enOptionOverride)
  debugMsg(sMsgBuf00,1,1);

} // checkProcedureProcessingStatus()
//==!=====================================================================!==
//==! Get flow activity list info: get the list associated with the flow
//==! phase (initialize, repeat, or terminate phase) for the device being
//==! processed.
//==!
//==! 2005/06/21: Added lRptFlowCount parameter to indicate which repeat flow
//==! to use. The standard allows one initialize phase, followed by multiple
//==! repeat phases, and a single terminate phase in a flow. lRptFlowCount
//==! will indicate which repeat phase to execute.
//==!
//==! Function Parameters: pCDDvar = device being processed
//==!                      lFlowPhase = flow activity phase
//==!                        0 = initialize phase
//==!                        1 = repeat phase
//==!                        2 = terminate phase
//==!                      lRptFlowCount = used only for repeat phase;
//==!                        specifies which repeat phase to use (see
//==!                        description)
//==!                      pFlowObjVar = pointer to flow list of pCDDvar
//==!                      plRptMax = pointer to return max flow repeat count
//==!                      pCFAvar = pointer to return flow activity list
//==!                      plGetData = pointer to return get data info
//==! Return Value       : none
//==!=====================================================================!==
void getFlowActivityListInfo(CDeviceData *pCDDvar,long lFlowPhase,
           long lRptFlowCount,
           CxFlowObj *pFlowObjVar,
           long *plRptMax,CFlowAct **pCFAvar,long *plGetData)
{ long lX;
  CMultFlow *pCMFvar;

  *plRptMax = 1;
  switch(lFlowPhase)
  { case 0:
    { *pCFAvar = pFlowObjVar->m_pFlowItem->m_pCFA_Initialize;
      *plGetData = pFlowObjVar->m_pFlowItem->m_lUseDataName_Initialize;
//==! Look for the INITIALIZE subheader in the data file only if the
//==! initialize part of the flow requires data.
      if (pFlowObjVar->m_pFlowItem->m_lUseDataName_Initialize > 0)
      { pCDDvar->m_pCIscData->readIscDataSubHeader(
                       pFlowObjVar->m_pFlowItem->m_csDataName,"initialize",0,
                       &(pCDDvar->m_CMemIscDataContext));
      } // IF
      break;
    } // CASE 0
    case 1:
    { pCMFvar = pFlowObjVar->m_pFlowItem->m_pCMF_Repeat;
      if (pCMFvar == NULL)
      { *pCFAvar = NULL;
        *plGetData = 0;
        *plRptMax = 0;
      } // IF
      else
      { for (lX = 0; lX < lRptFlowCount; ++lX)
        { if (pCMFvar == NULL)
          { *pCFAvar = NULL;
            *plGetData = 0;
            *plRptMax = 0;
            break;
          } // IF
          *pCFAvar = pCMFvar->m_pCFA_Repeat;
          *plGetData = pCMFvar->m_lUseDataName_Repeat;
          *plRptMax = pCMFvar->m_lRptCnt;
          pCMFvar = pCMFvar->m_pNextCMultFlow;
        } // FOR lX < lRptFlowCount
      } // ELSE
//==! Look for the REPEAT subheader in the data file only if the
//==! repeat part of the flow requires data.
      if (*plGetData > 0)
      { pCDDvar->m_pCIscData->readIscDataSubHeader(
                       pFlowObjVar->m_pFlowItem->m_csDataName,"repeat",
                       lRptFlowCount,&(pCDDvar->m_CMemIscDataContext));
      } // IF
      break;
    } // CASE 1
    case 2:
    default:
    { *pCFAvar = pFlowObjVar->m_pFlowItem->m_pCFA_Terminate;
      *plGetData = pFlowObjVar->m_pFlowItem->m_lUseDataName_Terminate;
//==! Look for the TERMINATE subheader in the data file only if the
//==! terminate part of the flow requires data.
      if (pFlowObjVar->m_pFlowItem->m_lUseDataName_Terminate > 0)
      { pCDDvar->m_pCIscData->readIscDataSubHeader(
                       pFlowObjVar->m_pFlowItem->m_csDataName,"terminate",0,
                       &(pCDDvar->m_CMemIscDataContext));
      } // IF
      break;
    } // CASE 2/default
  } // SWITCH (lFlowPhase)

} // getFlowActivityListInfo()
//==!=====================================================================!==
//==! Process the procedure.
//==!
//==! Function Parameters: pCDD = pointer to device list
//==!                      pCDDobj = device being processed
//==!                      pCIscProcObj = pointer to procedure
//==! Return Value       : 1 if successful, 0 otherwise
//==!=====================================================================!==
int processProcedure(CDeviceData *pCDD,CDeviceData *pCDDobj,
                     CIscProc *pCIscProcObj)
{ CxFlowObj *pFlowObjVar;
  long lX,lY,lZ,lRptMax,lGetData;
  CFlowAct *pCFAptr,*pCFAvar;
  int iFlowActStat,iDataAvail,iFlag0;

//==! Process flows in this procedure.
  for (pFlowObjVar = pCIscProcObj->m_pProcFlows;
       pFlowObjVar != NULL;
       pFlowObjVar = pFlowObjVar->m_pNextCxFlowObj)
  { sprintf(sMsgBuf00,"\n      ==> %s(%s)",
            pFlowObjVar->m_pFlowItem->m_csFlowName,
            pFlowObjVar->m_csFlowParam != NULL
              ? pFlowObjVar->m_csFlowParam : " ");
    debugMsg(sMsgBuf00,1,1);

//==! Check if data name is used in this flow. If > 0, then position the
//==! dataport pointer to the start of the data array name before executing
//==! the flows.
    if (pFlowObjVar->m_pFlowItem->m_lUseDataName > 0)
    { if (pCDDobj->m_pCIscData->readIscDataHeader(1,pCDD,
                        pCDDobj->m_pCBsdlData,
                        pFlowObjVar->m_pFlowItem->m_csDataName,
                        &(pCDDobj->m_CMemIscDataContext)) == 0)
      { continue;
      } // IF
    } // IF

    if (pCDDobj->m_iIscDataFormat == 1)
    { if (checkFlowObj(pFlowObjVar,-1,0) == 1)
      { pCDDobj->writeIscData(2,pFlowObjVar->m_pFlowItem->m_csDataName,-1);
      } // IF
    } // IF

    for (lX = 0,lZ = 0; lX < 3;)
    { if (lX == 1)
      { ++lZ;
      } // IF
      else
      { lZ = 0;
      } // ELSE
      getFlowActivityListInfo(pCDDobj,lX,lZ,pFlowObjVar,&lRptMax,
                              &pCFAptr,&lGetData);

      if (pCFAptr == NULL)
      { ++lX;    // NOTE_0014
        continue;
      } // IF

      if (checkFlowObj(pFlowObjVar,lX,lZ) == 1)
      { pCDDobj->writeIscData(3,NULL,lX);
      } // IF

      if (lRptMax > 1)
      { sprintf(sMsgOnly,"\n\t=> repeat flow max %ld times\n",lRptMax);
        debugMsg(sMsgOnly,0,1);
      } // IF

      for (lY = 0,iDataAvail = 0; lY < lRptMax; ++lY)
      { if ((iOpt_RptCntDispOff == 0) && (lRptMax > 1) && (lY%100 == 0))
        { sprintf(sMsgBuf00,"\r  ==> %ld\t",lY);
          debugMsg(sMsgBuf00,0,1);
        } // IF lRptMax > 1
        for (pCFAvar = pCFAptr;
             pCFAvar != NULL;
             pCFAvar = pCFAvar->m_pNextCFlowAct)
        { iFlowActStat = processFlowActivity(pCDDobj,pCFAvar,
                         pFlowObjVar->m_pFlowItem,&iDataAvail,lY,0);
          if (iFlowActStat == 1)
          { if (pFlowObjVar->m_pFlowItem != NULL)
            { if (pFlowObjVar->m_pFlowItem->m_iExitOnError == 1)
              { sprintf(sMsgBuf00,"\n      ==> flow %s(%s) has exit_on_error",
                        pFlowObjVar->m_pFlowItem->m_csFlowName,
                        pFlowObjVar->m_csFlowParam != NULL
                           ? pFlowObjVar->m_csFlowParam : " ");
                if (iGlobOptSkipExitOnError == 0)
                { iGlobExitOnError = 1;
                } // IF
                else
                { strcat(sMsgBuf00," (skipped)");
                } // ELSE
                debugMsg(sMsgBuf00,1,1);
              } // IF
            } // IF
            return 0;
          } // IF
        } // FOR pCFAvar != NULL

        iFlag0 = 0;
        if ((lY == (lRptMax-1)) && (lGetData > 0))
        { if (iGlobOptUseMemIscData == 0)
          { pCDDobj->m_pCIscData->m_pCDP_IscData->iParserGetChar = 1;
            if (pCDDobj->m_pCIscData->m_pCDP_IscData->iCurChar != ')')
            { iFlag0 = 1;
            } // IF
          } // IF iGlobOptUseMemIscData == 0
          else
          { if (pCDDobj->m_CMemIscDataContext.m_pCMIDR_current != NULL)
            { iFlag0 = 1;
            } // IF
          } // ELSE
        } // IF

        if (iFlag0 == 1)
        { jMessage(DISP_AND_EXIT,
                   "Data block has too many data records",
                   1,pCDDobj->m_pCIscData->m_pCDP_IscData);
        } // IF iFlag0 = 1

        if (lGetData > 0 && iDataAvail == 2)
        { ++lY;
          break;
        } // IF
      } // FOR lY < lRptMax

//==! Print terminal count only for lRptMax > 1
      if (lRptMax > 1)
      { sprintf(sMsgBuf00,"\r\t   - Processed %ld of %ld repeat steps\t",
                lY,lRptMax);
        debugMsg(sMsgBuf00,0,1);
      } // IF

      if (checkFlowObj(pFlowObjVar,lX,lZ) == 1)
      { pCDDobj->writeIscData(3,NULL,-2);
      } // IF

//==! Move to the next phase if the current phase is not the repeat phase.
//==! The only condition that can cause function to move from repeat phase to
//==! the next (i.e., terminate) phase is marked by NOTE_0014.
      if (lX != 1)
      { ++lX;
      } // IF

    } // FOR lX < 3

    if (checkFlowObj(pFlowObjVar,-1,0) == 1)
    { pCDDobj->writeIscData(4,NULL,0);
    } // IF

  } // FOR pFlowObjVar != NULL

  return 1;

} // processProcedure()
//==!=====================================================================!==
//==! Process the specified action.
//==!
//==! Function Parameters: pCDD = pointer to device list
//==!                      pCDDobj = device being processed
//==! Return Value       : 1 if successful, 0 otherwise
//==!=====================================================================!==
int processAction(CDeviceData *pCDD,CDeviceData *pCDDobj)
{ CIscAct *pCIAvar;
  CxProcObj *pProcObjVar;
  char sMode[8];
  int iSkipFlag;

  pCIAvar = findIscAction(pCDDobj);

  if (pCIAvar == NULL)
  { return 0;
  } // IF

  if (pCDDobj->m_pCDP_Out != NULL)
  { sprintf(sMode,"%s",iOptOverWritePrompt == 1 ? "c" : "w");
    pCDDobj->m_pCDP_Out->pFILE =
        xFileOpen(pCDDobj->m_pCDP_Out->psPortName,sMode); //==! NOTE_0007
    pCDDobj->writeIscData(0,NULL,-1);
  } // IF

//==! Process procedures in this action.
  for (pProcObjVar = pCIAvar->m_pActionProcs;
       pProcObjVar != NULL;
       pProcObjVar = pProcObjVar->m_pNextCxProcObj)
  { iGlobProgramState = dPROGSTATE_PROCESSACTION;

    checkProcedureProcessingStatus(pCDDobj,pProcObjVar,&iSkipFlag);
    if (iSkipFlag == 1)
    { continue;
    } // IF

//==! Process flows in this procedure.
    if (processProcedure(pCDD,pCDDobj,pProcObjVar->m_pProcItem) == 0)
    { break;
    } // IF
  } // FOR pProcObjVar != NULL

//==! If pProcObjVar != NULL then not all procedures processed successfully
  if (pProcObjVar != NULL)
  { if (iGlobOptUseErrorProc == 1)
    { if (pCDDobj->m_pCBsdlData != NULL)
      { if (pCDDobj->m_pCBsdlData->m_pCIscProcError == NULL)
        { sprintf(sMsgOnly,"\nPROC_ERROR_EXIT not defined for [%s]",
                  pCDDobj->m_csDeviceName);
          jMessage(DISP_SHOW_MSG,sMsgOnly,0,NULL);
        } // IF
        else
        { debugMsg("\n-----> Executing PROC_ERROR_EXIT -----",1,1);
          if (processProcedure(pCDD,pCDDobj,
                         pCDDobj->m_pCBsdlData->m_pCIscProcError) == 0)
          { sprintf(sMsgOnly,"PROC_ERROR_EXIT for [%s] unsuccessful",
                    pCDDobj->m_csDeviceName);
            jMessage(DISP_SHOW_MSG,sMsgOnly,0,NULL);
          } // IF processProcedure() == 0
        } // ELSE
      } // IF pCDDobj->m_pCBsdlData != NULL
    } // IF iGlobOptUseErrorProc == 1
  } // IF pProcObjVar != NULL

  if (pCDDobj->m_pCIscData != NULL)
  { if (pCDDobj->m_pCIscData->m_pCDP_IscData != NULL)
    { pCDDobj->m_pCIscData->m_pCDP_IscData->xCloseDataPort();
    } // IF
  } // IF

  if (pCDDobj->m_pCDP_Out != NULL)
  { pCDDobj->writeIscData(1,NULL,-1);
    xFclose(&(pCDDobj->m_pCDP_Out->pFILE));
  } // IF

  if (pProcObjVar == NULL)
  { return 1;
  } // IF
  else
  { return 0;
  } // ELSE

} // processAction()
//==!=====================================================================!==
//==! Get the maximum minimum wait time and the minimum maximum wait time.
//==!
//==! Function Parameters: pCDD = pointer to current device
//==!                      pCFW = pointer to wait object that will hold the
//==!                         max min wait time and min max wait time
//==!                      pCFA = pointer to CFlowAct obj containing wait
//==!                         time for the flow activity
//==! Return Value       : none
//==!=====================================================================!==
void getMaxMin_MinMax_Wait(CDeviceData *pCDD,CFlowWait *pCFW,CFlowAct *pCFA)
{ int iFlag;

  iFlag = 0;
  if (pCFA == NULL)
  { return;
  } // IF
  if (pCFA->m_pFlowWaitInfo == NULL)
  { return;
  } // IF
  if (pCFA->m_pFlowWaitInfo->m_lMinFlowWaitCycles != 0)
  { if ((pCFW->m_lMinFlowWaitCycles == 0) ||
        ((pCFW->m_lMinFlowWaitCycles != 0) &&
         (pCFA->m_pFlowWaitInfo->m_lMinFlowWaitCycles >
          pCFW->m_lMinFlowWaitCycles)
        )
       )
    { pCFW->m_lMinFlowWaitCycles = pCFA->m_pFlowWaitInfo
                                       ->m_lMinFlowWaitCycles;
    } // IF
  } // IF pCFA->m_pFlowWaitInfo->m_lMinFlowWaitCycles != 0

  if (pCFA->m_pFlowWaitInfo->m_lMaxFlowWaitCycles != 0)
  { if ((pCFW->m_lMaxFlowWaitCycles == 0) ||
        ((pCFW->m_lMaxFlowWaitCycles != 0) &&
         (pCFA->m_pFlowWaitInfo->m_lMaxFlowWaitCycles <
          pCFW->m_lMaxFlowWaitCycles)
        )
       )
    { pCFW->m_lMaxFlowWaitCycles = pCFA->m_pFlowWaitInfo
                                       ->m_lMaxFlowWaitCycles;
    } // IF
  } // IF pCFA->m_pFlowWaitInfo->m_lMaxFlowWaitCycles != 0

  if (pCFA->m_pFlowWaitInfo->m_fMinFlowWaitTimeSec != 0)
  { if ((pCFW->m_fMinFlowWaitTimeSec == 0) ||
        ((pCFW->m_fMinFlowWaitTimeSec != 0) &&
         (pCFA->m_pFlowWaitInfo->m_fMinFlowWaitTimeSec >
          pCFW->m_fMinFlowWaitTimeSec)
        )
       )
    { pCFW->m_fMinFlowWaitTimeSec = pCFA->m_pFlowWaitInfo
                                       ->m_fMinFlowWaitTimeSec;
    } // IF
  } // IF pCFA->m_pFlowWaitInfo->m_fMinFlowWaitTimeSec != 0

  if (pCFA->m_pFlowWaitInfo->m_fMaxFlowWaitTimeSec != 0)
  { if ((pCFW->m_fMaxFlowWaitTimeSec == 0) ||
        ((pCFW->m_fMaxFlowWaitTimeSec != 0) &&
         (pCFA->m_pFlowWaitInfo->m_fMaxFlowWaitTimeSec <
          pCFW->m_fMaxFlowWaitTimeSec)
        )
       )
    { pCFW->m_fMaxFlowWaitTimeSec = pCFA->m_pFlowWaitInfo
                                       ->m_fMaxFlowWaitTimeSec;
    } // IF
  } // IF pCFA->m_pFlowWaitInfo->m_fMaxFlowWaitTimeSec != 0

  if (pCFW->m_lMinFlowWaitCycles != 0 &&
      pCFW->m_lMaxFlowWaitCycles != 0)
  { if (pCFW->m_lMinFlowWaitCycles > pCFW->m_lMaxFlowWaitCycles)
    { sprintf(sMsgOnly,
   "\n*** ==> Wait TCK cycle max [%ld] < min [%ld] (concurrent op)",
              pCFW->m_lMaxFlowWaitCycles,pCFW->m_lMinFlowWaitCycles);
      jMessage(DISP_SHOW_MSG,sMsgOnly,0,NULL);
      iFlag = 1;
    } // IF
  } // IF
  if (pCFW->m_fMinFlowWaitTimeSec != 0 &&
      pCFW->m_fMaxFlowWaitTimeSec != 0)
  { if (pCFW->m_fMinFlowWaitTimeSec > pCFW->m_fMaxFlowWaitTimeSec)
    { sprintf(sMsgOnly,
    "\n*** ==> Wait time max [%.2f] < min time [%.2f] secs (concurrent op)",
              pCFW->m_fMaxFlowWaitTimeSec,pCFW->m_fMinFlowWaitTimeSec);
      jMessage(DISP_SHOW_MSG,sMsgOnly,0,NULL);
      iFlag = 1;
    } // IF
  } // IF

  if (iFlag == 1)
  { sprintf(sMsgOnly,
  "\n      = see wait time for device %s (action [%s %s])",
            pCDD->m_csDeviceName,pCDD->m_csAction,
            (pCDD->m_csActionData != NULL ? pCDD->m_csActionData : " "));
    jMessage(DISP_SHOW_MSG,sMsgOnly,0,NULL);
  } // IF iFlag == 1

} // getMaxMin_MinMax_Wait()
//==!=====================================================================!==
//==! Assemble data from the selected device registers in the pCDDlist into a
//==! single register pcOneReg. If iFlowType == -1, then assemble data from
//==! all the devices that have flows. If iFlowType == [0..3], then assemble
//==! data only from devices in the pCDD list that have this flow type. The
//==! other devices will be put into bypass.
//==!
//==! Function Parameters: pCDD = pointer to device list
//==!                      iRegSel = device register selector
//==!                      pcOneReg = pointer to register array (must be
//==!                            large enough to hold the assembled data)
//==!                      iFlowType = current device flow activity type
//==! Return Value       : assembled data size
//==!=====================================================================!==
long assembleData(CDeviceData *pCDDlist,int iRegSel,char *pcOneReg,
                  int iFlowType)
{ CDeviceData *pCDDvar;
  long lTotLength,lDataRegLength,lX;

  for (pCDDvar = pCDDlist,pcOneReg[0] = '\0',lTotLength = 0; pCDDvar != NULL;
       pCDDvar = pCDDvar->m_pNextCDDfromTdi)
  { if (pCDDvar->m_iSkipDev == 1)
    { continue;
    } // IF
    switch(iRegSel)
    { case 0:  // assemble instruction opcodes
      { if (iFlowType == -1 || iFlowType == pCDDvar->m_iHoldFlowType)
        { strcat(pcOneReg,pCDDvar->m_cSIR);
        } // IF
        else
        { strcat(pcOneReg,pCDDvar->m_cSIR_bypass);
        } // ELSE
        lTotLength += pCDDvar->m_lInstrRegLength;
        break;
      } // CASE 0
      case 1:  // assemble TDI data
      { if ((pCDDvar->m_pCProcContext->m_iBypass == 1) ||
            ((iFlowType != -1) && (iFlowType != pCDDvar->m_iHoldFlowType))
           )
        { strcat(pcOneReg,"1");   // add bypass register containing '1'
          ++lTotLength;
        } // IF
        else
        { if (pCDDvar->m_pCProcContext->m_pCFAloopVar != NULL)
          { lDataRegLength = getDataRegWidth(pCDDvar->m_pCBsdlData,
                    pCDDvar->m_pCProcContext->m_pCFAloopVar->m_lOpcodeIndex);

          } // IF
          else
          { lDataRegLength = getDataRegWidth(pCDDvar->m_pCBsdlData,
                    pCDDvar->m_pCProcContext->m_pCFAvar->m_lOpcodeIndex);
          } // ELSE
          lTotLength += lDataRegLength;

//==! Check if device's data register has all 0s or 1s. If not, then the
//==! register was probably used by another instruction which used it to
//==! store fewer bits. For example, if IDCODE was used previously to load 32
//==! bits but then an ISC_DATA_SHIFT instruction is used but we want to just
//==! load a 4096-bit long don't care value, then the following FOR loop will
//==! check the register to see if there were put 4096 '0's or '1's,
//==! otherwise it will put '1's in those positions and terminate the array
//==! with a '\0'.
          for (lX = 0; lX < lDataRegLength; ++lX)
          { if (pCDDvar->m_cTdi[lX] != '0' && pCDDvar->m_cTdi[lX] != '1')
            { pCDDvar->m_cTdi[lX] = '1';
            } // IF
          } // FOR lX < lDataRegLength
          pCDDvar->m_cTdi[lX] = '\0';
          strcat(pcOneReg,pCDDvar->m_cTdi);
        } // ELSE
        break;
      } // CASE 1
    } // SWITCH (iRegSel)
  } // FOR pCDDvar != NULL

  return lTotLength;

} // assembleData()
//==!=====================================================================!==
//==! Process one flow action. iMode determines what part of the flow
//==! activity to perform:
//==! iMode == 0: do the flow activity data preparation steps
//==!             - get instruction opcode
//==!             - get update data (if specified)
//==!             - get wait time
//==!          1: evaluate capture field using TDO data
//==!
//==! The iDataStatus variable pointer returns a value that is interpreted as
//==! follows:
//==! - 1: if there is still data available in the ISC data file
//==! - 2: if no more data in ISC data file
//=!
//==! plTdoIndex is pointer to index into cTdoData[] to get TDO data to be
//==! evaluated when performing device's processCaptureField().
//==!
//==! Function Parameters: pCDDobj = device being processed
//==!                      pCFWobj = pointer to wait data structure
//==!                      pCFAobj = flow activity object being processed
//==!                      pCIFobj = pointer to CIscFlow object containing
//==!                         the variables used by this flow activity
//==!                      iDataStatus = variable pointer to return value
//==!                         (see function description)
//==!                         - return 1 if there is still data available in
//==!                           the ISC data file
//==!                         - return 2 if no more data in ISC data file
//==!                      lCurRptCount = current repeat count (i.e., how
//==!                         many times this flow was processed)
//==!                      plUpdate = pointer to variable to indicate whether
//==!                         an update operation needs to be performed
//==!                      plCapture = pointer to variable to indicate whether
//==!                         a capture operation needs to be performed
//==!                      plTdoIndex = see function description
//==! Return Value       : 0 or 1 as described above
//==!=====================================================================!==
int processOneFlowActivity(CDeviceData *pCDDobj,CFlowWait *pCFWobj,
                        CFlowAct *pCFAobj,CIscFlow *pCIFobj,int *iDataStatus,
                        long lCurRptCount,
                        long *plUpdate,long *plCapture,long *plTdoIndex)
{
  iGlobProgramState = dPROGSTATE_PROCESSFLOWACTIVITY;

  pCDDobj->m_iFlowType = 2;

  if (pCFAobj->m_pLoopActions != NULL)
  { if (pCDDobj->m_pCProcContext->m_iPhase == 0)
    { pCDDobj->m_pCProcContext->m_lLoopMin = pCFAobj->m_lMinLoopCount;
      pCDDobj->m_pCProcContext->m_lLoopMax = pCFAobj->m_lMaxLoopCount;
      if (iOpt_FlowLoopDbg == 1)
      { sprintf(sMsgOnly,"\n\t  --> [%s] loop max: %ld times",
                pCDDobj->m_csDeviceName,
                pCDDobj->m_pCProcContext->m_lLoopMax);
        debugMsg(sMsgOnly,1,1);
      } // IF iOpt_FlowLoopDbg == 1
      pCDDobj->m_pCProcContext->m_lLoopVar = 0;
    } // IF pCDDobj->m_pCProcContext->m_iPhase == 0

    for (; pCDDobj->m_pCProcContext->m_lLoopVar <
                  pCDDobj->m_pCProcContext->m_lLoopMax;
         ++(pCDDobj->m_pCProcContext->m_lLoopVar))
    { if (pCDDobj->m_pCProcContext->m_iPhase == 0)
      { if (iOpt_FlowLoopDbg == 1)
        { sprintf(sMsgBuf03,"\n== [%s] Flow loop iteration [%ld] --",
                  pCDDobj->m_csDeviceName,
                  pCDDobj->m_pCProcContext->m_lLoopVar);
          debugMsg(sMsgBuf03,1,1);
        } // IF
        pCDDobj->m_pCProcContext->m_pCFAloopVar = pCFAobj->m_pLoopActions;
        pCDDobj->m_pCProcContext->m_l_OST = 0;
      } // IF pCDDobj->m_pCProcContext->m_iPhase == 0
      for (; pCDDobj->m_pCProcContext->m_pCFAloopVar != NULL;
           pCDDobj->m_pCProcContext->m_pCFAloopVar =
                pCDDobj->m_pCProcContext->m_pCFAloopVar->m_pNextCFlowAct)
      { if (pCDDobj->m_pCProcContext->m_iPhase == 0)
        { sprintf(sMsgOnly,"\n\t- opcode [%s: %s (Loop %ld | Rpt %ld)]",
                  pCDDobj->m_csDeviceName,
                  pCDDobj->getOpcodeInfo(
                    pCDDobj->m_pCProcContext->m_pCFAloopVar->m_lOpcodeIndex,
                    pCDDobj->m_cSIR),
                  pCDDobj->m_pCProcContext->m_lLoopVar,lCurRptCount);
          if (iOpt_ProgDebug == 1)
          { debugMsg(sMsgOnly,2,1);
          } // IF
          *plUpdate += pCDDobj->m_pCProcContext->m_pCFAloopVar
                              ->processUpdateField(1,
                           pCDDobj->m_cTdi,pCDDobj->m_cTdo,
                           pCDDobj,pCDDobj->m_pCProcContext->m_pFlowObjVar
                                          ->m_pFlowItem,
                           &(pCDDobj->m_pCProcContext->m_iDataAvail));
          getMaxMin_MinMax_Wait(pCDDobj,pCFWobj,
                                pCDDobj->m_pCProcContext->m_pCFAloopVar);
          *plCapture += (pCDDobj->m_pCProcContext->m_pCFAloopVar
                                ->m_pCF_Capture != NULL ? 1 : 0);
          pCDDobj->m_pCProcContext->m_iPhase = 1;

          pCDDobj->m_iFlowType =
                pCDDobj->m_pCProcContext->m_pCFAloopVar->getFlowType();
          return 0;
        } // pCDDobj->m_pCProcContext->m_iPhase == 0
        else
        { *plTdoIndex -= getDataRegWidth(pCDDobj->m_pCBsdlData,
                     pCDDobj->m_pCProcContext->m_pCFAloopVar->m_lOpcodeIndex);
          pCDDobj->m_pCProcContext->m_l_OST +=
               pCDDobj->m_pCProcContext->m_pCFAloopVar
                      ->processCaptureField(1,cTdiData,
                      &(cTdoData[*plTdoIndex]),pCDDobj,pCIFobj,
                      lCurRptCount,&(*iDataStatus),1);
          pCDDobj->m_pCProcContext->m_iPhase = 0;
        } // ELSE
      } // FOR pCDDobj->m_pCProcContext->m_pCFAloopVar != NULL
      if ((pCDDobj->m_pCProcContext->m_lLoopVar >=
                pCDDobj->m_pCProcContext->m_lLoopMin) &&
          (pCDDobj->m_pCProcContext->m_l_OST == 0))
      { break;
      } // IF
      pCDDobj->m_pCProcContext->m_iPhase = 0;
    } // FOR pCDDobj->m_pCProcContext->m_lLoopVar < ->m_lLoopMax
    if ((pCDDobj->m_pCProcContext->m_lLoopVar >=
         pCDDobj->m_pCProcContext->m_lLoopMax) &&
        (pCDDobj->m_pCProcContext->m_l_OST > 0))
    { sprintf(sMsgOnly,"\n*** ==> [%s]: max flow loop count [%ld] reached",
              pCDDobj->m_csDeviceName,pCDDobj->m_pCProcContext->m_lLoopMax);
      jMessage(DISP_SHOW_MSG,sMsgOnly,0,NULL);
      pCDDobj->m_pCProcContext->m_iBypass = 1;
      return 1;
    } // IF
    if (iOpt_FlowLoopDbg == 1)
    { sprintf(sMsgOnly,"\n\t  --> [%s] loop iterations: %ld of max %ld",
              pCDDobj->m_csDeviceName,pCDDobj->m_pCProcContext->m_lLoopVar,
              pCDDobj->m_pCProcContext->m_lLoopMax);
      debugMsg(sMsgOnly,1,1);
    } // IF iOpt_FlowLoopDbg == 1
    return 0;
  } // IF pCFAobj->m_pLoopActions != NULL

  if (pCDDobj->m_pCProcContext->m_iPhase == 0)
  { sprintf(sMsgOnly,"\n\t- opcode [%s: %s (Rpt %ld)]",
            pCDDobj->m_csDeviceName,
            pCDDobj->getOpcodeInfo(pCFAobj->m_lOpcodeIndex,pCDDobj->m_cSIR),
            lCurRptCount);
    if (iOpt_ProgDebug == 1)
    { debugMsg(sMsgOnly,2,1);
    } // IF

    *plUpdate += pCFAobj->processUpdateField(1,
                     pCDDobj->m_cTdi,pCDDobj->m_cTdo,
                     pCDDobj,pCDDobj->m_pCProcContext->m_pFlowObjVar
                                    ->m_pFlowItem,
                     &(pCDDobj->m_pCProcContext->m_iDataAvail));
    getMaxMin_MinMax_Wait(pCDDobj,pCFWobj,pCFAobj);
    *plCapture += (pCFAobj->m_pCF_Capture != NULL ? 1 : 0);
    pCDDobj->m_pCProcContext->m_iPhase = 1;

    pCDDobj->m_iFlowType = pCFAobj->getFlowType();
    return 0;
  } // IF pCDDobj->m_pCProcContext->m_iPhase == 0
  else
  { *plTdoIndex -= getDataRegWidth(pCDDobj->m_pCBsdlData,
             pCFAobj->m_lOpcodeIndex);
    pCDDobj->m_pCProcContext->m_iPhase = 0;
    return pCFAobj->processCaptureField(1,cTdiData,
                  &(cTdoData[*plTdoIndex]),pCDDobj,pCIFobj,
                  lCurRptCount,&(*iDataStatus),0);
  } // ELSE

} // processOneFlowActivity()
//==!=====================================================================!==
//==! Check if CDeviceData list has at least one flow of type iTargetFlowType.
//==! If so then return 1 else return 0.
//==!
//==! Function Parameters: pCDD = pointer to device list
//==!                      iTargetFlowType = flow type to look for
//==! Return Value       : see function description
//==!=====================================================================!==
int findFlowType(CDeviceData *pCDD,int iTargetFlowType)
{ CDeviceData *pCDDvar;

  for (pCDDvar = pCDD;
       pCDDvar != NULL;
       pCDDvar = pCDDvar->m_pNextCDeviceData)
  { if (pCDDvar->m_iSkipDev == 1 ||
        pCDDvar->m_pCProcContext->m_iBypass == 1)
    { continue;
    } // IF
    if (pCDDvar->m_iHoldFlowType == iTargetFlowType)
    { return 1;
    } // IF
  } // FOR pCDDvar != NULL
  return 0;

} // findFlowType()
//==!=====================================================================!==
//==! Check if current flow for parts in device list all have the same flow
//==! type. If so then return 1 else return 0.
//==!
//==! Function Parameters: pCDD = pointer to device list
//==!                      piFlowType = pointer to return common flow type
//==! Return Value       : 1 if parts all have the same flow type, else 0
//==!=====================================================================!==
int isSameFlowType(CDeviceData *pCDD,int *piFlowType)
{ CDeviceData *pCDDvar;

  for (pCDDvar = pCDD,*piFlowType = -1;
       pCDDvar != NULL;
       pCDDvar = pCDDvar->m_pNextCDeviceData)
  { if (pCDDvar->m_iSkipDev == 1 ||
        pCDDvar->m_pCProcContext->m_iBypass == 1)
    { continue;
    } // IF
    if (*piFlowType == -1)
    { *piFlowType = pCDDvar->m_iFlowType;
    } // IF
    else
    { if (*piFlowType != pCDDvar->m_iFlowType)
      { return 0;
      } // IF
    } // ELSE
  } // FOR pCDDvar != NULL
  return 1;

} // isSameFlowType()
//==!=====================================================================!==
//==! Process an action one step at a time. lTdoIndex specifies index into
//==! cTdoData[] to get the TDO data to be used by the device when processing
//==! capture data in processOneFlowActivity().
//==!
//==! Function Parameters: pCDD = pointer to device list
//==!                      pCDDobj = device being processed
//==!                      pCFWobj = pointer to wait time object
//==!                      plUpdate = pointer to long var; if non-0 then
//==!                              an update operation must be performed
//==!                      plCapture = pointer to long var; if non-0 then
//==!                              a capture operation must be performed
//==!                      plTdoIndex = pointer to index variable for
//==!                              cTdoData when evaluating captured data
//==!                              (see function description)
//==! Return Value       : 1 if successful, 0 otherwise
//==!=====================================================================!==
int processOneDevAction(CDeviceData *pCDD,CDeviceData *pCDDobj,
          CFlowWait *pCFWobj,long *plUpdate,long *plCapture,long *plTdoIndex)
{
//==! Process procedures in this action.
  if (pCDDobj->m_pCProcContext->m_iPhase == 0)
  { pCDDobj->m_pCProcContext->m_pProcObj =
       pCDDobj->m_pCProcContext->m_pCIAvar->m_pActionProcs;
  } // IF pCDDobj->m_pCProcContext->m_iPhase == 0
  for (; pCDDobj->m_pCProcContext->m_pProcObj != NULL;
         pCDDobj->m_pCProcContext->m_pProcObj =
             pCDDobj->m_pCProcContext->m_pProcObj->m_pNextCxProcObj)
  { iGlobProgramState = dPROGSTATE_PROCESSACTION;
    if (pCDDobj->m_pCProcContext->m_iPhase == 0)
    { checkProcedureProcessingStatus(pCDDobj,
            pCDDobj->m_pCProcContext->m_pProcObj,
            &(pCDDobj->m_pCProcContext->m_iSkipFlag));
      if (pCDDobj->m_pCProcContext->m_iSkipFlag == 1)
      { continue;
      } // IF
      pCDDobj->m_pCProcContext->m_pFlowObjVar =
             pCDDobj->m_pCProcContext->m_pProcObj
                    ->m_pProcItem->m_pProcFlows;
    } // IF pCDDobj->m_pCProcContext->m_iPhase == 0

//==! Process flows in this procedure.
    for (; pCDDobj->m_pCProcContext->m_pFlowObjVar != NULL;
         pCDDobj->m_pCProcContext->m_pFlowObjVar =
              pCDDobj->m_pCProcContext->m_pFlowObjVar->m_pNextCxFlowObj)
    { if (pCDDobj->m_pCProcContext->m_iPhase == 0)
      { sprintf(sMsgBuf00,"\n      ==> [%s] %s(%s)",pCDDobj->m_csDeviceName,
                pCDDobj->m_pCProcContext->m_pFlowObjVar
                       ->m_pFlowItem->m_csFlowName,
                pCDDobj->m_pCProcContext->m_pFlowObjVar
                       ->m_csFlowParam != NULL
                  ? pCDDobj->m_pCProcContext->m_pFlowObjVar
                           ->m_csFlowParam : " ");
        debugMsg(sMsgBuf00,1,1);

//==! Check if data name is used in this flow. If > 0, then position the
//==! dataport pointer to the start of the data array name before executing
//==! the flows.
        if (pCDDobj->m_pCProcContext->m_pFlowObjVar
                  ->m_pFlowItem->m_lUseDataName > 0)
        { if (pCDDobj->m_pCIscData->readIscDataHeader(1,pCDD,
                            pCDDobj->m_pCBsdlData,
                            pCDDobj->m_pCProcContext->m_pFlowObjVar
                                   ->m_pFlowItem->m_csDataName,
                            &(pCDDobj->m_CMemIscDataContext)) == 0)
          { continue;
          } // IF
        } // IF
        pCDDobj->m_pCProcContext->m_lX = 0;
        pCDDobj->m_pCProcContext->m_lZ = 0;

        if (pCDDobj->m_iIscDataFormat == 1)
        { if (checkFlowObj(pCDDobj->m_pCProcContext->m_pFlowObjVar,-1,0) == 1)
          { pCDDobj->writeIscData(2,pCDDobj->m_pCProcContext->m_pFlowObjVar
                                           ->m_pFlowItem->m_csDataName,-1);
          } // IF
        } // IF
      } // IF pCDDobj->m_pCProcContext->m_iPhase == 0

      for (; pCDDobj->m_pCProcContext->m_lX < 3;)
      { if (pCDDobj->m_pCProcContext->m_iPhase == 0)
        { if (pCDDobj->m_pCProcContext->m_lX == 1)
          { ++(pCDDobj->m_pCProcContext->m_lZ);
          } // IF
          else
          { pCDDobj->m_pCProcContext->m_lZ = 0;
          } // ELSE
          getFlowActivityListInfo(pCDDobj,
                 pCDDobj->m_pCProcContext->m_lX,
                 pCDDobj->m_pCProcContext->m_lZ,
                 pCDDobj->m_pCProcContext->m_pFlowObjVar,
                 &(pCDDobj->m_pCProcContext->m_lRptMax),
                 &(pCDDobj->m_pCProcContext->m_pCFAptr),
                 &(pCDDobj->m_pCProcContext->m_lGetData));

          if (pCDDobj->m_pCProcContext->m_pCFAptr == NULL)
          { ++(pCDDobj->m_pCProcContext->m_lX);  // NOTE_0015
            continue;
          } // IF

          if (checkFlowObj(pCDDobj->m_pCProcContext->m_pFlowObjVar,
                           pCDDobj->m_pCProcContext->m_lX,
                           pCDDobj->m_pCProcContext->m_lZ) == 1)
          { pCDDobj->writeIscData(3,NULL,pCDDobj->m_pCProcContext->m_lX);
          } // IF

          if (pCDDobj->m_pCProcContext->m_lRptMax > 1)
          { sprintf(sMsgOnly,"\n\t=> [%s] repeat flow max %ld times\n",
                    pCDDobj->m_csDeviceName,
                    pCDDobj->m_pCProcContext->m_lRptMax);
            debugMsg(sMsgOnly,0,1);
          } // IF

          pCDDobj->m_pCProcContext->m_lY = 0;
          pCDDobj->m_pCProcContext->m_iDataAvail = 0;
        } // IF pCDDobj->m_pCProcContext->m_iPhase == 0

        for (; pCDDobj->m_pCProcContext->m_lY <
                    pCDDobj->m_pCProcContext->m_lRptMax;
             ++(pCDDobj->m_pCProcContext->m_lY))
        { if (pCDDobj->m_pCProcContext->m_iPhase == 0)
          { if ((pCDDobj->m_pCProcContext->m_lRptMax > 1)
                && ((pCDDobj->m_pCProcContext->m_lY%100 == 0) ||
                   ((pCDDobj->m_pCProcContext->m_lY + 1) >=
                     pCDDobj->m_pCProcContext->m_lRptMax))
                && (iOpt_RptCntDispOff == 0))
            { sprintf(sMsgBuf00,"\r  ==> ( %ld )",
                      pCDDobj->m_pCProcContext->m_lY);
              debugMsg(sMsgBuf00,0,1);
            } // IF pCDDobj->m_pCProcContext->m_lRptMax > 1
            pCDDobj->m_pCProcContext->m_pCFAvar =
                    pCDDobj->m_pCProcContext->m_pCFAptr;
          } // IF pCDDobj->m_pCProcContext->m_iPhase == 0
          for (; pCDDobj->m_pCProcContext->m_pCFAvar != NULL;
               pCDDobj->m_pCProcContext->m_pCFAvar =
                    pCDDobj->m_pCProcContext->m_pCFAvar->m_pNextCFlowAct)
          { pCDDobj->m_pCProcContext->m_iFlowActStat =
                 processOneFlowActivity(pCDDobj,pCFWobj,
                     pCDDobj->m_pCProcContext->m_pCFAvar,
                     pCDDobj->m_pCProcContext->m_pFlowObjVar->m_pFlowItem,
                     &(pCDDobj->m_pCProcContext->m_iDataAvail),
                     pCDDobj->m_pCProcContext->m_lY,
                     &(*plUpdate),&(*plCapture),&(*plTdoIndex));
            if (pCDDobj->m_pCProcContext->m_iPhase == 1)
            { return 1;
            } // IF

            if (pCDDobj->m_pCProcContext->m_iFlowActStat == 1)
            { if (pCDDobj->m_pCDP_Out != NULL)
              { pCDDobj->writeIscData(1,NULL,-1);
                xFclose(&(pCDDobj->m_pCDP_Out->pFILE));
              } // IF
              if (pCDDobj->m_pCIscData != NULL)
              { if (pCDDobj->m_pCIscData->m_pCDP_IscData != NULL)
                { pCDDobj->m_pCIscData->m_pCDP_IscData->xCloseDataPort();
                } // IF
              } // IF
              pCDDobj->m_pCProcContext->m_iBypass = 1;
              pCDDobj->m_iProcessingStatus = 0;

              if (pCDDobj->m_pCProcContext->m_pFlowObjVar
                         ->m_pFlowItem->m_iExitOnError == 1)
              { sprintf(sMsgBuf00,
                        "\n      ==> [%s] flow %s(%s) has exit_on_error",
                        pCDDobj->m_csDeviceName,
                        pCDDobj->m_pCProcContext->m_pFlowObjVar
                               ->m_pFlowItem->m_csFlowName,
                        pCDDobj->m_pCProcContext->m_pFlowObjVar
                               ->m_csFlowParam != NULL
                           ? pCDDobj->m_pCProcContext->m_pFlowObjVar
                                    ->m_csFlowParam : " ");
                if (iGlobOptSkipExitOnError == 0)
                { iGlobExitOnError = 1;
                } // IF
                else
                { strcat(sMsgBuf00," (skipped)");
                } // ELSE
                debugMsg(sMsgBuf00,1,1);
              } // IF m_pFlowItem->m_iExitOnError == 1
              return 0;
            } // IF
          } // FOR pCDDobj->m_pCProcContext->m_pCFAvar != NULL

          pCDDobj->m_pCProcContext->m_iFlag0 = 0;
          if ((pCDDobj->m_pCProcContext->m_lY ==
                  (pCDDobj->m_pCProcContext->m_lRptMax-1)) &&
              (pCDDobj->m_pCProcContext->m_lGetData > 0))
          { if (iGlobOptUseMemIscData == 0)
            { pCDDobj->m_pCIscData->m_pCDP_IscData->iParserGetChar = 1;
              if (pCDDobj->m_pCIscData->m_pCDP_IscData->iCurChar != ')')
              { pCDDobj->m_pCProcContext->m_iFlag0 = 1;
              } // IF
            } // IF iGlobOptUseMemIscData == 0
            else
            { if (pCDDobj->m_CMemIscDataContext.m_pCMIDR_current != NULL)
              { pCDDobj->m_pCProcContext->m_iFlag0 = 1;
              } // IF
            } // ELSE
          } // IF

          if (pCDDobj->m_pCProcContext->m_iFlag0 == 1)
          { jMessage(DISP_AND_EXIT,
                     "Data block has too many data records",
                     1,pCDDobj->m_pCIscData->m_pCDP_IscData);
          } // IF pCDDobj->m_pCProcContext->m_iFlag0 = 1

          if (pCDDobj->m_pCProcContext->m_lGetData > 0 &&
              pCDDobj->m_pCProcContext->m_iDataAvail == 2)
          { ++(pCDDobj->m_pCProcContext->m_lY);
            break;
          } // IF
        } // FOR pCDDobj->m_pCProcContext->m_lY < ->lRptMax

//==! Print terminal count only for lRptMax > 1
        if (pCDDobj->m_pCProcContext->m_lRptMax > 1)
        { sprintf(sMsgBuf00,
                  "\n\t   - [%s] Processed %ld of %ld repeat steps",
                  pCDDobj->m_csDeviceName,
                  pCDDobj->m_pCProcContext->m_lY,
                  pCDDobj->m_pCProcContext->m_lRptMax);
          debugMsg(sMsgBuf00,0,1);
        } // IF

        if (checkFlowObj(pCDDobj->m_pCProcContext->m_pFlowObjVar,
                         pCDDobj->m_pCProcContext->m_lX,
                         pCDDobj->m_pCProcContext->m_lZ) == 1)
        { pCDDobj->writeIscData(3,NULL,-2);
        } // IF

//==! Move to the next phase if the current phase is not the repeat phase.
//==! The only condition that can cause function to move from repeat phase to
//==! the next (i.e., terminate) phase is marked by NOTE_0015.
        if (pCDDobj->m_pCProcContext->m_lX != 1)
        { ++(pCDDobj->m_pCProcContext->m_lX);
        } // IF

      } // FOR pCDDobj->m_pCProcContext->m_lX < 3

      if (checkFlowObj(pCDDobj->m_pCProcContext->m_pFlowObjVar,-1,0) == 1)
      { pCDDobj->writeIscData(4,NULL,0);
      } // IF

    } // FOR pCDDobj->m_pCProcContext->m_pFlowObjVar != NULL

  } // FOR pCDDobj->m_pCProcContext->m_pProcObjVar != NULL

  sprintf(sMsgOnly,
          "(*)-> Action [%s: %s(%s)] = [ DONE ]",
          pCDDobj->m_csDeviceName,pCDDobj->m_csAction,
          (pCDDobj->m_csActionData != NULL ? pCDDobj->m_csActionData : " "));
  dispMsg(sMsgOnly,DISP_PRE_LF,1,1);

  pCDDobj->m_pCProcContext->m_iBypass = 1;

  return 0;

} // processOneDevAction()
//==!=====================================================================!==
//==! Prepare and shift out data for flow type iTargetFlowType. If
//==! iTargetFlowType is -1, then combine all flow types together (i.e.,
//==! process them in parallel).
//==!
//==! Function Parameters: pCDD_Tdi_Tdo = pointer to device list arranged
//==!                         from TDI to TDO
//==!                      iFlowTypeVar = flow type to process
//==!                      lDoUpdateFlag > 0 if update data required (used
//==!                         only if iFlowTypeVar == -1
//==!                      lDoCaptureFlag > 0 if capture data required (used
//==!                         only if iFlowTypeVar == -1
//==!                      pCFWparam = CFlowWait wait time parameter
//==! Return Value       : max assembled data reg width
//==!=====================================================================!==
long shiftFlowData(CDeviceData *pCDD_Tdi_Tdo,int iFlowTypeVar,
            long lDoUpdateFlag,long lDoCaptureFlag,CFlowWait *pCFWparam)
{ long lDataRegWidth;

  lDataRegWidth = assembleData(pCDD_Tdi_Tdo,0,cTdiData,iFlowTypeVar);
  jtagShift(TAP_IRSHIFT,lDataRegWidth,
            cTdiData,cTdoData,0,(lDoUpdateFlag > 0 ? 1:0));

  if (iOpt_ProgDebug == 1)
  { sprintf(sMsgOnly,"\nSIR [%ld]\n  - TDI [",lDataRegWidth);
    debugMsg(sMsgOnly,1,1);
    debugStrHex(cTdiData,lDataRegWidth,1,1);
    debugMsg("]",1,1);
  } // IF iOpt_ProgDebug == 1

//==! If at least one device needs update data, then combine data from all
//==! selected data registers in each JTAG device into cTdiData before
//==! shifting out.
  if (lDoUpdateFlag > 0)
  { lDataRegWidth = assembleData(pCDD_Tdi_Tdo,1,cTdiData,iFlowTypeVar);
    jtagShift(TAP_DRSHIFT,lDataRegWidth,cTdiData,cTdoData,0,0);

    if (iOpt_ProgDebug == 1)
    { sprintf(sMsgOnly,"\nSDR [%ld]\n  - TDI [",lDataRegWidth);
      debugMsg(sMsgOnly,1,1);
      debugStrHex(cTdiData,lDataRegWidth,1,1);
      debugMsg("]",1,1);
    } // IF iOpt_ProgDebug == 1
  } // IF lDoUpdateFlag > 0

//==! wait maximum min wait time from each device in JTAG chain
  doFlowWait(pCFWparam);

  if (lDoCaptureFlag > 0)  // NOTE_0009
  { lDataRegWidth = assembleData(pCDD_Tdi_Tdo,1,cTdiData,iFlowTypeVar);
    jtagShift(TAP_DRSHIFT,lDataRegWidth,cTdiData,cTdoData,1,1);
    if (iOpt_ProgDebug == 1)
    { sprintf(sMsgOnly,"\nSDR [%ld]\n  - I/O TD",lDataRegWidth);
      debugMsg(sMsgOnly,1,1);
      if (lDoUpdateFlag > 0)
      { debugMsg("I [",1,1);
        debugStrHex(cTdiData,lDataRegWidth,1,1);
        debugMsg("] TD",1,1);
      } // IF
      debugMsg("O [",1,1);
      debugStrHex(cTdoData,lDataRegWidth,1,1);
      debugMsg("]",1,1);
    } // IF iOpt_ProgDebug == 1
  } // IF lDoCaptureFlag > 0

  return lDataRegWidth;

} // shiftFlowData()
//==!=====================================================================!==
//==! Process device actions concurrently.
//==!
//==! Function Parameters: pCDD = pointer to device list
//==!                      lProcIteration = current processing iteration
//==! Return Value       : 1 if successful, 0 otherwise
//==!=====================================================================!==
int processActionsConcurrently(CDeviceData *pCDD,long lProcIteration)
{ CDeviceData *pCDDvar;
  CFlowWait recCFlowWait;
  char sMode[8];
  long lDoUpdate,lDoCapture,lAssembledDataSize,lCurDevTdoIndex,
       lDevicesToProcess,lVar0,lVar1;
  int iFlowTypeVar,iConcurrentProcStatus,iProcessingFlag;

  for (pCDDvar = pCDD,iConcurrentProcStatus = 1;
       pCDDvar != NULL;
       pCDDvar = pCDDvar->m_pNextCDeviceData)
  { pCDDvar->m_iProcessingStatus = 1; // initialize processing status to PASS
    if (pCDDvar->m_iSkipDev == 1)
    { continue;
    } // IF
    pCDDvar->m_pCProcContext->m_pCIAvar = findIscAction(pCDDvar);
    pCDDvar->m_pCProcContext->m_iPhase = 0;
    pCDDvar->m_pCProcContext->m_iBypass = 0;
    if (pCDDvar->m_pCProcContext->m_pCIAvar == NULL)
    { pCDDvar->m_pCProcContext->m_iBypass = 1; // no action so bypass device
      pCDDvar->loadBypassOp();
      sprintf(sMsgOnly,"    => [%s (%s)]: - bypass -",pCDDvar->m_csDeviceName,
              pCDDvar->m_pCBsdlData->csBsdlFileName);
      dispMsg(sMsgOnly,DISP_PRE_LF,1,1);
      continue;
    } // IF pCDDvar->m_pCProcContext->m_pCIAvar == NULL

    sprintf(sMsgOnly,"    => [%s (%s)]: %s",pCDDvar->m_csDeviceName,
              pCDDvar->m_pCBsdlData->csBsdlFileName,pCDDvar->m_csAction);
    if (pCDDvar->m_csActionData != NULL)
    { sprintf(sMsgBuf01,"(%s)",pCDDvar->m_csActionData);
      strcat(sMsgOnly,sMsgBuf01);
    } // IF
    dispMsg(sMsgOnly,DISP_PRE_LF,1,1);

//==! Check if action is exclusive/proprietary. If proprietary, then action
//==! should be processed sequentially because it may not be fully compliant
//==! with the 1149 or 1532 standards.
    if (pCDDvar->m_pCProcContext->m_pCIAvar->m_iExclusivity == 1)
    { sprintf(sMsgOnly,"     *--> Action [%s: %s(%s)] is proprietary",
          pCDDvar->m_csDeviceName,pCDDvar->m_csAction,
          (pCDDvar->m_csActionData != NULL ? pCDDvar->m_csActionData : " "));
      dispMsg(sMsgOnly,DISP_PRE_LF,1,1);
      dispMsg(
"       => chain processing cancelled; process chain in sequential mode",
              DISP_PRE_LF | DISP_POST_LF,1,1);
      pCDDvar->m_pCProcContext->m_iBypass = 1;
      iConcurrentProcStatus = 0;
    } // IF
    if (pCDDvar->m_pCDP_Out != NULL)
    { sprintf(sMode,"%s",iOptOverWritePrompt == 1 ? "c" : "w");
      pCDDvar->m_pCDP_Out->pFILE =
          xFileOpen(pCDDvar->m_pCDP_Out->psPortName,sMode); //==! NOTE_0007
      pCDDvar->writeIscData(0,NULL,-1);
    } // IF
  } // FOR pCDDvar != NULL
  dispMsg("    ----------- start concurrent processing -----------",
          DISP_PRE_LF,1,1);

  if (iConcurrentProcStatus == 0)
  { return 0;
  } // IF

  recCFlowWait.resetWaitTimes();
  lDoUpdate = 0;
  lDoCapture = 0;
  lAssembledDataSize = 0;

//==! Keep looping until procedures/flows for all devices are done.
  while (1)
  { lCurDevTdoIndex = lAssembledDataSize;

    for (pCDDvar = pCDD,lDevicesToProcess = 0;
         pCDDvar != NULL;
         pCDDvar = pCDDvar->m_pNextCDeviceData)
    {
//==! Check if this device is to be skipped. This is usually done in a
//==! device programming environment where a JTAG command file is defined
//==! that corresponds to the maximum number of device sockets on the device
//==! programmer. If a device is not in the socket, then this device in the
//==! list is skipped completely, i.e., do not put it into bypass.
      if (pCDDvar->m_iSkipDev == 1)
      { continue;
      } // IF

      if (pCDDvar->m_pCProcContext->m_iBypass == 1) // bypass device
      { --lCurDevTdoIndex;
        pCDDvar->loadBypassOp();
        continue;
      } // IF

      lDevicesToProcess += processOneDevAction(pCDD,pCDDvar,&recCFlowWait,
            &lDoUpdate,&lDoCapture,&lCurDevTdoIndex);

//==! Check if device already put into bypass; i.e., all device's actions
//==! completed.
      if (pCDDvar->m_pCProcContext->m_iBypass == 1)
      { pCDDvar->loadBypassOp();
      } // IF
    } // FOR pCDDvar != NULL

    if (lDevicesToProcess == 0 || iGlobExitOnError == 1)
    { break;
    } // IF

    for (iProcessingFlag = 0; iProcessingFlag == 0; )
    {
//==! Store flow type into another variable which will not change until the
//==! next time the chain is processed again. This temporary variable is
//==! required because the member variable m_iFlowType could be changed after
//==! a call to processOneDevAction() which gets the next flow activity for
//==! the device. This variable is used if flows cannot be processed
//==! concurrently. In this case, each flow type is processed separately
//==! which requires going through the chain 4 times. When one flow type is
//==! processed, devices with other flow types are put into bypass.
      for (pCDDvar = pCDD,lDevicesToProcess = 0;
           pCDDvar != NULL;
           pCDDvar = pCDDvar->m_pNextCDeviceData)
      { pCDDvar->m_iHoldFlowType = pCDDvar->m_iFlowType;
      } // FOR pCDDvar != NULL

//==! If current flows for all devices are the same type then process them
//==! concurrently.
//==!
//==! Combine opcodes from all devices in JTAG chain into cTdiData before
//==! shifting out. If there is update data to shift, then load the opcodes
//==! but stop in UPDATE-IR (see jtagShift() parameter description). The
//==! update data shift will go from UPDATE_IR to SELECT-DR (skipping RTI),
//==! then end in RTI after SHIFT-DR. This TAP state transition sequence is
//==! required so that previously loaded opcodes will not start executing
//==! until after the data (if required) is loaded. If there is no input data
//==! to shift, the JTAG IR shift operation will load the opcode and stop in
//==! RTI to begin opcode execution.
      if (iGlobOptMaxConcurrent == 1 ||
          isSameFlowType(pCDD,&iFlowTypeVar) == 1)
      { if (iOpt_ProgDebug == 1)
        { sprintf(sMsgOnly,"-----> common flow type [%ld]",iFlowTypeVar);
          dispMsg(sMsgOnly,DISP_PRE_LF,1,1);
        } // IF

        lAssembledDataSize = shiftFlowData(pgCDD_FromTdi,-1,lDoUpdate,
                                           lDoCapture,&recCFlowWait);
        iProcessingFlag = 1;
      } // IF iGlobOptMaxConcurrent == 1 || isSameFlowType()
      else
      {
//==! Process flows types one by one; put parts which have different flow
//==! types into bypass mode.
        for (iFlowTypeVar = 0; iFlowTypeVar < 4; ++iFlowTypeVar)
        { if (findFlowType(pCDD,iFlowTypeVar) == 0)
          { continue;   // continue if no flow of type iFlowTypeVar
          } // IF

//==! Note that the processing flow will show the opcode being processed, but
//==! the opcode message (as of 2004/03/26 implementation) will actually have
//==! occurred at the last processOneDevAction() call.
          if (iOpt_ProgDebug == 1)
          { sprintf(sMsgOnly,"-----> processing flow type [%ld]",
                    iFlowTypeVar);
            dispMsg(sMsgOnly,DISP_PRE_LF,1,1);
          } // IF

          lVar0 = (iFlowTypeVar == 0 || iFlowTypeVar == 2 ? 1 : 0);
          lVar1 = (iFlowTypeVar == 0 || iFlowTypeVar == 3 ? 1 : 0);

          lCurDevTdoIndex = shiftFlowData(pgCDD_FromTdi,iFlowTypeVar,
                      lVar0,lVar1,&recCFlowWait);

          for (pCDDvar = pCDD;
               pCDDvar != NULL;
               pCDDvar = pCDDvar->m_pNextCDeviceData)
          { if (pCDDvar->m_iSkipDev == 1)
            { continue;
            } // IF

            if (pCDDvar->m_pCProcContext->m_iBypass == 1) // bypass device
            { --lCurDevTdoIndex;
              pCDDvar->loadBypassOp();
              continue;
            } // IF

            if (pCDDvar->m_iHoldFlowType != iFlowTypeVar)
            { --lCurDevTdoIndex;
              continue;
            } // IF

            if (pCDDvar->m_pCProcContext->m_iPhase != 1)
            { sprintf(sMsgOnly,"Parallel-PhaseErr[%s]",
                      pCDDvar->m_csDeviceName);
              jMessage(DISP_AND_EXIT,sMsgOnly,0,NULL);
            } // IF

            lDevicesToProcess += processOneDevAction(pCDD,pCDDvar,
                  &recCFlowWait,&lVar0,&lVar1,&lCurDevTdoIndex);

//==! Check if device already put into bypass; i.e., all device's actions
//==! completed.
            if (pCDDvar->m_pCProcContext->m_iBypass == 1)
            { pCDDvar->loadBypassOp();
            } // IF
          } // FOR pCDDvar != NULL

        } // FOR iFlowTypeVar < 4

        if (lDevicesToProcess == 0)
        { break;
        } // IF

        for (pCDDvar = pCDD,lDoUpdate = 0,lDoCapture = 0;
             pCDDvar != NULL;
             pCDDvar = pCDDvar->m_pNextCDeviceData)
        { if (pCDDvar->m_iSkipDev == 1)
          { continue;
          } // IF
          if (pCDDvar->m_pCProcContext->m_iBypass == 1) // bypass device
          { continue;
          } // IF

          switch(pCDDvar->m_iFlowType)
          { case 0:
            { ++lDoUpdate;
              ++lDoCapture;
              break;
            } // CASE 0
            case 1:
            { break;
            } // CASE 1
            case 2:
            { ++lDoUpdate;
              break;
            } // CASE 2
            case 3:
            { ++lDoCapture;
              break;
            } // CASE 3
            default:
            { sprintf(sMsgOnly,"Invalid Ftype[%s:%d]",
                      pCDDvar->m_csDeviceName,pCDDvar->m_iFlowType);
              jMessage(DISP_AND_EXIT,sMsgOnly,0,NULL);
              break;
            } // DEFAULT
          } // SWITCH pCDDvar->m_iFlowType

        } // FOR pCDDvar != NULL

        iProcessingFlag = 0;
      } // ELSE
    } // FOR iProcessingFlag == 0

    recCFlowWait.resetWaitTimes();
    lDoUpdate = 0;
    lDoCapture = 0;
  } // WHILE (1)

  sprintf(sMsgOnly,"---[ Concurrent processing summary (iteration %5ld) ]---",
          lProcIteration);
  dispMsg(sMsgOnly,DISP_PRE_LF,1,1);
  for (pCDDvar = pCDD; pCDDvar != NULL; pCDDvar = pCDDvar->m_pNextCDeviceData)
  { if (pCDDvar->m_pCIscData != NULL)
    { if (pCDDvar->m_pCIscData->m_pCDP_IscData != NULL)
      { pCDDvar->m_pCIscData->m_pCDP_IscData->xCloseDataPort();
      } // IF
    } // IF
    if (pCDDvar->m_pCDP_Out != NULL)
    { pCDDvar->writeIscData(1,NULL,-1);
      xFclose(&(pCDDvar->m_pCDP_Out->pFILE));
    } // IF
    sprintf(sMsgOnly,"   *** [%s",pCDDvar->m_csDeviceName);
    dispMsg(sMsgOnly,DISP_PRE_LF,1,1);
    if (pCDDvar->m_iSkipDev == 1)
    { dispMsg("]: skipped",DISP_USE_CUR_POS,1,1);
      continue;
    } // IF
    if (pCDDvar->m_pCProcContext->m_pCIAvar == NULL)
    { dispMsg("]: BYPASS",DISP_USE_CUR_POS,1,1);
      continue;
    } // IF
    if (pCDDvar->m_iProcessingStatus == 0)  // at least one action failed
    { iConcurrentProcStatus = 0;
    } // IF

    sprintf(sMsgOnly,": %s(%s)] = ",pCDDvar->m_csAction,
        (pCDDvar->m_csActionData != NULL ? pCDDvar->m_csActionData : " "));
    if (pCDDvar->m_iProcessingStatus == 0)
    { strcat(sMsgOnly,"FAIL");
    } // IF
    else
    { if ((iGlobExitOnError == 0) ||
          (pCDDvar->m_pCProcContext->m_iBypass == 1))
      { strcat(sMsgOnly,"PASS");
      } // IF
      else
      { if (iGlobOptSkipExitOnError == 0)
        { strcat(sMsgOnly,"incomplete due to EXIT_ON_ERROR");
        } // IF
        else
        { strcat(sMsgOnly,"incomplete (EXIT_ON_ERROR skipped)");
        } // ELSE
      } // ELSE
    } // ELSE
    dispMsg(sMsgOnly,DISP_USE_CUR_POS,1,1);

    if ((iGlobOptUseErrorProc == 1) &&
         ((pCDDvar->m_iProcessingStatus == 0) ||
          (pCDDvar->m_pCProcContext->m_iBypass == 0)
         )
       )
    {
// should never occur because action information is contained in a BSDL file
      if (pCDDvar->m_pCBsdlData == NULL)
      { continue;
      } // IF

      debugMsg("\n-----> Executing PROC_ERROR_EXIT -----",1,1);
      if (pCDDvar->m_pCBsdlData->m_pCIscProcError == NULL)
      { sprintf(sMsgOnly,"\nPROC_ERROR_EXIT not defined for [%s]",
                pCDDvar->m_csDeviceName);
        jMessage(DISP_SHOW_MSG,sMsgOnly,0,NULL);
      } // IF
      else
      { if (processProcedure(pCDD,pCDDvar,
                       pCDDvar->m_pCBsdlData->m_pCIscProcError) == 0)
        { sprintf(sMsgOnly,"PROC_ERROR_EXIT for [%s] unsuccessful",
                  pCDDvar->m_csDeviceName);
          jMessage(DISP_SHOW_MSG,sMsgOnly,0,NULL);
        } // IF processProcedure() == 0
      } // ELSE
    } // IF iGlobOptUseErrorProc == 1 && pCDDvar->m_iProcessingStatus == 0

  } // FOR pCDDvar != NULL

  return iConcurrentProcStatus;

} // processActionsConcurrently()
//==!=====================================================================!==
//==! Calculate/derive header and trailer information from the pCDDobj's
//==! position in the pCDD list. This function modifies the global variables
//==! lHdr, lHir, lTdr, lTir. Note that the device list pointer pCDD starts
//==! with parts closest to TDO.
//==!
//==! Function Parameters: pCDD = pointer to device list
//==!                      pCDDobj = pointer to device being processed
//==! Return Value       : none
//==!=====================================================================!==
void calcHeaderTrailer(CDeviceData *pCDD,CDeviceData *pCDDobj)
{ CDeviceData *pCDDvar;
  int iPartFound;

  for (pCDDvar = pCDD,lHir = 0,lHdr = 0,lTir = 0,lTdr = 0,iPartFound = 0;
       pCDDvar != NULL;
       pCDDvar = pCDDvar->m_pNextCDeviceData)
  { if (pCDDvar->m_iSkipDev == 1)
    { continue;
    } // IF
    if (pCDDvar == pCDDobj)
    { iPartFound = 1;
      continue;
    } // IF
    if (iPartFound == 0)
    { ++lHdr;
      lHir += pCDDvar->m_lInstrRegLength;
    } // IF
    else
    { ++lTdr;
      lTir += pCDDvar->m_lInstrRegLength;
    } // ELSE
  } // FOR pCDDvar != NULL

} // calcHeaderTrailer()
//==!=====================================================================!==
//==! Prepare/allocate data buffers used to shift JTAG data.
//==!
//==! Function Parameters: pCDD = pointer to device list
//==!                      iProcMode = processing mode
//==!                        0: sequential
//==!                        1: concurrent
//==! Return Value       : none
//==!=====================================================================!==
void prepareDataBuffers(CDeviceData *pCDD,int iProcMode)
{ CDeviceData *pCDDvar;
  CIscAct *pCIAvar;
  long lMaxDataRegSize,lMaxOpcodeRegSize,lVar,lCnt;

  cTdiData = NULL;
  cTdoData = NULL;
  cTdoCopy = NULL;
  cGlobTapDataReg = NULL;
  for (pCDDvar = pCDD,lMaxDataRegSize = 0,lMaxOpcodeRegSize = 0,
                      lVar = 0,lCnt = 0;
       pCDDvar != NULL;
       pCDDvar = pCDDvar->m_pNextCDeviceData,++lCnt)
  { if (pCDDvar->m_iSkipDev == 1)
    { continue;
    } // IF

    lMaxOpcodeRegSize += pCDDvar->m_lInstrRegLength;
    if (iOpt_ProgDebug == 1)
    { sprintf(sMsgOnly,"-[%3ld] %s (%s) IR(%ld)",lCnt,
            pCDDvar->m_csDeviceName,
            pCDDvar->m_pCBsdlData != NULL
               ? pCDDvar->m_pCBsdlData->csBsdlFileName : "*",
               pCDDvar->m_lInstrRegLength);
      dispMsg(sMsgOnly,DISP_PRE_LF,1,1);
    } // IF

    if (pCDDvar->m_csAction == NULL)  
    { ++lMaxDataRegSize;       // Use BYPASS reg if no action specified
      if (iProcMode == 1)
      { pCDDvar->createTapReg(pCDDvar->m_lInstrRegLength,1);
      } // IF
      continue;
    } // IF pCDDvar->m_csAction == NULL
    pCIAvar = findIscAction(pCDDvar);
    if (pCIAvar == NULL)       // Use BYPASS reg if no action specified
    { ++lMaxDataRegSize;
      if (iProcMode == 1)
      { pCDDvar->createTapReg(pCDDvar->m_lInstrRegLength,1);
      } // IF
      continue;
    } // IF

    if (iOpt_ProgDebug == 1)
    { sprintf(sMsgOnly,": %s",pCDDvar->m_csAction);
      if (pCDDvar->m_csActionData != NULL)
      { sprintf(sMsgBuf01,"(%s)",pCDDvar->m_csActionData);
        strcat(sMsgOnly,sMsgBuf01);
      } // IF
      sprintf(sMsgBuf01," DR(%ld)",pCIAvar->m_lMaxRegSizeUsed);
      strcat(sMsgOnly,sMsgBuf01);
      dispMsg(sMsgOnly,DISP_USE_CUR_POS,1,1);
    } // IF

//==! In sequential processing mode, find the longest register. If a register
//==! is shorter than the current longest, then increment the lMaxDataRegSize
//==! to account for the BYPASS register which this device will be put in.
//==
//==! In concurrent processing mode, add the longest register size used by
//==! the selected action for each device.
    if (iProcMode == 0)
    { ++lMaxDataRegSize;
      if (lVar < pCIAvar->m_lMaxRegSizeUsed)
      { lVar = pCIAvar->m_lMaxRegSizeUsed;
      } // IF
      else
      { ++lMaxDataRegSize;
      } // ELSE
    } // IF
    else
    { lMaxDataRegSize += pCIAvar->m_lMaxRegSizeUsed;
      if (iProcMode == 1)
      { pCDDvar->createTapReg(pCDDvar->m_lInstrRegLength,
                              pCIAvar->m_lMaxRegSizeUsed);
      } // IF
    } // ELSE
  } // FOR pCDDvar != NULL

//==! In sequential mode, the largest register size will be the chain length
//==! (i.e., 1 bypass register per device) plus the max register length - 1
//==! (replace one of the bypass registers with the longest register).
  if (iProcMode == 0)
  { lMaxDataRegSize += lVar-1;
  } // IF

  lVar = lMaxDataRegSize;
  if (lMaxOpcodeRegSize > lMaxDataRegSize)
  { lVar = lMaxOpcodeRegSize;
  } // IF

  if (iOpt_ProgDebug == 1)
  { sprintf(sMsgOnly,"-> Max Reg Size: IR %6ld | DR %6ld (%s mode) TDI (%ld)",
            lMaxOpcodeRegSize,lMaxDataRegSize,
            iProcMode == 0 ? "sequential" : "concurrent",
            lVar);
    dispMsg(sMsgOnly,DISP_PRE_LF,1,1);
  } // IF

  cTdiData = new char [lVar+1];
  cTdoData = new char [lVar+1];
  cTdoCopy = new char [lVar+1];
  cGlobTapDataReg = new char [lVar+1];

} // prepareDataBuffers()
//==!=====================================================================!==
//==! Deallocate data buffers used to shift JTAG data in sequential
//==! processing mode. Data buffers allocated for concurrent processing mode
//==! will be deallocated along with the device data structure.
//==!
//==! Function Parameters: pCDD = pointer to device list
//==! Return Value       : none
//==!=====================================================================!==
void freeDataBuffers(CDeviceData *pCDD)
{ CDeviceData *pCDDvar;

  for (pCDDvar = pCDD; pCDDvar != NULL; pCDDvar = pCDDvar->m_pNextCDeviceData)
  { freeArray(pCDDvar->m_cSIR);
    freeArray(pCDDvar->m_cTdi);
    freeArray(pCDDvar->m_cTdo);
    freeArray(pCDDvar->m_cTdoCopy);
  } // FOR pCDDvar != NULL

  freeArray(cTdiData);
  freeArray(cTdoData);
  freeArray(cTdoCopy);
  freeArray(cGlobTapDataReg);

} // freeDataBuffers()
//==!=====================================================================!==
//==! Process JTAG 1532 device list.
//==!
//==! Function Parameters: pCDD = pointer to device list
//==!                      lProcIteration = current processing iteration
//==! Return Value       : 1 if successful, 0 otherwise
//==!=====================================================================!==
int processDeviceList(CDeviceData *pCDD,long lProcIteration)
{ CDeviceData *pCDDvar;
  int iFuncStat;
  unsigned short usTransCrc;

  iGlobProgramState = dPROGSTATE_PROCESSDEVICE_SEQUENTIAL;

  iFuncStat = 1;

  doSyncJtagReset(1);  // goto RTI after synchronous JTAG reset

  if (iGlobProcMode == 0)  //==! Process devices sequentially
  { for (pCDDvar = pCDD; pCDDvar != NULL;
         pCDDvar = pCDDvar->m_pNextCDeviceData)
    { if (pCDDvar->m_iSkipDev == 1)
      { sprintf(sMsgOnly,"-+> Skipping device [%s]",
                pCDDvar->m_csDeviceName);
        dispMsg(sMsgOnly,DISP_PRE_LF,1,1);
        continue;
      } // IF pCDDvar->m_iSkipDev == 1
      if (pCDDvar->m_csAction == NULL)
      { sprintf(sMsgOnly,"-*> No action specified for [%s]",
                pCDDvar->m_csDeviceName);
        dispMsg(sMsgOnly,DISP_PRE_LF,1,1);
        continue;
      } // IF pCDDvar->m_csAction == NULL
      calcHeaderTrailer(pCDD,pCDDvar);
      dispMsg("  /~~~~~~~~~~~~~~~~~~\\",DISP_PRE_LF,1,1);
      sprintf(sMsgOnly,
"--> Performing action [%s (%s)]: %s",pCDDvar->m_csDeviceName,
              pCDDvar->m_pCBsdlData->csBsdlFileName,pCDDvar->m_csAction);
      if (pCDDvar->m_csActionData != NULL)
      { sprintf(sMsgBuf01,"(%s)",pCDDvar->m_csActionData);
        strcat(sMsgOnly,sMsgBuf01);
      } // IF
      dispMsg(sMsgOnly,DISP_PRE_LF,1,1);
      dispMsg("  \\__________________/",DISP_PRE_LF,1,1);
      if (pCDDvar->m_pCDP_Out != NULL)
      { sprintf(sMsgOnly,"    = outputs from action to file [%s] %s",
                pCDDvar->m_pCDP_Out->psPortName,
                (pCDDvar->m_iIscDataFormat == 1 ? "in ISC format" : " "));
        dispMsg(sMsgOnly,DISP_PRE_LF,1,1);
      } // IF

      iFuncStat = processAction(pCDD,pCDDvar);

      dispMsg(" /~~~~\\--------------------------------------------------",
              DISP_PRE_LF,1,1);
      if (iFuncStat == 0)
      { dispMsg("[ FAIL ",DISP_PRE_LF,1,1);
      } // IF
      else
      { dispMsg("[ PASS ",DISP_PRE_LF,1,1);
      } // ELSE

      sprintf(sMsgOnly,"]: Action [%s (%s)]: %s",
              pCDDvar->m_csDeviceName,
              pCDDvar->m_pCBsdlData->csBsdlFileName,pCDDvar->m_csAction);
      if (pCDDvar->m_csActionData != NULL)
      { sprintf(sMsgBuf01,"(%s)",pCDDvar->m_csActionData);
        strcat(sMsgOnly,sMsgBuf01);
      } // IF
      dispMsg(sMsgOnly,DISP_USE_CUR_POS,1,1);

//==! Highlight the FAIL/PASS text in the previous line
      dispMsg(" \\____/--------------------------------------------------",
              DISP_PRE_LF,1,1);

      if ((iFuncStat == 0 && iGlobOptContinueOnFail == 0) ||
          (iGlobExitOnError == 1))
      { break;
      } // IF

    } // FOR pCDDvar != NULL

  } // IF iGlobProcMode == 0
  else
  { sprintf(sMsgOnly,"--> Performing actions concurrently");
    dispMsg(sMsgOnly,DISP_PRE_LF,1,1);
    iFuncStat = processActionsConcurrently(pCDD,lProcIteration);

    dispMsg(" /~~~~\\--------------------------------------------------",
            DISP_PRE_LF,1,1);
    if (iFuncStat == 0)
    { dispMsg("[ FAIL ",DISP_PRE_LF,1,1);
    } // IF
    else
    { dispMsg("[ PASS ",DISP_PRE_LF,1,1);
    } // ELSE
    sprintf(sMsgOnly,"]: concurrent processing mode (iteration %ld)",
            lProcIteration);
    dispMsg(sMsgOnly,DISP_USE_CUR_POS,1,1);
    dispMsg(" \\____/--------------------------------------------------",
            DISP_PRE_LF,1,1);

  } // ELSE

//==! Generate a transmission CRC if the device specified an ISC-formatted
//==! output file.
  for (pCDDvar = pCDD; pCDDvar != NULL;
       pCDDvar = pCDDvar->m_pNextCDeviceData)
  { if ((pCDDvar->m_iSkipDev == 1)      ||
        (pCDDvar->m_csAction == NULL)   ||
        (pCDDvar->m_pCDP_Out == NULL)   ||
        (pCDDvar->m_iIscDataFormat == 0)
       )
    { continue;
    } // IF

    usTransCrc = calcIscCRC(pCDDvar->m_pCDP_Out);

    fflush(pCDDvar->m_pCDP_Out->pFILE);
    xFclose(&(pCDDvar->m_pCDP_Out->pFILE));
    pCDDvar->m_pCDP_Out->pFILE = fopen(pCDDvar->m_pCDP_Out->psPortName,"a");
    if (pCDDvar->m_pCDP_Out->pFILE == NULL)
    { sprintf(sMsgOnly,"- could not reopen [%s] to write transmission CRC",
              pCDDvar->m_pCDP_Out->psPortName);
      jMessage(DISP_SHOW_MSG,sMsgOnly,0,NULL);
      continue;
    } // IF
    fprintf(pCDDvar->m_pCDP_Out->pFILE,"\n%04X",usTransCrc);
    xFclose(&(pCDDvar->m_pCDP_Out->pFILE));
  } // FOR pCDDvar != NULL

  doSyncJtagReset(0);  // do synchronous JTAG reset and wait in TLR
  return iFuncStat;

} // processDeviceList()
//==!=====================================================================!==
//==! end of module XRD_DATA.CPP
//==!=====================================================================!==

