/***************************************************************************/
/* J Drive (tm) Engine                                                     */
/* Copyright 2003-2005, Xilinx, Inc.                                       */
/* The use, reproduction, and distribution of this code is subject         */
/* to the J Drive license requirements and restrictions.                   */
/***************************************************************************/
//==!=====================================================================!==
//==! See XPROJECT.H for project file list and project version history.
//==!=====================================================================!==
//==! Revision history: XROUTINE.CPP
//==! 2003/05/02: MCS reader/writer routines
//==!=====================================================================!==
#include "xcommon.h"
#include "x_bsdl.h"
//==!=====================================================================!==
//==! Get a long value from the input data port by reading iByteCnt chars and
//==! then converting to a long value.
//==!
//==! Function Parameters: pDP_In = input cDataPort
//==!                      iByteCnt = number of bytes forming the number
//==!                          being read
//==!                      uiByteSum = pointer to an unsigned int to hold the
//==!                          unsigned sum of bytes that make up this value
//==! Return Value       : long value
//==!=====================================================================!==
long getLValue(cDataPort *pDP_In,int iByteCnt,unsigned int *uiByteSum)
{ int iVar;
  long lVar,lByteVar;

  for (lVar = 0,lByteVar = 0,*uiByteSum = 0,iVar = iByteCnt-1;
       iVar >= 0; --iVar)
  { getDpChar(pDP_In,1);
    if (!isxdigit((char)pDP_In->iCurChar))
    { jMessage(DISP_AND_EXIT,"Invalid input file",1,pDP_In);
    } // IF
    lVar = lVar + (chrtohex((char)pDP_In->iCurChar) << (iVar*4));
    lByteVar = lByteVar + (chrtohex((char)pDP_In->iCurChar) << ((iVar%2)*4));
    if (iVar%2 == 0)
    { *uiByteSum += lByteVar;
      lByteVar = 0;
    } // IF
  } // FOR iVar >= 0

  return lVar;

} // getLValue()
//==!=====================================================================!==
//==! Convert a char to a hex value. The char has to be a hex char.
//==!
//==! Function Parameters: cChr = hex char to convert
//==! Return Value       : ASCII value of hex char
//==!=====================================================================!==
unsigned char convChrtoHex(char cChr,cDataPort *pDP_In)
{
  if (!isxdigit(cChr))
  { sprintf(sMsgBuf00,"Unknown hex value [%c]",cChr);
    jMessage(DISP_AND_EXIT,sMsgBuf00,pDP_In != NULL ? 1 : 0,pDP_In);
  } // IF
  return chrtohex(cChr);

} // convChrtoHex()
//==!=====================================================================!==
//==! Read an MCS record from the input data port. Calculate checksum and
//==! compare with the checksum at the end of the record. Checksum computed
//==! by doing an unsigned 8 bit addition of the:
//==!   = byte count
//==!   = byte addr
//==!   = record type
//==!   = number of data bytes defined by the byte count
//==! and then doing a 2s-complement conversion.
//==!
//==! Function Parameters: pDP_In = input cDataPort
//==!                      pRec = pointer to a record
//==! Return Value       : 1 if record read, 0 if no more records
//==!=====================================================================!==
int readMCSrec(cDataPort *pDP_In,CMCSrecord *pRec)
{ long lVar;
  unsigned int uiCheckSum,uiRecCheckSum,uiVar;

  xParseData(pDP_In,XDATA_READ);
  if (pDP_In->m_iEndOfDataStream == 1)  //==! No more records
  { return 0;
  } // IF
  if (pDP_In->sData[0] != ':')
  { sprintf(sMsgBuf00,"Invalid char [%s]; may not be an MCS file",
       pDP_In->sData);
    jMessage(DISP_AND_EXIT,sMsgBuf00,1,pDP_In);
  } // IF

  pRec->lByteCount = getLValue(pDP_In,2,&uiCheckSum);
  pRec->lByteAddr  = getLValue(pDP_In,4,&uiVar);
  uiCheckSum += uiVar;
  pRec->iRecType   = (int)(getLValue(pDP_In,2,&uiVar));
  uiCheckSum += uiVar;
  pRec->pUCarray = new unsigned char [pRec->lByteCount];
  for (lVar = 0; lVar < pRec->lByteCount; ++lVar)
  { getDpChar(pDP_In,1);
    pRec->pUCarray[lVar] = convChrtoHex((char)pDP_In->iCurChar,pDP_In) << 4;
    getDpChar(pDP_In,1);
    pRec->pUCarray[lVar] = pRec->pUCarray[lVar] |
                           convChrtoHex((char)pDP_In->iCurChar,pDP_In);
    uiCheckSum = 0x00FF & (uiCheckSum + pRec->pUCarray[lVar]);
  } // FOR lVar < pRec->lByteCount
  uiCheckSum = 0x00FF & (~(uiCheckSum) + 1);

  if (pRec->iRecType == 2 || pRec->iRecType == 4)
  { if (pRec->lByteCount != 2)
    { jMessage(DISP_AND_EXIT,"Rec type must have 2 data bytes",1,pDP_In);
    } // IF
    pRec->lOffset = (((long)(pRec->pUCarray[0])) << 8) |
                      (long)(pRec->pUCarray[1]);
  } // IF

  uiRecCheckSum = (unsigned int)getLValue(pDP_In,2,&uiVar);
  if (uiRecCheckSum != uiCheckSum)
  { sprintf(sMsgBuf00,
       "MCS record checksum [%02X] != Computed checksum [%02X]",
        uiRecCheckSum,uiCheckSum);
    jMessage(DISP_SHOW_MSG,sMsgBuf00,1,pDP_In);
  } // IF

  return 1;

} // readMCSrec()
//==!=====================================================================!==
//==! Compute checksum by doing an 8-bit unsigned addition on bytes sBuf
//==! and pRec. Bytes in sBuf have to be converted to values, while the
//==! bytes in pRec->pUCarray (if pRec != NULL) have already been converted
//==! (just add the bytes to the checksum).
//==!
//==! Function Parameters: pDP_In = bitstream cDataPort
//==! Return Value       : unsigned int checksum
//==!=====================================================================!==
unsigned int calcMCSCheckSum(char *sBuf,CMCSrecord *pRec)
{ unsigned int uiCheckSum;
  unsigned char ucChr;
  long lVar0,lVar1;

  uiCheckSum = 0;
  if (sBuf != NULL)
  { lVar0 = strlen(sBuf);
    if (lVar0%2 != 0)
    { sprintf(sMsgBuf00,"Incomplete byte [%s]",sBuf);
      jMessage(DISP_AND_EXIT,sMsgBuf00,0,NULL);
    } // IF
    for (lVar1 = 0; lVar1 < lVar0; lVar1 += 2)
    { ucChr = convChrtoHex(sBuf[lVar1],NULL) << 4;
      ucChr = ucChr | convChrtoHex(sBuf[lVar1+1],NULL);
      uiCheckSum = 0x00FF & (uiCheckSum + ucChr);
    } // FOR lVar1 < lVar0
  } // IF sBuf != NULL

  if (pRec != NULL)
  { for (lVar0 = 0; lVar0 < pRec->lByteCount; ++lVar0)
    { uiCheckSum = 0x00FF & (uiCheckSum + pRec->pUCarray[lVar0]);
    } // FOR lVar0 < pRec->lByteCount
  } // IF pRec != NULL

  return 0x00FF & (~(uiCheckSum)+1);

} // calcMCSCheckSum()
//==!=====================================================================!==
//==! Get number of data bytes in the bitstream/MCS file.
//==!
//==! Function Parameters: pDP_In = bitstream cDataPort
//==! Return Value       : long bitstream bytecount
//==!=====================================================================!==
long getBitstreamByteSize(cDataPort *pDP_In)
{ CMCSrecord pCMCSvar;
  long lTotByteCount,lCurByteAddr,lAbsByteAddr,
       lLinearBaseAddr,lSegmentBaseAddr;

  for (lTotByteCount = lAbsByteAddr = lLinearBaseAddr = 0,
          lSegmentBaseAddr = lCurByteAddr = 0;
       readMCSrec(pDP_In,&pCMCSvar) == 1; )
  { switch(pCMCSvar.iRecType)
    { case 0: //==! Data record
        lCurByteAddr = lAbsByteAddr;
        lAbsByteAddr = lLinearBaseAddr + lSegmentBaseAddr
                     + pCMCSvar.lByteAddr;
        if (lAbsByteAddr < lCurByteAddr)
        { sprintf(sMsgBuf00,"Overlapping bytes starting at addr [%X]",
                  lAbsByteAddr);
          jMessage(DISP_AND_EXIT,sMsgBuf00,1,pDP_In);
        } // IF
//==! Check if the address of the new record starts where the previous or
//==! current record ended. If true, then the bytes are contiguous, else
//==! there is a gap in the byte map. In the latter case, the bytes in the
//==! gap are assumed to be 0s and just add the difference to the total
//==! byte count.
        lTotByteCount += pCMCSvar.lByteCount;
        if (lAbsByteAddr > lCurByteAddr)
        { lTotByteCount += lAbsByteAddr - lCurByteAddr;
          sprintf(sMsgBuf00,"Gap in byte addr [%X]->[%X]",
                  lCurByteAddr,lAbsByteAddr);
          jMessage(DISP_SHOW_MSG,sMsgBuf00,1,pDP_In);
        } // IF
        lAbsByteAddr += pCMCSvar.lByteCount;
        break;
      case 1: //==! End of MCS file
        freeArray(pCMCSvar.pUCarray);
        return lTotByteCount;
      case 2: //==! extended segment address record (defines address bits 19
              //==! to  4)
        lLinearBaseAddr = 0;
        lSegmentBaseAddr = pCMCSvar.lOffset << 4;
        break;
      case 3: //==! start record (ignored by most device programmers)
        break;
      case 4: //==! extended linear address record (defines address bits 31
              //==! to 16)
        lLinearBaseAddr = pCMCSvar.lOffset << 16;
        lSegmentBaseAddr = 0;
        break;
      case 5: //==! start linear address record (ignored)
        break;
      default:
        sprintf(sMsgBuf00,"Invalid record type [%d]",pCMCSvar.iRecType);
        jMessage(DISP_AND_EXIT,sMsgBuf00,1,pDP_In);
        break;
    } // SWITCH (pCMCSvar.iRecType)
    freeArray(pCMCSvar.pUCarray);
  } // FOR readMCSrec() == 1
  return lTotByteCount;

} // getBitstreamByteSize()
//==!=====================================================================!==
//==! Get a byte (i.e., 8-bit symbol) from the MCS input file. This routine
//==! assumes that the bytes in the MCS file are in contiguous locations and
//==! start from location 0. Input data port must already be open.
//==!
//==! Function Parameters: pMCSfile = pointer to an MCS file structure
//==! Return Value       : 1 if successful, 0 otherwise
//==!=====================================================================!==
int getMCSByte(CMCSfile *pMCSfile)
{ CMCSrecord pCMCStmp;


  if (pMCSfile->pCMCSinfo == NULL)
  { pMCSfile->pCMCSinfo = new CMCSrecord;
    pMCSfile->pCMCSinfo->lByteCount = 0;
    pMCSfile->pCMCSinfo->lBufIndex = 0;
    pMCSfile->pCMCSinfo->iDataAvail = 1;
  } // IF pMCSfile->pCMCSinfo == NULL

  if (pMCSfile->pCMCSinfo->iDataAvail == 0)
  { pMCSfile->ucByte = 0;    //==! Default to a NULL byte
    return 0;
  } // IF

//==! Buffer still has some bytes. Current version assumes MCS rec read
//==! to always have an even number of hex chars.
  if (pMCSfile->pCMCSinfo->lBufIndex < pMCSfile->pCMCSinfo->lByteCount)
  { pMCSfile->ucByte = pMCSfile->pCMCSinfo->pUCarray[
                             pMCSfile->pCMCSinfo->lBufIndex
                                                    ];
    ++(pMCSfile->pCMCSinfo->lBufIndex);
    ++(pMCSfile->pCMCSinfo->lCurAddr);
    return 1;
  } // IF

//==! Deallocate array before reading the next MCS record.
  freeArray(pMCSfile->pCMCSinfo->pUCarray);

  while (readMCSrec(pMCSfile->pDP,&pCMCStmp) == 1)
  { switch(pCMCStmp.iRecType)
    { case 0:
        pMCSfile->pCMCSinfo->lBufIndex = 0;
        pMCSfile->pCMCSinfo->lByteCount = pCMCStmp.lByteCount;
        pMCSfile->pCMCSinfo->lByteAddr = pCMCStmp.lByteAddr +
                   pMCSfile->pCMCSinfo->lCurLinearOffset;
        pMCSfile->pCMCSinfo->lCurAddr = pMCSfile->pCMCSinfo->lByteAddr;
        pMCSfile->pCMCSinfo->iRecType = pCMCStmp.iRecType;
        pMCSfile->pCMCSinfo->pUCarray = pCMCStmp.pUCarray;
        pCMCStmp.pUCarray = NULL;
        isInRange(1,pMCSfile->pCMCSinfo->lBufIndex,
                   0,pMCSfile->pCMCSinfo->lByteCount,
                   "0MCSByteCnt",DISP_AND_EXIT);
        pMCSfile->ucByte = pMCSfile->pCMCSinfo->pUCarray[
                                pMCSfile->pCMCSinfo->lBufIndex
                                                        ];
        ++(pMCSfile->pCMCSinfo->lBufIndex);
        return 1;
      case 2:
        pMCSfile->pCMCSinfo->lCurLinearOffset = pCMCStmp.lOffset << 4;
        break;
      case 4:
        pMCSfile->pCMCSinfo->lCurLinearOffset = pCMCStmp.lOffset << 16;
        break;
      case 1: //==! End of MCS file found
        pMCSfile->pCMCSinfo->iDataAvail = 0;
        return 0;
      case 3: //==! Ignore these record types
      case 5:
        break;
    } // SWITCH
//==! Deallocate array before reading the next MCS record.
    freeArray(pCMCStmp.pUCarray);
  } // WHILE (readMCSrec() == 1)
  return 0;

} // getMCSByte()
//==!=====================================================================!==
//==! Write byte to output data port using MCS format using addressing data in
//==! pCMCSrecInfo. Print an MCS record when pCMCSinfo->lByteCount bytes have
//==! been read. Output data port is already opened as an ASCII output file.
//==!
//==! If iMode = 0  print out MCS record when it contains lByteCount bytes
//==!            1: print out remaining bytes as the last MCS record
//==!
//==! Function Parameters: pDP_Out = output data port
//==!                      pCMCSrecInfo = pointer to MCS output record info
//==!                          (e.g., output record size, current record
//==!                          address, etc.)
//==!                      ucByte = byte to print out
//==!                      iMode = see above description
//==! Return Value       : none
//==!=====================================================================!==
void writeMCSbyte(cDataPort *pDP_Out,CMCSrecord *pCMCSrecInfo,
           unsigned char ucByte,int iMode)
{ long lVar00;

  if (iMode == 0)
  { if (pCMCSrecInfo->lBufIndex >= pCMCSrecInfo->lByteCount)
    { if (((pCMCSrecInfo->lCurAddr & 0xFFFF0000) >> 16) >
            unsigned(pCMCSrecInfo->lCurLinearOffset))
      { ++(pCMCSrecInfo->lCurLinearOffset);
//==! Use extended linear addressing mode (type 04).
        sprintf(sMsgBuf00,"02000004%04X",pCMCSrecInfo->lCurLinearOffset);
        fprintf(pDP_Out->pFILE,":%s%02X\n",sMsgBuf00,
            calcMCSCheckSum(sMsgBuf00,NULL));
        sprintf(sMsgBuf00,"Byte addr [%010X] hex",pCMCSrecInfo->lCurAddr);
        dispMsg(sMsgBuf00,DISP_PRE_CR,1,0);
      } // IF
      sprintf(sMsgBuf00,"%02X%04X00",pCMCSrecInfo->lByteCount,
              pCMCSrecInfo->lCurAddr & 0x0000FFFF);
      fprintf(pDP_Out->pFILE,":%s",sMsgBuf00);
      for (lVar00 = 0; lVar00 < pCMCSrecInfo->lByteCount; ++lVar00)
      { fprintf(pDP_Out->pFILE,"%02X",pCMCSrecInfo->pUCarray[lVar00]);
      } // FOR lVar00 < pCMCSrecInfo->lByteCount
      fprintf(pDP_Out->pFILE,"%02X\n",
              calcMCSCheckSum(sMsgBuf00,pCMCSrecInfo));
      pCMCSrecInfo->lCurAddr += pCMCSrecInfo->lByteCount;
      pCMCSrecInfo->lBufIndex = 0;
    } // IF lBufIndex >= pCMCSrecInfo->lByteCount
    pCMCSrecInfo->pUCarray[pCMCSrecInfo->lBufIndex] = ucByte;
    ++(pCMCSrecInfo->lBufIndex);
  } // IF iMode == 0
  else
  {
//==! Print any bytes left in the buffer; occurs when the lByteCount-byte
//==! buffer is partially-filled.
    if (pCMCSrecInfo->lBufIndex > 0)
    { for (lVar00 = pCMCSrecInfo->lBufIndex;
           lVar00 < pCMCSrecInfo->lByteCount;
           ++lVar00)
      { pCMCSrecInfo->pUCarray[lVar00] = 0;
      } // FOR lVar00 < pCMCSrecInfo->pByteCount

      if (((pCMCSrecInfo->lCurAddr & 0xFFFF0000) >> 16) >
            unsigned(pCMCSrecInfo->lCurLinearOffset))
      { ++(pCMCSrecInfo->lCurLinearOffset);
        sprintf(sMsgBuf00,"02000004%04X",pCMCSrecInfo->lCurLinearOffset);
        fprintf(pDP_Out->pFILE,":%s%02X\n",sMsgBuf00,
        calcMCSCheckSum(sMsgBuf00,NULL));
      } // IF
      sprintf(sMsgBuf00,"%02X%04X00",pCMCSrecInfo->lBufIndex,
             pCMCSrecInfo->lCurAddr & 0x0000FFFF);
      fprintf(pDP_Out->pFILE,":%s",sMsgBuf00);
      for (lVar00 = 0; lVar00 < pCMCSrecInfo->lBufIndex; ++lVar00)
      { fprintf(pDP_Out->pFILE,"%02X",pCMCSrecInfo->pUCarray[lVar00]);
      } // FOR lVar00 < lBufIndex
      fprintf(pDP_Out->pFILE,"%02X\n",
          calcMCSCheckSum(sMsgBuf00,pCMCSrecInfo));
    } // IF

    sprintf(sMsgBuf00,"00000001");  //==! Put end of file record
    fprintf(pDP_Out->pFILE,":%s%02X\n",sMsgBuf00,
            calcMCSCheckSum(sMsgBuf00,NULL));
    freeArray(pCMCSrecInfo->pUCarray);
    dispMsg(" ",DISP_PRE_LF,1,1);
  } // ELSE

} // writeMCSbyte()
//==!=====================================================================!==
//==! end of module XROUTINE.CPP
//==!=====================================================================!==

