/***************************************************************************/
/* J Drive (tm) Engine                                                     */
/* Copyright 2003-2005, Xilinx, Inc.                                       */
/* The use, reproduction, and distribution of this code is subject         */
/* to the J Drive license requirements and restrictions.                   */
/***************************************************************************/
//==!=====================================================================!==
//==! File X_BSDL.H version 1.0 April 11, 2003
//==! Contains defines for Xilinx BSDL file reader.
//==!
//==! See XPROJECT.H for a list of files required for this project.
//==!=====================================================================!==
//==! Program development history, program/project name and version number
//==! are in XPROJECT.H.
//==!=====================================================================!==
//==! Header files for program.
//==!=====================================================================!==
#include "xproject.h"         //==! Contains program name and version number
#include "windows.h"          //==! Required for DWORD type definition used
#include "winbase.h"          //==! by Sleep() in waitLoop()
#include <math.h>             //==! Required by ceil()
#include "ports.h"
//==!=====================================================================!==
//==! Defines/macros for program
//==!=====================================================================!==

//==!=====================================================================!==
//==! Program-specific definitions
//==!=====================================================================!==
//==! JTAG TAP states
//==!=====================================================================!==
#define TAP_RESET             160
#define TAP_IDLE              161
#define TAP_DRSELECT          162
#define TAP_DRCAPTURE         163
#define TAP_DRSHIFT           164
#define TAP_DREXIT1           165
#define TAP_DRPAUSE           166
#define TAP_DREXIT2           167
#define TAP_DRUPDATE          168
#define TAP_IRSELECT          169
#define TAP_IRCAPTURE         170
#define TAP_IRSHIFT           171
#define TAP_IREXIT1           172
#define TAP_IRPAUSE           173
#define TAP_IREXIT2           174
#define TAP_IRUPDATE          175
#define TAP_UNKNOWN           176   //==! Unknown TAP state
//==!===================================================================!==
//==! JTAG TAP opcodes
//==!===================================================================!==
#define OP_TMS         0x02
#define OP_TDI         0x03
#define OP_RUNTEST     0x05
#define OP_NOP         0x00
#define OP_END         0x07
#define OP_PROG        0x08
#define OP_JUMP        0x10
#define OP_TDI_TDO     0x04
#define OP_TDI_N       0x06
#define OP_ACETOOLS    0xAB
//==!=====================================================================!==
//==! CxDataExpr definitions:
//==! - integer and char constants used in a data expression
//==!=====================================================================!==
#define dNOT_SET            100
#define dHEX_CONSTANT       101
#define dINP_SPECIFIER      102
#define d_VAR_EXPR          103
#define dc_GET_VAR           '.'
#define dc_INPUT_OP          '?'
#define dc_OUTPUT_OP         '!'
#define dc_IO_OP             '@'      //==! 1532 IO operator "?!"
#define dc_ASSIGN_OP         '='
#define dc_COMPLEMENT_OP     '~'
#define dc_ADD_OP            '+'
#define dc_SUB_OP            '-'
#define dc_LEFT_SHIFT_OP     '('      //==! Left-shift operator "<<"
#define dc_RIGHT_SHIFT_OP    ')'      //==! Right-shift operation ">>"

//==!=====================================================================!==
//==! CMemIscDataRowSection definitions:
//==! - integer and char constants used for memory data array class
//==!=====================================================================!==
#define dCMIDRS_SIZE         2048

//==!=====================================================================!==
//==! Values used by iGlobProgramState (can also be defined as enumerated
//==! values) NOTE_0009.
//==!=====================================================================!==
#define dPROGSTATE_INITIALIZE                10
#define dPROGSTATE_READCMDFILE               11
#define dPROGSTATE_CHECKCMDOPTIONS           12
#define dPROGSTATE_READBSDL                  13
#define dPROGSTATE_READBSDLPORT              14
#define dPROGSTATE_READBSDLATTR              15
#define dPROGSTATE_READISCDATA               16
#define dPROGSTATE_PROCESSACTION             17
#define dPROGSTATE_PROCESSDEVICE_SEQUENTIAL  18
#define dPROGSTATE_PROCESSDEVICE_CONCURRENT  19
#define dPROGSTATE_PROCESSFLOWACTIVITY       20
#define dPROGSTATE_PROCESSUPDATEFIELD        21
#define dPROGSTATE_PROCESSCAPTUREFIELD       22

//==!=====================================================================!==
//==! Valid values for enPinType in CPortData
//==!=====================================================================!==
enum EnumPortType
{ ePortNoType, ePortInp, ePortOut, ePortIO, ePortLinkage, ePortBuffer };
//==!=====================================================================!==
//==! Valid values for enBscResult in CBscData
//==!=====================================================================!==
enum EnumBscResult
{ eBscNoResult, eBscResultZ, eBscResultPull0, eBscResultPull1,
  eBscResultWeak0, eBscResultWeak1, eBscResultKeeper };
//==!=====================================================================!==
//==! Valid values for doHeaderTrailer()
//==!=====================================================================!==
enum EnumTapType
{ eHeader, eTrailer, eIreg, eDreg };
//==!=====================================================================!==
//==! Enumerated type for ISC data file override settings
//==!=====================================================================!==
enum EnumProcSetting
{ ePROC_NO_SETTING, ePROC_ON, ePROC_OFF, ePROC_OPTIONAL, ePROC_RECOMMENDED };
//==!=====================================================================!==
//==! Valid values for enBscFunc in CBscData
//==!=====================================================================!==
enum EnumBscFunction
{ eNoBscFunc, eBscInput, eBscClock, eBscOutput2, eBscOutput3, eBscControl,
  eBscControlr, eBscInternal, eBscBidir, eBscObserve_Only };
//==!=====================================================================!==
//==! Constants and global variables used for data and file CRC computations
//==!=====================================================================!==
//==! Following are used in data file transmission CRC computation
#define CCITT_CRC  0x8408      //==! bit-mask for CCITT CRC polynomial
#define CHAR_BIT   8           //==! number of bits per byte
DEF_DEC unsigned short usCRC_16register;   //==! global 16-bit shift register

//==! Following are used in data CRC algorithm
#define Polynomial 0x04C11DB7  //==! CRC-32 polynomial (Ethernet)
#define High_Bit   0x80000000  //==! Bit 31
#define Low_Bit    0x00000001  //==! Bit 0
DEF_DEC unsigned int uiCRC_32register;  //==! 32-bit CRC accumulator
//==!=====================================================================!==
#define PC_SYSTEM  1  // Set to 0 if compiling with other embedded code
#define PC_DEBUG   0  // Set to 1 to turn on debugging statements.
                      // As of 2000/07/28: added code to print out TAP
                      //                   transitions.
#define USE_BURST  1  // 2002/10/31: Set to 0 to use callport() to drive
                      // and read TAP values. If 1 then use burstData() as
                      // much as possible to increase data transfer speed.
//==!=====================================================================!==
DEF_DEC int g_verbose;
//==!========================================================================
#define chrtohex(cChr) (isalpha(cChr) ? tolower(cChr) - 'a' + 10 : cChr - '0')
#define hextochr(cChr) (cChr - 10 >= 0 ? cChr - 10 + 'a' : cChr + '0')
//==!=====================================================================!==
//==! Constants
//==!=====================================================================!==
//==! The following 32-element array contains masks for each of the 32
//==! possible bit positions. Ex: 0x00000001 has 1 in bit pos 0.
const unsigned long culGlobBitPosMask[32] = {
    1      , 1 <<  1, 1 <<  2, 1 <<  3, 1 <<  4, 1 <<  5, 1 <<  6, 1 <<  7,
    1 <<  8, 1 <<  9, 1 << 10, 1 << 11, 1 << 12, 1 << 13, 1 << 14, 1 << 15,
    1 << 16, 1 << 17, 1 << 18, 1 << 19, 1 << 20, 1 << 21, 1 << 22, 1 << 23,
    1 << 24, 1 << 25, 1 << 26, 1 << 27, 1 << 28, 1 << 29, 1 << 30, 1 << 31
};

//==!=====================================================================!==
//==! Forward declarations
//==!=====================================================================!==
class CIscFlow;
class CMemIscDataSrc;
class CBsdlData;
class CMemIscDataContext;
class CDeviceData;
//==!========================================================================
//==! Classes for BSDL reader
//==!========================================================================
class CMCSrecord
{
public:
  long lByteCount;
  long lByteAddr;
  long lOffset,
       lBufIndex,
       lCurAddr,
       lCurLinearOffset;
  int  iRecType,iDataAvail;
  unsigned char *pUCarray;

  CMCSrecord();
  ~CMCSrecord();
  void resetMCSrec();
};
//==!=====================================================================!==
//==! CMCSfile class has pointers to a cDataPort and a CMCSrec record. The
//==! CMCSfile pointer is provided just in case a linked list is needed;
//==! The lCMCSfileCnt var is used to keep track of how many CMCSfiles there
//==! are in the list or array. iBitPos indicates which bit position the bits
//==! in this file will go to. lCMCSfileCnt and iBitPos were originally used
//==! in XINTERLV utility.
//==!=====================================================================!==
class CMCSfile
{
public:
  long lCMCSfileCnt;
  int  iBitPos;
  unsigned char ucByte;
  cDataPort *pDP;
  CMCSrecord *pCMCSinfo;
  CMCSfile *pNextCMCSfile;

  CMCSfile();
  ~CMCSfile();
  void reopenMCS(char *sMode);
};
//==!=====================================================================!==
//==! CBscData = boundary scan cell data
//==!=====================================================================!==
class CBscData
{
public:
  long lPortNum;               //==! Port that BSC is associated with; this
                               //==! is the index into the port arry
                               //==! CBsdl->pCPortDataInfo[]
  long lBscIndex;              //==! bsc number/index
  long lBscCellType;           //==! bsc type; if 0 then BC_n else AC_n
  long lBscCellNum;            //==! bsc type number n from cell type BC_n
                               //==! where n = 0 to 99 (note that boundary
                               //==! scan handbook states that BSDL has
                               //==! reserved names BC_0 to BC_99)
  EnumBscFunction enBscFunc;   //==! bsc function
  int iSafeVal;                //==! Cell safe value
  long lCtrlCell;              //==! Control cell for this cell
  int iDisVal;                 //==! Value to disable this cell
  EnumBscResult enDisResult;   //==! Value of disabled cell

  CBscData();
  ~CBscData();
};

//==!=====================================================================!==
//==! CPortData = BSDL pin data class to hold pin information
//==!=====================================================================!==
class CPortData
{
public:
  char *csPortName,       //==! Port name in ASCII
       *csBasePortName;   //==! Base port name is the same as csPortName for
                          //==! scalar ports. Port vector names will not have
                          //==! the vector array index (e.g., base name is
                          //==! "IO" while csPortName can be "IO(25)")
  long lPortNameCnt;      //==! Number of ports with this name (usually for
                          //==! GND and VCC)
  int  iPortNameDir;      //==! 0: port vector in [max..min] order, else
                          //==! [min..max]
  long lMinNameArrayIndex,
       lMaxNameArrayIndex;
  long lPortLocation;     //==! port number; if device uses site numbers,
                          //==! then this is a user-assigned number while
                          //==! csPortLocation will contain the site number
                          //==! read from the BSDL file; else csPortLocation
                          //==! will contain the same port location info but
                          //==! in ASCII string format
  char *csPortLocation;   //==! Site number or ASCII of lPortLocation
  EnumPortType enPortType;    //==! Port type (input,output,IO,linkage)
  long lInCell,           //==! Input BSC index (if input or IO type)
       lOutCell,          //==! Output BSC index (if output or IO)
       lCtrlCell;         //==! Output control BSC index (if output or IO)
  CPortData *pNextCPortData;

  CPortData();
  ~CPortData();
  void removePin(CPortData **pCPD);
};

//==!=====================================================================!==
//==! SList = structure to create linked list of longs
//==!=====================================================================!==
struct SList
{ long lVal;
  SList *pNextSList;

  SList();
  ~SList();
};

//==!=====================================================================!==
//==! CBsdlDataReg = BSDL data register class. The opcodes that use a data
//==! register are contained in pOpList which is a linked list of indices
//==! into pCBsdlOpInfo[].
//==! = pOpList -> linked list of opcode indices (values index into
//==!   pCBdlOpInfo[] in the CBsdlData object); these are the opcodes that
//==!   use this data register
//==!=====================================================================!==
class CBsdlDataReg
{
public:
  char *csRegName;             //==! register name
  long lRegLength;             //==! register length
  SList *pOpList;              //==! see above description
  CBsdlDataReg *pNextCBsdlDataReg;

  CBsdlDataReg();
  ~CBsdlDataReg();
};

//==!=====================================================================!==
//==! CxCStr = Xilinx string object class that can be made part of a singly
//==! linked list.
//==!=====================================================================!==
class CxCStr
{
public:
  char *pStr;
  CxCStr *pNextCxCStr;

  CxCStr();
  ~CxCStr();
};
  
//==!=====================================================================!==
//==! CBsdlOp = BSDL opcode class. The data register associated with the
//==! instruction is determined by using lDataRegIndex as an index into the
//==! array CBsdlData.pCBsdlRegInfo[]
//==!=====================================================================!==
class CBsdlOp
{
public:
  char *csOpcodeName;          //==! opcode name
  CxCStr *m_CxCStr_OpcodeList; //==! char list of opcode bit encodings (i.e.,
                               //==! an opcode can have multiple bit
                               //==! encodings), but software will use the
                               //==! last one in the BSDL list.
  long lDataRegIndex;          //==! index into CBsdlData.pBsdlRegInfo[]
  int  iPrivate;               //==! 0: public, 1: private
  CBsdlOp *pNextCBsdlOp;

  CBsdlOp();
  ~CBsdlOp();
};

//==!=====================================================================!==
//==! CxDataVar = Xilinx variable object class for instantiating a variable
//==! and keeping track of its value. Instantiations can be made part of a
//==! singly linked list.
//==!=====================================================================!==
class CxDataVar
{
public:
  char *m_pcVarName;            //==! variable name
  long m_lVarWidth;             //==! variable width
  unsigned long
     *m_ulVarValue,             //==! current variable value
     *m_ulInitValue;            //==! variable initialize value
  CxDataVar *m_pNextCxDataVar;

  CxDataVar();
  ~CxDataVar();
};

//==!=====================================================================!==
//==! CFlowWait = Flow wait class defines wait specification
//==!=====================================================================!==
class CFlowWait
{
public:
  long m_lMinFlowWaitCycles,     //==! Min flow wait cycles
       m_lMaxFlowWaitCycles;     //==! Max flow wait cycles
  float m_fMinFlowWaitTimeSec,   //==! Min flow wait time in seconds
        m_fMaxFlowWaitTimeSec;   //==! Max flow wait time in seconds

  CFlowWait();
  ~CFlowWait();
  void resetWaitTimes();
};

//==!=====================================================================!==
//==! CxDataExpr = data expression field class. Used to define data
//==! expressions for update and capture data shifts.
//==! - m_iFieldType:
//==!   = dHEX_CONSTANT: constant/fixed hex value in m_pulDataField[]
//==!   = dINP_SPECIFIER: get m_lFieldWidth-wide value according to input
//==!     specifier ? or ?! in m_cOperator and store value in m_pulDataField
//==!   = d_VAR_EXPR: a var expression, if m_cOperator:
//==!     - dc_GET_VAR/'.': get value from variable in m_CxDataVarArray[]
//==!       indexed by m_lVarIndex and store in m_pulDataField
//==!     = dc_INPUT_OP/'?' or dc_IO_OP/'@'/'?!': get value from input source
//==!       and store in variable in m_CxDataVarArray[] indexed by m_lVarIndex
//==!     - dc_ASSIGN_OP/'=': store m_pulDataField into variable indexed by
//==!       m_lVarIndex
//==!     - dc_COMPLEMENT_OP/'~': invert m_CxDataVarArray[m_lVarIndex]
//==!     - dc_ADD_OP/'+' or dc_SUB_OP/'-': perform arithmetic operation on
//==!       variable indexed by m_lVarIndex with the value in m_ulOperand
//==!     - dc_LEFT_SHIFT_OP/'('/"<<" and dc_RIGHT_SHIFT_OP/')'/">>": perform
//==!       left or right shift by m_ulOperand bits with '0's shifted in side
//==!       of shift direction
//==!=====================================================================!==
class CxDataExpr
{
public:
  long m_lDE_len;                //==! Evaluated data expression width
  int m_iFieldType;
  long m_lVarIndex;
  unsigned long
     *m_pulDataField;            //==! Work area to hold expression
                                 //==! calculations using variable value from
                                 //==! the variable array indexed by
                                 //==! m_lVarIndex, or holds constant value
                                 //==! if m_iFieldType = dHEX_CONSTANT
  char m_cOperator;
  unsigned long m_ulOperand;

  CxDataExpr();
  ~CxDataExpr();
  void resetDataArea();
  void setDataBit(int iBitVal,long lBitPos);
  int  getDataBit(long lBitPos);
  void showFieldData(cDataPort *pCDP_FieldDest);
  void updateVar(CIscFlow *pCIFobj,int iMode);
  void evalOperator(CIscFlow *pCIFobj,int iMode);
};
  
//==!=====================================================================!==
//==! CxUpdateField = Update field class. Used to define data for the
//==! update data shifts. An update field is m_lFieldWidth long and consists
//==! of one data expression object.
//==!=====================================================================!==
class CxUpdateField
{
public:
  long m_lFieldWidth;                //==! Width of this update field
  CxDataExpr *m_pCDE_data;           //==! data expression

  CxUpdateField *m_pNextCxUpdateField;

  CxUpdateField();
  ~CxUpdateField();
};

//==!=====================================================================!==
//==! CxCaptureField = Capture field class. Used to define data for the
//==! capture data shifts.
//==!=====================================================================!==
class CxCaptureField
{
public:
  long m_lFieldWidth;                //==! Width of this capture field
  char m_cExpDataOutOp,
       m_cMaskDataOutOp,
       m_cCrcTag,
       m_cOstTag;
  CxDataExpr
     *m_pCDE_expdata,                //==! Expected data
     *m_pCDE_maskdata,               //==! Mask data
     *m_pCDE_CapturedData;           //==! Captured/raw data

  CxCaptureField *m_pNextCxCaptureField;

  CxCaptureField();
  ~CxCaptureField();
};

//==!=====================================================================!==
//==! CFlowAct = Flow action class.
//==!=====================================================================!==
class CFlowAct
{
public:
  long m_lMinLoopCount,          //==! Min loop count
       m_lMaxLoopCount;          //==! Max loop count
  CFlowAct *m_pLoopActions;      //==! flow actions if min/max > 0

  long m_lOpcodeIndex;           //==! Index into CBsdlData->CBsdlOpInfo[]
  CFlowWait *m_pFlowWaitInfo;    //==! pointer to wait specification

  long m_lUpdateWidth;           //==! Total update field width
  CxUpdateField *m_pUF_Update;   //==! Update field info

  long m_lCaptureWidth;          //==! Total capture field width
  CxCaptureField *m_pCF_Capture;
                                 //==! Capture field info
  CFlowAct *m_pNextCFlowAct;     //==! ptr to next flow action

  CFlowAct();
  ~CFlowAct();
  int getFlowType();             //==! See NOTE_0010 for details
  long processUpdateField(int iMode,char *pcTdiBuf,
             char *pcTdoBuf,CDeviceData *pCDD,CIscFlow *pCIFobj,int *iStat);
  int processCaptureField(int iMode,char *pcTdiBuf,char *pcTdoBuf,
             CDeviceData *pCDD,CIscFlow *pCIFobj,long lCurRptCount,
             int *iStatus,int iInLoop);
};
//==!=====================================================================!==
//==! CMultFlow = Multiple flow pointer. Used mainly to hold multiple repeat
//==! flows as defined in IEEE 1532.
//==!=====================================================================!==
class CMultFlow
{
public:
  long m_lRptCnt;               //==! Number of times to repeat the flow
  CFlowAct *m_pCFA_Repeat;      //==! Repeat flow pointer
  long m_lUseDataName_Repeat;   //==! needs data from ISC data file
  int m_iRptHasOutOp;           //==! 1 if m_pCFA_Repeat uses output op
  CMultFlow *m_pNextCMultFlow;  //==! Pointer to next flow

  CMultFlow();
  ~CMultFlow();
};

//==!=====================================================================!==
//==! CIscFlow = ISC_FLOW class. Stores ISC FLOW information.
//==!=====================================================================!==
class CIscFlow
{
public:
  char *m_csFlowName,            //==! flow name
       *m_csDataName;            //==! data to use with this flow
  int m_iUnprocessed,            //==! 1 if unprocessed tag detected
      m_iExitOnError;            //==! 1 if exit_on_error tag detected
  long m_lMaxRegSizeUsed,        //==! Max reg size used by activities in
                                 //==! this flow
      m_lUseDataName,            //==! If > 0, then csDataName refers to a
                                 //==! data record that is used by the flow
                                 //==! activities in this flow list
      m_lUseDataName_Initialize, //==! If > 1 then INITIALIZE flow phase
      m_lUseDataName_Terminate;  //==! needs data from ISC data file (same
                                 //==! for TERMINATE).

  CFlowAct *m_pCFA_Initialize,   //==! initialize flow
           *m_pCFA_Terminate;    //==! terminate flow

  CMultFlow *m_pCMF_Repeat;      //==! one or more repeat flows

  int m_iInitHasOutOp,           //==! 1 if m_pCFA_Initialize uses output op
      m_iTermHasOutOp;           //==! 1 if m_pCFA_Terminate uses output op

  long m_lCxDataVarArraySize;    //==! Max array size of m_CxDataVarArray[]
  CxDataVar **m_CxDataVarArray;  //==! array of variables used in this flow;
                                 //==! the variables' scope is limited to
                                 //==! this flow, so another flow could have
                                 //==! variables with the same name
  CxDataVar *m_pCDVlist,         //==! Variable information in singly-linked
                                 //==! list format
            *m_pCDVlistTail;     //==! Tail of m_pCDVlist
  CIscFlow *m_pNextCIscFlow;     //==! ptr to next flow

  CIscFlow();
  ~CIscFlow();
  CxDataExpr *getDataExpr(cDataPort *pDP,long lDataWidth);
  long getUpdateFieldList(cDataPort *pDP,CxUpdateField **pUF,
          long *lGetDataFlag,int *iHasOutputOp);
  void getFlowActWaitSpec(cDataPort *pDP,CFlowWait **pCFW,CBsdlData *pCBD);
  long getCaptureFieldList(cDataPort *pDP,CxCaptureField **pCF,
          long *lGetDataFlag,int *iHasOutputOp);
  void arrangeDataVars();
  long readFlowActivityList(cDataPort *pDP,int iMode,
             CBsdlData *pCBD,CFlowAct **pCFAhead,int *iHasOutputOp);

};

//==!=====================================================================!==
//==! CxFlowObj = class/container to hold flows
//==!=====================================================================!==
class CxFlowObj
{
public:
  char *m_csFlowParam;           //==! Optional flow parameter
  CIscFlow *m_pFlowItem;         //==! pointer to flow in m_pCIscFlowInfo
  CxFlowObj *m_pNextCxFlowObj;   //==! next flow object

  CxFlowObj();
  ~CxFlowObj();
};

//==!=====================================================================!==
//==! CIscProc = ISC_PROC class. Stores ISC Procedure information.
//==!=====================================================================!==
class CIscProc
{
public:
  char *m_csProcName,            //==! procedure name
       *m_csProcDataName;        //==! data to use with this procedure (just
                                 //==! a tag, the flow contains the real data
                                 //==! tag that must have a match in the ISC
                                 //==! data file)
  long m_lMaxRegSizeUsed;        //==! Max reg size used by flows in this
                                 //==! procedure.
  CxFlowObj *m_pProcFlows;       //==! Flows associated with this procedure
  CIscProc *m_pNextCIscProc;     //==! ptr to next procedure

  CIscProc();
  ~CIscProc();
};

//==!=====================================================================!==
//==! CxProcObj = Class/Container to hold procedures
//==!=====================================================================!==
class CxProcObj
{
public:
  char *m_csProcParam;           //==! Optional procedure parameter
  int m_iProcExclusivity;        //==! Procedure-level exclusivity
                                 //==! 0: not proprietary
                                 //==! 1: proprietary
  EnumProcSetting                //==! Procedure option settings
      m_enOptionSpec,            //==! - specification/setting from BSDL file
      m_enOptionOverride;        //==! - setting from ISC data file

  CIscProc *m_pProcItem;         //==! pointer to procedure in
                                 //==! m_pCIscProcInfo
  CxProcObj *m_pNextCxProcObj;   //==! next procedure object

  CxProcObj();
  ~CxProcObj();
};

//==!=====================================================================!==
//==! CIscAct = ISC_ACTION class. Stores ISC Action information.
//==!=====================================================================!==
class CIscAct
{
public:
  char *m_csActName,             //==! action name
       *m_csActDataName;         //==! data to use with this action (this
                                 //==! name must match name supplied by
                                 //==! calling program just in case there are
                                 //==! multiple actions with the same name
                                 //==! but which differ only by data name)
  long m_lMaxRegSizeUsed;        //==! Max reg size used by procedures in
                                 //==! this action.
  int m_iExclusivity;            //==! 1 if "proprietary" tag specified
  CxProcObj *m_pActionProcs;     //==! procedures associated with this action
  CIscAct *m_pNextCIscAct;       //==! ptr to next action

  CIscAct();
  ~CIscAct();
};

//==!=====================================================================!==
//==! CBsdlData = BSDL class to store select items of information from a
//==! BSDL file
//==!=====================================================================!==
class CBsdlData
{
public:
  char *csBsdlFileName,        //==! BSDL file name
       *csEntityName,          //==! Entity name
       *csPortMapName;         //==! Name used with PIN_MAP_STRING

  char *m_csTapClkName;        //==! TCK name from TAP_SCAN_CLK attribute
  float m_fTckFreq;            //==! TCK frequency (default = 1 MHz)

  CxCStr
       *m_CxCStr_IdcodeList,   //==! 32-bit JTAG IDCODE(s) in BSDL file
       *m_CxCStr_InstrCaptureList,
                               //==! Instruction capture bit pattern
       *m_CxCStr_IscBlankUsercodeList,
       *m_CxCStr_UsercodeList; //==! 32-bit usercodes
                               //==! 32-bit blank usercode values
  int iPortMapType;            //==! Port type
  long  lNumPorts,             //==! number of ports
        lInstrLength,          //==! JTAG instruction length
        lBscLength,            //==! boundary scan chain length
        lNumOpcodes,           //==! size of pCBsdlOpInfo[]
        lNumDataRegs,          //==! size of pCbsdlRegInfo[]
        lImplementState;       //==! 1 if implemented, 0 otherwise
  long lJtagStdLevel;          //==! JTAG STD level

  CBsdlOp **pCBsdlOpInfo;
  CBsdlDataReg **pCBsdlDataRegInfo,
        *m_pCBsdlDataRegListHead;  //==! List of BSDL data registers that
                               //==! will be moved to pCBsdlDataRegInfo[]

//==! Following are indices into pCBsdlOpInfo[] for the following commonly
//==! used instructions. Other instructions can be found by searching for the
//==! opcode name in pCBsdlOpInfo[].
  long lOpcodeSamplePreload,
       lOpcodeExtest,
       lOpcodeBypass,
       lOpcodeUsercode,
       lOpcodeUser1,           //==! opcode for Xilinx USER1 instruction
       lOpcodeUser2,           //==! opcode for Xilinx USER2 instruction
       lOpcodeIdcode;          //==! opcode for idcode

  CPortData **pCPortDataInfo;  //==! array of port pointers
  CPortData *pCPortDataList;

//==! 2004/05/28: Copy the ISC_PIN_BEHAVIOR instruction to m_csPinBehavior.
//==! If m_iPinBehvaiorCheck = 1, then check if the instruction in
//==! m_csPinBehavior is defined in the opcode list. This is necessary
//==! to allow the ISC_PIN_BEHAVIOR attribute to be defined before the
//==! INSTRUCTION_OPCODE attribute.
  char *m_csPinBehavior;
  int m_iPinBehaviorCheck;

//==!------------------------------------------------------------------------
//==! 2004/01/09: Changed pBscData from pointer to array to pointer to array
//==! of pointers to accomodate embedded OS and development environment used
//==! by Data I/O called ENEA. This member variable is the only one where
//==! program tries to allocate an array of objects of user-defined type
//==! (CBscData). All other object array allocations were for predefined
//==! types (i.e., char, long, etc.). This apparently does not cause problems
//==! for their embedded system.
//==!
//==! The pointer to array pointer definition just needs to be initialized
//==! and deallocated accordingly (see NOTE_0008).
//==!------------------------------------------------------------------------
  CBscData  **pBscData;        //==! pointer array of BSC cell pointers

  CIscFlow *m_pCIscFlowInfo;   //==! ISC Flow info
  CIscProc *m_pCIscProcInfo,   //==! ISC Procedure list
           *m_pCIscProcError;  //==! points to PROC_ERROR_EXIT procedure
  CIscAct  *m_pCIscActInfo;    //==! ISC Action list

  CBsdlData *m_pNextCBsdlData; //==! Pointer to next BSDL file

  CBsdlData();
  ~CBsdlData();
  int readBsdl(cDataPort *pDP);
  void semanticCheckBsdl();
  void readBsdlAttributes(cDataPort *pDP);
  void checkAttrEntity(cDataPort *pDP);
  EnumBscFunction getEnBscType(cDataPort *pDP,char *sBscTypeStr);
  EnumBscResult getEnBscResult(cDataPort *pDP,char *sBscResultStr);
  int getIsafeVal(cDataPort *pDP,char *sSafeValStr);
  long lAssignDataRegToOpcode(long lVal,char *sOpcode,cDataPort *pDP);
  void getBscInfo(cDataPort *pDP,long lBscArrayIndex);
  int verifyOverride(char *csAction,char *csActionData,
                     char *csProc,char *csProcData,
                     EnumProcSetting eProcSetting,cDataPort *pDP);
  long readBsdlPort(cDataPort *pDP_In);
  void readPortMap(cDataPort *pDP_In);
  long lAssignRegToOpcode(long lRegIndex,char *sOpName);
  long getOpcodeIndex(int iMode,char *sOpcode,cDataPort *pDP);
  void checkRegOpcodeAssociation(int iMode,char *sOpcode,
         char *sDataReg,cDataPort *pDP,long *plDataRegCnt,
         CBsdlDataReg **pm_pCBsdlDataRegListHead,
         CBsdlDataReg **pCBsdlDataRegTail);
  void readIscFlows(cDataPort *pDP);
  CIscFlow *getPtrToFlow(char *sFlowName,char *sFlowData,int iMode,
           cDataPort *pDP);
  void readIscProcedures(cDataPort *pDP);
  CIscProc *getPtrToProc(char *sProcName,char *sProcData,int iMode,
           cDataPort *pDP);
  void readIscActions(cDataPort *pDP);
  CFlowAct *readFlowActivityList(cDataPort *pDP,int iMode);

  void display1532Bsdl();
  void check1149StdLevel(cDataPort *pDP);

};

//==!=====================================================================!==
//==! CxCxSetting = holds procedure name, optional procedure data name, and
//==! procedure option setting
//==!=====================================================================!==
class CxCSetting
{
public:
  char *m_csProcName,
       *m_csProcData;
  EnumProcSetting m_enProcOption;
  CxCSetting *m_pNextCxCSetting;

  CxCSetting();
  ~CxCSetting();
};

//==!=====================================================================!==
//==! CxCxVarData = class to hold action and procedure option setting info
//==!=====================================================================!==
class CxCVarData
{
public:
  char *m_csActionName,
       *m_csActionDataName;
  CxCSetting *m_pCxCS_ProcSetting;
  CxCVarData *m_pNextCxCVarData;

  CxCVarData();
  ~CxCVarData();
};

//==!=====================================================================!==
//==! CIscData = class to hold 1532 ISC data
//==! - m_pCxCVD_OverrideList contains action-procedure plus override
//==!   settings that can change the default procedure options. For example,
//==!   an action that can call muktiple procedures:
//==!
//==!   erase = (proc_verify(idcode) recommended,
//==!            proc_enable,
//==!            proc_erase,
//==!            proc_disable),
//==!
//==! where "recommended" and "optional" mean (from 1532 standard):
//==!
//==! "The option "optional" menas the proecdire need no generally be
//==!  performed. A user interface that gives a user selection control of the
//==!  procedures making up an action will disable this procedure by default.
//==!  A user interface that does not give control of the procedures making
//==!  up an action does not perform this procedure.
//==!
//==!  The option "recommended" means the procedure need not be performed,
//==!  but is generally needed. A user interface that gives a user selection
//==!  control of the procedures making up a sequence enables this procedure
//==!  by default. A user interface that does not give control of the
//==!  procedures making up a sequence performs this procedure."
//==!
//==! The ISC data file can have an override specification list:
//==!
//==!    override proc_erase program OFF,
//==!    override proc_program(security) program ON
//==!
//==! which means:
//==! - when performing the PROGRAM action, the prc_erase procedure is NOT
//==!   performed (proc_erase must have been defined with the optional
//==!   or recommended option specification in the ISC_ACTION section)
//==! - when performing the PROGRAM action, the proc_program(security)
//==!   procedure is performed
//==!
//==!=====================================================================!==
class CIscData
{
public:
  cDataPort *m_pCDP_IscData; //==! data port pointer to ISC data; contains
                             //==! ISC data file name
  char *m_csStdVersion,
       *m_csCreationDate,    //==! file creation date in yyyymmdd.hhmm format
       *m_csCreatorString;   //==! NULL-terminated string usually describing
                             //==! tool that generated this file
  long m_lBracketLevel;      //==! Keeps track of brackets in the ISC file
  CxCStr *m_pCxCStr_EntityList;
                             //==! list of BSDL entities that can use this
                             //==! file
  CxCVarData *m_pCxCVD_OverrideList;
                             //==! list of override specifications (see
                             //==! class description)

  CIscData(char *cIscDataFileName);
  ~CIscData();
  void addSetting(char *csAction,char *csActionData,
                  char *csProc,char *csProcData,EnumProcSetting eProcSetting,
                  cDataPort *pDP,CBsdlData *pCBDvar);
  void processOverrideList(int iMode,CBsdlData *pCBDvar);
  int readIscDataHeader(int iMode,CDeviceData *pCDD,CBsdlData *pCBDvar,
                  char *csDataArray,
                  CMemIscDataContext *p_CMemIscDataContext_obj);
  void readIscDataSubHeader(char *csDataArray,char *csSubHeader,
                  long lMultDataHeaderIndex,
                  CMemIscDataContext *p_CMiscDataContext_obj);
  void copyCIscDataInfo(CDeviceData *pCDD);
  void testCreationDate(char *cDateString);
  EnumProcSetting cnvProcSet(char *sSetting);
  int checkDataRecord(int iMode,char *csDataArray,CMemIscDataSrc **p_pCMIDS);
};

//==!=====================================================================!==
//==! CMemIscDataRowSection = class to hold section (ie., part) of data row
//==! - data read from file as hex chars
//==! - up to m_lDataCnt chars are read into an array (can be less in the
//==!   last array)
//==! - more row section objects can be allocated to hold additional data
//==! - row sections are linked into a doubly-linked list with one direction
//==!   pointing to the next m_lDataCnt chars in the LSB direction and the
//==!   other in the MSB direction
//==!=====================================================================!==
class CMemIscDataRowSection
{
public:
  char *m_csHexData;        //==! hex data array
  long m_lDataCnt;          //==! num hex chars in this data row section
  CMemIscDataRowSection
    *m_pNextCMIDRS_LSB,     //==! next data row section in LSB direction
    *m_pNextCMIDRS_MSB;     //==! next data row section in MSB direction

  CMemIscDataRowSection(int iMode);
  ~CMemIscDataRowSection();
};
//==!=====================================================================!==
//==! CMemIscDataRow = class to hold row of data
//==! - ISC data row can consist of multiple sections
//==!   = Example: A row can be 4096 bits long (1024 hex chars). Instead of
//==!     allocating a single 1024 char array, allocate four 256 byte arrays
//==!     and link them together. See definition of CMemIscDataRowSection to
//==!     see how the row is stored such that the LSB or MSB can be easily
//==!     accessed via the head or tail pointers in this class/structure
//==! - class contains pointer to next row of data (separated by commas in
//==!   the ISC data file)
//==!=====================================================================!==
class CMemIscDataRow
{
public:
  long m_lTotDataCnt;       //==! total number of hex chars in this row
  CMemIscDataRowSection
    *m_pCMIDRS_Head,        //==! pointer to head of row section list
    *m_pCMIDRS_Tail;        //==! pointer to tail of row section list
  CMemIscDataRow
    *m_pNextCMemIscDataRow;  //==! next row of data

  CMemIscDataRow();
  ~CMemIscDataRow();
  void addHexChar(int iChar);
  void optimizeSpace();
};
//==!=====================================================================!==
//==! CMultDataSection = class representing multiple ISC data array sections
//==! in memory. For example, the 1532 data format allows multiple REPEAT
//==! sections in a data array.
//==!=====================================================================!==
class CMultDataSection
{
public:
  CMemIscDataRow
    *m_pCMIDR_DataRow;         //==! pointer to head of data rows in section
  CMultDataSection
    *m_pNextCMultDataSection;  //==! pointer to next data section

  CMultDataSection();
  ~CMultDataSection();

};
//==!=====================================================================!==
//==! CMemIscDataArray = class to hold ISC data array in memory
//==!=====================================================================!==
class CMemIscDataArray
{
public:
  char *m_csDataArrayName;  //==! data array name
  CMemIscDataRow
    *m_pCMIDR_Init,         //==! pointer to initialize data rows
    *m_pCMIDR_Term;         //==! pointer to head of terminate data rows
  CMultDataSection
    *m_pCMultDataSect_Rpt;  //==! Pointer to one or more REPEAT data
                            //==! sections
  CMemIscDataArray
    *m_pNextCMemIscDataArray;
                            //==! next data array; i.e., ISC data file can
                            //==! have multiple data arrays
  CMemIscDataArray();
  ~CMemIscDataArray();

};
//==!=====================================================================!==
//==! CMemIscDataSrc = class to hold ISC data arrays in memory
//==!=====================================================================!==
class CMemIscDataSrc
{
public:
  char *m_csDataSrcName;    //==! data src name (i.e., filename if ISC data
                            //==! is from a file)
  CMemIscDataArray
    *m_pCMIDA_list;         //==! pointer to list of arrays in this ISC
                            //==! data source
  CMemIscDataSrc
    *m_pNextCMemIscDataSrc; //==! next memory ISC data object

  CMemIscDataSrc();
  ~CMemIscDataSrc();

};
//==!========================================================================
//==! CMemIscDataContext = class to hold context information for ISC data in
//==! memory
//==!========================================================================
class CMemIscDataContext
{
public:
  CMemIscDataSrc *m_pCMIDS;
  CMemIscDataArray *m_pCMIDArray;
  CMemIscDataRow *m_pCMIDR_current;
  long m_lCMIDR_index;

  CMemIscDataContext();
  ~CMemIscDataContext();
};

//==!=====================================================================!==
//==! CProcContext = class to hold processing context information. Used
//==! mainly during concurrent action processing mode. The variables are
//==! the same as found in the function processingActions() which processes
//==! the action for a single device.
//==!=====================================================================!==
class CProcContext
{
public:
  int m_iBypass,            //==! 1 to bypass device
      m_iPhase,             //==! 0: do update phase, 1: do capture phase
      m_iSkipFlag,
      m_iFlowActStat,
      m_iDataAvail,m_iFlag0,
      m_iProcessingLoop;
  long m_lX,m_lY,m_lZ,m_lRptMax,
      m_lGetData,
      m_lLoopMin,
      m_lLoopMax,
      m_lLoopVar,
      m_l_OST;
  CIscAct *m_pCIAvar;
  CxProcObj *m_pProcObj;
  CxFlowObj *m_pFlowObjVar;
  CFlowAct *m_pCFAptr,*m_pCFAvar,*m_pCFAloopVar,*m_pCFA_LoopActions;

  CProcContext();
  ~CProcContext();
};
//==!=====================================================================!==
//==! CDeviceData = class to hold device information
//==!=====================================================================!==
class CDeviceData
{
public:
  char *m_csDeviceName,     //==! device name
       *m_csAction,         //==! action to perform on this device
       *m_csActionData;     //==! data for specified action
  CBsdlData *m_pCBsdlData;  //==! pointer to BSDL file in global pgCBD_list
                            //==! for this device
  cDataPort *m_pCDP_Out;    //==! Data port destination for any field output
                            //==! operations
  long m_lInstrRegLength;   //==! JTAG instruction reg length if m_pCBsdlData
                            //==! not defined
  int m_iSkipDev,           //==! 1 if this device is to be skipped; used in
                            //==! device programmer environments where a JTAG
                            //==! command file describes the maximum number
                            //==! of device sockets, but if there is no device
                            //==! in the socket, then this device list entry
                            //==! is skipped completely and not put into
                            //==! bypass mode
     m_iProcessingStatus,   //==! 1 if successful, else 0 (used during
                            //==! concurrent processing)
     m_iIscDataFormat;      //==! If 1 then write data from output operator
                            //==! to output file m_pCDP_Out using ISC data
                            //==! format
  CProcContext *m_pCProcContext;
                            //==! structure to hold processing context (used
                            //==! during concurrent processing only)
//==! Following variables are used during concurrent processing mode
  char *m_cSIR,             //==! device opcode to shift
       *m_cSIR_bypass,      //==! bypass opcode
       *m_cTdi,             //==! TDI (usually SDR data) to shift
       *m_cTdo,*m_cTdoCopy; //==! TDO and copy of TDO data
  int m_iFlowType,          //==! Specifies one of 4 possible flow types
                            //==! currently being processed (NOTE_0010).
                            //==!   0 = (Opcode,Update Data,Wait,Capture Data)
                            //==!   1 = (Opcode,            Wait             )
                            //==!   2 = (Opcode,Update Data,Wait             )
                            //==!   3 = (Opcode,            Wait,Capture Data)
                            //==! Note that during concurrent processing, it
                            //==! may be necessary to process devices in the
                            //==! chain with the same flow type while the
                            //==! others are temporarily placed in bypass
                            //==! (using the m_cSIR_bypass value). This safer
                            //==! concurrent processing mode is the default
                            //==! but is overridden with -m max_concurrent
      m_iHoldFlowType;      //==! Holds m_iFlowType while processing the
                            //==! device chain a max of 4 separate times
                            //==! (once for each flow)

//==! 2003/12/19: Added CMemIscDataContext object (not a pointer) to hold
//==! context information for accessing the ISC data in memory for this
//==! object. Context information is required because multiple devices
//==! can point to and access the same CMemIscDataSrc object, with each
//==! object reading from a different row object at the same time.
  CMemIscDataContext m_CMemIscDataContext;

  CIscData *m_pCIscData;    //==! ISC data file for this device
//==!========================================================================

  CDeviceData
       *m_pNextCDeviceData, //==! pointer to next device in JTAG chain from
                            //==! device closest to TDO to TDI
       *m_pNextCDDfromTdi;  //==! pointer to next device in JTAG chain from
                            //==! device closest to TDI to TDO; used mainly
                            //==! during concurrent processing mode to
                            //==! assemble data (opcode or data) into one
                            //==! register before shifting out

  CDeviceData();
  ~CDeviceData();
  char *getOpcodeInfo(long lOpIndex,char *pcsOpBits);
  void createTapReg(long l_IR,long l_DR);
  void loadBypassOp();
  void writeIscData(int iMode,char *cStrData,long lParam);
};

//==!=====================================================================!==
//==! Global variables for program
//==!=====================================================================!==
DEF_DEC int iOpt_SkipIdCheck,
            iOpt_ProgDebug,
            iOpt_RptCntDispOff,    //==! Set to 1 using -xopt 32 to turn off
                                   //==! repeat count display while processing
                                   //==! the repeat phase of a flow
            iOpt_FlowLoopDbg,
            iOpt_MinOneClk,        //==! Set to 1 using -min1clk to force
                                   //==! at least one TCK pulse 0-1-0 when
                                   //==! waiting X seconds
            iOpt_PrintBsdl,
            iOpt_ShowAllOpts,      //==! 1 to show all program options
            iOpt_ShowCmdFileOpts,  //==! 1 to show ISC command file options
            iOpt_CheckInputFiles,  //==! If 1 then check input files only
                                   //==! (default = 1); use -checkinpfiles to
                                   //==! set to 0
            iGlobOptUseMemIscData, //==! If 1 then store ISC data into mem
                                   //==! and use from mem to enhance
                                   //==! performance
            iGlobOptUseErrorProc,  //==! If 1 (default) then use mandatory
                                   //==! procedure PROC_ERROR_EXIT on errors
                                   //==! when processing actions; set to 0
                                   //==! using -skip_error_proc to just put
                                   //==! the part into bypass on error
            iGlobOptSkipExitOnError,
                                   //==! Skip EXIT_ON_ERROR if 1 (default = 0)
                                   //==! set to 1 using -skip_exit_on_error
            iGlobExitOnError,      //==! 1 if any flow with EXIT_ON_ERROR
                                   //==! attribute fails
            iGlobOptDisplayBsdlWarnings,
                                   //==! Set to 1 using -bsdlwarnings to
                                   //==! display BSDL DESIGN_WARNINGS or
                                   //==! ISC_DESIGN_WARNINGS (default = 0)
            iGlobOptContinueOnFail,  //==! set to 1 using -continue_on_fail
                                   //==! to continue processing even if an
                                   //==! error/failure was encountered
//==! iGlobProgramState is set to a value (search for NOTE_0009 for list)
//==! that can be used by another program to determine what state or function
//==! the program is in (requested by Data I/O).
            iGlobProgramState,

//==! If iGlobOptMaxConcurrent = 0, then process only those flows which have
//==! the same type (see NOTE_0010). If 1, then concurrently process the
//==! flows, using the previous or don't care values for flows that do not
//==! specify input data or output data in their flow activity specification.
//==! If iGlobOptMaxConcurrent = 0, it is possible to go through a chain
//==! 4 times to process flows of type 0, then 1, then 2, and then 3, with
//==! devices that have a different flow type (from the one being processed)
//==! being temporarily put into bypass. This mode allows concurrent
//==! processing of flows from JTAG devices that are not fully 1532-compliant. to be
//==! Use sequential processing mode if concurrent mode does not work.
            iGlobOptMaxConcurrent,  // Set using -m max_concurrent option
            iGlobOptCheckBsdl,      //==! Set using -checkbsdl option

            iGlobParportPin17Setting;
                            

// iGlobReadByte is used to assemble data read from the flash before
// writing it to an output file. iGlobReadBitCnt keeps track of how many
// valid bits of the byte have been read; print out byte when the cnt
// reaches 8.
DEF_DEC int iGlobReadByte,
    iGlobReadBitCnt,
    iStatus,
    iPortStatus,
    iMatchAll,
    iReadDevice,
    iReadUserArea,
    iReadUserAreaDone;
DEF_DEC long lMismatchCnt,  // Keeps track of bit/byte mismatches
    lCntBytesRead,
    lCntBytesWritten,
    lHdr,lHir,lTdr,lTir,
    lOptMaxBitErr,          // Tracks max number of bit mismatches
    lMaxIrChainLength,      // Specifies maximum IR bit length to use in
                            // JTAG integrity test to put all devices in the
                            // JTAG chain to BYPASS before attempting to
                            // determine the number of devices in the chain.
                            // Default is 200 and can be changed using
                            // "-irchainlength X".
    lMaxDrChainLength,      // Specifies maximum DR bit length to use in
                            // JTAG integrity test to determine max number of
                            // devices in a JTAG chain. Can be changed using
                            // "-drchainlength X".
    lTdiTdoInstrCnt;        // Keeps track of number of JTAG cmds processed

#if PC_DEBUG == 1
int iTapState,iTapPrint;
#endif

DEF_DEC CMCSrecord *pCMCSobject;
DEF_DEC int iChar,
    iByte,iByteOpcode,iTdi,iTdo,iMask,iGetDataFlag;
DEF_DEC long lMicroSecCount;
DEF_DEC int iOptBscanTests,iOptContScan,iOptCalibrateTimer,iOptByteOrder;
DEF_DEC CMCSfile *pMcsFile;
//==!=====================================================================!==
//==! 2003/01/20:
//==! - cArrayBurstReadTdo[] holds TDO data read back from burstData()
//==!=====================================================================!==
DEF_DEC char
    cArrayBurstReadTdo[MAX_BYTE_BURST_SIZE];

DEF_DEC CBsdlData *pBsdlData;
DEF_DEC CDeviceData
           *pgCDD_list,        //==! Global device data list
           *pgCDD_FromTdi;     //==! Global device data list starting from TDI
DEF_DEC CBsdlData *pgCBD_list;    //==! Global BSDL file list
DEF_DEC CMemIscDataSrc *pgCMIDS_list; //==! Global mem ISC data source list
DEF_DEC CPortData *pTempCPortData;
DEF_DEC int iGlobProcMode;  //==! 0: sequential, 1: concurrent
DEF_DEC char *psGlobBsdlStr;  // holds name of BSDL file to check
DEF_DEC char *cTdiData,*cTdoData,*cTdoCopy,
          *cGlobTapDataReg;    // global TAP data reg used in concurrent mode
                               // - IR and DR data from each device is
                               //   assembled into this register before
                               // shifting out to the JTAG chain
DEF_DEC long lGlob_IscCounter; // Debug: used to format data being sent to
                               // calcDataCrc()
//==!=====================================================================!==
//==! Function prototypes
//==!=====================================================================!==
void debugStrHex(char *sBinData,long lLen,int iMode,int iDest);

void initProgVariables();
void freeProgMem();
void xstrcat(char **sDest,char *sSrc);

void doSyncJtagReset(int iMode);
void doHeaderTrailer(EnumTapType enMode,EnumTapType enReg);
void jtagShift(int iMode,long lShiftLen,char *pcDataIn,char *pcDataOut,
               int iTdoCheck,int iWaitInUpdatexR);

void debugMsg(char *sMsg,int iMode,int iDest);
void formatData(int iMode,long lDataLen,char *pcData,char *pcDataBuf);
void dispInfo(int iMode,char *sMsg);
unsigned int calcMCSCheckSum(char *sBuf,CMCSrecord *pRec);
int getMCSByte(CMCSfile *pMCSfile);
void writeMCSbyte(cDataPort *pDP_Out,CMCSrecord *pCMCSrecInfo,
           unsigned char ucByte,int iMode);
long getBitstreamByteSize(cDataPort *pDP_In);

void prepareDataBuffers(CDeviceData *pCDD,int iProcMode);
void freeDataBuffers(CDeviceData *pCDD);
long getDataRegWidth(CBsdlData *pCBD,long lOpcodeIndex);

void xGetData(cDataPort *pDP,int iMode,EnumPARSE_ATTR eGetType,char *sGVmsg);
int readCmdFile(cDataPort *pDP);
int isIscDataRequired(CDeviceData *pCDDobj);
int checkCmdOptions(CDeviceData *pDeviceList);
int processDeviceList(CDeviceData *pCDD,long lProcIteration);

//==!=====================================================================!==
//==! end of X_BSDL.H
//==!=====================================================================!==

