/***************************************************************************/
/* J Drive (tm) Engine                                                     */
/* Copyright 2003-2005, Xilinx, Inc.                                       */
/* The use, reproduction, and distribution of this code is subject         */
/* to the J Drive license requirements and restrictions.                   */
/***************************************************************************/
//==!=====================================================================!==
//==! See XPROJECT.H for project file list and project version history.
//==!=====================================================================!==
//==! Revision history: XRD_BSDL.CPP
//==! 2003/04/11: BSDL file reader (1149.1 and 1532-2002). Some attributes
//==!             are skipped and not syntactically checked.
//==!=====================================================================!==
#include "xcommon.h"
#include "x_bsdl.h"
//==!=====================================================================!==
//==! Look for cChr in the BSDL file pDP.
//==!
//==! Function Parameters: pDP = pointer to BSDL cDataPort
//==!                      cChr = char to look for
//==! Return Value       : none
//==!=====================================================================!==
void findBsdlChar(cDataPort *pDP,char cChr)
{

  while (1)
  { getDpChar(pDP,1);
    if (ispunct(pDP->iCurChar))
    { if (isComment(pDP) == 1)
      { continue;
      } // IF
    } // IF
    if (pDP->iCurChar != '"')
    { sprintf(sMsgOnly,"Expected [%c] in BSDL file",pDP->iCurChar);
      jMessage(DISP_AND_EXIT,sMsgOnly,1,pDP);
    } // IF
    return;
  } // WHILE (1)

} // findBsdlChar()
//==!=====================================================================!==
//==! Return next character from BSDL data port through the cDataPort's
//==! iCurChar variable. If iMode = 0, then return the first char read, else
//==! return the first non-whitespace character. If a double quote is
//==! encountered, then look for either the '&' or ';' chars to continue with
//==! the string on the next line, or to terminate the string, respectively.
//==!
//==! Ex: "1876: FFFFFFFF" &   is equivalent to "1876: FFFFFFFFFFFF         "
//==!     "FFFF          " &
//==!
//==! while
//==!     "1876: FFFFFFFF" &   is equivalent to "1876: FFFFFFFF FFFF        "
//==!     " FFFF         " &
//==!
//==! Function Parameters: pDP = pointer to cDataPort
//==!                      iMode = see function description
//==! Return Value       : none
//==!=====================================================================!==
void getDpBsdlChar(cDataPort *pDP,int iMode)
{

  while (1)
  { getDpChar(pDP,0);
    if (pDP->iCurChar == '"')
    { ++pDP->iBracketLevel;
      getDpChar(pDP,1);
      if (pDP->iCurChar == '&')
      { pDP->iBracketLevel -= 2;
        if (pDP->iBracketLevel != 0)
        { sprintf(pDP->sData,"%c",pDP->iCurChar);
          jMessage(DISP_AND_EXIT,"Mismatched BSDL string brackets",1,pDP);
        } // IF
        findBsdlChar(pDP,'"');
        ++pDP->iBracketLevel;
      } // IF
      else
      { if (pDP->iCurChar == ';')
        { pDP->iBsdlEntityEnd = 1;
        } // IF
        return;
      } // ELSE
    } // IF
    else
    { if (pDP->iCurChar == ';')
      { pDP->iBsdlEntityEnd = 1;
      } // IF
      return;
    } // ELSE
  } // WHILE (1)

} // getDpBsdlChar()
//==!=====================================================================!==
//==! Get next data from BSDL file. This function supports the following
//==! BSDL format:
//==!     "TCK:C3, " &
//==!     "TDI:A11," &
//==!     "TDO:A12 " &  <-- note ',' is on next line
//==!     ",TMS:B1," &
//==! If function sees the trailing '"' character, it will look for the '&'
//==! and then the following '"' before getting the next data token.
//==!
//==! Function Parameters: pDP = input data port pointer
//==!                      iMode = used by xParseData() (see description)
//==! Return Value       : none
//==!=====================================================================!==
void xDataBsdl(cDataPort *pDP,int iMode)
{ int iX;
  long lNewSize;
  cCharStr *pcCStail,*pcCSvar;

//==! Set to 1 to read BSDL tokens terminated by the '"'. Set to 0 to use
//==! getDpBsdlChar() to read the next char in the next BSDL string enclosed
//==! by double quotes (see getDpBsdlChar() description).
#if 0
  while (1)
  { pDP->iBsdlEntityEnd = 0;
    xParseData(pDP,iMode);
    if (!strcmp(pDP->sData,"\""))
    { ++pDP->iBracketLevel;
      xParseData(pDP,XDATA_READ);
      if (!strcmp(pDP->sData,"&"))
      { pDP->iBracketLevel -= 2;
        if (pDP->iBracketLevel != 0)
        { jMessage(DISP_AND_EXIT,"Mismatched brackets",1,pDP);
        } // IF
        findStr(pDP,"\"",0);
        ++pDP->iBracketLevel;
      } // IF
      else
      { if (pDP->sData[0] == ';')
        { pDP->iBsdlEntityEnd = 1;
        } // IF
        return;
      } // ELSE
    } // IF
    else
    { return;
    } // ELSE
  } // WHILE (1)
#else
  if (pDP->iParserGetData == 1)
  { pDP->iBsdlEntityEnd = 0;
    pDP->m_pcCharStr->m_csStr[0] = '\0';
    while (pDP->m_iEndOfDataStream == 0)
    { getDpBsdlChar(pDP,0);

//==! Alphanumeric character read
      if (isalnum(pDP->iCurChar) || pDP->iCurChar == '_')
      { if (isdigit(pDP->iCurChar))     //==! data starts with a number
        { xParseNumData(pDP,(char)pDP->iCurChar);
        } // IF
        else
        { for (iX = 0,pcCStail = pDP->m_pcCharStr;
               pDP->m_iEndOfDataStream == 0;
               ++iX, getDpBsdlChar(pDP,0))
          { if (isalnum(pDP->iCurChar) || pDP->iCurChar == '_')
            { if (iX >= MAX_XDATA_SIZE)
              { pcCStail->m_csStr[MAX_XDATA_SIZE] = '\0';
                pcCStail->m_lSize = MAX_XDATA_SIZE;
                pcCStail->m_pNextcCharStr = new cCharStr;
                pcCStail = pcCStail->m_pNextcCharStr;
                iX = 0;
              } // IF iX >= MAX_XDATA_SIZE
              pcCStail->m_csStr[iX] = pDP->iCurChar;
            } // IF
            else
            { break;
            } // ELSE
          } // FOR iX
          pcCStail->m_csStr[iX] = '\0';
          pcCStail->m_lSize = iX;

          pDP->eDataAttr = eATTR_VAR;
          if (pcCStail != pDP->m_pcCharStr)
          { for (lNewSize = 0, pcCStail = pDP->m_pcCharStr;
                 pcCStail != NULL;
                 pcCStail = pcCStail->m_pNextcCharStr)
            { lNewSize += pcCStail->m_lSize;
            } // FOR pcCStail != NULL

            if (lNewSize > pDP->m_lDpDataSize)
            { freeArray(pDP->sData);
              freeArray(pDP->sOrigData);
              pDP->sData = new char [lNewSize+1];
              pDP->sOrigData = new char [lNewSize+1];
              pDP->m_lDpDataSize = lNewSize;
            } // IF
          } // IF pcCStail != pDP->m_pcCharStr

          for (pDP->sOrigData[0] = '\0', pcCStail = pDP->m_pcCharStr;
               pcCStail != NULL;
               pcCStail = pcCStail->m_pNextcCharStr)
          { strcat(pDP->sOrigData,pcCStail->m_csStr);
          } // FOR pcCStail != NULL

//==! Check if the data array was resized. If so, then delete any of the
//==! linked list arrays that were used to temporarily hold the long string.
          if (pDP->m_pcCharStr->m_pNextcCharStr != NULL)
          { pcCStail = pDP->m_pcCharStr->m_pNextcCharStr;
            pDP->m_pcCharStr->m_pNextcCharStr = NULL;
            while (pcCStail != NULL)
            { pcCSvar = pcCStail->m_pNextcCharStr;
              pcCStail->m_pNextcCharStr = NULL;
              freeData(pcCStail);
              pcCStail = pcCSvar;
            } // WHILE pcCStail != NULL
          } // IF pDP->m_pcCharStr->m_pNextcCharStr != NULL

          if (pDP->m_iEndOfDataStream == 0)
          { pDP->iParserGetChar = 0;
          } // IF
        } // ELSE
      } // IF isalnum()
      else
      { sprintf(pDP->m_pcCharStr->m_csStr,"%c",pDP->iCurChar);
        if (ispunct(pDP->iCurChar))
        { pDP->eDataAttr = eATTR_PUNCT;
          if (isComment(pDP) == 1)
          { continue;
          } // IF
        } // IF ispunct()
        else //==! All other characters treated as white-space.
        { if (isspace(pDP->iCurChar))
          { pDP->eDataAttr = eATTR_WSPACE;
          } // IF
          else
          { pDP->eDataAttr = eATTR_OTHERS;
          } // ELSE
        } // ELSE
        strcpy(pDP->sOrigData,pDP->m_pcCharStr->m_csStr);
      } // ELSE

      strcpy(pDP->sData,pDP->sOrigData);

      if (pDP->eDataAttr != eATTR_WSPACE)
      { if (pDP->eDataAttr == eATTR_VAR)
        { strCaseConvert(pDP->sData,strlen(pDP->sData),0);
        } // IF
        return;
      } // IF
      else
      { if (iMode == XDATA_READ_RAW)
        { return;
        } // IF
      } // ELSE

    } // WHILE pDP->m_iEndOfDataStream == 0
  } // IF iCallParser == 1
  else
  { pDP->iParserGetData = 1;
  } // ELSE

#endif

} // xDataBsdl()
//==!=====================================================================!==
//==! Get a number from the BSDL file. This number can be on the next line of
//==! the BSDL file via the concatenation operation. Read the number by
//==! getting the next data string and then pushing it back into the read
//==! buffer for the lReadDecimal() call.
//==!
//==! Function Parameters: pDP = input data port pointer
//==!                      iMode = if 0 then number should be >= 0
//==! Return Value       : none
//==!=====================================================================!==
long lReadDecimalBsdl(cDataPort *pDP,int iMode)
{ long lValue;

  xDataBsdl(pDP,XDATA_READ);
  pDP->iParserGetData = 0;
  lValue = lReadDecimal(pDP);
  if (lValue < 0 && iMode == 0)
  { jMessage(DISP_AND_EXIT,"Positive number expected",1,pDP);
  } // IF
  return lValue;

} // lReadDecimalBsdl()
//==!=====================================================================!==
//==! Check if the BSDL entity was terminated properly. Specifcally, if an
//==! attribute entity is a string, it should be termined by a ';' after the
//==! quote-enclosed string.
//==!
//==! Ex: entity is "  ";    <== correct
//==!     entity is " ;"     <== incorrect
//==!
//==! Function Parameters: pDP = input data port pointer
//==! Return Value       : none
//==!=====================================================================!==
void checkBsdlEntityEnd(cDataPort *pDP)
{
  if (pDP->iBsdlEntityEnd == 0)
  { jMessage(DISP_AND_EXIT,"Invalid entity termination format",1,pDP);
  } // IF

} // checkBsdlEntityEnd()
//==!=====================================================================!==
//==! Search for the next string or token in the cDataPort. This is similar
//==! to findStr() but accounts for string concatenation via the '&'
//==! character in BSDL files.
//==!
//==! Function Parameters: pDP = input data port pointer
//==!                      sTarget = string to search for
//==!                      iMode = 0: check the next data token only
//==!                              1: check until sTarget found
//==! Return Value       : none
//==!=====================================================================!==
void findStrBsdl(cDataPort *pDP,char *sTarget,int iMode)
{

  while (1)
  { xDataBsdl(pDP,XDATA_READ);
    if (xstrcmp(pDP->sData,sTarget) == 0)
    { return;
    } // IF
    if (iMode == 0)
    { sprintf(sMsgBuf00,"[%s] not found",sTarget);
      jMessage(DISP_AND_EXIT,sMsgBuf00,1,pDP);
      break;
    } // IF iMode == 0
  } // WHILE (1)

} // findStrBsdl()
//==!=====================================================================!==
//==! Get data of specified type from cDataPort. Display error message if
//==! expected data type not found.
//==!
//==! Function Parameters: pDP = input data port
//==!                      iMode = 0: use xParseData(), 1: use xDataBsdl()
//==!                      eGetType = expected data type
//==!                      sGVmsg = error message
//==! Return Value       : none
//==!=====================================================================!==
void xGetData(cDataPort *pDP,int iMode,EnumPARSE_ATTR eGetType,char *sGVmsg)
{
  if (iMode == 0)
  { xParseData(pDP,XDATA_READ);
  } // IF
  else
  { xDataBsdl(pDP,XDATA_READ);
  } // ELSE
  if (pDP->eDataAttr != eGetType)
  { jMessage(DISP_AND_EXIT,sGVmsg,1,pDP);
  } // IF

} // xGetData()
//==!=====================================================================!==
//==! CxCStr constructor. Set default values for the char string object which
//==! can be added to a list.
//==!
//==! Function Parameters: none
//==! Return Value       : none
//==!=====================================================================!==
CxCStr::CxCStr()
{

  pStr = NULL;
  pNextCxCStr = NULL;

} // CxCStr::CxCStr()
//==!=====================================================================!==
//==! CxCStr destructor.
//==!
//==! Function Parameters: none
//==! Return Value       : none
//==!=====================================================================!==
CxCStr::~CxCStr()
{

  freeArray(pStr);
  pNextCxCStr = NULL;

} // CxCStr::!CxCStr()
//==!=====================================================================!==
//==! CPortData constructor. Set default values for a device pin.
//==!
//==! Function Parameters: none
//==! Return Value       : none
//==!=====================================================================!==
CPortData::CPortData()
{

  csPortName = NULL;
  csBasePortName = NULL;
  lPortNameCnt = 0;
  iPortNameDir = -2;      //==! Initial value; must later be set to one of
                          //==! the following values:
                          //==! -1: pin vector order will be  [max..min]
                          //==!  1: to make pin order go from [min..max]
  lMinNameArrayIndex = -1;
  lMaxNameArrayIndex = -1;
  lPortLocation = -1;
  csPortLocation = NULL;
  enPortType = ePortNoType;
  lInCell = -1;
  lOutCell = -1;
  lCtrlCell = -1;
  pNextCPortData = NULL;

} // CPortData::CPortData()
//==!=====================================================================!==
//==! CPortData destructor.
//==!
//==! Function Parameters: none
//==! Return Value       : none
//==!=====================================================================!==
CPortData::~CPortData()
{

  freeArray(csPortName);
  freeArray(csBasePortName);
  lPortNameCnt = 0;
  iPortNameDir = -2;       //==! initial value
  lMinNameArrayIndex = -1;
  lMaxNameArrayIndex = -1;
  lPortLocation = -1;
  freeArray(csPortLocation);
  enPortType = ePortNoType;
  lInCell = -1;
  lOutCell = -1;
  lCtrlCell = -1;
  pNextCPortData = NULL;

} // CPortData::~CPortData()
//==!=====================================================================!==
//==! CPortData::removePin() will remove all pins in this pin list.
//==!
//==! Function Parameters: pCPD = pointer to CPortData list pointer
//==! Return Value       : none
//==!=====================================================================!==
void CPortData::removePin(CPortData **pCPD)
{ CPortData *pCPDtmp;

  while (*pCPD != NULL)
  { pCPDtmp = (*pCPD)->pNextCPortData;
    (*pCPD)->pNextCPortData = NULL;
    delete (*pCPD);
    *pCPD = pCPDtmp;
  } // WHILE *pCPD != NULL

} // CPortData::removePin()
//==!=====================================================================!==
//==! CBscData constructor. Set default values for a boundary scan cell
//==! object
//==!
//==! Function Parameters: none
//==! Return Value       : none
//==!=====================================================================!==
CBscData::CBscData()
{

  lPortNum = -1;
  lBscIndex = -1;
  lBscCellNum = -1;
  lBscCellType = 0;
  enBscFunc = eNoBscFunc;
  iSafeVal = -1;
  lCtrlCell = -1;
  iDisVal = -1;
  enDisResult = eBscNoResult;

} // CBscData::CBscData()
//==!=====================================================================!==
//==! CBscData destructor. Deallocate and reinitialize.
//==!
//==! Function Parameters: none
//==! Return Value       : none
//==!=====================================================================!==
CBscData::~CBscData()
{

  lPortNum = -1;
  lBscIndex = -1;
  lBscCellNum = -1;
  lBscCellType = 0;
  enBscFunc = eNoBscFunc;
  iSafeVal = -1;
  lCtrlCell = -1;
  iDisVal = -1;
  enDisResult = eBscNoResult;

} // CBscData::~CBscData()
//==!=====================================================================!==
//==! SList constructor.
//==!
//==! Function Parameters: none
//==! Return Value       : none
//==!=====================================================================!==
SList::SList()
{

  lVal = -1;
  pNextSList = NULL;

} // SList::SList()
//==!=====================================================================!==
//==! SList destructor.
//==!
//==! Function Parameters: none
//==! Return Value       : none
//==!=====================================================================!==
SList::~SList()
{

} // SList::~SList()
//==!=====================================================================!==
//==! CBsdlDataReg constructor.
//==!
//==! Function Parameters: none
//==! Return Value       : none
//==!=====================================================================!==
CBsdlDataReg::CBsdlDataReg()
{

  csRegName = NULL;
  lRegLength = 0;
  pOpList = NULL;
  pNextCBsdlDataReg = NULL;

} // CBsdlDataReg::CBsdlDataReg()
//==!=====================================================================!==
//==! CBsdlDataReg destructor.
//==!
//==! Function Parameters: none
//==! Return Value       : none
//==!=====================================================================!==
CBsdlDataReg::~CBsdlDataReg()
{ SList *pSList;

  freeArray(csRegName);
  lRegLength = -1;
  while(pOpList != NULL)
  { pSList = pOpList->pNextSList;
    pOpList->pNextSList = NULL;
    freeData(pOpList);
    pOpList = pSList;
  } // WHILE
  pNextCBsdlDataReg = NULL;

} // CBsdlDataReg::~CBsdlDataReg()
//==!=====================================================================!==
//==! CBsdlOp constructor.
//==!
//==! Function Parameters: none
//==! Return Value       : none
//==!=====================================================================!==
CBsdlOp::CBsdlOp()
{

  csOpcodeName = NULL;
  m_CxCStr_OpcodeList = NULL;
  lDataRegIndex = -1;
  iPrivate = -1;
  pNextCBsdlOp = NULL;

} // CBsdlOp::CBsdlOp()
//==!=====================================================================!==
//==! CBsdlOp destructor.
//==!
//==! Function Parameters: none
//==! Return Value       : none
//==!=====================================================================!==
CBsdlOp::~CBsdlOp()
{ CxCStr *pCxCStr_Var;

  freeArray(csOpcodeName);
  while (m_CxCStr_OpcodeList != NULL)
  { pCxCStr_Var = m_CxCStr_OpcodeList->pNextCxCStr;
    m_CxCStr_OpcodeList->pNextCxCStr = NULL;
    freeData(m_CxCStr_OpcodeList);
    m_CxCStr_OpcodeList = pCxCStr_Var;
  } // WHILE m_CxCStr_OpcodeList != NULL
  lDataRegIndex = -1;
  iPrivate = -1;
  pNextCBsdlOp = NULL;

} // CBsdlOp::~CBsdlOp()
//==!=====================================================================!==
//==! CxDataVar constructor.
//==!
//==! Function Parameters: none
//==! Return Value       : none
//==!=====================================================================!==
CxDataVar::CxDataVar()
{
  m_pcVarName = NULL;
  m_lVarWidth = 0;
  m_ulVarValue = NULL;
  m_ulInitValue = NULL;
  m_pNextCxDataVar = NULL;

} // CxDataVar::CxDataVar()
//==!=====================================================================!==
//==! CxDataVar destructor.
//==!
//==! Function Parameters: none
//==! Return Value       : none
//==!=====================================================================!==
CxDataVar::~CxDataVar()
{

  freeArray(m_pcVarName);
  m_lVarWidth = 0;
  freeArray(m_ulVarValue);
  freeArray(m_ulInitValue);
  m_pNextCxDataVar = NULL;

} // CxDataVar::~CxDataVar()
//==!=====================================================================!==
//==! CxDataExpr constructor.
//==!
//==! Function Parameters: none
//==! Return Value       : none
//==!=====================================================================!==
CxDataExpr::CxDataExpr()
{
  m_lDE_len = 0;
  m_iFieldType = dNOT_SET;
  m_lVarIndex = -1;
  m_pulDataField = NULL;
  m_cOperator = NULL;
  m_ulOperand = 0;

} // CxDataExpr::CxDataExpr()
//==!=====================================================================!==
//==! CxDataExpr destructor.
//==!
//==! Function Parameters: none
//==! Return Value       : none
//==!=====================================================================!==
CxDataExpr::~CxDataExpr()
{
  m_lDE_len = 0;
  m_iFieldType = dNOT_SET;
  m_lVarIndex = -1;
  freeArray(m_pulDataField);
  m_cOperator = NULL;
  m_ulOperand = 0;

} // CxDataExpr::~CxDataExpr()
//==!=====================================================================!==
//==! CxDataExpr member function to reset the data area.
//==!
//==! Function Parameters: none
//==! Return Value       : none
//==!=====================================================================!==
void CxDataExpr::resetDataArea()
{ long lVar0,lVar1;

  lVar1 = (m_lDE_len+31)/32;
  for (lVar0 = 0; lVar0 < lVar1; ++lVar0)
  { m_pulDataField[lVar0] = 0;
  } // FOR lVar0 < lVar1

} // CxDataExpr::resetDataArea()
//==!=====================================================================!==
//==! CxDataExpr member function to set bit position in the data area to
//==! iBitVal.
//==!
//==! Function Parameters: iBitVal = bit val to set bit to
//==!                      lBitPos = bit position to set
//==! Return Value       : none
//==!=====================================================================!==
void CxDataExpr::setDataBit(int iBitVal,long lBitPos)
{ long lArrayIndex,lBitIndex;

  lArrayIndex = lBitPos/32;
  lBitIndex = lBitPos%32;

  if (iBitVal == 1)        //==! Set bit to 1
  { m_pulDataField[lArrayIndex] |= culGlobBitPosMask[lBitIndex];
  } // IF iBitVal == 1
  else                     //==! Set bit to 0
  { m_pulDataField[lArrayIndex] &= ~(culGlobBitPosMask[lBitIndex]);
  } // ELSE

} // CxDataExpr::setDataBit()
//==!=====================================================================!==
//==! CxDataExpr member function to get value of specified bit position
//==! in the data area.
//==!
//==! Function Parameters: lBitPos = bit position to read
//==! Return Value       : value in array[lBitPos]
//==!=====================================================================!==
int CxDataExpr::getDataBit(long lBitPos)
{ long lArrayIndex,lBitIndex;

  lArrayIndex = lBitPos/32;
  lBitIndex = lBitPos%32;

  if ((m_pulDataField[lArrayIndex] & culGlobBitPosMask[lBitIndex]) != 0)
  { return 1;
  } // IF
  return 0;

} // CxDataExpr::getDataBit()
//==!=====================================================================!==
//==! Show/print field data. Print only the upper part of the word required
//==! by the number of bits. For example, if field bit length = 46, then
//==! this is represented by two 32-bit words 1 and 0. When printing word 1,
//==! print only the lower 14 bits:
//==! => (46%32 + 3) = (14+3)/4, then 14+3/4 = 4
//==! => 14 bits are represented by 4 hex digits
//==!
//==! Note that after printing the first 14 bits, then 32 bits of the next
//==! word (word 0) is printed using the full 8 hex digits, which is why
//==! lBitIndex = 8 in the FOR loop iteration section.
//==!
//==! The 46 bit word is represented by 12 (= 4+8) hex digits.
//==!
//==! Function Parameters: pCDP_FieldDest = port to send field data to
//==! Return Value       : none
//==!=====================================================================!==
void CxDataExpr::showFieldData(cDataPort *pCDP_FieldDest)
{ long lArrayIndex,lBitIndex,lCnt,lDispLoc;

  if (pCDP_FieldDest == NULL)
  { return;
  } // IF

  lArrayIndex = (m_lDE_len-1)/32;
  lBitIndex = (m_lDE_len%32 + 3)/4;
  if (lBitIndex == 0)
  { lBitIndex = 8;
  } // IF

  if (iGlobOptBufferedWriteOp == 0)
  { if (pCDP_FieldDest->m_iDataFormat == 1)
    { if (pCDP_FieldDest->m_iFirstData == 0)
      { fprintf(pCDP_FieldDest->pFILE,",\n");
      } // IF
      pCDP_FieldDest->m_iFirstData = 0;
    } // IF

    for (lCnt = lArrayIndex,lDispLoc = 0; lCnt >= 0; --lCnt,lBitIndex = 8)
    { lDispLoc += fprintf(pCDP_FieldDest->pFILE,
                          "%0*X ",lBitIndex,m_pulDataField[lCnt]);
      if (lDispLoc >= 64)
      { fprintf(pCDP_FieldDest->pFILE,"\n");
        lDispLoc = 0;
      } // IF
    } // FOR lCnt >= 0

    if (lDispLoc != 0)
    { fprintf(pCDP_FieldDest->pFILE,"\n");
    } // IF
  } // IF iGlobOptBufferedWriteOp == 0
  else
  { if (pCDP_FieldDest->m_iDataFormat == 1)
    { if (pCDP_FieldDest->m_iFirstData == 0)
      { pCDP_FieldDest->writeToCDP(0,",\n");
      } // IF
      pCDP_FieldDest->m_iFirstData = 0;
    } // IF

    for (lCnt = lArrayIndex,lDispLoc = 0; lCnt >= 0; --lCnt,lBitIndex = 8)
    { lDispLoc += sprintf(csTmpWriteBuf,"%0*X ",
                          lBitIndex,m_pulDataField[lCnt]);
      if (lDispLoc >= 64)
      { strcat(csTmpWriteBuf,"\n");
        lDispLoc = 0;
      } // IF
      pCDP_FieldDest->writeToCDP(0,csTmpWriteBuf);
    } // FOR lCnt >= 0

    if (lDispLoc != 0)
    { pCDP_FieldDest->writeToCDP(0,"\n");
    } // IF
  } // ELSE

} // CxDataExpr::showFieldData()
//==!=====================================================================!==
//==! CxDataExpr member function to update variable and data expression areas
//==! based on iMode:
//==!
//==! iMode = 0 then:
//==!   - update data expression data area and data area of variable in data
//==!     var array pCIFobj->m_CxDataVarArray[] indexed by m_lVarIndex with
//==!     initialization value m_ulInitValue (if defined, else init to 0)
//==! iMode = 1 then:
//==!   - update data expression data area with current value of variable
//==!     indexed by m_lVarIndex
//==! iMode = 2 then:
//==!   - update variable data area with data expression data area (i.e.,
//==!     store updated/evaluated value in data expression data area to the
//==!     variable data area)
//==!
//==! Function Parameters: pCIFobj = pointer to CIscFlow object
//==!                      iMode = see function description
//==! Return Value       : none
//==!=====================================================================!==
void CxDataExpr::updateVar(CIscFlow *pCIFobj,int iMode)
{ long lVar0,lVar1;

  isInRange(1,m_lVarIndex,0,pCIFobj->m_lCxDataVarArraySize-1,"CDE::gCVV",
            DISP_AND_EXIT);
  if (m_lDE_len != pCIFobj->m_CxDataVarArray[m_lVarIndex]->m_lVarWidth)
  { jMessage(DISP_AND_EXIT,"CdeLen!=ArrayLen",0,NULL);
  } // IF

  lVar1 = (m_lDE_len+31)/32;
  for (lVar0 = 0; lVar0 < lVar1; ++lVar0)
  { switch (iMode)
    { case 0:
      { if (pCIFobj->m_CxDataVarArray[m_lVarIndex]->m_ulInitValue == NULL)
        { m_pulDataField[lVar0] = 0;
          pCIFobj->m_CxDataVarArray[m_lVarIndex]->m_ulVarValue[lVar0] = 0;
        } // IF
        else
        { m_pulDataField[lVar0] =
               pCIFobj->m_CxDataVarArray[m_lVarIndex]->m_ulInitValue[lVar0];
          pCIFobj->m_CxDataVarArray[m_lVarIndex]->m_ulVarValue[lVar0] =
               pCIFobj->m_CxDataVarArray[m_lVarIndex]->m_ulInitValue[lVar0];
        } // ELSE
        break;
      } // CASE 0
      case 1:
      { m_pulDataField[lVar0] =
               pCIFobj->m_CxDataVarArray[m_lVarIndex]->m_ulVarValue[lVar0];
        break;
      } // CASE 1
      case 2:
      { pCIFobj->m_CxDataVarArray[m_lVarIndex]->m_ulVarValue[lVar0] =
               m_pulDataField[lVar0];
      } // CASE 2
    } // SWITCH (iMode)
  } // FOR lVar0 < lVar1

} // CxDataExpr::updateVar()
//==!=====================================================================!==
//==! CxDataExpr member function to do the operation specified by the data
//==! expression operator: complement/invert, add, or subtract.
//==!
//==! Function Parameters: pCIFobj = pointer to CIscFlow object
//==!                      iMode = 1: update variable data area with
//==!                         evaluated value in data expression work area
//==! Return Value       : none
//==!=====================================================================!==
void CxDataExpr::evalOperator(CIscFlow *pCIFobj,int iMode)
{ long lVar0,lVar1;
  int iCopyFlag,iVal;
  double dbl_Val;

  isInRange(1,m_lVarIndex,0,pCIFobj->m_lCxDataVarArraySize-1,"CDE::eO",
            DISP_AND_EXIT);
  if (m_lDE_len != pCIFobj->m_CxDataVarArray[m_lVarIndex]->m_lVarWidth)
  { jMessage(DISP_AND_EXIT,"CdeLen!=ArrayLen(1)",0,NULL);
  } // IF

  lVar1 = (m_lDE_len+31)/32;
  switch (m_cOperator)
  { case dc_COMPLEMENT_OP:
    { for (lVar0 = 0; lVar0 < lVar1; ++lVar0)
      { m_pulDataField[lVar0] =
             ~(pCIFobj->m_CxDataVarArray[m_lVarIndex]->m_ulVarValue[lVar0]);
      } // FOR lVar0 < lVar1
      break;
    } // CASE m_cOperator = dc_COMPLEMENT_OP

//==! ADDition operation is performed as follows:
//==! - add first 32 bits of current variable value from variable array
//==!   to m_ulOperand after converting to double type
//==!   = check if result < 2^32
//==!     = if yes, then no overflow occurred (i.e., addition does not
//==!       propagate to the next set of 32 bits); set iCopyFlag to copy the
//==!       the remaining words
//==!     = if no, then 32-bit overflow occurred, so we have to add 1 to the
//==!       next 32 bits of the current variable value
    case dc_ADD_OP:
    { for (lVar0 = 0,iCopyFlag = 0; lVar0 < lVar1; ++lVar0)
      { if (iCopyFlag == 1)
        { m_pulDataField[lVar0] =
                    pCIFobj->m_CxDataVarArray[m_lVarIndex]
                           ->m_ulVarValue[lVar0];
          continue;
        } // IF
        if (lVar0 == 0)
        { dbl_Val = (double)(pCIFobj->m_CxDataVarArray[m_lVarIndex]
                                    ->m_ulVarValue[lVar0]) +
                    (double)m_ulOperand;
//==! Could have cast dbl_Val as unsigned long, but loss of precision could
//==! occur depending on what compiler/system this will run on.
          m_pulDataField[lVar0] =
                    pCIFobj->m_CxDataVarArray[m_lVarIndex]
                           ->m_ulVarValue[lVar0] + m_ulOperand;
        } // IF
        else
        { dbl_Val = (double)(pCIFobj->m_CxDataVarArray[m_lVarIndex]
                                    ->m_ulVarValue[lVar0]) + 1;
          m_pulDataField[lVar0] =
                    pCIFobj->m_CxDataVarArray[m_lVarIndex]
                           ->m_ulVarValue[lVar0] + 1;
        } // ELSE

//==! 2003/11/04: Used to be (dbl_Val < 0x100000000) but some embedded system
//==! compilers do not support constants greater than 4 bytes.
        if (dbl_Val <= 0xFFFFFFFF)
        { iCopyFlag = 1;
        } // IF
      } // FOR lVar0 < lVar1
      break;
    } // CASE m_cOperator = dc_ADD_OP

//==! SUBtraction operation is performed as follows:
//==! - cast as double type and then subtract m_ulOperand from first 32 bits
//==!   of current variable value from variable array
//==!   = check if result >= 0
//==!     = if yes, then no underflow occurred (i.e., subtraction does not
//==!       propagate to the next set of 32 bits); set iCopyFlag to copy the
//==!       remaining words
//==!     = if no, then 32-bit underflow occurred, so we have to add 1 to the
//==!       next 32 bits of the current variable value
    case dc_SUB_OP:
    { for (lVar0 = 0,iCopyFlag = 0; lVar0 < lVar1; ++lVar0)
      { if (iCopyFlag == 1)
        { m_pulDataField[lVar0] =
                    pCIFobj->m_CxDataVarArray[m_lVarIndex]
                           ->m_ulVarValue[lVar0];
          continue;
        } // IF
        if (lVar0 == 0)
        { dbl_Val = (double)(pCIFobj->m_CxDataVarArray[m_lVarIndex]
                                    ->m_ulVarValue[lVar0]) -
                    (double)m_ulOperand;
//==! Could have cast dbl_Val as unsigned long, but loss of precision could
//==! occur depending on what compiler/system this will run on.
          m_pulDataField[lVar0] =
                    pCIFobj->m_CxDataVarArray[m_lVarIndex]
                           ->m_ulVarValue[lVar0] - m_ulOperand;
        } // IF
        else
        { dbl_Val = (double)(pCIFobj->m_CxDataVarArray[m_lVarIndex]
                                    ->m_ulVarValue[lVar0]) - 1;
          m_pulDataField[lVar0] =
                    pCIFobj->m_CxDataVarArray[m_lVarIndex]
                           ->m_ulVarValue[lVar0] - 1;
        } // ELSE
        if (dbl_Val >= 0)
        { iCopyFlag = 1;
        } // IF
      } // FOR lVar0 < lVar1
      break;
    } // CASE m_cOperator = dc_SUB_OP

    case dc_LEFT_SHIFT_OP:
    { updateVar(pCIFobj,1);
      for (lVar0 = m_lDE_len-m_ulOperand-1; lVar0 >= 0; --lVar0)
      { iVal = getDataBit(lVar0);
        setDataBit(iVal,lVar0+m_ulOperand);
      } // FOR lVar0 >= 0
      for (lVar0 = 0; lVar0 < (long)m_ulOperand; ++lVar0)
      { setDataBit(0,lVar0);
      } // FOR lVar0 < m_ulOperand
      break;
    } // CASE m_cOperator = dc_LEFT_SHIFT_OP

    case dc_RIGHT_SHIFT_OP:
    { updateVar(pCIFobj,1);
      for (lVar0 = m_ulOperand; lVar0 < m_lDE_len; ++lVar0)
      { iVal = getDataBit(lVar0);
        setDataBit(iVal,lVar0-m_ulOperand);
      } // FOR lVar0 < m_lDE_len
      for (lVar0 = m_lDE_len-m_ulOperand; lVar0 < m_lDE_len; ++lVar0)
      { setDataBit(0,lVar0);
      } // FOR lVar0 < l_lDE_len
      break;
    } // CASE m_cOperator = dc_LEFT_SHIFT_OP

    default:  // Unknown operator
    { sprintf(sMsgOnly,"CDE::eO[%c]?",m_cOperator);
      jMessage(DISP_AND_EXIT,sMsgOnly,0,NULL);
    } // DEFAULT
  } // SWITCH (m_cOperator)

  if (iMode == 1)
  { updateVar(pCIFobj,2);
  } // IF

} // CxDataExpr::evalOperator()
//==!=====================================================================!==
//==! CFlowWait constructor.
//==!
//==! Function Parameters: none
//==! Return Value       : none
//==!=====================================================================!==
CFlowWait::CFlowWait()
{
  m_lMinFlowWaitCycles = 0;
  m_lMaxFlowWaitCycles = 0;
  m_fMinFlowWaitTimeSec = 0;
  m_fMaxFlowWaitTimeSec = 0;

} // CFlowWait::CFlowWait()
//==!=====================================================================!==
//==! CFlowWait destructor.
//==!
//==! Function Parameters: none
//==! Return Value       : none
//==!=====================================================================!==
CFlowWait::~CFlowWait()
{
  m_lMinFlowWaitCycles = 0;
  m_lMaxFlowWaitCycles = 0;
  m_fMinFlowWaitTimeSec = 0;
  m_fMaxFlowWaitTimeSec = 0;

} // CFlowWait::~CFlowWait()
//==!=====================================================================!==
//==! CFlowWait member function to reset variables.
//==!
//==! Function Parameters: none
//==! Return Value       : none
//==!=====================================================================!==
void CFlowWait::resetWaitTimes()
{
  m_lMinFlowWaitCycles = 0;
  m_lMaxFlowWaitCycles = 0;
  m_fMinFlowWaitTimeSec = 0;
  m_fMaxFlowWaitTimeSec = 0;

} // CFlowWait::CFlowWait()
//==!=====================================================================!==
//==! CxUpdateField constructor.
//==!
//==! Function Parameters: none
//==! Return Value       : none
//==!=====================================================================!==
CxUpdateField::CxUpdateField()
{

  m_lFieldWidth = 0;
  m_pCDE_data = NULL;
  m_pNextCxUpdateField = NULL;

} // CxUpdateField::CxUpdateField()
//==!=====================================================================!==
//==! CxUpdateField destructor.
//==!
//==! Function Parameters: none
//==! Return Value       : none
//==!=====================================================================!==
CxUpdateField::~CxUpdateField()
{

  m_lFieldWidth = 0;
  freeData(m_pCDE_data);
  m_pNextCxUpdateField = NULL;

} // CxUpdateField::~CxUpdateField()
//==!=====================================================================!==
//==! CxCaptureField constructor.
//==!
//==! Function Parameters: none
//==! Return Value       : none
//==!=====================================================================!==
CxCaptureField::CxCaptureField()
{

  m_lFieldWidth = 0;
  m_cExpDataOutOp = NULL;
  m_cMaskDataOutOp = NULL;
  m_cCrcTag = NULL;
  m_cOstTag = NULL;
  m_pCDE_expdata = NULL;
  m_pCDE_maskdata = NULL;
  m_pCDE_CapturedData = NULL;
  m_pNextCxCaptureField = NULL;

} // CxCaptureField::CxCaptureField()
//==!=====================================================================!==
//==! CxCaptureField destructor.
//==!
//==! Function Parameters: none
//==! Return Value       : none
//==!=====================================================================!==
CxCaptureField::~CxCaptureField()
{

  m_lFieldWidth = 0;
  m_cExpDataOutOp = NULL;
  m_cMaskDataOutOp = NULL;
  m_cCrcTag = NULL;
  m_cOstTag = NULL;
  freeData(m_pCDE_expdata);
  freeData(m_pCDE_maskdata);
  freeData(m_pCDE_CapturedData);
  m_pNextCxCaptureField = NULL;

} // CxCaptureField::~CxCaptureField()
//==!=====================================================================!==
//==! CFlowAct constructor.
//==!
//==! Function Parameters: none
//==! Return Value       : none
//==!=====================================================================!==
CFlowAct::CFlowAct()
{
  m_lMinLoopCount = 0;
  m_lMaxLoopCount = 0;
  m_pLoopActions = NULL;
  m_lOpcodeIndex = -1;
  m_pFlowWaitInfo = NULL;
  m_lUpdateWidth = 0;
  m_pUF_Update = NULL;
  m_lCaptureWidth = 0;
  m_pCF_Capture = NULL;
  m_pNextCFlowAct = NULL;

} // CFlowAct::CFlowAct()
//==!=====================================================================!==
//==! CFlowAct destructor.
//==!
//==! Function Parameters: none
//==! Return Value       : none
//==!=====================================================================!==
CFlowAct::~CFlowAct()
{ CFlowAct *pVarCFA;
  CxUpdateField *pVarUF;
  CxCaptureField *pVarCF;

  m_lMinLoopCount = 0;
  m_lMaxLoopCount = 0;

  while (m_pLoopActions != NULL)
  { pVarCFA = m_pLoopActions->m_pNextCFlowAct;
    freeData(m_pLoopActions);
    m_pLoopActions = pVarCFA;
  } // WHILE m_pLoopActions != NULL

  m_lOpcodeIndex = -1;
  freeData(m_pFlowWaitInfo);
  m_lUpdateWidth = 0;

  while (m_pUF_Update != NULL)
  { pVarUF = m_pUF_Update->m_pNextCxUpdateField;
    freeData(m_pUF_Update);
    m_pUF_Update = pVarUF;
  } // WHILE m_pUF_Update != NULL

  m_lCaptureWidth = 0;
  while (m_pCF_Capture != NULL)
  { pVarCF = m_pCF_Capture->m_pNextCxCaptureField;
    freeData(m_pCF_Capture);
    m_pCF_Capture = pVarCF;
  } // WHILE m_pCF_Capture != NULL

  m_pNextCFlowAct = NULL;

} // CFlowAct::~CFlowAct()
//==!=====================================================================!==
//==! Get lBitLen bits from pCDDobj (device being processed) and put into
//==! pcBuf[]. The appropriate data source from pCDDobj (either cDataPort or
//==! from memory) will be positioned for the read operation and will have
//==! the data in hex format.
//==!
//==! Function Parameters: lBitLen = number of bits to get from pCDD
//==!                      pcBuf = pointer to buffer
//==!                      pCDDobj = pointer to CDeviceData object (to get
//==!                         input data from if required)
//==! Return Value       : 0: no data found
//==!                      1: data found
//==!                      2: last data found
//==!=====================================================================!==
int getIscData(long lBitLen,char *pcBuf,CDeviceData *pCDDobj)
{ long lHexMax,lVar0,lVar1,lVar2,ulValue,lFlag,lFlagSkip,lDataSectionIndex;
  int iLastData;
  CMemIscDataRowSection *pCMIDRS_var;

  iLastData = 1;
  lHexMax = (lBitLen+3)/4;

  if (iGlobOptUseMemIscData == 0)
  { getDpChar(pCDDobj->m_pCIscData->m_pCDP_IscData,1);
    if (pCDDobj->m_pCIscData->m_pCDP_IscData->iCurChar == ')')
    { return 0;
    } // IF
    if (pCDDobj->m_pCIscData->m_pCDP_IscData->iCurChar != '(' &&
        pCDDobj->m_pCIscData->m_pCDP_IscData->iCurChar != ',')
    { sprintf(sMsgOnly,"Incorrect ISC data port position: [%c] line [%ld]",
              pCDDobj->m_pCIscData->m_pCDP_IscData->iCurChar,
              pCDDobj->m_pCIscData->m_pCDP_IscData->lLineNum);
      jMessage(DISP_AND_EXIT,sMsgOnly,0,NULL);
    } // IF
  } // IF iGlobOptUseMemIscData == 0
  else
  { if (pCDDobj->m_CMemIscDataContext.m_pCMIDR_current == NULL)
    { jMessage(DISP_AND_EXIT,"No ISC data in mem",0,NULL);
      return 0;
    } // IF
  } // ELSE

  if (iGlobOptUseMemIscData == 0)
  { for (lVar0 = 0,lFlag = 0; lFlag == 0;)
    { getDpChar(pCDDobj->m_pCIscData->m_pCDP_IscData,1);
      switch(pCDDobj->m_pCIscData->m_pCDP_IscData->iCurChar)
      { case ',':
        case ')':
          cTdoCopy[lVar0] = '\0';
          lFlag = 1;
          if (pCDDobj->m_pCIscData->m_pCDP_IscData->iCurChar == ')')
          { iLastData = 2;
          } // IF
          pCDDobj->m_pCIscData->m_pCDP_IscData->iParserGetChar = 0;
          break;
        default:
          if (!isxdigit(pCDDobj->m_pCIscData->m_pCDP_IscData->iCurChar))
          { sprintf(sMsgOnly,"Invalid hex data char [%c] on line [%ld]",
                    pCDDobj->m_pCIscData->m_pCDP_IscData->iCurChar,
                    pCDDobj->m_pCIscData->m_pCDP_IscData->lLineNum);
            jMessage(DISP_AND_EXIT,sMsgOnly,0,NULL);
          } // IF
          if (lVar0 >= lHexMax)
          { sprintf(sMsgOnly,
                    "Line [%ld]: data hex string > [%ld] bits required",
                    pCDDobj->m_pCIscData->m_pCDP_IscData->lLineNum,lBitLen);
            jMessage(DISP_AND_EXIT,sMsgOnly,0,NULL);
          } // IF
          cTdoCopy[lVar0] = pCDDobj->m_pCIscData->m_pCDP_IscData->iCurChar;
          ++lVar0;
          break;
      } // SWITCH iCurChar
    } // FOR lFlag == 0

//==! Convert HEX chars in cTdoCopy[] and expand into binary in pcBuf[].
//==! The following nested FOR loops will check for proper array indexing in
//==! the case where the bit length is not a multiple of 2.
//==! Ex. The hex data string 3A will be put into a 7-bit array as 011 1010.
//==! Ex. Hex data 0004A is put into 7-bit array as 100 1010 (upper hex
//==!     chars can be 0s and will be ignored).
//==! Ex. Hex data A is put into 7-bit array as 000 1010
    for (lFlag = lVar0-1,lVar1 = lBitLen-1; lFlag >= 0; --lFlag)
    { ulValue = chrtohex(cTdoCopy[lFlag]);
      if (lVar1 < 0)
      { if (ulValue != 0)
        { sprintf(sMsgOnly,
                  "MSB hex [%c] cannot be stored in [%ld] bit buffer",
                  cTdoCopy[lFlag],lBitLen);
          jMessage(DISP_AND_EXIT,sMsgOnly,0,NULL);
        } // IF
        continue;
      } // IF
      for (lVar2 = 0; lVar2 < 4; ++lVar2)
      { if ((ulValue & (0x01 << lVar2)) != 0)
        { if (lVar1 < 0) // check if remaining bits in current chrtohex are 0
          { sprintf(sMsgOnly,
                    "Cannot store hex char %ld of [%s] in [%ld]-wide field",
                    lFlag,cTdoCopy,lBitLen);
            jMessage(DISP_AND_EXIT,sMsgOnly,0,NULL);
          } // IF
          pcBuf[lVar1] = '1';
        } // IF
        else
        { if (lVar1 >= 0)
          { pcBuf[lVar1] = '0';
          } // IF
          else
          { continue;
          } // ELSE
        } // ELSE
        --lVar1;
      } // FOR lVar2 < 4
    } // FOR lFlag >= 0

  } // IF iGlobOptUseMemIscData == 0
  else
  { for (lVar1 = lBitLen-1,lFlag = 0,lFlagSkip = 0;
         lVar1 >= 0 && lFlag == 0 && lFlagSkip == 0;)
    { //==! Data row pointer should not be NULL
      if (pCDDobj->m_CMemIscDataContext.m_pCMIDR_current == NULL)
      { jMessage(DISP_SHOW_MSG,"Incorrect CMIDR pointer",1,NULL);
        break;
      } // IF
      for (pCMIDRS_var = pCDDobj->m_CMemIscDataContext.m_pCMIDR_current
                                ->m_pCMIDRS_Head;
           pCMIDRS_var != NULL && lFlagSkip == 0;
           pCMIDRS_var = pCMIDRS_var->m_pNextCMIDRS_MSB)
      { for (lDataSectionIndex = pCMIDRS_var->m_lDataCnt-1;
             lDataSectionIndex >= 0 && lFlagSkip == 0;
             --lDataSectionIndex)
        { ulValue = chrtohex(pCMIDRS_var->m_csHexData[lDataSectionIndex]);
          if (lVar1 >= 0)
          { for (lVar2 = 0; lVar2 < 4; ++lVar2)
            { if ((ulValue & culGlobBitPosMask[lVar2]) != 0)
              { if (lVar1 >= 0)
                { pcBuf[lVar1] = '1';
                } // IF
                else
                { sprintf(sMsgOnly,
"Non-0 bits in [%ld] hex char row ignored for device\n\t\
requiring only [%ld] bits",
                          pCDDobj->m_CMemIscDataContext.m_pCMIDR_current
                                 ->m_lTotDataCnt,lBitLen);
                  jMessage(DISP_SHOW_MSG,sMsgOnly,1,NULL);
                  lFlagSkip = 1;
                  continue;
                } // ELSE
              } // IF
              else
              { if (lVar1 >= 0)
                { pcBuf[lVar1] = '0';
                } // IF
                else
                { continue;
                } // ELSE
              } // ELSE
              --lVar1;
            } // FOR lVar2 < 4
          } // IF lVar1 >= 0
          else
          { if (ulValue > 0)
            { sprintf(sMsgOnly,
"Non-0 chars in [%ld] hex char row ignored for device\n\t\
requiring only [%ld] bits",
                      pCDDobj->m_CMemIscDataContext.m_pCMIDR_current
                             ->m_lTotDataCnt,lBitLen);
              jMessage(DISP_SHOW_MSG,sMsgOnly,1,NULL);
              lFlagSkip = 1;
            } // IF
          } // ELSE
        } // FOR lDataSectionIndex >= 0

      } // FOR pCMIDRS_var != NULL
//==! Check if there are any more data rows in this array section (i.e.,
//==! initialize, repeat, or terminate section)
      pCDDobj->m_CMemIscDataContext.m_pCMIDR_current =
               pCDDobj->m_CMemIscDataContext.m_pCMIDR_current
                      ->m_pNextCMemIscDataRow;
      if (pCDDobj->m_CMemIscDataContext.m_pCMIDR_current == NULL)
      { iLastData = 2;
      } // IF
      lFlag = 1;
    } // FOR lVar1 >= 0 && iFlag == 0
  } // ELSE

//==! Fill remaining locations with 0s.
  for (;lVar1 >= 0; --lVar1)
  { pcBuf[lVar1] = '0';
  } // FOR lVar1 >= 0

  pcBuf[lBitLen] = '\0';
  return iLastData;

} // getIscData()
//==!=====================================================================!==
//==! Get ISC data for the flow phase (update or capture).
//==!
//==! Function Parameters: pCDEobj = data expression object
//==!                      pCDD = CDeviceData pointer (to get input data from
//==!                         if required)
//==!                      pCIFobj = ptr to CIscFlow obj to access variables
//==!                         used by flows
//==!                      pcBuf = pointer to buffer to put data in
//==!                      lBufi = pcBuf index to start putting data in
//==!                        - buffer could be 400 chars long, but the data
//==!                          expression requires 20 bits starting at
//==!                          lBufi = 8
//==!                      lTotDataWidth = max pcBuf size
//==!                      iMode = 0: leave data in pcBuf
//==!                                 - pcBuf usually points to TDI
//==!                              1: copy pcBuf data into data expression
//==!                                 work area
//==!                                 = pcBuf is usually TDO and TDO data is
//==!                                   moved into the data expression work
//==!                                   area for comparison
//==!                      iStat = variable pointer to return status of call
//==!                        to getIscData()
//==! Return Value       : none
//==!=====================================================================!==
void getFlowPhaseData(CxDataExpr *pCDEobj,CDeviceData *pCDD,
                      CIscFlow *pCIFobj,char *pcBuf,long lBufi,
                      long lTotDataWidth,int iMode,int *iStat)
{ long lX,lY,lBufIndex;

  switch(pCDEobj->m_iFieldType)
  { case dHEX_CONSTANT:
    { break;
    } // CASE dHEX_CONSTANT
    case dINP_SPECIFIER:
    { lBufIndex = lTotDataWidth-lBufi-1;
      *iStat = getIscData(pCDEobj->m_lDE_len,&pcBuf[lBufIndex],pCDD);
      if (*iStat == 0)
      { sprintf(sMsgOnly,
                "Not enough data in array (%s) for flow %s(%s)",
                pCIFobj->m_csDataName,
                pCIFobj->m_csFlowName,pCIFobj->m_csDataName);
        jMessage(DISP_AND_EXIT,sMsgOnly,1,
                 pCDD->m_pCIscData->m_pCDP_IscData);
      } // IF
      if (iMode == 1)
      { for (lX = pCDEobj->m_lDE_len-1,lY = 0; lX >= 0 ; --lX,++lY)
        { pCDEobj->setDataBit(pcBuf[lBufIndex+lX]-'0',lY);
        } // FOR lX >= 0
      } // IF iMode == 1
      break;
    } // CASE dINP_SPECIFIER
    case d_VAR_EXPR:
    { switch(pCDEobj->m_cOperator)
      { case dc_INPUT_OP:
        case dc_IO_OP:
        { lBufIndex = lTotDataWidth-lBufi-1;
          *iStat = getIscData(pCDEobj->m_lDE_len,&pcBuf[lBufIndex],pCDD);
          if (*iStat == 0)
          { sprintf(sMsgOnly,
                    "Not enough data in array (%s) for flow %s(%s)",
                    pCIFobj->m_csDataName,
                    pCIFobj->m_csFlowName,pCIFobj->m_csDataName);
            jMessage(DISP_AND_EXIT,sMsgOnly,1,
                     pCDD->m_pCIscData->m_pCDP_IscData);
          } // IF

          lX = strlen(pcBuf);

          if (iMode == 1)
          { for (lX = pCDEobj->m_lDE_len-1,lY = 0; lX >= 0 ; --lX,++lY)
            { pCDEobj->setDataBit(pcBuf[lBufIndex+lX]-'0',lY);
            } // FOR lX >= 0
            pCDEobj->updateVar(pCIFobj,2);
          } // IF iMode == 1
          break;
        } // CASE m_cOperator = dc_INPUT_OP || dc_IO_OP
//==! Assign or initialize data expression work area and variable data area
        case dc_ASSIGN_OP:
        { pCDEobj->updateVar(pCIFobj,0);
          break;
        } // CASE m_cOperator = dc_ASSIGN_OP

//==! Load variable data area into data expression work area
        case dc_GET_VAR:
        { pCDEobj->updateVar(pCIFobj,1);
          break;
        } // CASE m_cOperator = dc_GET_VAR

        case dc_COMPLEMENT_OP:
        case dc_ADD_OP:
        case dc_SUB_OP:
        case dc_LEFT_SHIFT_OP:
        case dc_RIGHT_SHIFT_OP:
        { pCDEobj->evalOperator(pCIFobj,1);
          break;
        } // CASE m_cOperator = + - ~ << >>

        default:
        { sprintf(sMsgOnly,"gFPDop[%c]?",pCDEobj->m_cOperator);
          jMessage(DISP_AND_EXIT,sMsgOnly,0,NULL);
          break;
        } // DEFAULT
      } // SWITCH (pCDEobj->m_cOperator)
      break;
    } // CASE d_VAR_EXPR
    default:
    { jMessage(DISP_AND_EXIT,"gFPD(FType)",0,NULL);
      break;
    } // DEFAULT
  } // SWITCH pCDEobj->m_iFieldType

} // getFlowPhaseData()
//==!=====================================================================!==
//==! Get the flow type (see NOTE_0010) based on whether the action being
//==! processed has update or capture fields.
//==!
//==! Function Parameters: none
//==! Return Value       : See NOTE_0010 for return values
//==!=====================================================================!==
int CFlowAct::getFlowType()
{
  if (m_pUF_Update != NULL && m_pCF_Capture != NULL)
  { return 0;
  } // IF
  if (m_pUF_Update == NULL && m_pCF_Capture == NULL)
  { return 1;
  } // IF
  if (m_pUF_Update != NULL && m_pCF_Capture == NULL)
  { return 2;
  } // IF
  if (m_pUF_Update == NULL && m_pCF_Capture != NULL)
  { return 3;
  } // IF

  return -1;  // should never happen

} // CFlowAct:: getFlowType()
//==!=====================================================================!==
//==! CFlowAct member function to process the update field. iMode specifies
//==! device flow processing:
//==! iMode = 0: function is called during sequential device flow processing
//==!            = assemble data then shift out
//==!         1: concurrent device flow processing
//==!            = assemble data only; will get shifted out after update data
//==!              for all devices in the chain have been assembled
//==!
//==! Function Parameters: iMode = see function description
//==!                      pcTdiBuf = pointer to TDI buffer. Note that the
//==!                         TDI buffer has LSB at the max char buffer
//==!                         index, so a 10 element char buffer will have
//==!                         LSB data (data element 0) at location 9, so
//==!                         data element 3 is translated into 10-3-1 = 6:
//==!
//==!                         char index 0 1 2 3 4 5 6 7 8 9
//==!                                   MSB               LSB
//==!                      virtual index 9 8 7 6 5 4 3 2 1 0
//==!                                                ^
//==!                                         -------+
//==!                      pcTdoBuf = pointer to TDO buffer
//==!                      pCDD = CDeviceData pointer (to get input data from
//==!                         if required)
//==!                      pCIFobj = ptr to CIscFlow obj to access variables
//==!                         used by flows
//==!                      iStat = variable pointer to return state of
//==!                         getFlowPhaseData(); note that it will return
//==!                         the current state if not set in
//==!                         getFlowPhaseData()
//==! Return Value       : update field width for the action if update data
//==!                      required else 0
//==!=====================================================================!==
long CFlowAct::processUpdateField(int iMode,char *pcTdiBuf,char *pcTdoBuf,
                      CDeviceData *pCDD,CIscFlow *pCIFobj,int *iStat)
{ CxUpdateField *pUFvar;
  long lX,lVar,lFieldSize,lArrayIndex;

  if (m_pUF_Update == NULL)
  { return 0;
  } // IF m_pUF_Update == NULL

  iGlobProgramState = dPROGSTATE_PROCESSUPDATEFIELD;

  for (pUFvar = m_pUF_Update,lVar = m_lUpdateWidth-1;
       pUFvar != NULL;
       pUFvar = pUFvar->m_pNextCxUpdateField)
  { lArrayIndex = (pUFvar->m_pCDE_data->m_lDE_len + 31)/32;

//==! Get update data (if necessary)
    getFlowPhaseData(pUFvar->m_pCDE_data,pCDD,pCIFobj,
                     pcTdiBuf,lVar,m_lUpdateWidth,1,&(*iStat));

//==! Display update data
    if (pUFvar->m_pCDE_data->m_cOperator == dc_IO_OP ||
        pUFvar->m_pCDE_data->m_cOperator == dc_OUTPUT_OP)
    { pUFvar->m_pCDE_data->showFieldData(pCDD->m_pCDP_Out);
    } // IF

//==! Put m_pCDE_data into TDI buffer unless already loaded in case where
//==! iFieldType = dINP_SPECIFIER.
    if (pUFvar->m_pCDE_data->m_iFieldType != dINP_SPECIFIER)
    { for (lFieldSize = 0;
           lFieldSize < pUFvar->m_pCDE_data->m_lDE_len;
           ++lFieldSize)
      { lX = m_lUpdateWidth-lVar-1+pUFvar->m_lFieldWidth-lFieldSize-1;
        pcTdiBuf[lX] = pUFvar->m_pCDE_data->getDataBit(lFieldSize) + '0';
      } // FOR lFieldSize < pUFvar->m_pCDE_data->m_lDE_len
    } // IF

    lVar -= pUFvar->m_lFieldWidth;

  } // FOR pUFvar != NULL

  pcTdiBuf[m_lUpdateWidth] = '\0';

  if (iMode == 0)
  { jtagShift(TAP_DRSHIFT,m_lUpdateWidth,pcTdiBuf,pcTdoBuf,0,0);
    if (iOpt_FlowLoopDbg == 1)
    { debugMsg("\nInput    TDI [",1,1);
      debugStrHex(pcTdiBuf,m_lUpdateWidth,1,1);
      debugMsg("]",1,1);
    } // IF
  } // IF
  return m_lUpdateWidth;

} // CFlowAct::processUpdateField()
//==!=====================================================================!==
//==! CFlowAct member function to process the capture field.
//==!
//==! Function Parameters: iMode = 0: shift data out (used during sequential
//==!                                 processing mode)
//==!                              1: do not shift data out but just process
//==!                                 pcTdoBuf[] data (used during concurrent
//==!                                 processing mode)
//==!                      pcTdiBuf = pointer to TDI buffer
//==!                      pcTdoBuf = pointer to TDO buffer. Note that the
//==!                         TDO buffer has LSB at the max char buffer
//==!                         index, so a 10 element char buffer will have
//==!                         LSB data (data element 0) at location 9, so
//==!                         data element 3 is translated into 10-3-1 = 6:
//==!
//==!                         char index 0 1 2 3 4 5 6 7 8 9
//==!                                   MSB               LSB
//==!                      virtual index 9 8 7 6 5 4 3 2 1 0
//==!                                                ^
//==!                                         -------+
//==!
//==!                      pCDD = CDeviceData pointer (to get input data from
//==!                         if required)
//==!                      pCIFobj = ptr to CIscFlow obj to access variables
//==!                         used by flows
//==!                      lCurRptCount = current repeat count (i.e., how
//==!                         many times this flow was processed in a repeat
//==!                         loop)
//==!                      iStatus = variable pointer to return state of
//==!                         getFlowPhaseData(); note that it will return
//==!                         the current state if not set in
//==!                         getFlowPhaseData()
//==! Return Value       : 0 if successful, 1 otherwise
//==!=====================================================================!==
int CFlowAct::processCaptureField(int iMode,char *pcTdiBuf,char *pcTdoBuf,
                      CDeviceData *pCDD,CIscFlow *pCIFobj,long lCurRptCount,
                      int *iStatus,int iInLoop)
{ int iErrDisplayed = 0,iFuncStatus = 0;
  CxCaptureField *pCFvar;
  long lVar,lCnt,lFieldSize,lArrayIndex,lErrCount,lSubFieldCounter;

  iGlobProgramState = dPROGSTATE_PROCESSCAPTUREFIELD;

//==! lSubFieldCounter variable keeps track of which subfield is being
//==! processed. For example, the following flow activity statement has 2
//==! fields in the capture section: field 1 = 20 bits, field 2 = 12 bits.
//==!
//==!   (IDCODE         wait 1   20:!05033!*FFFFF,12:!063!*FFF)," &

  lErrCount = 0;

  if (m_pCF_Capture == NULL)
  { return iFuncStatus;
  } // IF m_pCF_Capture == NULL

//==! Shift DR data only in sequential mode (iMode == 0). During concurrent
//==! processing, data from the selected data register in all devices in the
//==! chain will be shifted out at once (see NOTE_0009), but this same
//==! routine is used to process the TDO data. The pcTdoBuf[] will point to
//==! the TDO data for the device currently being processed.
  if (iMode == 0)
  { //==! Do not go to RTI at the end of the DR-SHIFT.
    jtagShift(TAP_DRSHIFT,m_lCaptureWidth,NULL,pcTdoBuf,1,1);
  } // IF iMode == 0

  if (iOpt_FlowLoopDbg == 1)
  { debugMsg("\nCaptured TDO [",1,1);
    debugStrHex(pcTdoBuf,m_lCaptureWidth,1,1);
    debugMsg("]",1,1);
  } // IF

  for (pCFvar = m_pCF_Capture,lSubFieldCounter = 1,lVar = m_lCaptureWidth-1;
       pCFvar != NULL;
       pCFvar = pCFvar->m_pNextCxCaptureField,++lSubFieldCounter)
  { lArrayIndex = (pCFvar->m_pCDE_CapturedData->m_lDE_len + 31)/32;
//==! put captured TDO data into m_pCDE_CapturedData before
//==! processing/comparing
    for (lFieldSize = 0;
         lFieldSize < pCFvar->m_pCDE_CapturedData->m_lDE_len;
         ++lFieldSize)
    { pCFvar->m_pCDE_CapturedData->setDataBit(
         (pcTdoBuf[m_lCaptureWidth-lVar-1 + lFieldSize] == '1' ? 1 : 0),
         pCFvar->m_pCDE_CapturedData->m_lDE_len - lFieldSize - 1);
    } // FOR lFieldSize < pCFvar->m_pCDE_CapturedData->m_lDE_len

//==! Output raw captured data
    if (pCFvar->m_cExpDataOutOp == dc_OUTPUT_OP)
    { pCFvar->m_pCDE_CapturedData->showFieldData(pCDD->m_pCDP_Out);
    } // IF

//==! Compare (i.e., EXOR) expected w/ captured data
    if (pCFvar->m_pCDE_expdata != NULL) 
    { getFlowPhaseData(pCFvar->m_pCDE_expdata,pCDD,pCIFobj,
                       pcTdiBuf,lVar,m_lCaptureWidth,1,&(*iStatus));

      if (iOpt_FlowLoopDbg == 1)
      { for (lCnt = 0; lCnt < lArrayIndex; ++lCnt)
        { sprintf(sMsgBuf03,"\nCapture [%08X] Exp [%08x] = %ld (%ld)",
                  pCFvar->m_pCDE_CapturedData->m_pulDataField[lCnt],
                  pCFvar->m_pCDE_expdata->m_pulDataField[lCnt],lCnt,
                  pCFvar->m_pCDE_CapturedData->m_lDE_len);
          debugMsg(sMsgBuf03,1,1);
        } // FOR lCnt < lArrayIndex
      } // IF

      for (lCnt = 0; lCnt < lArrayIndex; ++lCnt)
      { pCFvar->m_pCDE_CapturedData->m_pulDataField[lCnt] =
             pCFvar->m_pCDE_CapturedData->m_pulDataField[lCnt]
           ^ pCFvar->m_pCDE_expdata->m_pulDataField[lCnt];
      } // FOR lCnt < lArrayIndex

//==! Display expected data
      if (pCFvar->m_pCDE_expdata->m_cOperator == dc_OUTPUT_OP ||
          pCFvar->m_pCDE_expdata->m_cOperator == dc_IO_OP)
      { pCFvar->m_pCDE_expdata->showFieldData(pCDD->m_pCDP_Out);
      } // IF
    } // IF pCFvar->m_pCDE_expdata != NULL

    if (pCFvar->m_pCDE_maskdata != NULL)
    { getFlowPhaseData(pCFvar->m_pCDE_maskdata,pCDD,pCIFobj,
                       pcTdiBuf,lVar,m_lCaptureWidth,1,&(*iStatus));

//==! Do not mask or logically AND if dNOT_SET (i.e., no mask specified)
      if (pCFvar->m_pCDE_maskdata->m_iFieldType != dNOT_SET)
      { for (lCnt = 0; lCnt < lArrayIndex; ++lCnt)
        { pCFvar->m_pCDE_CapturedData->m_pulDataField[lCnt] =
               pCFvar->m_pCDE_CapturedData->m_pulDataField[lCnt]
             & pCFvar->m_pCDE_maskdata->m_pulDataField[lCnt];
        } // FOR lCnt < lArrayIndex
      } // IF

//==! Output masked compared data
      if (pCFvar->m_cMaskDataOutOp == dc_OUTPUT_OP)
      { pCFvar->m_pCDE_CapturedData->showFieldData(pCDD->m_pCDP_Out);
      } // IF

//==! Display mask data
      if (pCFvar->m_pCDE_maskdata->m_cOperator == dc_OUTPUT_OP ||
          pCFvar->m_pCDE_maskdata->m_cOperator == dc_IO_OP)
      { pCFvar->m_pCDE_maskdata->showFieldData(pCDD->m_pCDP_Out);
      } // IF
    } // IFpCFvar->m_pCDE_maskdata != NULL

//==! If expected data was specified, then check if captured data after all
//==! the masking done above results in all 0s.
    if (pCFvar->m_pCDE_expdata != NULL) 
    { for (lCnt = 0; lCnt < lArrayIndex; ++lCnt)
      { if (pCFvar->m_pCDE_CapturedData->m_pulDataField[lCnt] != 0)
        { for (lFieldSize = 0; lFieldSize < 32; ++lFieldSize)
          { if ((pCFvar->m_pCDE_CapturedData->m_pulDataField[lCnt] &
                culGlobBitPosMask[lFieldSize]) != 0)
            { ++lErrCount;
              iFuncStatus = 1;
              if (lErrCount > lOptMaxBitErr)
              { if (iErrDisplayed == 0)
                { sprintf(sMsgOnly,
"\n=> Other mismatches suppressed (more than %ld mismatches)",lOptMaxBitErr);
                  jMessage(DISP_SHOW_MSG,sMsgOnly,0,NULL);
                  iErrDisplayed = 1; // Display message only once
                } // IF
                continue;
              } // IF

              sprintf(sMsgBuf00,
                "\n[%s] mismatch: bit [%ld] in [%ld:0] bit field",
                pCDD->m_csDeviceName,
                lFieldSize + (lCnt*32) + lVar-pCFvar->m_lFieldWidth+1,
                m_lCaptureWidth-1);

              if (iInLoop == 1)
              { if (iOpt_ProgDebug == 1)
                { debugMsg(sMsgBuf00,2,1);
                } // IF
              } // IF
              else
              { debugMsg(sMsgBuf00,1,1);
              } // ELSE

              if (lCurRptCount > 1)
              { sprintf(sMsgBuf00," (rpt count [%ld])",lCurRptCount);
                debugMsg(sMsgBuf00,1,1);
              } // IF
              if ((iOpt_ProgDebug == 1) &&
                  (m_lCaptureWidth != pCFvar->m_lFieldWidth))
              { sprintf(sMsgBuf00,
                  "\r= bit [%ld] in [%ld:0] bit subfield # %ld",
                  lFieldSize + (lCnt*32),
                  pCFvar->m_lFieldWidth-1,lSubFieldCounter);
                debugMsg(sMsgBuf00,1,1);
              } // IF m_lCaptureWidth != m_lFieldWidth
            } // IF bit mismatch
          } // FOR lFieldSize < 32
        } // IF
      } // FOR lCnt < lArrayIndex
    } // IF

    lVar -= pCFvar->m_lFieldWidth;

  } // FOR pCFvar != NULL

  return iFuncStatus;

} // CFlowAct::processCaptureField()
//==!=====================================================================!==
//==! CMultFlow constructor.
//==!
//==! Function Parameters: none
//==! Return Value       : none
//==!=====================================================================!==
CMultFlow::CMultFlow()
{
  m_lRptCnt = 0;
  m_pCFA_Repeat = NULL;
  m_lUseDataName_Repeat = 0;
  m_iRptHasOutOp = 0;
  m_pNextCMultFlow = NULL;

} // CMultFlow::CMultFlow()
//==!=====================================================================!==
//==! CMultFlow destructor.
//==!
//==! Function Parameters: none
//==! Return Value       : none
//==!=====================================================================!==
CMultFlow::~CMultFlow()
{ CFlowAct *p_VarCFA;

  m_lRptCnt = 0;
  while (m_pCFA_Repeat != NULL)
  { p_VarCFA = m_pCFA_Repeat->m_pNextCFlowAct;
    freeData(m_pCFA_Repeat);
    m_pCFA_Repeat = p_VarCFA;
  } // WHILE m_pCFA_Repeat != NULL
  m_lUseDataName_Repeat = 0;
  m_iRptHasOutOp = 0;
  m_pNextCMultFlow = NULL;

} // CMultFlow::~CMultFlow()
//==!=====================================================================!==
//==! CIscFlow constructor.
//==!
//==! Function Parameters: none
//==! Return Value       : none
//==!=====================================================================!==
CIscFlow::CIscFlow()
{
  m_csFlowName = NULL;
  m_csDataName = NULL;
  m_iUnprocessed = 0;
  m_iExitOnError = 0;
  m_lMaxRegSizeUsed = 0;

  m_lUseDataName = 0;
  m_lUseDataName_Initialize = 0;
  m_lUseDataName_Terminate = 0;

  m_pCFA_Initialize = NULL;
  m_pCFA_Terminate = NULL;

  m_pCMF_Repeat = NULL;

  m_iInitHasOutOp = 0;
  m_iTermHasOutOp = 0;

  m_lCxDataVarArraySize = 0;
  m_CxDataVarArray = NULL;
  m_pCDVlist = NULL;
  m_pCDVlistTail = NULL;
  m_pNextCIscFlow = NULL;

} // CIscFlow::CIscFlow()
//==!=====================================================================!==
//==! CIscFlow destructor.
//==!
//==! Function Parameters: none
//==! Return Value       : none
//==!=====================================================================!==
CIscFlow::~CIscFlow()
{ CFlowAct *p_VarCFA;
  CMultFlow *p_VarCMF;
  long lVar;
  CxDataVar *p_VarCDV;

  freeArray(m_csFlowName);
  freeArray(m_csDataName);
  m_iUnprocessed = 0;
  m_iExitOnError = 0;
  m_lMaxRegSizeUsed = 0;

  m_lUseDataName = 0;
  m_lUseDataName_Initialize = 0;
  m_lUseDataName_Terminate = 0;

  while (m_pCFA_Initialize != NULL)
  { p_VarCFA = m_pCFA_Initialize->m_pNextCFlowAct;
    freeData(m_pCFA_Initialize);
    m_pCFA_Initialize = p_VarCFA;
  } // WHILE m_pCFA_Initialize != NULL

  while (m_pCFA_Terminate != NULL)
  { p_VarCFA = m_pCFA_Terminate->m_pNextCFlowAct;
    freeData(m_pCFA_Terminate);
    m_pCFA_Terminate = p_VarCFA;
  } // WHILE m_pCFA_Terminate != NULL

  while (m_pCMF_Repeat != NULL)
  { p_VarCMF = m_pCMF_Repeat->m_pNextCMultFlow;
    freeData(m_pCMF_Repeat);
    m_pCMF_Repeat = p_VarCMF;
  } // WHILE m_pCMF_Repeat != NULL

  m_iInitHasOutOp = 0;
  m_iTermHasOutOp = 0;

  if (m_lCxDataVarArraySize > 0)
  { while (m_pCDVlist != NULL)
    { p_VarCDV = m_pCDVlist->m_pNextCxDataVar;
      freeData(m_pCDVlist);
      m_pCDVlist = p_VarCDV;
    } // WHILE m_pCDVlist != NULL

    for (lVar = 0; lVar < m_lCxDataVarArraySize; ++lVar)
    { m_CxDataVarArray[lVar] = NULL;
    } // FOR lVar < m_lCxDataVarArraySize
    freeArray(m_CxDataVarArray);

    m_lCxDataVarArraySize = 0;
  } // IF

  m_pCDVlist = NULL;

  m_pNextCIscFlow = NULL;

} // CIscFlow::~CIscFlow()
//==!=====================================================================!==
//==! CxFlowObj constructor.
//==!
//==! Function Parameters: none
//==! Return Value       : none
//==!=====================================================================!==
CxFlowObj::CxFlowObj()
{
  m_csFlowParam = NULL;
  m_pFlowItem = NULL;
  m_pNextCxFlowObj = NULL;

} // CxFlowObj::CxFlowObj()
//==!=====================================================================!==
//==! CxFlowObj destructor.
//==!
//==! Function Parameters: none
//==! Return Value       : none
//==!=====================================================================!==
CxFlowObj::~CxFlowObj()
{
  freeArray(m_csFlowParam);
//==! m_pFlowItem points to a flow object in the BSDL data object pointer
//==! CBsdlData->m_pCiscFlowInfo; this flow object will be deallocated
//==! when this list is deallocated.
  m_pFlowItem = NULL;
  m_pNextCxFlowObj = NULL;

} // CxFlowObj::~CxFlowObj()
//==!=====================================================================!==
//==! CIscProc constructor.
//==!
//==! Function Parameters: none
//==! Return Value       : none
//==!=====================================================================!==
CIscProc::CIscProc()
{
  m_csProcName = NULL;
  m_csProcDataName = NULL;
  m_lMaxRegSizeUsed = 0;
  m_pProcFlows = NULL;
  m_pNextCIscProc = NULL;

} // CIscProc::CIscProc()
//==!=====================================================================!==
//==! CIscProc destructor.
//==!
//==! Function Parameters: none
//==! Return Value       : none
//==!=====================================================================!==
CIscProc::~CIscProc()
{ CxFlowObj *p_VarFlowObj;

  freeArray(m_csProcName);
  freeArray(m_csProcDataName);
  m_lMaxRegSizeUsed = 0;
  while (m_pProcFlows != NULL)
  { p_VarFlowObj = m_pProcFlows->m_pNextCxFlowObj;
    freeData(m_pProcFlows);
    m_pProcFlows = p_VarFlowObj;
  } // WHILE m_pProcFlows != NULL
  m_pNextCIscProc = NULL;

} // CIscProc::~CIscProc()
//==!=====================================================================!==
//==! CxProcObj constructor.
//==!
//==! Function Parameters: none
//==! Return Value       : none
//==!=====================================================================!==
CxProcObj::CxProcObj()
{
  m_csProcParam = NULL;
  m_iProcExclusivity = 0;
  m_enOptionSpec = ePROC_NO_SETTING;
  m_enOptionOverride = ePROC_NO_SETTING;
  m_pProcItem = NULL;
  m_pNextCxProcObj = NULL;

} // CxProcObj::CxProcObj()
//==!=====================================================================!==
//==! CxProcObj destructor.
//==!
//==! Function Parameters: none
//==! Return Value       : none
//==!=====================================================================!==
CxProcObj::~CxProcObj()
{
  freeArray(m_csProcParam);
  m_iProcExclusivity = 0;
  m_enOptionSpec = ePROC_NO_SETTING;
  m_enOptionOverride = ePROC_NO_SETTING;
//==! m_pProcItem points to a procedure object in the BSDL data object pointer
//==! CBsdlData->m_pCiscProcInfo; this procedure object will be deallocated
//==! when this list is deallocated.
  m_pProcItem = NULL;
  m_pNextCxProcObj = NULL;

} // CxProcObj::~CxProcObj()
//==!=====================================================================!==
//==! CIscAct constructor. Set default values for a BSDL ISC action object
//==!
//==! Function Parameters: none
//==! Return Value       : none
//==!=====================================================================!==
CIscAct::CIscAct()
{
  m_csActName = NULL;
  m_csActDataName = NULL;
  m_lMaxRegSizeUsed = 0;
  m_iExclusivity = 0;
  m_pActionProcs = NULL;
  m_pNextCIscAct = NULL;

} // CIscAct::CIscAct()
//==!=====================================================================!==
//==! CIscAct destructor.
//==!
//==! Function Parameters: none
//==! Return Value       : none
//==!=====================================================================!==
CIscAct::~CIscAct()
{ CxProcObj *p_VarCxProcObj;

  freeArray(m_csActName);
  freeArray(m_csActDataName);
  m_lMaxRegSizeUsed = 0;
  m_iExclusivity = 0;

  while (m_pActionProcs != NULL)
  { p_VarCxProcObj = m_pActionProcs->m_pNextCxProcObj;
    freeData(m_pActionProcs);
    m_pActionProcs = p_VarCxProcObj;
  } // WHILE m_pActionProcs != NULL

  m_pNextCIscAct = NULL;

} // CIscAct::~CIscAct()
//==!=====================================================================!==
//==! CBsdlData constructor. Set default values for a BSDL object
//==!
//==! Function Parameters: none
//==! Return Value       : none
//==!=====================================================================!==
CBsdlData::CBsdlData()
{

  csBsdlFileName = NULL;
  csEntityName = NULL;
  csPortMapName = NULL;
  m_csTapClkName = NULL;
  m_fTckFreq = 1e6;
  m_CxCStr_IdcodeList = NULL;
  m_CxCStr_InstrCaptureList = NULL;
  m_CxCStr_IscBlankUsercodeList = NULL;
  m_CxCStr_UsercodeList = NULL;

  lJtagStdLevel = 2001;  // default to 2001 std level
  lNumOpcodes = 0;
  lNumDataRegs = 0;
  pCBsdlOpInfo = NULL;
  pCBsdlDataRegInfo = NULL;
  m_pCBsdlDataRegListHead = NULL;
  pCPortDataList = NULL;

  lOpcodeSamplePreload = -1;
  lOpcodeExtest = -1;
  lOpcodeBypass = -1;
  lOpcodeUsercode = -1;
  lOpcodeUser1 = -1;
  lOpcodeUser2 = -1;
  lOpcodeIdcode = -1;

  m_iPinBehaviorCheck = 0;
  m_csPinBehavior = NULL;

  pCPortDataInfo = NULL;
  pBscData = NULL;
  iPortMapType = -1;
  lNumPorts = -1;
  lInstrLength = -1;
  lBscLength = -1;
  lImplementState = -1;

  m_pCIscFlowInfo = NULL;
  m_pCIscProcInfo = NULL;
  m_pCIscProcError = NULL;
  m_pCIscActInfo = NULL;

  m_pNextCBsdlData = NULL;

} // CBsdlData::CBsdlData()
//==!=====================================================================!==
//==! CBsdlData destructor.
//==!
//==! Function Parameters: none
//==! Return Value       : none
//==!=====================================================================!==
CBsdlData::~CBsdlData()
{ CPortData *pCPDtmp;
  long lVar;
  CxCStr *pCxCStr_Var;
  CIscFlow *p_VarIscFlow;
  CIscProc *p_VarIscProc;
  CIscAct *p_VarIscAct;

  freeArray(csBsdlFileName);
  freeArray(csEntityName);
  freeArray(csPortMapName);
  freeArray(m_csTapClkName);
  m_fTckFreq = 0;

  while (m_CxCStr_IdcodeList != NULL)
  { pCxCStr_Var = m_CxCStr_IdcodeList->pNextCxCStr;
    m_CxCStr_IdcodeList->pNextCxCStr = NULL;
    freeData(m_CxCStr_IdcodeList);
    m_CxCStr_IdcodeList = pCxCStr_Var;
  } // WHILE m_CxCStr_IdcodeList != NULL

  while (m_CxCStr_InstrCaptureList != NULL)
  { pCxCStr_Var = m_CxCStr_InstrCaptureList->pNextCxCStr;
    m_CxCStr_InstrCaptureList->pNextCxCStr = NULL;
    freeData(m_CxCStr_InstrCaptureList);
    m_CxCStr_InstrCaptureList = pCxCStr_Var;
  } // WHILE m_CxCStr_InstrCaptureList != NULL

  while (m_CxCStr_IscBlankUsercodeList != NULL)
  { pCxCStr_Var = m_CxCStr_IscBlankUsercodeList->pNextCxCStr;
    m_CxCStr_IscBlankUsercodeList->pNextCxCStr = NULL;
    freeData(m_CxCStr_IscBlankUsercodeList);
    m_CxCStr_IscBlankUsercodeList = pCxCStr_Var;
  } // WHILE m_CxCStr_IscBlankUsercodeList != NULL

  while (m_CxCStr_UsercodeList != NULL)
  { pCxCStr_Var = m_CxCStr_UsercodeList->pNextCxCStr;
    m_CxCStr_UsercodeList->pNextCxCStr = NULL;
    freeData(m_CxCStr_UsercodeList);
    m_CxCStr_UsercodeList = pCxCStr_Var;
  } // WHILE m_CxCStr_UsercodeList != NULL

  if (lNumOpcodes > 0)
  { for (lVar = 0; lVar < lNumOpcodes; ++lVar)
    { freeData(pCBsdlOpInfo[lVar]);
    } // lVar < lNumOpcodes
    freeArray(pCBsdlOpInfo);
  } // IF lNumOpcodes > 0
  if (lNumDataRegs > 0)
  { for (lVar = 0; lVar < lNumDataRegs; ++lVar)
    { freeData(pCBsdlDataRegInfo[lVar]);
    } // FOR lVar < lNumDataRegs;
    freeArray(pCBsdlDataRegInfo);
  } // IF lNumRegs > 0

  m_iPinBehaviorCheck = 0;
  freeArray(m_csPinBehavior);

//==! Do not delete the pCPortDataInfo[] pointer elements because these will
//==! be cleared when the pCPortDataList linked list is cleared. Just clear
//==! the pointer array.

  if (pCPortDataInfo != NULL)
  { for (lVar = 0; lVar < lNumPorts; ++lVar)
    { pCPortDataInfo[lVar] = NULL;
    } // FOR lVar < lNumPins
    freeArray(pCPortDataInfo);
  } // IF pCPortDataInfo != NULL
  while (pCPortDataList != NULL)
  { pCPDtmp = pCPortDataList->pNextCPortData;
    pCPortDataList->pNextCPortData = NULL;
    freeData(pCPortDataList);
    pCPortDataList = pCPDtmp;
  } // WHILE pCPortDataList != NULL

  if (lBscLength > 0)
  { if (pBscData != NULL)   //==! See NOTE_0008
    { for (lVar = 0; lVar < lBscLength; ++lVar)
      { freeData(pBscData[lVar]);
      } // FOR
      freeArray(pBscData);
    } // IF pBscData != NULL
  } // IF lBscLength > 0

  iPortMapType = -1;
  lNumPorts = -1;
  lInstrLength = -1;
  lBscLength = -1;
  lImplementState = -1;

  while (m_pCIscFlowInfo != NULL)
  { p_VarIscFlow = m_pCIscFlowInfo->m_pNextCIscFlow;
    freeData(m_pCIscFlowInfo);
    m_pCIscFlowInfo = p_VarIscFlow;
  } // WHILE m_pCIscFlowInfo != NULL

  while (m_pCIscProcInfo != NULL)
  { p_VarIscProc = m_pCIscProcInfo->m_pNextCIscProc;
    freeData(m_pCIscProcInfo);
    m_pCIscProcInfo = p_VarIscProc;
  } // WHILE m_pCIscProcInfo != NULL
  m_pCIscProcError = NULL;  // already deleted when m_pCiscProcInfo deleted

  while (m_pCIscActInfo != NULL)
  { p_VarIscAct = m_pCIscActInfo->m_pNextCIscAct;
    freeData(m_pCIscActInfo);
    m_pCIscActInfo = p_VarIscAct;
  } // WHILE m_pCIscActInfo != NULL

} // CBsdlData::~CBsdlData()
//==!=====================================================================!==
//==! Find the port with the specified name in the port list. Look in the
//==! port name field (iMode = 0) or the base port name field (iMode = 1).
//==!
//==! Function Parameters: pCPDlist = ports in linked list format
//==!                      iMode = field to search (see function description)
//==!                      sPortName = port name to search for
//==! Return Value       : none
//==!=====================================================================!==
CPortData *findPort(CPortData *pGlobCPD,int iMode,char *sPortName)
{ CPortData *pCPD;

  for (pCPD = pGlobCPD; pCPD != NULL; pCPD = pCPD->pNextCPortData)
  { if (iMode == 0)
    { if (!strcmp(sPortName,pCPD->csPortName))
      { return pCPD;
      } // IF
    } // IF
    else
    { if (!strcmp(sPortName,pCPD->csBasePortName))
      { return pCPD;
      } // IF
    } // ELSE
  } // FOR
  return NULL;

} // findPort()
//==!=====================================================================!==
//==! Check BSDL port type
//==!
//==! Function Parameters: sPortType = port type in ASCII format
//==! Return Value       : Enumerated port type
//==!=====================================================================!==
EnumPortType checkPortType(char *sPortType)
{

  if (sPortType == NULL)
  { return ePortNoType;
  } // IF

  if (!strcmp(sPortType,"in"))
  { return ePortInp;
  } // IF
  if (!strcmp(sPortType,"inout"))
  { return ePortIO;
  } // IF
  if (!strcmp(sPortType,"out"))
  { return ePortOut;
  } // IF
  if (!strcmp(sPortType,"linkage"))
  { return ePortLinkage;
  } // IF
  if (!strcmp(sPortType,"buffer"))
  { return ePortBuffer;
  } // IF

  return ePortNoType;

} // checkPortType()
//==!=====================================================================!==
//==! Get BSDL port type string.
//==!
//==! Function Parameters: enumPortParam = port type
//==! Return Value       : Pointer to char string
//==!=====================================================================!==
char *getPortTypeStr(EnumPortType enumPortParam)
{

  if (enumPortParam == ePortInp)
  { return "in";
  } // IF
  if (enumPortParam == ePortIO)
  { return "io";
  } // IF
  if (enumPortParam == ePortOut)
  { return "out";
  } // IF
  if (enumPortParam == ePortLinkage)
  { return "linkage";
  } // IF
  if (enumPortParam == ePortBuffer)
  { return "buffer";
  } // IF

  return "unknownPortType";

} // getPortTypeStr()
//==!=====================================================================!==
//==! Get BSDL port type
//==!
//==! Function Parameters: pDP = BSDL data port pointer
//==!                      pCPD = pointer to CPortData pointer
//==! Return Value       : none
//==!=====================================================================!==
void getBsdlPortInfo(cDataPort *pDP,CPortData **pCPD)
{

  xDataBsdl(pDP,XDATA_READ);
  (*pCPD)->enPortType = checkPortType(pDP->sData);
  if ((*pCPD)->enPortType == ePortNoType)
  { freeData(*pCPD);
    sprintf(sMsgBuf00,"Unknown port type [%s]",pDP->sOrigData);
    jMessage(DISP_AND_EXIT,sMsgBuf00,1,NULL);
  } // IF

  xDataBsdl(pDP,XDATA_READ);
  if (!strcmp(pDP->sData,"bit"))
  { (*pCPD)->lPortNameCnt = 1;
  } // IF
  else
  { if (!strcmp(pDP->sData,"bit_vector"))
    { findStrBsdl(pDP,"(",0);
      (*pCPD)->lMaxNameArrayIndex = lReadDecimal(pDP);
      xDataBsdl(pDP,XDATA_READ);
      if (!strcmp(pDP->sData,"to"))
      { (*pCPD)->iPortNameDir = 1;
      } // IF
      else
      { if (!strcmp(pDP->sData,"downto"))
        { (*pCPD)->iPortNameDir = -1;
        } // IF
        else
        { jMessage(DISP_AND_EXIT,"Expecting to/downto",1,pDP);
        } // ELSE
      } // ELSE
      (*pCPD)->lMinNameArrayIndex = lReadDecimal(pDP);
      findStrBsdl(pDP,")",0);
      (*pCPD)->lPortNameCnt = abs((*pCPD)->lMaxNameArrayIndex -
                                  (*pCPD)->lMinNameArrayIndex) + 1;
      if ((*pCPD)->lMinNameArrayIndex > (*pCPD)->lMaxNameArrayIndex)
      { if ((*pCPD)->iPortNameDir == -1)
        { jMessage(DISP_AND_EXIT,"Use 'to' for [min..max]",1,pDP);
        } // IF
      } // IF
      else
      { if ((*pCPD)->iPortNameDir == 1)
        { jMessage(DISP_AND_EXIT,"Use 'downto' for [max..min]",1,pDP);
        } // IF
      } // ELSE
    } // IF
    else
    { freeData(*pCPD);
      sprintf(sMsgBuf00,"Unknown port size [%s]",pDP->sOrigData);
      jMessage(DISP_AND_EXIT,sMsgBuf00,1,NULL);
    } // ELSE
  } // ELSE

  xDataBsdl(pDP,XDATA_READ);
  if (!strcmp(pDP->sData,")"))
  { pDP->iParserGetData = 0;
  } // IF

} // getBsdlPortInfo()
//==!=====================================================================!==
//==! Search pCBsdlOpInfo array for sOpcode.
//==!
//==! Function Parameters: sOpcode = opcode string
//==!                      pDP = source of sOpcode
//==!                      iMode = if 1 then:
//==!                        - return -1 if opcode list not yet defined
//==!                        - return -2 if opcode not found
//==! Return Value       : index into pCvsdlOpInfo[]
//==!=====================================================================!==
long CBsdlData::getOpcodeIndex(int iMode,char *sOpcode,cDataPort *pDP)
{ long lVar0;

  if (lNumOpcodes <= 0 || pCBsdlOpInfo == NULL)
  { if (iMode == 1)
    { return -1;
    } // IF
    jMessage(DISP_AND_EXIT,"No opcodes!",1,pDP);  // should never happen
  } // IF
  for (lVar0 = 0; lVar0 < lNumOpcodes; ++lVar0)
  { if (!strcmp(pCBsdlOpInfo[lVar0]->csOpcodeName,sOpcode))
    { return lVar0;
    } // IF
  } // FOR lVar0 < lNumOpcodes
  if (iMode == 1)
  { return -2;
  } // IF
  sprintf(sMsgOnly,"[%s] not found",sOpcode);
  jMessage(DISP_AND_EXIT,sMsgOnly,1,pDP);
  return -1;

} // CBsdlData::getOpcodeIndex()
//==!=====================================================================!==
//==! Verify if the override setting for the action + procedure + (optional)
//==! procedure data is valid. Note that if there is no option specification
//==! in the BSDL file (i.e., procedure is not defined as optional or
//==! recommended, then no override setting (on/off/optional/recommended) is
//==! allowed.
//==!
//==! Function Parameters: csAction = action
//==!                      csActionData = optional action data
//==!                      csProc = procedure
//==!                      csProcData = optional procedure data
//==!                      eProcSetting = procedure setting
//==!                      pDP = pointer to data port
//==! Return Value       : 1 if override allowed, 0 otherwise
//==!=====================================================================!==
int CBsdlData::verifyOverride(char *csAction,char *csActionData,
                   char *csProc,char *csProcData,
                   EnumProcSetting eProcSetting,cDataPort *pDP)
{ CIscAct *pCAct;
  CxProcObj *pCPO;

  for (pCAct = m_pCIscActInfo; pCAct != NULL; pCAct = pCAct->m_pNextCIscAct)
  { if (strcmp(pCAct->m_csActName,csAction))
    { continue;
    } // IF
    if (pCAct->m_csActDataName != NULL && csActionData != NULL)
    { if (strcmp(pCAct->m_csActDataName,csActionData))
      { continue;
      } // IF
    } // IF
    for (pCPO = pCAct->m_pActionProcs; pCPO != NULL;
         pCPO = pCPO->m_pNextCxProcObj)
    {
//==! There should always be a procedure associated with the action
      isPtrNull(pCPO->m_pProcItem,"AP_PI");

      if (strcmp(pCPO->m_pProcItem->m_csProcName,csProc))
      { continue;
      } // IF
      if (pCPO->m_pProcItem->m_csProcDataName != NULL &&
          csProcData != NULL)
      { if (strcmp(pCPO->m_pProcItem->m_csProcDataName,
                   csProcData))
        { continue;
        } // IF
      } // IF

//==! Check if procedure is mandatory
      if (pCPO->m_enOptionSpec == ePROC_NO_SETTING)
      { sprintf(sMsgOnly,
                "Action [%s%c%s%c] Proc [%s%c%s%c] in [%s] is mandatory",
                csAction,
                (csActionData == NULL ? ' ' : '('),
                (csActionData == NULL ? " " : csActionData),
                (csActionData == NULL ? ' ' : ')'),
                csProc,
                (csProcData == NULL ? ' ' : '('),
                (csProcData == NULL ? " " : csProcData),
                (csProcData == NULL ? ' ' : ')'),
                csBsdlFileName);
        jMessage(DISP_SHOW_MSG,sMsgOnly,1,pDP);
        return 0;
      } // IF m_enOptionSpec == ePROC_NO_SETTING (i.e., mandatory)

//==! Update override value only if different from m_enOptionSpec (e.g.,
//==! RECOMMENDED to OPTIONAL or OPTIONAL to ON)
      if (pCPO->m_enOptionSpec != eProcSetting)
      { pCPO->m_enOptionOverride = eProcSetting;
      } // IF
      return 1;
    } // FOR pCPO != NULL

    sprintf(sMsgOnly,
            "No action [%s%c%s%c] proc [%s%c%s%c] in [%s] to override",
            csAction,
            (csActionData == NULL ? ' ' : '('),
            (csActionData == NULL ? " " : csActionData),
            (csActionData == NULL ? ' ' : ')'),
            csProc,
            (csProcData == NULL ? ' ' : '('),
            (csProcData == NULL ? " " : csProcData),
            (csProcData == NULL ? ' ' : ')'),
            csBsdlFileName);
    jMessage(DISP_SHOW_MSG,sMsgOnly,1,pDP);
    return 0;
  } // FOR pCAct != NULL
  sprintf(sMsgOnly,"No action [%s%c%s%c] in [%s] to override",
          csAction,
          (csActionData == NULL ? ' ' : '('),
          (csActionData == NULL ? " " : csActionData),
          (csActionData == NULL ? ' ' : ')'),
          csBsdlFileName);
  jMessage(DISP_SHOW_MSG,sMsgOnly,1,pDP);
  return 0;

} // CBsdlData::verifyOverride()
//==!=====================================================================!==
//==! Read BSDL port statement.
//==!
//==! Function Parameters: pDP_In = BSDL data port pointer
//==! Return Value       : number of ports in port statement
//==!=====================================================================!==
long CBsdlData::readBsdlPort(cDataPort *pDP_In)
{ long lPortCount,lVar,lCurrentPortCnt;
  char *pcsTempPortName;
  CPortData *pCPDtail;

  iGlobProgramState = dPROGSTATE_READBSDLPORT;

  pCPDtail = NULL;

  findStrBsdl(pDP_In,"(",0);
  for (xDataBsdl(pDP_In,XDATA_READ),lPortCount = 0,pCPortDataList = NULL;
       pDP_In->sData[0] != ')'; xDataBsdl(pDP_In,XDATA_READ))
  { if (pDP_In->eDataAttr != eATTR_VAR)
    { jMessage(DISP_AND_EXIT,"Expecting a port name",1,pDP_In);
    } // IF
    pcsTempPortName = NULL;
    copyStr(&pcsTempPortName,pDP_In->sData);
    pTempCPortData = new CPortData;
    findStrBsdl(pDP_In,":",0);
    getBsdlPortInfo(pDP_In,&pTempCPortData);
    lPortCount += pTempCPortData->lPortNameCnt;
    if (pTempCPortData->lPortNameCnt == 1)
    { if (findPort(pCPortDataList,0,pcsTempPortName) != NULL)
      { sprintf(sMsgBuf00,"[%s] already defined",pcsTempPortName);
        freeData(pTempCPortData);
        freeArray(pcsTempPortName);
        jMessage(DISP_AND_EXIT,sMsgBuf00,1,pDP_In);
        return -1;
      } // IF
      pTempCPortData->csPortName = pcsTempPortName;
      copyStr(&(pTempCPortData->csBasePortName),pcsTempPortName);
      pcsTempPortName = NULL;
      if (pCPortDataList == NULL)
      { pCPortDataList = pTempCPortData;
        pCPDtail = pTempCPortData;
      } // IF
      else
      { pCPDtail->pNextCPortData = pTempCPortData;
        pCPDtail = pTempCPortData;
      } // ELSE
      pTempCPortData = NULL;
    } // IF pTempCPortData->lPinNameCnt = 1
    else
    { lCurrentPortCnt = pTempCPortData->lPortNameCnt;
      for (lVar = 0; lVar < lCurrentPortCnt; ++lVar)
      { sprintf(sMsgBuf00,"%s(%d)",pcsTempPortName,
                pTempCPortData->lMaxNameArrayIndex +
                   (pTempCPortData->iPortNameDir*lVar));
        if (findPort(pCPortDataList,0,sMsgBuf00) != NULL)
        { freeData(pTempCPortData);
          freeArray(pcsTempPortName);
          sprintf(sMsgBuf01,"[%s] already defined",sMsgBuf00);
          jMessage(DISP_AND_EXIT,sMsgBuf01,1,pDP_In);
          return -1;
        } // IF
        copyStr(&(pTempCPortData->csPortName),sMsgBuf00);
        copyStr(&(pTempCPortData->csBasePortName),pcsTempPortName);
        if (pCPortDataList == NULL)
        { pCPortDataList = pTempCPortData;
          pCPDtail = pTempCPortData;
        } // IF
        else
        { pCPDtail->pNextCPortData = pTempCPortData;
          pCPDtail = pTempCPortData;
        } // ELSE
        if (lVar+1 < lCurrentPortCnt)
        { pTempCPortData = new CPortData;
          pTempCPortData->lPortNameCnt = pCPDtail->lPortNameCnt;
          pTempCPortData->iPortNameDir = pCPDtail->iPortNameDir;
          pTempCPortData->lMinNameArrayIndex =
                              pCPDtail->lMinNameArrayIndex;
          pTempCPortData->lMaxNameArrayIndex =
                              pCPDtail->lMaxNameArrayIndex;
          pTempCPortData->enPortType = pCPDtail->enPortType;  // NOTE_0011
        } // IF
        else
        { pTempCPortData = NULL;
        } // ELSE
      } // FOR lVar < lCurrentPortCnt
      freeArray(pcsTempPortName);
    } // ELSE
  } // FOR sData[0] != ')'
  findStrBsdl(pDP_In,";",0);

  return lPortCount;

} // CBsdl::readBsdlPort()
//==!=====================================================================!==
//==! Read BSDL port map. Assign the port mappings (i.e., pin or site number
//==! assignments to the appropriate port. After reading port mappings,
//==! check that all ports have been mapped by checking that lPortLocation
//==! and csPortLocation fields in the BSDL object for all the ports have
//==! values in them.
//==!
//==! 2003/10/14: Number of ports in entity section may not include all
//==! possible pins (e.g., VCC, GND, NC) as per 1149.1 STD. We therefore
//==! cannot use the pin number as an index into the number of ports read.
//==! Ex: If only 8 ports are defined, one port can be assigned pin number
//==! 35. Use lPortsUsed instead (see NOTE_0006).
//==!
//==! Function Parameters: pDP_In = BSDL data port pointer
//==! Return Value       : none
//==!=====================================================================!==
void CBsdlData::readPortMap(cDataPort *pDP_In)
{ long lPortsUsed,lVar0,lVar1,lVar2;
  char *pcsPortName;
  CPortData *pCPD_0,*pCPD_1;

  lPortsUsed = 0;
  findStr(pDP_In,"\"",0);
  ++pDP_In->iBracketLevel;
  while (1)
  { xGetData(pDP_In,1,eATTR_VAR,"Expecting port name");
    copyStr(&pcsPortName,pDP_In->sData);
    findStrBsdl(pDP_In,":",0);
    xDataBsdl(pDP_In,XDATA_READ);
    if (!strcmp(pDP_In->sData,"("))
    { pCPD_0 = findPort(pCPortDataList,1,pcsPortName);
      if (pCPD_0 == NULL)
      { sprintf(sMsgBuf00,"Cannot find port [%s]",pcsPortName);
        freeArray(pcsPortName);
        jMessage(DISP_AND_EXIT,sMsgBuf00,1,pDP_In);
      } // IF pCPD_0 == NULL
      if (pCPD_0->lPortLocation != -1)
      { sprintf(sMsgBuf00,"[%s] already assigned to %s",pcsPortName,
                pCPD_0->csPortLocation);
        freeArray(pcsPortName);
        jMessage(DISP_AND_EXIT,sMsgBuf00,1,pDP_In);
      } // IF
      lVar0 = pCPD_0->lPortNameCnt;
      if (lVar0 <= 1)
      { sprintf(sMsgBuf00,"[%s] is not a port vector",pcsPortName);
        freeArray(pcsPortName);
        jMessage(DISP_AND_EXIT,sMsgBuf00,1,pDP_In);
      } // IF
      for (lVar1 = 0; lVar1 < lVar0; ++lVar1)
      { sprintf(sMsgBuf01,"%s(%d)",pcsPortName,
                pCPD_0->lMaxNameArrayIndex + (pCPD_0->iPortNameDir*lVar1));
        pCPD_1 = findPort(pCPortDataList,0,sMsgBuf01);
        if (pCPD_1 == NULL)
        { sprintf(sMsgBuf00,"Cannot find port vector element [%s]",sMsgBuf01);
          freeArray(pcsPortName);
          jMessage(DISP_AND_EXIT,sMsgBuf00,1,pDP_In);
        } // IF pCPD_1 == NULL
        xDataBsdl(pDP_In,XDATA_READ);
        if (pDP_In->eDataAttr == eATTR_NUM)
        { if (iPortMapType == 1)
          { sprintf(sMsgBuf00,"Entity [%s] already using site numbers",
                    csEntityName);
            jMessage(DISP_SHOW_MSG,sMsgBuf00,1,pDP_In);
          } // IF
          else
          { iPortMapType = 0;
          } // ELSE
        } // IF
        else
        { if (pDP_In->eDataAttr != eATTR_VAR)
          { sprintf(sMsgBuf00,"Invalid port map [%s]",pDP_In->sData);
            freeArray(pcsPortName);
            jMessage(DISP_AND_EXIT,sMsgBuf00,1,pDP_In);
          } // IF
          if (iPortMapType == 0)
          { sprintf(sMsgBuf00,"Entity [%s] already using pin numbers",
                    csEntityName);
            jMessage(DISP_SHOW_MSG,sMsgBuf00,1,pDP_In);
          } // IF;
          else
          { iPortMapType = 1;
          } // ELSE
        } // ELSE

//==! See NOTE_0006
        for (lVar2 = 0; lVar2 < lPortsUsed; ++lVar2)
        { if (!strcmp(pCPortDataInfo[lVar2]->csPortLocation,
               pDP_In->sData))
          { sprintf(sMsgBuf00,"Port mapping [%s] already used by [%s]",
                    pDP_In->sData,
                    pCPortDataInfo[lVar2]->csPortName);
            freeArray(pcsPortName);
            jMessage(DISP_AND_EXIT,sMsgBuf00,1,pDP_In);
          } // IF
        } // FOR lVar2 < lPortsUsed
        if (pCPortDataInfo[lPortsUsed] != NULL)
        { sprintf(sMsgBuf00,
           "Port mapping [%s] already used (do not mix site and pin numbers)",
            pDP_In->sData);
          freeArray(pcsPortName);
          jMessage(DISP_AND_EXIT,sMsgBuf00,1,pDP_In);
        } // IF
        pCPortDataInfo[lPortsUsed] = pCPD_1;
        pCPD_1->lPortLocation = lPortsUsed;
        copyStr(&(pCPD_1->csPortLocation),pDP_In->sData);
        ++lPortsUsed;

        xDataBsdl(pDP_In,XDATA_READ);
        if ((lVar1 + 1) < lVar0)
        { if (strcmp(pDP_In->sData,","))
          { sprintf(sMsgBuf00,
              "incomplete vector mapping (only %ld of %ld specified)",
               lVar1+1,lVar0);
            freeArray(pcsPortName);
            jMessage(DISP_AND_EXIT,sMsgBuf00,1,pDP_In);
          } // IF
        } // IF
        else
        { if (strcmp(pDP_In->sData,")"))
          { sprintf(sMsgBuf00,"port vector [%s] has only [%ld] elements",
                    pcsPortName,lVar0);
            freeArray(pcsPortName);
            jMessage(DISP_AND_EXIT,sMsgBuf00,1,pDP_In);
          } // IF
        } // ELSE
      } // FOR lVar1 < lVar0
      freeArray(pcsPortName);
    } // IF sData = '('
    else
    { pCPD_0 = findPort(pCPortDataList,0,pcsPortName);
      if (pCPD_0 == NULL)
      { sprintf(sMsgBuf00,"Cannot find [%s]",pcsPortName);
        freeArray(pcsPortName);
        jMessage(DISP_AND_EXIT,sMsgBuf00,1,pDP_In);
      } // IF
      if (pCPD_0->lPortLocation != -1)
      { sprintf(sMsgBuf00,"[%s] already assigned to %s",pcsPortName,
                pCPD_0->csPortLocation);
        freeArray(pcsPortName);
        jMessage(DISP_AND_EXIT,sMsgBuf00,1,pDP_In);
      } // IF
      if (pDP_In->eDataAttr == eATTR_NUM)
      { if (iPortMapType == 1)
        { sprintf(sMsgBuf00,"Entity [%s] already using site numbers",
                  csEntityName);
          jMessage(DISP_SHOW_MSG,sMsgBuf00,1,pDP_In);
        } // IF
        else
        { iPortMapType = 0;
        } // ELSE
      } // IF
      else
      { if (pDP_In->eDataAttr != eATTR_VAR)
        { sprintf(sMsgBuf00,"Invalid port map [%s]",pDP_In->sData);
          freeArray(pcsPortName);
          jMessage(DISP_AND_EXIT,sMsgBuf00,1,pDP_In);
        } // IF
        if (iPortMapType == 0)
        { sprintf(sMsgBuf00,"Entity [%s] already using pin numbers",
                  csEntityName);
          jMessage(DISP_SHOW_MSG,sMsgBuf00,1,pDP_In);
        } // IF;
        else
        { iPortMapType = 1;
        } // ELSE
      } // ELSE

//==! See NOTE_0006
      for (lVar2 = 0; lVar2 < lPortsUsed; ++lVar2)
      { if (!strcmp(pCPortDataInfo[lVar2]->csPortLocation,
             pDP_In->sData))
        { sprintf(sMsgBuf00,"Port mapping [%s] already used by [%s]",
                  pDP_In->sData,
                  pCPortDataInfo[lVar2]->csPortName);
          freeArray(pcsPortName);
          jMessage(DISP_AND_EXIT,sMsgBuf00,1,pDP_In);
        } // IF
      } // FOR lVar2 < lPortsUsed
      if (pCPortDataInfo[lPortsUsed] != NULL)
      { sprintf(sMsgBuf00,
         "Port mapping [%s] already used (do not mix site and pin numbers)",
          pDP_In->sData);
        freeArray(pcsPortName);
        jMessage(DISP_AND_EXIT,sMsgBuf00,1,pDP_In);
      } // IF
      pCPortDataInfo[lPortsUsed] = pCPD_0;
      pCPD_0->lPortLocation = lPortsUsed;
      copyStr(&(pCPD_0->csPortLocation),pDP_In->sData);
      ++lPortsUsed;

    } // ELSE

    freeArray(pcsPortName);
    xDataBsdl(pDP_In,XDATA_READ);
    if (!strcmp(pDP_In->sData,";"))
    { break;
    } // IF
  } // WHILE (1)

//==! Check that all ports have been mapped by looking through the global
//==! port linked list.
//==! 2004/09/22: Ports of type linkage not assigned/mapped to any pin/site
//==! will be flagged as warnings, but other port types (in, out, inout,
//==! buffer) must be assigned to pins/sites.
  for (pCPD_0 = pCPortDataList,lVar0 = 0; pCPD_0 != NULL;
       pCPD_0 = pCPD_0->pNextCPortData)
  { if (pCPD_0->lPortLocation == -1)
    { sprintf(sMsgBuf00,"Warning: [%s] port [%s] not mapped to any pin/site",
              getPortTypeStr(pCPD_0->enPortType),pCPD_0->csPortName);
      if (pCPD_0->enPortType != ePortLinkage)
      { ++lVar0;
      } // IF
      jMessage(DISP_SHOW_MSG,sMsgBuf00,0,NULL);
    } // IF
  } // FOR pCPD_0 != NULL
  if (lVar0 > 0)
  { jMessage(DISP_AND_EXIT,
             "In/Out/InOut/Buffer ports must be assigned to pins/sites",
             0,NULL);
  } // IF

} // readPortMap()
//==!=====================================================================!==
//==! prtOpcode(): print BSDL opcode(s) to file
//==!
//==! Function Parameters: pFILE = output file
//==!                      pBsdlObj = BSDL object
//==!                      sOpcode = opcode string
//==!                      lOpIndex = opcode index into BSDL object array
//==! Return Value       : none
//==!=====================================================================!==
void prtOpcode(FILE *pFILE,CBsdlData *pBsdlObj,char *sOpcode,long lOpIndex)
{ CxCStr *pCxCStr_Var;

  fprintf(pFILE,"  => %s\t",sOpcode);
  if (isInRange(0,lOpIndex,0,pBsdlObj->lNumOpcodes-1,"prtOpR",DISP_AND_EXIT) == 1)
  { isPtrNull(pBsdlObj->pCBsdlOpInfo[lOpIndex],"prtOp"); // Should not be NULL
    for (pCxCStr_Var = pBsdlObj->pCBsdlOpInfo[lOpIndex]->m_CxCStr_OpcodeList;
         pCxCStr_Var != NULL;
         pCxCStr_Var = pCxCStr_Var->pNextCxCStr)
    { fprintf(pFILE,"[%s] ",pCxCStr_Var->pStr);
    } // FOR pCxCStr_Var != NULL
  } // IF
  else
  { fprintf(pFILE,"[*** none ***]");
  } // ELSE
  fprintf(pFILE,"\n");

} // prtOpcode()
//==!=====================================================================!==
//==! bsdlDataOut(): write debug data out
//==!
//==! Function Parameters: pBsdlObj
//==! Return Value       : none
//==!=====================================================================!==
void bsdlDataOut(CBsdlData *pBsdlObj)
{ long lVar;
  CxCStr *pCxCStr_Var;
  SList *pSListVar;

  if (pDP_Output == NULL)
  { return;
  } // IF
  if (pDP_Output->pFILE == NULL)
  { return;
  } // IF
  fprintf(pDP_Output->pFILE,"\n===== Info from BSDL file [%s] =====\n",
          pDP_Input->psPortName);
  fprintf(pDP_Output->pFILE,
    "Entity [%s] PortMapName [%s]\n- ports(%ld)\n- ir(%ld)\n- bsc(%ld)\n",
          pBsdlObj->csEntityName,
          pBsdlObj->csPortMapName,
          pBsdlObj->lNumPorts,
          pBsdlObj->lInstrLength,
          pBsdlObj->lBscLength);
  fprintf(pDP_Output->pFILE,"= dev IDCODE");
  if (pBsdlObj->m_CxCStr_IdcodeList == NULL)
  { fprintf(pDP_Output->pFILE," (*** none ***)\n");
  } // IF
  else
  { for (pCxCStr_Var = pBsdlObj->m_CxCStr_IdcodeList;
         pCxCStr_Var != NULL;
         pCxCStr_Var = pCxCStr_Var->pNextCxCStr)
    { fprintf(pDP_Output->pFILE,"  - [%s]\n",pCxCStr_Var->pStr);
    } // FOR pCxCStr_Var != NULL
  } // ELSE

  fprintf(pDP_Output->pFILE,"= Opcodes:\n");
  for (lVar = 0; lVar < pBsdlObj->lNumOpcodes; ++lVar)
  { isPtrNull(pBsdlObj->pCBsdlOpInfo[lVar],"wrtOp"); // Should not be NULL
    fprintf(pDP_Output->pFILE,"  [%2ld] [%18s] = ",
            lVar,pBsdlObj->pCBsdlOpInfo[lVar]->csOpcodeName);
    for (pCxCStr_Var = pBsdlObj->pCBsdlOpInfo[lVar]->m_CxCStr_OpcodeList;
         pCxCStr_Var != NULL;
         pCxCStr_Var = pCxCStr_Var->pNextCxCStr)
    { fprintf(pDP_Output->pFILE,"[%s] ",pCxCStr_Var->pStr);
    } // FOR pCxCStr_Var != NULL
    fprintf(pDP_Output->pFILE,"\n");
  } // FOR lVar < pBsdlObj->lNumOpcodes

  fprintf(pDP_Output->pFILE,"= Frequently-used Opcodes:\n");
  prtOpcode(pDP_Output->pFILE,pBsdlObj,"Sample"  ,pBsdlObj->lOpcodeSamplePreload);
  prtOpcode(pDP_Output->pFILE,pBsdlObj,"Extest"  ,pBsdlObj->lOpcodeExtest);
  prtOpcode(pDP_Output->pFILE,pBsdlObj,"Idcode"  ,pBsdlObj->lOpcodeIdcode);
  prtOpcode(pDP_Output->pFILE,pBsdlObj,"Usercode",pBsdlObj->lOpcodeUsercode);
  prtOpcode(pDP_Output->pFILE,pBsdlObj,"User1"   ,pBsdlObj->lOpcodeUser1);
  prtOpcode(pDP_Output->pFILE,pBsdlObj,"User2"   ,pBsdlObj->lOpcodeUser2);

  if (pBsdlObj->pCBsdlDataRegInfo != NULL)
  { fprintf(pDP_Output->pFILE,"= Register access attribute info:\n");
    for (lVar = 0; lVar < pBsdlObj->lNumDataRegs; ++lVar)
    { fprintf(pDP_Output->pFILE,"  - [%2ld] : %s[%ld] = ",
              lVar,pBsdlObj->pCBsdlDataRegInfo[lVar]->csRegName,
              pBsdlObj->pCBsdlDataRegInfo[lVar]->lRegLength);
      for (pSListVar = pBsdlObj->pCBsdlDataRegInfo[lVar]->pOpList;
           pSListVar != NULL; pSListVar = pSListVar->pNextSList)
      { fprintf(pDP_Output->pFILE,"%s",
                pBsdlObj->pCBsdlOpInfo[pSListVar->lVal]->csOpcodeName);
        if (pSListVar->pNextSList != NULL)
        { fprintf(pDP_Output->pFILE,",");
        } // IF
      } // FOR pSList != NULL
      fprintf(pDP_Output->pFILE,";\n");
    } // FOR lVar < pBsdlObj->lNumDataRegs
  } // IF pBsdlObj->pCBsdlDataRegInfo != NULL

  for (lVar = 0; lVar < pBsdlObj->lNumPorts; ++lVar)
  { fprintf(pDP_Output->pFILE,"[%3ld]",lVar);
    if (pBsdlObj->pCPortDataInfo[lVar] != NULL)
    { fprintf(pDP_Output->pFILE,
            " %2d %2d %3ld:%3ld %3ld %5s i %3d o %3d c %3d [%20s (%20s)] ",
            pBsdlObj->pCPortDataInfo[lVar]->lPortNameCnt,
            pBsdlObj->pCPortDataInfo[lVar]->iPortNameDir,
            pBsdlObj->pCPortDataInfo[lVar]->lMaxNameArrayIndex,
            pBsdlObj->pCPortDataInfo[lVar]->lMinNameArrayIndex,
            pBsdlObj->pCPortDataInfo[lVar]->lPortLocation,
            (pBsdlObj->pCPortDataInfo[lVar]->csPortLocation == NULL
                ? " "
                : pBsdlObj->pCPortDataInfo[lVar]->csPortLocation
            ),
            pBsdlObj->pCPortDataInfo[lVar]->lInCell,
            pBsdlObj->pCPortDataInfo[lVar]->lOutCell,
            pBsdlObj->pCPortDataInfo[lVar]->lCtrlCell,
            pBsdlObj->pCPortDataInfo[lVar]->csPortName,
            pBsdlObj->pCPortDataInfo[lVar]->csBasePortName
           );
    } // IF pBsdlObj->pCPortDataInfo[lVar] != NULL
    fprintf(pDP_Output->pFILE,"\n");
  } // FOR lVar < pBsdl->lNumPorts
  fprintf(pDP_Output->pFILE,"------ Boundary Scan Register chain ------\n");
  for (lVar = 0; lVar < pBsdlObj->lBscLength; ++lVar)
  { if (pBsdlObj->pBscData[lVar]->lBscIndex != lVar)
    { jMessage(DISP_AND_EXIT,"BSC index err",0,NULL);
    } // IF
    fprintf(pDP_Output->pFILE,"[%4ld] %cC_%02ld [%26s]",
            lVar,(pBsdlObj->pBscData[lVar]->lBscCellType == 0 ? 'B' : 'A'),
            pBsdlObj->pBscData[lVar]->lBscCellNum,
            (pBsdlObj->pBscData[lVar]->lPortNum == -1
               ? " "
               : pBsdlObj
                   ->pCPortDataInfo[pBsdlObj->pBscData[lVar]->lPortNum]
                   ->csPortName
            )
           );
    switch(pBsdlObj->pBscData[lVar]->enBscFunc)
    { case eBscInput:
        fprintf(pDP_Output->pFILE,"%9s","input");
        break;
      case eBscClock:
        fprintf(pDP_Output->pFILE,"%9s","clock");
        break;
      case eBscOutput2:
        fprintf(pDP_Output->pFILE,"%9s","output2");
        break;
      case eBscOutput3:
        fprintf(pDP_Output->pFILE,"%9s","output3");
        break;
      case eBscControl:
        fprintf(pDP_Output->pFILE,"%9s","control");
        break;
      case eBscControlr:
        fprintf(pDP_Output->pFILE,"%9s","controlr");
        break;
      case eBscInternal:
        fprintf(pDP_Output->pFILE,"%9s","int");
        break;
      case eBscBidir:
        fprintf(pDP_Output->pFILE,"%9s","bidir");
        break;
      case eBscObserve_Only:
        fprintf(pDP_Output->pFILE,"%9s","observe");
        break;
      default:
        fprintf(pDP_Output->pFILE,"%9s","--- ? ---");
        break;
    } // SWITCH
    fprintf(pDP_Output->pFILE," Safe[%c] Dv[%c]",
            (pBsdlObj->pBscData[lVar]->iSafeVal == -1
                 ? ' ' : (char)(pBsdlObj->pBscData[lVar]->iSafeVal + '0')),
            (pBsdlObj->pBscData[lVar]->iDisVal == -1
                 ? ' ' : (char)(pBsdlObj->pBscData[lVar]->iDisVal + '0'))
           );
    if (pBsdlObj->pBscData[lVar]->lCtrlCell >= 0)
    { fprintf(pDP_Output->pFILE," C [%5ld]",
              pBsdlObj->pBscData[lVar]->lCtrlCell);
    } // IF
    fprintf(pDP_Output->pFILE,"\n");
  } // FOR lVar < pBsdlObj->lBscLength
  fprintf(pDP_Output->pFILE,"===== end of BSDL file info ========");

} // bsdlDataOut()
//==!=====================================================================!==
//==! checkAttrEntity(): check that the entity name in the attribute matches
//==! the name stored in the BSDL object
//==!
//==! Function Parameters: pDP = BSDL data port
//==! Return Value       : none
//==!=====================================================================!==
void CBsdlData::checkAttrEntity(cDataPort *pDP)
{

  findStrBsdl(pDP,"of",0);
  xDataBsdl(pDP,XDATA_READ);
  if (strcmp(pDP->sData,csEntityName))
  { sprintf(sMsgBuf00,"[%s] does not match entity [%s]",
            pDP->sData,csEntityName);
    jMessage(DISP_AND_EXIT,sMsgBuf00,1,pDP);
  } // IF
  findStrBsdl(pDP,":",0);
  findStrBsdl(pDP,"entity",0);
  findStrBsdl(pDP,"is",0);

} // CBsdlData::checkAttrEntity()
//==!=====================================================================!==
//==! validateData(): Check if data (of size lDataSize) contains only the
//==! characters found in sValChars.
//==!
//==! Function Parameters: sData = data to validate
//==!                      lDataSize = size of sData (sData can non-NULL
//==!                         terminated)
//==!                      sValChars = valid characters
//==!                      pDP = cDataPort that data was read from
//==! Return Value       : none
//==!=====================================================================!==
void validateData(char *sData,long lDataSize,char *sValChars,
                   cDataPort *pDP)
{ long lVar;

  isPtrNull((void *)(sData),"valData0");
  isPtrNull((void *)(sValChars),"valData1");

  for (lVar = 0; lVar < lDataSize; ++lVar)
  { if (strchr(sValChars,sData[lVar]) == NULL)
    { sprintf(sMsgBuf00,"[%s] must contain only chars [%s]",
              sData,sValChars);
      jMessage(DISP_AND_EXIT,sMsgBuf00,1,pDP);
    } // IF
  } // FOR lVar < lDataSize

} // validateData()
//==!=====================================================================!==
//==! getEnBscType(): get enumerated BSC type given the BSC function name in
//==! ASCII.
//==!
//==! Function Parameters: pDP = input cDataPort
//==!                      sBscTypeStr = BSC type string
//==! Return Value       : enumerated BSC function value
//==!=====================================================================!==
EnumBscFunction CBsdlData::getEnBscType(cDataPort *pDP,char *sBscTypeStr)
{

  if (!xstrcmp(sBscTypeStr,"input"))
  { return eBscInput;
  } // IF
  if (!xstrcmp(sBscTypeStr,"clock"))
  { return eBscClock;
  } // IF
  if (!xstrcmp(sBscTypeStr,"output2"))
  { return eBscOutput2;
  } // IF
  if (!xstrcmp(sBscTypeStr,"output3"))
  { return eBscOutput3;
  } // IF
  if (!xstrcmp(sBscTypeStr,"control"))
  { return eBscControl;
  } // IF
  if (!xstrcmp(sBscTypeStr,"controlr"))
  { return eBscControlr;
  } // IF
  if (!xstrcmp(sBscTypeStr,"internal"))
  { return eBscInternal;
  } // IF
  if (!xstrcmp(sBscTypeStr,"bidir"))
  { return eBscBidir;
  } // IF
  if (!xstrcmp(sBscTypeStr,"observe_only"))
  { return eBscObserve_Only;
  } // IF

  sprintf(sMsgBuf00,
          "[%s]: unknown BSC function (will default to no BSC function)",
          sBscTypeStr);
  jMessage(DISP_SHOW_MSG,sMsgBuf00,1,pDP);
  return eNoBscFunc;

} // CBsdlData::getEnBscType()
//==!=====================================================================!==
//==! getEnBscResult(): get enumerated driver result given the result in
//==! ASCII form.
//==!
//==! Function Parameters: pDP = input cDataPort
//==!                      sBscResultStr = BSC type string
//==! Return Value       : enumerated BSC function value
//==!=====================================================================!==
EnumBscResult CBsdlData::getEnBscResult(cDataPort *pDP,char *sBscResultStr)
{

  if (!strcmp(sBscResultStr,"z"))
  { return eBscResultZ;
  } // IF
  if (!strcmp(sBscResultStr,"pull0"))
  { return eBscResultPull0;
  } // IF
  if (!strcmp(sBscResultStr,"pull1"))
  { return eBscResultPull1;
  } // IF
  if (!strcmp(sBscResultStr,"weak0"))
  { return eBscResultWeak0;
  } // IF
  if (!strcmp(sBscResultStr,"weak1"))
  { return eBscResultWeak1;
  } // IF
  if (!strcmp(sBscResultStr,"keeper"))
  { return eBscResultKeeper;
  } // IF

  sprintf(sMsgBuf00,
          "[%s]: unknown disable result (will default to keeper)",
          sBscResultStr);
  jMessage(DISP_SHOW_MSG,sMsgBuf00,1,pDP);
  return eBscResultKeeper;

} // CBsdlData::getEnBscResult()
//==!=====================================================================!==
//==! getIsafeVal(): get safe value for cell.
//==!
//==! Function Parameters: pDP = input cDataPort
//==!                      sSafeValStr = safe value in ASCII
//==! Return Value       : -1 for 'X'
//==!                       0 for '0'
//==!                       1 for '1'
//==!=====================================================================!==
int CBsdlData::getIsafeVal(cDataPort *pDP,char *sSafeValStr)
{

  if (!xstrcmp(sSafeValStr,"x"))
  { return -1;
  } // IF
  if (!xstrcmp(sSafeValStr,"0"))
  { return 0;
  } // IF
  if (!xstrcmp(sSafeValStr,"1"))
  { return 1;
  } // IF

  sprintf(sMsgBuf00,
          "[%s]: unknown safe value; will default to X",
          sSafeValStr);
  jMessage(DISP_SHOW_MSG,sMsgBuf00,1,pDP);
  return -1;

} // CBsdlData::getIsafeVal()
//==!=====================================================================!==
//==! getBscInfo(): get BSC info from the cDataPort for BSC[lBscArrayIndex]
//==!
//==! Function Parameters: pDP = input cDataPort
//==!                      lBscArrayIndex = BSC cell to get info for
//==! Return Value       : none
//==!=====================================================================!==
void CBsdlData::getBscInfo(cDataPort *pDP,long lBscArrayIndex)
{ long lVar0,lVar1;
  char *pCh;

//==! Check if BSC[lBscArrayIndex] already defined
  if (pBscData[lBscArrayIndex]->lBscIndex != -1)
  { sprintf(sMsgBuf00,"BSC [%ld] already defined",lBscArrayIndex);
    jMessage(DISP_AND_EXIT,sMsgBuf00,1,NULL);
  } // IF

  findStrBsdl(pDP,"(",0);

//==! Get BSC cell type (BC_0 to BC_99 or AC_0 to AC_99)
  xGetData(pDP,1,eATTR_VAR,"Expecting BSC type");
  lVar1 = strlen(pDP->sData);
  if ((strncmp(pDP->sData,"bc_",3) && strncmp(pDP->sData,"ac_",3)) ||
      (lVar1 < 4))
  { jMessage(DISP_AND_EXIT,"Expecting BC_0 to BC_99 or AC_0 to AC_99",1,pDP);
    return;
  } // IF
  for (lVar0 = 3; lVar0 < lVar1; ++lVar0)
  { if (!isdigit(pDP->sData[lVar0]))
    { jMessage(DISP_AND_EXIT,"Use decimal numbers for bsc type",1,pDP);
      return;
    } // IF
  } // FOR lVar0 < lVar1
  pCh = &(pDP->sData[3]);
  lVar0 = atol(pCh);
  pBscData[lBscArrayIndex]->lBscIndex = lBscArrayIndex;

  pBscData[lBscArrayIndex]->lBscCellType = 0;
  if (!strncmp(pDP->sData,"ac_",3))
  { pBscData[lBscArrayIndex]->lBscCellType = 1;
  } // IF

  pBscData[lBscArrayIndex]->lBscCellNum = 0;
  if (isInRange(1,lVar0,0,99,"Invalid BSC number; will use 0",
                 DISP_SHOW_MSG) == 1)
  { pBscData[lBscArrayIndex]->lBscCellNum = lVar0;
  } // IF
  findStrBsdl(pDP,",",0);

//==! Get BSC name and search for it in the BSDL CPortDataInfo array
  for (sMsgBuf02[0] = '\0',xDataBsdl(pDP,XDATA_READ);
       strcmp(pDP->sData,",");
       xDataBsdl(pDP,XDATA_READ))
  { strcat(sMsgBuf02,pDP->sData);
  } // FOR sData != ','
  if (strcmp(sMsgBuf02,"*"))
  { for (lVar0 = 0; lVar0 < lNumPorts; ++lVar0)
    { if (pCPortDataInfo[lVar0] != NULL)
      { if (!strcmp(sMsgBuf02,pCPortDataInfo[lVar0]->csPortName))
        { break;
        } // IF
      } // IF
    } // FOR lVar0 < lNumPorts
    if (lVar0 >= lNumPorts)
    { sprintf(sMsgBuf00,"[%s] not in port list",sMsgBuf02);
      jMessage(DISP_AND_EXIT,sMsgBuf00,1,pDP);
      return;
    } // IF
    pBscData[lBscArrayIndex]->lPortNum = lVar0;
  } // IF

//==! Get BSC function
  xDataBsdl(pDP,XDATA_READ);
  pBscData[lBscArrayIndex]->enBscFunc = getEnBscType(pDP,pDP->sData);
  findStrBsdl(pDP,",",0);

//==! Get safe cell value
  xDataBsdl(pDP,XDATA_READ);
  pBscData[lBscArrayIndex]->iSafeVal = getIsafeVal(pDP,pDP->sData);

//==! Continue reading 3 more pieces of data (i.e., the control cell index,
//==! the disable value, and the disable result) if the BSC is an output or
//==! bidirectional cell.
  switch(pBscData[lBscArrayIndex]->enBscFunc)
  { case eBscOutput3:
    case eBscBidir:
      findStrBsdl(pDP,",",0);
//==! Get control cell index: this is index of control cell associated with
//==! an output or bidirectional pin.
      xGetData(pDP,1,eATTR_NUM,"Expected a control cell number");
      pBscData[lBscArrayIndex]->lCtrlCell = atol(pDP->sData);
      findStrBsdl(pDP,",",0);

//==! Get control cell disable value: this is value to load into the control
//==! BSC to disable the associated output or bidirectional pin.
      xGetData(pDP,1,eATTR_NUM,"Expected a control cell number");
      pBscData[lBscArrayIndex]->iDisVal = (int)atol(pDP->sData);
      if (isInRange(1,pBscData[lBscArrayIndex]->iDisVal,0,1,
          "DisableVal must be 0 or 1",DISP_AND_EXIT) == 0)
      { return;
      } // IF
      findStrBsdl(pDP,",",0);

//==! Get disable result: this is behaviour on pin when driver is disabled by
//==! its associated control cell.
      xDataBsdl(pDP,XDATA_READ);
      pBscData[lBscArrayIndex]->enDisResult =
             getEnBscResult(pDP,pDP->sData);
      break;
  } // SWITCH (pBscData[lBscArrayIndex]->enBscFunc)

  findStrBsdl(pDP,")",0);

  switch(pBscData[lBscArrayIndex]->enBscFunc)
  { case eBscInput:
      if (pCPortDataInfo[lVar0]->lInCell != -1)
      { sprintf(sMsgBuf00,
                "BSC [%ld]: Port [%s] already assigned INPUT BSC [%ld]",
                lBscArrayIndex,
                pCPortDataInfo[lVar0]->csPortName,
                pCPortDataInfo[lVar0]->lInCell);
        jMessage(DISP_AND_EXIT,sMsgBuf00,1,NULL);
      } // IF
      pCPortDataInfo[lVar0]->lInCell = lBscArrayIndex;
      break;
    case eBscClock:
      if (pCPortDataInfo[lVar0]->lInCell != -1)
      { sprintf(sMsgBuf00,
                "BSC [%ld]: Port [%s] already assigned CLOCK BSC [%ld]",
                lBscArrayIndex,
                pCPortDataInfo[lVar0]->csPortName,
                pCPortDataInfo[lVar0]->lInCell);
        jMessage(DISP_AND_EXIT,sMsgBuf00,1,NULL);
      } // IF
      pCPortDataInfo[lVar0]->lInCell = lBscArrayIndex;
      break;
    case eBscOutput2:
      if (pCPortDataInfo[lVar0]->lOutCell != -1)
      { sprintf(sMsgBuf00,
                "BSC [%ld]: Port [%s] already assigned OUTPUT2 BSC [%ld]",
                lBscArrayIndex,
                pCPortDataInfo[lVar0]->csPortName,
                pCPortDataInfo[lVar0]->lOutCell);
        jMessage(DISP_AND_EXIT,sMsgBuf00,1,NULL);
      } // IF
      pCPortDataInfo[lVar0]->lOutCell = lBscArrayIndex;
      break;
    case eBscOutput3:
      if (pCPortDataInfo[lVar0]->lOutCell != -1)
      { sprintf(sMsgBuf00,
                "BSC [%ld]: Port [%s] already assigned OUTPUT3 BSC [%ld]",
                lBscArrayIndex,
                pCPortDataInfo[lVar0]->csPortName,
                pCPortDataInfo[lVar0]->lOutCell);
        jMessage(DISP_AND_EXIT,sMsgBuf00,1,NULL);
      } // IF
      pCPortDataInfo[lVar0]->lOutCell = lBscArrayIndex;
      pCPortDataInfo[lVar0]->lCtrlCell =
                       pBscData[lBscArrayIndex]->lCtrlCell;
      break;
    case eBscBidir:
      if (pCPortDataInfo[lVar0]->lOutCell != -1)
      { sprintf(sMsgBuf00,
                "BSC [%ld]: Port [%s] already assigned BIDIR BSC [%ld]",
                lBscArrayIndex,
                pCPortDataInfo[lVar0]->csPortName,
                pCPortDataInfo[lVar0]->lOutCell);
        jMessage(DISP_AND_EXIT,sMsgBuf00,1,NULL);
      } // IF
      pCPortDataInfo[lVar0]->lOutCell = lBscArrayIndex;
      pCPortDataInfo[lVar0]->lCtrlCell =
                       pBscData[lBscArrayIndex]->lCtrlCell;
      break;

//==! Control BSC index assigned using output/bidir control cell index
    case eBscControl: 
      break;
    case eBscControlr:
      break;
    case eBscInternal:  //==! Internal BSCs are not associated with any port
      break;

    case eBscObserve_Only:
      if (pCPortDataInfo[lVar0]->lOutCell != -1)
      { sprintf(sMsgBuf00,
                "BSC [%ld]: Port [%s] already assigned OBSERVE_ONLY BSC [%ld]",
                lBscArrayIndex,
                pCPortDataInfo[lVar0]->csPortName,
                pCPortDataInfo[lVar0]->lOutCell);
        jMessage(DISP_AND_EXIT,sMsgBuf00,1,NULL);
      } // IF
      pCPortDataInfo[lVar0]->lOutCell = lBscArrayIndex;
      break;
    default:
      jMessage(DISP_AND_EXIT,"getBscInfoType",0,NULL);
      break;
  } // SWITCH

} // CBsdlData::getBscInfo()
//==!=====================================================================!==
//==! CBsdlData::checkRegOpcodeAssociation() will check if the opcode-data
//==! register association exists and if it is correct. If iMode = 1, then
//==! the opcode sOpcode is mandatory and should exist in the device's BSDL
//==! opcode list, else return. If the opcode exists, then check if it has
//==! already been assigned to the data register sDataReg. If sDataReg exists
//==! then manually associate the opcode to sDataReg (if not already done).
//==! If sDataReg does not exist, then manually create it and then assign
//==! sOpcode to it.
//==!
//==! For example: The BYPASS opcode does not have to be explicitly
//==! associated with the BYPASS register, so this function will make that
//==! connection explicitly.
//==!
//==! Function Parameters: iMode = see function description
//==!                      sOpcode = opcode string
//==!                      sDataReg = data reg for opcode
//==!                      pDP = cDataPort pointer to BSDL file
//==!                      plDataRegCnt = pointer to max number of data regs
//==!                      pm_CBsdlDataRegListHead = pointer to head of
//==!                           data register linked list
//==!                      pCBsdlDataRegTail = pointer to data register tail
//==!                           pointer
//==! Return Value       : none
//==!=====================================================================!==
void CBsdlData::checkRegOpcodeAssociation(int iMode,char *sOpcode,
         char *sDataReg,cDataPort *pDP,long *plDataRegCnt,
         CBsdlDataReg **pm_pCBsdlDataRegListHead,
         CBsdlDataReg **pCBsdlDataRegTail)
{ long lVar,lX,lY;
  SList *pSListVar;
  CBsdlDataReg *pCBsdlDataRegVar,*pCBsdlDataRegCurrentAssignment,
       *pCBsdlDataRegTo_sDataReg;

  lVar = getOpcodeIndex(1,sOpcode,pDP);
  if (lVar < 0)
  { if (iMode == 1)
    { sprintf(sMsgOnly,"Mandatory opcode [%s] undefined",sOpcode);
      jMessage(DISP_SHOW_MSG,sMsgOnly,1,pDP);
    } // IF
    return;
  } // IF opcode index not found

//==! Check if sDataReg already defined and also look for the data reg (if
//==! any) currently associated with sOpcode. Note that when this function is
//==! called the data register informaton is still in singly-linked list
//==! format.
  for (lX = 0,lY = -1,pCBsdlDataRegCurrentAssignment = NULL,
           pCBsdlDataRegTo_sDataReg = NULL,
           pCBsdlDataRegVar = m_pCBsdlDataRegListHead;
        pCBsdlDataRegVar != NULL;++lX)
  { if (!strcmp(pCBsdlDataRegVar->csRegName,sDataReg))
    { if (lY != -1)
      { jMessage(DISP_SHOW_MSG,
                 "DRegIndex multiply defined",0,NULL); // should never happen
      } // IF
      lY = lX;  // means sDataReg already defined
      pCBsdlDataRegTo_sDataReg = pCBsdlDataRegVar;
    } // IF csRegName = sDataReg
    if (pCBsdlOpInfo[lVar]->lDataRegIndex == lX)
    { if (pCBsdlDataRegCurrentAssignment != NULL)
      { jMessage(DISP_AND_EXIT,"DRegCurrent multiply defined",0,NULL);
      } // IF
//==! current opcode-to-data reg assignment
      pCBsdlDataRegCurrentAssignment = pCBsdlDataRegVar;
    } // IF
    pCBsdlDataRegVar = pCBsdlDataRegVar->pNextCBsdlDataReg;
  } // FOR pCBsdlDataRegVar != NULL

//==! If opcode already assigned to a data register, then confirm that it is
//==! to the correct data register (e.g., bypass to bypass, idcode to
//==! device_id, extest to boundary, etc).
  if (pCBsdlOpInfo[lVar]->lDataRegIndex >= 0)
  { isPtrNull(pCBsdlDataRegCurrentAssignment,"checkRegOpcode");
    if (pCBsdlOpInfo[lVar]->lDataRegIndex != lY)
    { sprintf(sMsgOnly,"Error: Cannot match opcode [%s] to reg [%s]",
              sOpcode,sDataReg);
      jMessage(DISP_SHOW_MSG,sMsgOnly,0,NULL);
      sprintf(sMsgOnly,"       - [%s] incorrectly matched to data reg [%s]",
              sOpcode,pCBsdlDataRegCurrentAssignment->csRegName);
      jMessage(DISP_AND_EXIT,sMsgOnly,0,NULL);
    } // IF opcode not matched to sDataReg
    return;
  } // IF pCBsdlOpInfo[lVar]->lDataRegIndex >= 0

//==! If sDataReg already defined then add sOpcode to the list of opcodes
//==! that refer to this data register.
  if (lY != -1)  
  { pCBsdlOpInfo[lVar]->lDataRegIndex = lY;
    for (pSListVar = pCBsdlDataRegTo_sDataReg->pOpList;
         pSListVar != NULL; pSListVar = pSListVar->pNextSList)
    {
//==! The opcode indexed by lVar should not be in the list of opcodes that
//==! use this data register.
      if (pSListVar->lVal == lVar)
      { jMessage(DISP_AND_EXIT,"Dreg-Opc",0,NULL);
      } // IF
    } // FOR pSListVar != NULL
    pSListVar = new SList;
    pSListVar->lVal = lVar;   // add opcode index to sDataReg opcode list
    pSListVar->pNextSList = pCBsdlDataRegTo_sDataReg->pOpList;
    pCBsdlDataRegTo_sDataReg->pOpList = pSListVar;
    return;
  } // IF lY != -1

//==! Add the sDataReg to the list of data registers and assign sOpcode to
//==! the list of opcodes associated with sDataReg.
  ++(*plDataRegCnt);  // increment to next data reg index
  pCBsdlDataRegVar = new CBsdlDataReg;    // create new data reg object
  copyStr(&(pCBsdlDataRegVar->csRegName),sDataReg);
  if (*pm_pCBsdlDataRegListHead == NULL)  // add to data reg list
  { *pm_pCBsdlDataRegListHead = pCBsdlDataRegVar;
  } // IF
  else
  { (*pCBsdlDataRegTail)->pNextCBsdlDataReg = pCBsdlDataRegVar;
  } // ELSE
  *pCBsdlDataRegTail = pCBsdlDataRegVar;
  pCBsdlOpInfo[lVar]->lDataRegIndex = *plDataRegCnt; // assign data reg index
  pSListVar = new SList;
  pSListVar->lVal = lVar;   // add opcode index to sDataReg opcode list
  pSListVar->pNextSList = pCBsdlDataRegVar->pOpList;
  pCBsdlDataRegVar->pOpList = pSListVar;

} // CBsdlData::checkRegOpcodeAssociation()
//==!=====================================================================!==
//==! lAssignDataRegToOpcode(): assigns data reg index to the opcode sOpcode
//==! in pCBsdlOpInfo[] that uses this data reg.
//==!
//==! Function Parameters: lVal = data register index into pCBsdlDataRegInfo[]
//==!                      sOpcode = opcode to assign data reg index to
//==!                      pDP = data port being used
//==! Return Value       : index of opcode in pCBsdlOpInfo[]
//==!=====================================================================!==
long CBsdlData::lAssignDataRegToOpcode(long lVal,char *sOpcode,cDataPort *pDP)
{ long lX,lY;
  CBsdlDataReg *pCBsdlDataRegVar;

  if (pCBsdlOpInfo == NULL)
  { jMessage(DISP_AND_EXIT,"No opcodes defined",pDP == NULL ? 0:1,pDP);
  } // IF

  for (lX = 0; lX < lNumOpcodes; ++lX)
  { if (!strcmp(pCBsdlOpInfo[lX]->csOpcodeName,sOpcode))
    { if (pCBsdlOpInfo[lX]->lDataRegIndex >= 0)
      {
//==! Opcode has already been assigned to a data reg, then at this point
//==! the REGISTER_ACCESS information is still in singly-linked list format,
//==! which is why we use a FOR loop to step through the linked list to find
//==! the data register name.
        for (lY = 0,pCBsdlDataRegVar = m_pCBsdlDataRegListHead;
             lY < pCBsdlOpInfo[lX]->lDataRegIndex; ++lY)
        { if (pCBsdlDataRegVar == NULL)
          { break;
          } // IF
          pCBsdlDataRegVar = pCBsdlDataRegVar->pNextCBsdlDataReg;
        } // FOR lY < pCBsdlOpInfo[lX]->lDataRegIndex
        isPtrNull((void *)pCBsdlDataRegVar,"RegChk");  // should never happen

        sprintf(sMsgOnly,
                "Warning: [%s] already associated with data register [%s]",
                sOpcode,
                pCBsdlDataRegVar->csRegName);
        jMessage(DISP_SHOW_MSG,sMsgOnly,1,pDP);
      } // IF
      else
      { pCBsdlOpInfo[lX]->lDataRegIndex = lVal;
      } // ELSE
      return lX;
    } // IF
  } // FOR lX < lNumOpcodes
  sprintf(sMsgOnly,"Opcode [%s] not defined",sOpcode);
  jMessage(DISP_AND_EXIT,sMsgOnly,pDP == NULL ? 0:1,pDP);
  return -1;

} // CBsdl::lAssignDataRegToOpcode()
//==!=====================================================================!==
//==! Create a lDataWidth-wide unsigned long array for holding hex constants
//==! or variable values.
//==!
//==! Ex: "FE83290123" will be stored in a 2 element array with the
//==! following values in locations 0 and 1:
//==! [0] = 1000 0011 0010 1001 0000 0001 0010 0011   (hex 8329 0123)
//==! [1] = 0000 0000 0000 0000 0000 0000 1111 1110   (hex 0000 00FE)
//==!
//==! Function Parameters: lDataWidth = maximum data width
//==! Return Value       : pointer to a data expression
//==!=====================================================================!==
unsigned long *createFieldSpace(long lDataWidth)
{ unsigned long *pulVar;
  long lWordMax,lCnt;

  lWordMax = (lDataWidth+31)/32;
  pulVar = new unsigned long [lWordMax];
  for (lCnt = 0; lCnt < lWordMax; ++lCnt)
  { pulVar[lCnt] = 0;
  } // FOR lCnt < lWordMax
  return pulVar;

} // createFieldSpace()
//==!=====================================================================!==
//==! Assign the hex value into pulField which points to an array of unsigned
//==! longs. Function also checks if the hex value can be represented in the
//==! width allocated for it.
//==!
//==! Ex: "FE83290123" will be converted into a 2 element array with the
//==! following elements in locations 0 and 1:
//==! [0] = 1000 0011 0010 1001 0000 0001 0010 0011   (hex 8329 0123)
//==! [1] = 0000 0000 0000 0000 0000 0000 1111 1110   (hex 0000 00FE)
//==!
//==! Function Parameters: pDP = input cDataPort containing hex string
//==!                      lDataWidth = maximum data width
//==!                      pulField = pointer to array of unsigned longs
//==! Return Value       : none
//==!=====================================================================!==
void assignFixedValue(cDataPort *pDP,long lDataWidth,unsigned long *pulField)
{ unsigned long ulValue;
  long lWordMax,lVar,lCnt,lVar0;

  lWordMax = (lDataWidth+31)/32;

  for (lCnt = strlen(pDP->sData)-1,lVar = 0,lVar0 = 0;
       lCnt >= 0; --lCnt)
  { if (!isxdigit(pDP->sData[lCnt]))
    { jMessage(DISP_AND_EXIT,"hex string must have hex chars only",1,pDP);
    } // IF
    ulValue = chrtohex(pDP->sData[lCnt]);
    if (lVar >= lWordMax)
    { if (ulValue == 0)
      { continue;
      } // IF
      else
      { sprintf(sMsgOnly,"hex [%s] cannot fit in [%ld] bit field",
                pDP->sData,lDataWidth);
        jMessage(DISP_AND_EXIT,sMsgOnly,1,pDP);
      } // ELSE
    } // IF

//==! Check if the last hex value can be represented in the last few bits
//==! because lDataWidth could be less than 4 bits (e.g., an 11 bit field
//==! cannot represent hex value "EFF")
    if (lVar == lWordMax-1)
    { if (lDataWidth/4 == lVar0)
      { if ((unsigned long)(pow((double)2,(double(lDataWidth%4)))-1) < ulValue)
        { sprintf(sMsgOnly,
          "MSB hex val [%X] cannot fit in left [%ld] bits of [%ld] bit field",
          ulValue,lDataWidth%4,lDataWidth);
          jMessage(DISP_AND_EXIT,sMsgOnly,1,pDP);
        } // IF
      } // IF
    } // IF

    if (lVar0 == 0)
    { pulField[lVar] = ulValue;
    } // IF
    else
    { pulField[lVar] |= (ulValue << 4*lVar0);
    } // ELSE
    if (++lVar0 >= 8)
    { lVar0 = 0;
      ++lVar;
    } // IF
  } // FOR lCnt >= 0

  for (++lVar; lVar < lWordMax; ++lVar)  //==! Reset remainder of array to 0
  { pulField[lVar] = 0;
  } // FOR lVar < lWordMax

} // assignFixedValue()
//==!=====================================================================!==
//==! Check if the string is a constant containing hex digits only.
//==!
//==! Function Parameters: pDP = input cDataPort
//==!                      lDataWidth = maximum data width
//==! Return Value       : pointer to a data expression
//==!=====================================================================!==
int checkHexConstant(char *sStr)
{ long lVar0,lVar1;

  for (lVar0 = 0,lVar1 = strlen(sStr); lVar0 < lVar1; ++lVar0)
  { if (!isxdigit(sStr[lVar0]))
    { return 0;
    } // IF isxdigit()
  } // FOR lVar0 < lVar1
  return 1;

} // checkHexConstant()
//==!=====================================================================!==
//==! Check if variable has been initialized already.
//==!
//==! Function Parameters: csVarName = variable name
//==!                      lFlag = 1 if variable encountered for the first
//==!                                timep
//==!                      DP = input cDataPort
//==! Return Value       : none
//==!=====================================================================!==
void checkVarInit(char *csVarName,long lFlag,cDataPort *pDP)
{
  if (lFlag == 1)
  { sprintf(sMsgOnly,"Variable [%s] must be initialized",csVarName);
    jMessage(DISP_AND_EXIT,sMsgOnly,1,pDP);
  } // IF

} // checkVarInit()
//==!=====================================================================!==
//==! Get a data expression.
//==!
//==! Function Parameters: pDP = input cDataPort
//==!                      lDataWidth = maximum data width
//==! Return Value       : pointer to a data expression
//==!=====================================================================!==
CxDataExpr *CIscFlow::getDataExpr(cDataPort *pDP,long lDataWidth)
{ CxDataExpr *pCDEvar;
  long lVar0,lVar1;
  CxDataVar *pCDVvar;

  if (pDP->sData[0] == '*')
  { return NULL;
  } // IF

  pCDVvar = NULL;
  pCDEvar = new CxDataExpr;
  pCDEvar->m_lDE_len = lDataWidth;
  pCDEvar->m_pulDataField = createFieldSpace(lDataWidth);

  xDataBsdl(pDP,XDATA_READ);
  if (checkHexConstant(pDP->sData) == 1)
  { pCDEvar->m_iFieldType = dHEX_CONSTANT;
    assignFixedValue(pDP,lDataWidth,pCDEvar->m_pulDataField);
    xDataBsdl(pDP,XDATA_READ);
    if (pDP->sData[0] != '?' && pDP->sData[0] != '!')
    { pDP->iParserGetData = 0;
      return pCDEvar;
    } // IF
  } // IF checkHexConstant(sData)

  if (!strcmp(pDP->sData,"!" ) ||
      !strcmp(pDP->sData,"?" ))
  { if (pCDEvar->m_iFieldType != dHEX_CONSTANT)
    { pCDEvar->m_iFieldType = dINP_SPECIFIER;
    } // IF
    if (pDP->sData[0] == '?')
    { xDataBsdl(pDP,XDATA_READ);
      if (pDP->sData[0] == '!')
      { pCDEvar->m_cOperator = dc_IO_OP;
      } // IF
      else
      { pCDEvar->m_cOperator = dc_INPUT_OP;
        pDP->iParserGetData = 0;
      } // ELSE
    } // IF pDP->sData[0] == '?'
    else
    { pCDEvar->m_cOperator = dc_OUTPUT_OP;
    } // ELSE
    return pCDEvar;
  } // IF input or IO specifier

  if (!strcmp(pDP->sData,"$"))
  { xDataBsdl(pDP,XDATA_READ);
    if (pDP->eDataAttr != eATTR_VAR)
    { jMessage(DISP_AND_EXIT,"variable name expected",1,pDP);
    } // IF eDataAttr != eATTR_VAR
    pCDEvar->m_iFieldType = d_VAR_EXPR;

    for (pCDVvar = m_pCDVlist,lVar0 = 0,lVar1 = 0;
         pCDVvar != NULL;
         pCDVvar = pCDVvar->m_pNextCxDataVar,++lVar1)
    { if (!strcmp(pCDVvar->m_pcVarName,pDP->sData))
      { if (pCDVvar->m_lVarWidth != lDataWidth)
        { sprintf(sMsgOnly,
              "[%ld] does not match previously defined field width [%ld]",
              pCDVvar->m_lVarWidth,lDataWidth);
          jMessage(DISP_AND_EXIT,sMsgOnly,1,pDP);
        } // IF
        pCDEvar->m_lVarIndex = lVar1; //==! Store variable index
        break;
      } // IF
    } // FOR pCDVvar != NULL
    if (pCDVvar == NULL)
    { pCDVvar = new CxDataVar;
      copyStr(&(pCDVvar->m_pcVarName),pDP->sData);
      pCDVvar->m_lVarWidth = lDataWidth;
      pCDVvar->m_ulVarValue = createFieldSpace(lDataWidth);
      if (m_pCDVlist == NULL)
      { m_pCDVlist = pCDVvar;
      } // IF
      else
      { m_pCDVlistTail->m_pNextCxDataVar = pCDVvar;
      } // ELSE
      m_pCDVlistTail = pCDVvar;

//==! lVar0 used here to indicate first instance/definition of variable
      lVar0 = 1;

      pCDEvar->m_lVarIndex = m_lCxDataVarArraySize; //==! Store variable index
      ++m_lCxDataVarArraySize;
    } // IF

    xDataBsdl(pDP,XDATA_READ);

    if (!strcmp(pDP->sData,"wait") || pDP->sData[0] == ',' ||
        !strcmp(pDP->sData,":"))
    { checkVarInit(pCDVvar->m_pcVarName,lVar0,pDP);
      pDP->iParserGetData = 0;
      pCDEvar->m_cOperator = dc_GET_VAR;
      return pCDEvar;
    } // IF sData == wait or ',' or ':'

    if (pDP->sData[0] == '=')
    { xDataBsdl(pDP,XDATA_READ);
      if (pCDVvar->m_ulInitValue != NULL)
      { sprintf(sMsgOnly,"Variable [%s] already initialized",
                pCDVvar->m_pcVarName);
        jMessage(DISP_AND_EXIT,sMsgOnly,1,pDP);
      } // IF
      pCDVvar->m_ulInitValue = createFieldSpace(lDataWidth);
      assignFixedValue(pDP,lDataWidth,pCDVvar->m_ulInitValue);
      pCDEvar->m_cOperator = dc_ASSIGN_OP;
      return pCDEvar;
    } // IF sData == '='

    if (!strcmp(pDP->sData,"!" ) ||
        !strcmp(pDP->sData,"?" ))
    { if (pDP->sData[0] == '?')
      { xDataBsdl(pDP,XDATA_READ);
        if (pDP->sData[0] == '!')
        { pCDEvar->m_cOperator = dc_IO_OP;
        } // IF
        else
        { pCDEvar->m_cOperator = dc_INPUT_OP;
          pDP->iParserGetData = 0;
        } // ELSE
      } // IF
      else
      { pCDEvar->m_cOperator = dc_OUTPUT_OP;
      } // ELSE
      return pCDEvar;
    } // IF input/output/IO specifier

    switch (pDP->sData[0])
    { case '~':
        checkVarInit(pCDVvar->m_pcVarName,lVar0,pDP);
        pCDEvar->m_cOperator = dc_COMPLEMENT_OP;
        return pCDEvar;
      case '+':
      case '-':
      case '<':
      case '>':
        checkVarInit(pCDVvar->m_pcVarName,lVar0,pDP);
        if (pDP->sData[0] == '+' || pDP->sData[0] == '-')
        { pCDEvar->m_cOperator = pDP->sData[0] == '+' ? dc_ADD_OP : dc_SUB_OP;
        } // IF
        else
        { lVar0 = 0;
          if (pDP->sData[0] == '>')
          { lVar0 = 1;
          } // IF
          xDataBsdl(pDP,XDATA_READ);
          if ((lVar0 == 0 && pDP->sData[0] != '<') ||
              (lVar0 == 1 && pDP->sData[0] != '>'))
          { jMessage(DISP_AND_EXIT,"Expected << or >> operator",1,pDP);
          } // IF
          pCDEvar->m_cOperator = dc_RIGHT_SHIFT_OP;
          if (lVar0 == 0)
          { pCDEvar->m_cOperator = dc_LEFT_SHIFT_OP;
          } // IF
        } // ELSE
        pCDEvar->m_ulOperand = (unsigned long)lReadDecimalBsdl(pDP,1);
        switch(pCDEvar->m_cOperator)
        { case dc_RIGHT_SHIFT_OP:
          case dc_LEFT_SHIFT_OP:
            if (pCDEvar->m_ulOperand > (unsigned long)lDataWidth - 1)
            { sprintf(sMsgOnly,
                 "Left/right shift count exceeds field width [%ld]-1",
                 lDataWidth);
              jMessage(DISP_AND_EXIT,sMsgOnly,1,pDP);
            } // IF
            break;
        } // SWITCH (m_cOperator)
        return pCDEvar;
      default:
        jMessage(DISP_AND_EXIT,"Unknown operator",1,pDP);
    } // SWITCH (sData[0])

  } // IF sData == 1532 variable

  jMessage(DISP_AND_EXIT,"Unrecognized symbol",1,pDP);
  return NULL;

} // CIscFlow::getDataExpr()
//==!=====================================================================!==
//==! Read the update field list.
//==!
//==! Function Parameters: pDP = input cDataPort
//==!                      pUF = pointer to CxUpdateField pointer
//==!                      lGetDataFlag = pointer to variable that gets
//==!                           incremented if the data expression for the
//==!                           flow requires data from a source
//==!                      iHasOutputOp = pointer to variable that gets set
//==!                           to 1 if field uses output or IO operator
//==! Return Value       : width of update field (i.e., total width of bits
//==!                      specified in update field list)
//==!=====================================================================!==
long CIscFlow::getUpdateFieldList(cDataPort *pDP,CxUpdateField **pUF,
                    long *lGetDataFlag,int *iHasOutputOp)
{ CxUpdateField *pCUFvar,*pCUFtail;
  long lTotWidth;

  if (*pUF != NULL)
  { jMessage(DISP_AND_EXIT,"gUFL",1,pDP);
  } // IF

  lTotWidth = 0;
  pCUFtail = NULL;

  xDataBsdl(pDP,XDATA_READ);

  while (strcmp(pDP->sData,"wait"))
  { pDP->iParserGetData = 0;
    pCUFvar = new CxUpdateField;
    pCUFvar->m_lFieldWidth = lReadDecimalBsdl(pDP,0);
    lTotWidth += pCUFvar->m_lFieldWidth;
    xDataBsdl(pDP,XDATA_READ);
    if (pDP->sData[0] == ':')
    { pCUFvar->m_pCDE_data = this->getDataExpr(pDP,pCUFvar->m_lFieldWidth);
      if (pCUFvar->m_pCDE_data != NULL)
      { switch(pCUFvar->m_pCDE_data->m_cOperator)
        { case dc_INPUT_OP:
          case dc_IO_OP:
          { ++(*lGetDataFlag);
            if (pCUFvar->m_pCDE_data->m_cOperator == dc_IO_OP)
            { *iHasOutputOp = 1;
            } // IF
            break;
          } // IF
        } // SWITCH (pCUFvar->m_pCDE_data->m_cOperator)
      } // IF
    } // IF sData == ':'

    if (*pUF == NULL)
    { *pUF = pCUFvar;
    } // IF
    else
    { pCUFtail->m_pNextCxUpdateField = pCUFvar;
    } // ELSE
    pCUFtail = pCUFvar;

    xDataBsdl(pDP,XDATA_READ);
    if (pDP->sData[0] != ',' && strcmp(pDP->sData,"wait"))
    { jMessage(DISP_AND_EXIT,"',' or wait expected",1,pDP);
    } // IF
    if (pDP->sData[0] == ',')
    { xDataBsdl(pDP,XDATA_READ);
    } // IF
  } // WHILE sData != wait

  pDP->iParserGetData = 0;

  return lTotWidth;

} // CIscFlow::getUpdateFieldList()
//==!=====================================================================!==
//==! Get flow activity wait specification.
//==!
//==! Function Parameters: pDP = input cDataPort
//==!                      pCFW = pointer to flow wait object pointer to
//==!                           return wait specification
//==!                      pCBD = pointer to CBsdlData object
//==! Return Value       : none
//==!=====================================================================!==
void CIscFlow::getFlowActWaitSpec(cDataPort *pDP,CFlowWait **pCFW,
                   CBsdlData *pCBD)
{ int iX;
  float fNumVar;
  long lNumVar;

  findStrBsdl(pDP,"wait",0);

  *pCFW = new CFlowWait;

  for (iX = 0; iX < 2; ++iX)
  { xDataBsdl(pDP,XDATA_READ);
    if (pDP->eDataAttr == eATTR_VAR)
    { if (pCBD->m_csTapClkName == NULL)
      { jMessage(DISP_AND_EXIT,"Missing TCK port ID",1,pDP);
      } // IF
      else
      { if (strcmp(pDP->sData,pCBD->m_csTapClkName))
        { sprintf(sMsgOnly,"Must use TCK port name [%s]",pCBD->m_csTapClkName);
          jMessage(DISP_AND_EXIT,sMsgOnly,1,pDP);
        } // IF
        if (iX == 0)
        { (*pCFW)->m_lMinFlowWaitCycles = lReadDecimalBsdl(pDP,0);
        } // IF
        else
        { (*pCFW)->m_lMaxFlowWaitCycles = lReadDecimalBsdl(pDP,0);
        } // ELSE
      } // ELSE

      xDataBsdl(pDP,XDATA_READ);
      if (pDP->sData[0] != ',')  //==! NOTE_0005
      { pDP->iParserGetData = 0;
        return;
      } // IF
      xDataBsdl(pDP,XDATA_READ);
    } // IF eDataAttr == eATTR_VAR

//==! Note that code execution flow can come to this point if sData[0] != ','
//==! test is not true (see NOTE_0005).

    if (pDP->eDataAttr == eATTR_NUM)
    { pDP->iParserGetData = 0;
      switch(getFloat(pDP,&fNumVar,&lNumVar))
      { case GF_FLOAT:
          if (iX == 0)
          { (*pCFW)->m_fMinFlowWaitTimeSec = fNumVar;
          } // IF
          else
          { (*pCFW)->m_fMaxFlowWaitTimeSec = fNumVar;
          } // ELSE
          break;
        case GF_LONG:
          if (iX == 0)
          { (*pCFW)->m_fMinFlowWaitTimeSec = (float)lNumVar;
          } // IF
          else
          { (*pCFW)->m_fMaxFlowWaitTimeSec = (float)lNumVar;
          } // ELSE
          break;
        default:
          jMessage(DISP_AND_EXIT,"Unknown wait spec format",1,pDP);
          break;
      } // SWITCH (getFloat())
    } // IF eDataAttr == eATTR_NUM
    else
    { jMessage(DISP_AND_EXIT,"Specify cycle, time, or both",1,pDP);
    } // ELSE

    if (iX == 0)
    { xDataBsdl(pDP,XDATA_READ);
      if (strcmp(pDP->sData,"min"))
      { pDP->iParserGetData = 0;
        return;
      } // IF
      xDataBsdl(pDP,XDATA_READ);
      if (pDP->sData[0] != ',')
      { pDP->iParserGetData = 0;
        return;
      } // IF
    } // IF
    else
    { findStrBsdl(pDP,"max",0);
      return;
    } // ELSE

  } // FOR iX < 2

} // CIscFlow::getFlowActWaitSpec()
//==!=====================================================================!==
//==! Read the capture field list.
//==!
//==! Function Parameters: pDP = input cDataPort
//==!                      pCF = pointer to CxCaptureField pointer
//==!                      lGetDataFlag = pointer to variable that gets
//==!                           incremented if the data expression for the
//==!                           flow requires data from a source
//==!                      iHasOutputOp = pointer to variable that gets set
//==!                           to 1 if field uses output or IO operator
//==! Return Value       : width of capture field (i.e., total width of bits
//==!                      specified in capture field list)
//==!=====================================================================!==
long CIscFlow::getCaptureFieldList(cDataPort *pDP,CxCaptureField **pCF,
                    long *lGetDataFlag,int *iHasOutputOp)
{ CxCaptureField *pCFvar,*pCFtail;
  long lTotWidth;

  if (*pCF != NULL)
  { jMessage(DISP_AND_EXIT,"gCFL",1,pDP);
  } // IF

  lTotWidth = 0;
  pCFtail = NULL;

  xDataBsdl(pDP,XDATA_READ);

  while (strcmp(pDP->sData,")"))
  { pDP->iParserGetData = 0;
    pCFvar = new CxCaptureField;
    pCFvar->m_lFieldWidth = lReadDecimalBsdl(pDP,0);
    lTotWidth += pCFvar->m_lFieldWidth;
    findStrBsdl(pDP,":",0);

    pCFvar->m_pCDE_CapturedData = new CxDataExpr;
    pCFvar->m_pCDE_CapturedData->m_lDE_len = pCFvar->m_lFieldWidth;
    pCFvar->m_pCDE_CapturedData->m_pulDataField
          = createFieldSpace(pCFvar->m_lFieldWidth);

    xDataBsdl(pDP,XDATA_READ);  //==! Get expected capture data specification
    if (!strcmp(pDP->sData,"!"))
    { pCFvar->m_cExpDataOutOp = dc_OUTPUT_OP;
      *iHasOutputOp = 1;
      xDataBsdl(pDP,XDATA_READ);
    } // IF

    if (pDP->sData[0] != ':' && pDP->sData[0] != ')' && pDP->sData[0] != ',')
    { pDP->iParserGetData = 0;
      pCFvar->m_pCDE_expdata = this->getDataExpr(pDP,pCFvar->m_lFieldWidth);
      if (pCFvar->m_pCDE_expdata != NULL)
      { switch(pCFvar->m_pCDE_expdata->m_cOperator)
        { case dc_INPUT_OP:
          case dc_IO_OP:
          { ++(*lGetDataFlag);
            if (pCFvar->m_pCDE_expdata->m_cOperator == dc_IO_OP)
            { *iHasOutputOp = 1;
            } // IF
            break;
          } // IF
        } // SWITCH (pCUFvar->m_pCDE_expdata->m_cOperator)
      } // IF

      xDataBsdl(pDP,XDATA_READ);
      if (pDP->sData[0] == '*')   //==! Get compare mask data specification
      { xDataBsdl(pDP,XDATA_READ);
        if (!strcmp(pDP->sData,"!"))
        { pCFvar->m_cMaskDataOutOp = dc_OUTPUT_OP;
          *iHasOutputOp = 1;
        } // IF
        else
        { pDP->iParserGetData = 0;
        } // ELSE
        pCFvar->m_pCDE_maskdata = this->getDataExpr(pDP,pCFvar->m_lFieldWidth);
        if (pCFvar->m_pCDE_maskdata != NULL)
        { switch(pCFvar->m_pCDE_maskdata->m_cOperator)
          { case dc_INPUT_OP:
            case dc_IO_OP:
            { ++(*lGetDataFlag);
              if (pCFvar->m_pCDE_maskdata->m_cOperator == dc_IO_OP)
              { *iHasOutputOp = 1;
              } // IF
              break;
            } // IF
          } // SWITCH (pCUFvar->m_pCDE_maskdata->m_cOperator)
        } // IF
        xDataBsdl(pDP,XDATA_READ);
      } // IF
    } // IF pDP->sData[0] != ':'

    while (pDP->sData[0] == ':' && pDP->sData[0] != ',')
    { xDataBsdl(pDP,XDATA_READ);
      if (!strcmp(pDP->sData,"crc") || !strcmp(pDP->sData,"ost"))
      { if (pDP->sData[0] == 'c')
        { if (pCFvar->m_cCrcTag != NULL)
          { jMessage(DISP_SHOW_MSG,"CRC already specified",1,pDP);
          } // IF
          pCFvar->m_cCrcTag = '1';
        } // IF
        else
        { if (pCFvar->m_cOstTag != NULL)
          { jMessage(DISP_SHOW_MSG,"OST already specified",1,pDP);
          } // IF
          pCFvar->m_cOstTag = '1';
        } // ELSE
        xDataBsdl(pDP,XDATA_READ);
      } // IF
      else
      { jMessage(DISP_AND_EXIT,"CRC or OST expected",1,pDP);
      } // ELSE
    } // WHILE sData == ':'

    pDP->iParserGetData = 0;

    if (*pCF == NULL)
    { *pCF = pCFvar;
    } // IF
    else
    { pCFtail->m_pNextCxCaptureField = pCFvar;
    } // ELSE
    pCFtail = pCFvar;

    xDataBsdl(pDP,XDATA_READ);
    if (pDP->sData[0] == ',')
    { xDataBsdl(pDP,XDATA_READ);
    } // IF
    else
    { if (pDP->sData[0] != ')')
      { jMessage(DISP_AND_EXIT,"',' or ')' expected",1,pDP);
      } // IF
    } // ELSE

  } // WHILE sData != ')'

  return lTotWidth;

} // CIscFlow::getCaptureFieldList()
//==!=====================================================================!==
//==! Get the width of the data register used by the opcode.
//==!
//==! Function Parameters: pCBD = BSDL data pointer
//==!                      lOpcodeIndex = opcode index
//==! Return Value       : data register width
//==!=====================================================================!==
long getDataRegWidth(CBsdlData *pCBD,long lOpcodeIndex)
{

  if (pCBD->pCBsdlOpInfo[lOpcodeIndex]->lDataRegIndex == -1)
  { sprintf(sMsgOnly,"No data register for opcode [%s]",
            pCBD->pCBsdlOpInfo[lOpcodeIndex]->csOpcodeName);
    jMessage(DISP_AND_EXIT,sMsgOnly,0,NULL);
  } // IF

  return pCBD->pCBsdlDataRegInfo[
                 pCBD->pCBsdlOpInfo[lOpcodeIndex]->lDataRegIndex
                                ]->lRegLength;

} // getDataRegWidth()
//==!=====================================================================!==
//==! Check that the field width matches the one required by the opcode's
//==! register.
//==!
//==! Function Parameters: pDP = pointer to data port
//==!                      sMsg = message identifying field name
//==!                      pCBD = BSDL data pointer
//==!                      lFieldWidth = field width to compare
//==!                      lOpcodeIndex = opcode index
//==! Return Value       : number of bits in data register for this opcode
//==!                      indexed by lOpcodeIndex
//==!=====================================================================!==
long checkFieldWidth(cDataPort *pDP,char *sMsg,CBsdlData *pCBD,
               long lFieldWidth,long lOpcodeIndex)
{ long lVal;

  if (pCBD->pCBsdlOpInfo[lOpcodeIndex]->lDataRegIndex == -1)
  { sprintf(sMsgOnly,
            "Data reg for %s not defined in REGISTER_ACCESS section",
            pCBD->pCBsdlOpInfo[lOpcodeIndex]->csOpcodeName);
    jMessage(DISP_AND_EXIT,sMsgOnly,1,pDP);
  } // IF

  lVal = pCBD->pCBsdlDataRegInfo[
                 pCBD->pCBsdlOpInfo[lOpcodeIndex]->lDataRegIndex
                                ]->lRegLength;

  if (lFieldWidth > 0)
  { if (lFieldWidth != lVal)
    { sprintf(sMsgOnly,
        "%s field width [%ld] != data reg [%s] width [%ld]",sMsg,lFieldWidth,
        pCBD->pCBsdlDataRegInfo[
              pCBD->pCBsdlOpInfo[lOpcodeIndex]
                  ->lDataRegIndex
                               ]->csRegName,
        lVal);
      jMessage(DISP_AND_EXIT,sMsgOnly,1,pDP);
    } // IF lFieldWidth != lVal
  } // IF lFieldWidth > 0

  return lVal;

} // checkFieldWidth()
//==!=====================================================================!==
//==! readFlowActivityList(): read ISC_FLOW activities
//==!
//==! Function Parameters: pDP = input cDataPort
//==!                      iMode = 1: activity list is part of a loop
//==!                              0: regular activity list
//==!                      pCBD = pointer to BSDL data object
//==!                      pCFAhead = pointer head of flow activity list
//==!                             pointer
//==!                      iHasOutputOp = pointer to variable to set to 1 if
//==!                             activity uses output or IO operator
//==! Return Value       : max register size used by this flow
//==!=====================================================================!==
long CIscFlow::readFlowActivityList(cDataPort *pDP,int iMode,
               CBsdlData *pCBD,CFlowAct **pCFAhead,int *iHasOutputOp)
{ CFlowAct *pOneFlowAct,*pLastFlowAct;
  long lFlowMaxRegSize,lOpcodeDataRegSize;

  *pCFAhead = NULL;
  pOneFlowAct = NULL;
  pLastFlowAct = NULL;

  lFlowMaxRegSize = 0;

  while (1)
  { xDataBsdl(pDP,XDATA_READ);
    if (!strcmp(pDP->sData,")"))
    { if (iMode == 0)
      { jMessage(DISP_AND_EXIT,"activity list is not part of a loop",1,pDP);
      } // IF
      break;
    } // IF
    if (!strcmp(pDP->sData,",") || !strcmp(pDP->sData,";") ||
        !strcmp(pDP->sData,"initialize") ||
        !strcmp(pDP->sData,"repeat"    ) ||
        !strcmp(pDP->sData,"terminate" )
       )
    { if (iMode == 1)
      { jMessage(DISP_AND_EXIT,"terminate loop (activity list) with ')'",1,pDP);
      } // IF
      break;
    } // IF

    if (pDP->sData[0] != '(' && strcmp(pDP->sData,"loop"))
    { jMessage(DISP_AND_EXIT,"'(' or LOOP expected",1,pDP);
    } // IF
    pOneFlowAct = new CFlowAct;

    if (!strcmp(pDP->sData,"loop"))
    { xDataBsdl(pDP,XDATA_READ);
      if (!strcmp(pDP->sData,"min"))
      { pOneFlowAct->m_lMinLoopCount = lReadDecimalBsdl(pDP,0);
        xDataBsdl(pDP,XDATA_READ);
      } // IF
      if (!strcmp(pDP->sData,"max"))
      { pOneFlowAct->m_lMaxLoopCount = lReadDecimalBsdl(pDP,0);
      } // IF
      else
      { jMessage(DISP_AND_EXIT,"Expecting MIN and/or MAX loopcount",1,pDP);
      } // ELSE

      if (pOneFlowAct->m_lMinLoopCount > pOneFlowAct->m_lMaxLoopCount)
      { jMessage(DISP_AND_EXIT,"Min loop count > Max loop count",1,pDP);
      } // IF

      findStrBsdl(pDP,"(",0);
      lFlowMaxRegSize = this->readFlowActivityList(pDP,1,pCBD,
                  &(pOneFlowAct->m_pLoopActions),&(*iHasOutputOp));
    } // IF pDP->sData = loop
    else
    { xGetData(pDP,1,eATTR_VAR,"instruction expected");
      pOneFlowAct->m_lOpcodeIndex = pCBD->getOpcodeIndex(0,pDP->sData,pDP);
      pOneFlowAct->m_lUpdateWidth = this->getUpdateFieldList(pDP,
                                          &(pOneFlowAct->m_pUF_Update),
                                          &m_lUseDataName,&(*iHasOutputOp));
      lOpcodeDataRegSize =
            checkFieldWidth(pDP,"Update",pCBD,pOneFlowAct->m_lUpdateWidth,
                      pOneFlowAct->m_lOpcodeIndex);

      this->getFlowActWaitSpec(pDP,&(pOneFlowAct->m_pFlowWaitInfo),pCBD);
      pOneFlowAct->m_lCaptureWidth = this->getCaptureFieldList(pDP,
                                          &(pOneFlowAct->m_pCF_Capture),
                                          &m_lUseDataName,&(*iHasOutputOp));
      checkFieldWidth(pDP,"Capture",pCBD,pOneFlowAct->m_lCaptureWidth,
                      pOneFlowAct->m_lOpcodeIndex);


      if (lOpcodeDataRegSize > lFlowMaxRegSize)
      { lFlowMaxRegSize = lOpcodeDataRegSize;
      } // IF
    } // ELSE

    if (*pCFAhead == NULL) // Attach flow activity object to end of flow list
    { *pCFAhead = pOneFlowAct;
    } // IF
    else
    { pLastFlowAct->m_pNextCFlowAct = pOneFlowAct;
    } // ELSE
    pLastFlowAct = pOneFlowAct;

  } // WHILE (1)

  return lFlowMaxRegSize;

} // CIscFlow::readFlowActivityList()
//==!=====================================================================!==
//==! Create an array of CxDataVar pointers and assign each array element to
//==! an object from the pCDVlist; variable access will be via an array index
//==! instead of traversing the pCDVlist linked list.
//==!
//==! Function Parameters: none
//==! Return Value       : none
//==!=====================================================================!==
void CIscFlow::arrangeDataVars()
{ long lVar0;
  CxDataVar *pCDVvar;

  if (m_lCxDataVarArraySize > 0)
  { m_CxDataVarArray = new CxDataVar * [m_lCxDataVarArraySize];
    for (lVar0 = 0,pCDVvar = m_pCDVlist;
         pCDVvar != NULL && lVar0 < m_lCxDataVarArraySize;
        )
    { m_CxDataVarArray[lVar0] = pCDVvar;
      ++lVar0;
      pCDVvar = pCDVvar->m_pNextCxDataVar;
      if ((lVar0 >= m_lCxDataVarArraySize && pCDVvar != NULL) ||
          (lVar0 <  m_lCxDataVarArraySize && pCDVvar == NULL))
      { jMessage(DISP_AND_EXIT,"CDVarray != CDVlist",0,NULL);
      } // IF
    } // FOR pCDVvar != NULL && lVar0 < m_lCxDataVarArraySize
  } // IF m_lCxDataVarArraySize > 0

} // CIscFlow::arrangeDataVars()
//==!=====================================================================!==
//==! readIscFlows(): read ISC_FLOW attribute section
//==!
//==! Function Parameters: pDP = input cDataPort
//==! Return Value       : none
//==!=====================================================================!==
void CBsdlData::readIscFlows(cDataPort *pDP)
{ CIscFlow *pCIFvar;
  CFlowAct **p_m_Flow;
  CMultFlow *p_CMFlow;
  long *p_lUseData,lMaxDataRegSize;
  int iRptBlk,*p_iPhaseHasOutOp;

  do
  { xGetData(pDP,1,eATTR_VAR,"Flow name expected");
    if (!xstrcmp(pDP->sData,"initialize") || !xstrcmp(pDP->sData,"repeat") ||
        !xstrcmp(pDP->sData,"terminate"))
    { sprintf(sMsgOnly,"Cannot use reserved word [%s] as a flow name",
              pDP->sData);
      jMessage(DISP_AND_EXIT,sMsgOnly,1,pDP);
    } // IF
    pCIFvar = new CIscFlow;
    copyStr(&(pCIFvar->m_csFlowName),pDP->sData);
    strcpy(sHoldData,pDP->sData);
    xDataBsdl(pDP,XDATA_READ);
    if (pDP->sData[0] == '(')
    { xGetData(pDP,1,eATTR_VAR,"Flow data name expected");
      copyStr(&(pCIFvar->m_csDataName),pDP->sData);
      if (getPtrToFlow(sHoldData,pDP->sData,1,NULL) != NULL)
      { sprintf(sMsgOnly,"Flow [%s(%s)] already defined",
                sHoldData,pDP->sData);
        jMessage(DISP_AND_EXIT,sMsgOnly,1,pDP);
      } // IF
      findStrBsdl(pDP,")",0);
      xDataBsdl(pDP,XDATA_READ);
      if (!strcmp(pDP->sData,"unprocessed"))
      { pCIFvar->m_iUnprocessed = 1;
        xDataBsdl(pDP,XDATA_READ);
      } // IF
      if (!strcmp(pDP->sData,"exit_on_error"))
      { pCIFvar->m_iExitOnError = 1;
        xDataBsdl(pDP,XDATA_READ);
      } // IF
    } // IF sData[0] = '('
    else
    { if (getPtrToFlow(sHoldData,NULL,1,NULL) != NULL)
      { sprintf(sMsgOnly,"Flow [%s] already defined",sHoldData);
        jMessage(DISP_AND_EXIT,sMsgOnly,1,pDP);
      } // IF
    } // ELSE

    while (1)
    {
//==! Initialize m_lUseDataName to 0 and increment if any flow in any of the
//==! 3 flow activity phases (INIT/REPEAT/TERMINATE) uses the '?' or '?!'
//==! operator to get data for the update or capture phase of the flow
//==! activity. Required so software will know to look for the INITIALIZE,
//==! REPEAT, or TERMINATE headers in the ISC data file only if the flow
//==! phase needs data.
      pCIFvar->m_lUseDataName = 0;

      iRptBlk = 0;
      p_CMFlow = NULL;

      if (!strcmp(pDP->sData,"initialize"))
      { if ((pCIFvar->m_pCFA_Terminate != NULL) ||
            (pCIFvar->m_pCMF_Repeat != NULL))
        { jMessage(DISP_AND_EXIT,
          "Initialize flow must come before Repeat/Terminate flow",1,pDP);
        } // IF
        p_m_Flow = &pCIFvar->m_pCFA_Initialize;
        p_lUseData = &(pCIFvar->m_lUseDataName_Initialize);
        p_iPhaseHasOutOp = &(pCIFvar->m_iInitHasOutOp);
      } // IF
      else
      { if (!strcmp(pDP->sData,"repeat"))
        { if (pCIFvar->m_pCFA_Terminate != NULL)
          { jMessage(DISP_AND_EXIT,
            "Repeat flow(s) must come before Terminate flow",1,pDP);
          } // IF
          if (pCIFvar->m_pCMF_Repeat == NULL)
          { pCIFvar->m_pCMF_Repeat = new CMultFlow;
            p_CMFlow = pCIFvar->m_pCMF_Repeat;
          } // IF
          else
          { for (p_CMFlow = pCIFvar->m_pCMF_Repeat;
                 p_CMFlow->m_pNextCMultFlow != NULL;
                 p_CMFlow = p_CMFlow->m_pNextCMultFlow)
            { ;
            } // FOR
            p_CMFlow->m_pNextCMultFlow = new CMultFlow;
            p_CMFlow = p_CMFlow->m_pNextCMultFlow;
          } // ELSE
          iRptBlk = 1;
          p_m_Flow = &(p_CMFlow->m_pCFA_Repeat);
          p_lUseData = &(p_CMFlow->m_lUseDataName_Repeat);
          p_iPhaseHasOutOp = &(p_CMFlow->m_iRptHasOutOp);
        } // IF
        else
        { if (!strcmp(pDP->sData,"terminate"))
          { p_m_Flow = &pCIFvar->m_pCFA_Terminate;
            p_lUseData = &(pCIFvar->m_lUseDataName_Terminate);
            p_iPhaseHasOutOp = &(pCIFvar->m_iTermHasOutOp);
          } // IF
          else
          { jMessage(DISP_AND_EXIT,"Initialize/Repeat/Terminate expected",1,pDP);
          } // ELSE
        } // ELSE
      } // ELSE

      if (*p_m_Flow != NULL)
      { sprintf(sMsgOnly,"[%s] activity list already defined",pDP->sData);
        jMessage(DISP_AND_EXIT,sMsgOnly,1,pDP);
      } // IF

      if (iRptBlk == 1)
      { xGetData(pDP,1,eATTR_NUM,"Expecting repeat count");
        isPtrNull(p_CMFlow,"p_CMFlow");
        p_CMFlow->m_lRptCnt = atol(pDP->sData);
      } // IF

      lMaxDataRegSize =
              pCIFvar->readFlowActivityList(pDP,0,this,&(*p_m_Flow),
              &(*p_iPhaseHasOutOp));
      if (lMaxDataRegSize > pCIFvar->m_lMaxRegSizeUsed)
      { pCIFvar->m_lMaxRegSizeUsed = lMaxDataRegSize;
      } // IF

      *p_lUseData = pCIFvar->m_lUseDataName;

      if (!strcmp(pDP->sData,",") || !strcmp(pDP->sData,";"))
      { break;
      } // IF

    } // WHILE (1)

    pCIFvar->m_lUseDataName =
             pCIFvar->m_lUseDataName_Initialize +
             pCIFvar->m_lUseDataName_Terminate;
    for (p_CMFlow = pCIFvar->m_pCMF_Repeat;
         p_CMFlow != NULL;
         p_CMFlow = p_CMFlow->m_pNextCMultFlow)
    { pCIFvar->m_lUseDataName += p_CMFlow->m_lUseDataName_Repeat;
    } // FOR p_CMFlow != NULL

    pCIFvar->arrangeDataVars();

    pCIFvar->m_pNextCIscFlow = m_pCIscFlowInfo;
    m_pCIscFlowInfo = pCIFvar;

//==! 2004/10/04: If flow requires data, then check if flow has a data name
//==! associated with it (NOTE_0012).
    if (pCIFvar->m_lUseDataName > 0 && pCIFvar->m_csDataName == NULL)
    { sprintf(sMsgOnly,"Flow [%s] has '?' or '?!' operator but no data name",
              pCIFvar->m_csFlowName);
      jMessage(DISP_AND_EXIT,sMsgOnly,1,pDP);
    } // IF pCIFvar->m_lUseDataName > 0

  } while (strcmp(pDP->sData,";"));

} // CBsdlData::readIscFlows()
//==!=====================================================================!==
//==! Search the CBsdlData m_pIscFlowInfo list for the flow called sFlowName
//==! and return a pointer to it.
//==!
//==! Function Parameters: sFlowName = flow to get pointer to
//==!                      sFlowData = data name for flow
//==!                      iMode = 0: error if flow not found
//==!                              1: return NULL if flow not found
//==!                      pDP = cDataPort pointer
//==! Return Value       : pointer to sFlowName
//==!=====================================================================!==
CIscFlow *CBsdlData::getPtrToFlow(char *sFlowName,char *sFlowData,int iMode,
        cDataPort *pDP)
{ CIscFlow *pCIFvar;

  isPtrNull((void *)(sFlowName),"gPTFn");
  for (pCIFvar = m_pCIscFlowInfo; pCIFvar != NULL;
       pCIFvar = pCIFvar->m_pNextCIscFlow)
  { if (!strcmp(sFlowName,pCIFvar->m_csFlowName))
    { if (sFlowData == NULL && pCIFvar->m_csDataName == NULL)
      { return pCIFvar;
      } // IF
      if (sFlowData != NULL && pCIFvar->m_csDataName != NULL)
      { if (!strcmp(sFlowData,pCIFvar->m_csDataName))
        { return pCIFvar;
        } // IF
      } // IF
    } // IF
  } // FOR pCIFvar != NULL

  if (iMode == 1)
  { return NULL;
  } // IF

  if (sFlowData != NULL)
  { sprintf(sMsgOnly,
            "Flow descriptor [%s(%s)] not in ISC_FLOW attribute list",
            sFlowName,sFlowData);
  } // IF
  else
  { sprintf(sMsgOnly,"Flow descriptor [%s] not in ISC_FLOW attribute list",
            sFlowName);
  } // ELSE
  jMessage(DISP_AND_EXIT,sMsgOnly,pDP == NULL ? 0:1,pDP);

  return NULL;

} // CBsdlData::getPtrToFlow()
//==!=====================================================================!==
//==! readIscProcedures(): read ISC_PROCEDURE attribute section
//==!
//==! Function Parameters: pDP = input cDataPort
//==! Return Value       : none
//==!=====================================================================!==
void CBsdlData::readIscProcedures(cDataPort *pDP)
{ CIscProc *pCIPvar;
  CxFlowObj *pCxFOvar,*pCxFOtail;
  long lMaxValue;

  do
  { pCxFOtail = NULL;
    xGetData(pDP,1,eATTR_VAR,"Procedure name expected");
    pCIPvar = new CIscProc;
    copyStr(&(pCIPvar->m_csProcName),pDP->sData);
    strcpy(sHoldData,pDP->sData);
    xDataBsdl(pDP,XDATA_READ);
    if (pDP->sData[0] == '(')
    { xGetData(pDP,1,eATTR_VAR,"Procedure data name expected");
      copyStr(&(pCIPvar->m_csProcDataName),pDP->sData);
      if (getPtrToProc(sHoldData,pDP->sData,1,NULL) != NULL)
      { sprintf(sMsgOnly,"Proc [%s(%s)] already defined",
                sHoldData,pDP->sData);
        jMessage(DISP_AND_EXIT,sMsgOnly,1,pDP);
      } // IF
      findStrBsdl(pDP,")",0);
    } // IF sData[0] = '('
    else
    { pDP->iParserGetData = 0;
      if (getPtrToProc(sHoldData,NULL,1,NULL) != NULL)
      { sprintf(sMsgOnly,"Proc [%s] already defined",sHoldData);
        jMessage(DISP_AND_EXIT,sMsgOnly,1,pDP);
      } // IF
    } // ELSE
    findStrBsdl(pDP,"=",0);
    findStrBsdl(pDP,"(",0);

    lMaxValue = 0;
    while (pDP->sData[0] != ')')
    { xGetData(pDP,1,eATTR_VAR,"Flow descriptor/name expected");
      strcpy(sHoldData,pDP->sData);
      xDataBsdl(pDP,XDATA_READ);
      pCxFOvar = new CxFlowObj;
      if (pDP->sData[0] == '(')
      { xGetData(pDP,1,eATTR_VAR,"Flow data parameter expected");
        copyStr(&(pCxFOvar->m_csFlowParam),pDP->sData);
        pCxFOvar->m_pFlowItem = getPtrToFlow(sHoldData,pDP->sData,0,pDP);
        findStrBsdl(pDP,")",0);
        xDataBsdl(pDP,XDATA_READ);
      } // IF sData[0] = '('
      else
      { pCxFOvar->m_pFlowItem = getPtrToFlow(sHoldData,NULL,0,pDP);
      } // ELSE

      if (pCxFOvar->m_pFlowItem != NULL)
      { if (lMaxValue < pCxFOvar->m_pFlowItem->m_lMaxRegSizeUsed)
        { lMaxValue = pCxFOvar->m_pFlowItem->m_lMaxRegSizeUsed;
        } // IF
      } // IF

      if (pCIPvar->m_pProcFlows == NULL)
      { pCIPvar->m_pProcFlows = pCxFOvar;
      } // IF
      else
      { pCxFOtail->m_pNextCxFlowObj = pCxFOvar;
      } // ELSE
      pCxFOtail = pCxFOvar;

      if (pDP->sData[0] != ',' && pDP->sData[0] != ')')
      { jMessage(DISP_AND_EXIT,"',' or ')' expected",1,pDP);
      } // IF

    } // WHILE (pDP->sData[0] != ')')

    pCIPvar->m_pNextCIscProc = m_pCIscProcInfo;
    pCIPvar->m_lMaxRegSizeUsed = lMaxValue;

    m_pCIscProcInfo = pCIPvar;

    xDataBsdl(pDP,XDATA_READ);
    if (pDP->sData[0] != ',' && pDP->sData[0] != ';')
    { jMessage(DISP_AND_EXIT,"',' or ';' expected",1,pDP);
    } // IF

  } while (strcmp(pDP->sData,";"));

//==! Check if the mandatory procedures are defined. Preceding WHILE loop
//==! already checks for multiply defined procedure names, but these 2
//==! procedures are expected to not have any arguments (i.e., the BSDL
//==! parser will look for "proc_error_exit" and not "proc_error_exit(arg)").
  for (pCIPvar = m_pCIscProcInfo, lMaxValue = 0; pCIPvar != NULL;
       pCIPvar = pCIPvar->m_pNextCIscProc)
  { if (!xstrcmp(pCIPvar->m_csProcName,"proc_error_exit"))
    { m_pCIscProcError = pCIPvar;
      lMaxValue = lMaxValue + 1;  // set bit 0
      continue;
    } // IF
    if (!xstrcmp(pCIPvar->m_csProcName,"proc_program_done"))
    { lMaxValue = lMaxValue + 2;  // set bit 1
      continue;
    } // IF
  } // FOR pCIPvar != NULL

//==! Flag missing mandatory procedures as warnings.
  if ((lMaxValue & 0x01) == 0)
  { jMessage(DISP_SHOW_MSG,"Mandatory proc_error_exit not defined",0,pDP);
  } // IF
  if ((lMaxValue & 0x02) == 0)
  { jMessage(DISP_SHOW_MSG,"Mandatory proc_program_done not defined",0,pDP);
  } // IF

} // CBsdlData::readIscProcedures()
//==!=====================================================================!==
//==! Search the CBsdlData m_pIscProcInfo list for the procedure called
//==! sProcName and return a pointer to it.
//==!
//==! Function Parameters: sProcName = procedure to get pointer to
//==!                      sProcData = optional procedure data name
//==!                      iMode = 0: error if procedure not found
//==!                              1: return NULL if procedure not found
//==!                      pDP = cDataPort pointer
//==! Return Value       : pointer to sProcName
//==!=====================================================================!==
CIscProc *CBsdlData::getPtrToProc(char *sProcName,char *sProcData,int iMode,
            cDataPort *pDP)
{ CIscProc *pCIPvar;

  isPtrNull((void *)(sProcName),"gPTPn");
  for (pCIPvar = m_pCIscProcInfo; pCIPvar != NULL;
       pCIPvar = pCIPvar->m_pNextCIscProc)
  { if (!strcmp(sProcName,pCIPvar->m_csProcName))
    { if (sProcData == NULL && pCIPvar->m_csProcDataName == NULL)
      { return pCIPvar;
      } // IF
      if (sProcData != NULL && pCIPvar->m_csProcDataName != NULL)
      { if (!strcmp(sProcData,pCIPvar->m_csProcDataName))
        { return pCIPvar;
        } // IF
      } // IF
    } // IF
  } // FOR pCIPvar != NULL

  if (iMode == 1)
  { return NULL;
  } // IF

  if (sProcData != NULL)
  { sprintf(sMsgOnly,
            "Procedure [%s(%s)] not in ISC_PROCEDURE attribute list",
            sProcName,sProcData);
  } // IF
  else
  { sprintf(sMsgOnly,"Procedure [%s] not in ISC_PROCEDURE attribute list",
            sProcName);
  } // ELSE
  jMessage(DISP_AND_EXIT,sMsgOnly,pDP == NULL ? 0:1,pDP);

  return NULL;

} // CBsdlData::getPtrToProc()
//==!=====================================================================!==
//==! readIscActions(): read ISC_ACTION attribute section
//==!
//==! Function Parameters: pDP = input cDataPort
//==! Return Value       : none
//==!=====================================================================!==
void CBsdlData::readIscActions(cDataPort *pDP)
{ CIscAct *pCIAvar;
  CxProcObj *pCxPOvar,*pCxPOtail;
  long lMaxValue;

  do
  { pCxPOtail = NULL;
    xGetData(pDP,1,eATTR_VAR,"Action name expected");
    pCIAvar = new CIscAct;
    copyStr(&(pCIAvar->m_csActName),pDP->sData);
    xDataBsdl(pDP,XDATA_READ);
    if (pDP->sData[0] == '(')
    { xGetData(pDP,1,eATTR_VAR,"Action data name expected");
      copyStr(&(pCIAvar->m_csActDataName),pDP->sData);
      findStrBsdl(pDP,")",0);
      xDataBsdl(pDP,XDATA_READ);
    } // IF sData[0] = '('

    if (pDP->eDataAttr == eATTR_VAR)
    { if (!strcmp(pDP->sData,"proprietary"))
      { pCIAvar->m_iExclusivity = 1;
      } // IF
      else
      { jMessage(DISP_AND_EXIT,"Invalid keyword or '=' expected",1,pDP);
      } // ELSE
    } // IF
    else
    { pDP->iParserGetData = 0;
    } // ELSE

    findStrBsdl(pDP,"=",0);
    findStrBsdl(pDP,"(",0);

    lMaxValue = 0;

    while (pDP->sData[0] != ')')
    { xGetData(pDP,1,eATTR_VAR,"Procedure descriptor/name expected");
      strcpy(sHoldData,pDP->sData);
      xDataBsdl(pDP,XDATA_READ);
      pCxPOvar = new CxProcObj;
      if (pDP->sData[0] == '(')
      { xGetData(pDP,1,eATTR_VAR,"Procedure data parameter expected");
        copyStr(&(pCxPOvar->m_csProcParam),pDP->sData);
        pCxPOvar->m_pProcItem = getPtrToProc(sHoldData,pDP->sData,0,pDP);
        findStrBsdl(pDP,")",0);
        xDataBsdl(pDP,XDATA_READ);
      } // IF sData[0] = '('
      else
      { pCxPOvar->m_pProcItem = getPtrToProc(sHoldData,NULL,0,pDP);
      } // ELSE

      if (pCxPOvar->m_pProcItem != NULL)
      { if (lMaxValue < pCxPOvar->m_pProcItem->m_lMaxRegSizeUsed)
        { lMaxValue = pCxPOvar->m_pProcItem->m_lMaxRegSizeUsed;
        } // IF
      } // IF

      while (pDP->eDataAttr == eATTR_VAR)
      { if (!strcmp(pDP->sData,"proprietary"))
        { if (pCxPOvar->m_iProcExclusivity != 0)
          { jMessage(DISP_SHOW_MSG,"Procedure exclusivity already specified",
                     1,pDP);
          } // IF
          pCxPOvar->m_iProcExclusivity = 1;
        } // IF
        else
        { if (!strcmp(pDP->sData,"optional") ||
              !strcmp(pDP->sData,"recommended"))
          { if (pCxPOvar->m_enOptionSpec != ePROC_NO_SETTING)
            { jMessage(DISP_SHOW_MSG,
              "Overriding previous procedure \"Optional | Recommended\" spec",
                1,pDP);
            } // IF
            if (!strcmp(pDP->sData,"optional"))
            { pCxPOvar->m_enOptionSpec = ePROC_OPTIONAL;
            } // IF
            else
            { pCxPOvar->m_enOptionSpec = ePROC_RECOMMENDED;
            } // ELSE
          } // IF
          else
          { jMessage(DISP_AND_EXIT,
            "Expecting ',' or PROPRIETARY | OPTIONAL | RECOMMENDED",1,pDP);
          } // ELSE
        } // ELSE
        xDataBsdl(pDP,XDATA_READ);
      } // WHILE (eDataAttr == eATTR_VAR

      if (pCIAvar->m_pActionProcs == NULL)
      { pCIAvar->m_pActionProcs = pCxPOvar;
      } // IF
      else
      { pCxPOtail->m_pNextCxProcObj = pCxPOvar;
      } // ELSE
      pCxPOtail = pCxPOvar;

      if (pDP->sData[0] != ',' && pDP->sData[0] != ')')
      { jMessage(DISP_AND_EXIT,"',' or ')' expected",1,pDP);
      } // IF

    } // WHILE (pDP->sData[0] != ')')

    pCIAvar->m_pNextCIscAct = m_pCIscActInfo;
    pCIAvar->m_lMaxRegSizeUsed = lMaxValue;

    m_pCIscActInfo = pCIAvar;

    xDataBsdl(pDP,XDATA_READ);
    if (pDP->sData[0] != ',' && pDP->sData[0] != ';')
    { jMessage(DISP_AND_EXIT,"',' or ';' expected",1,pDP);
    } // IF
  } while (strcmp(pDP->sData,";"));

} // CBsdlData::readIscActions()
//==!=====================================================================!==
//==! readBsdlAttributes(): read BSDL attribute statements.
//==!
//==! Function Parameters: pDP = input cDataPort
//==! Return Value       : none
//==!=====================================================================!==
void CBsdlData::readBsdlAttributes(cDataPort *pDP)
{ long lVar0,lVar1,lVar2;
  CBsdlOp *pCBsdlOpHead,*pCBsdlOpVar;
  CBsdlDataReg *pCBsdlDataRegTail,*pCBsdlDataRegVar;
  SList *pSListVar;
  CxCStr *pCxCStr_Var,**p_m_CxCStrPtrVar;
  float fNumVar;
  long lNumVar;

  iGlobProgramState = dPROGSTATE_READBSDLATTR;

  pCBsdlOpHead = NULL;

  xDataBsdl(pDP,XDATA_READ);

  if (!strcmp(pDP->sData,"tap_scan_clock"))
  { if (m_csTapClkName != NULL)
    { jMessage(DISP_AND_EXIT,"TAP_SCAN_CLK already specified",1,pDP);
    } // IF
    findStrBsdl(pDP,"of",0);
    xDataBsdl(pDP,XDATA_READ);
    copyStr(&m_csTapClkName,pDP->sData);
    findStrBsdl(pDP,":",0);
    findStrBsdl(pDP,"signal",0);
    findStrBsdl(pDP,"is",0);
    findStrBsdl(pDP,"(",0);

//==! 2003/11/04: Data IO's 32-bit embedded system seems to have problems
//==! with fscanf() of floating numbers; use internally-developed getFloat()
//==! instead.
#if 0
    fscanf(pDP->pFILE,"%e",&m_fTckFreq);
#else
    switch(getFloat(pDP,&fNumVar,&lNumVar))
    { case GF_FLOAT:
        m_fTckFreq = fNumVar;
        break;
      case GF_LONG:
        m_fTckFreq = (float)lNumVar;
        break;
      default:
        jMessage(DISP_AND_EXIT,"Unknown TAP_SCAN_CLK specification",1,pDP);
        break;
    } // SWITCH (getFloat())
#endif

    findStrBsdl(pDP,",",0);
    xDataBsdl(pDP,XDATA_READ);
    if (strcmp(pDP->sData,"low") && strcmp(pDP->sData,"both"))
    { jMessage(DISP_AND_EXIT,"Expecting LOW or BOTH",1,pDP);
    } // IF
    findStrBsdl(pDP,")",0);
    findStrBsdl(pDP,";",0);
    return;
  } // IF attribute = tap_scan_clk

  if (!strcmp(pDP->sData,"isc_status"))
  { if (lImplementState >= 0)
    { jMessage(DISP_AND_EXIT,"ISC_STATUS already defined",1,pDP);
    } // IF
    checkAttrEntity(pDP);
    xDataBsdl(pDP,XDATA_READ);
    lImplementState = 1;
    if (!strcmp(pDP->sData,"not"))
    { lImplementState = 0;
      xDataBsdl(pDP,XDATA_READ);
    } // IF
    if (strcmp(pDP->sData,"implemented"))
    { jMessage(DISP_AND_EXIT,"expecting [NOT] IMPLEMENTED",1,pDP);
    } // IF
    findStrBsdl(pDP,";",0);
    checkBsdlEntityEnd(pDP);
    return;
  } // IF attribute = isc_status

  if (!strcmp(pDP->sData,"instruction_length"))
  { if (lInstrLength >= 0)
    { jMessage(DISP_AND_EXIT,"Instr length already defined",1,pDP);
    } // IF
    checkAttrEntity(pDP);
    lInstrLength = lReadDecimal(pDP);
    findStrBsdl(pDP,";",0);
    return;
  } // IF attribute = instruction_length

  if (!strcmp(pDP->sData,"instruction_capture"))
  { if (lInstrLength <= 0)
    { jMessage(DISP_AND_EXIT,
               "INSTRUCTION_LENGTH attribute not yet defined",1,pDP);
    } // IF
    if (m_CxCStr_InstrCaptureList != NULL)
    { dispMsg("BSDL error: INSTRUCTION_CAPTURE already defined as:\n\t-> [",
              DISP_PRE_LF,1,1);
      for (pCxCStr_Var = m_CxCStr_InstrCaptureList; pCxCStr_Var != NULL;
           pCxCStr_Var = pCxCStr_Var->pNextCxCStr)
      { sprintf(sMsgOnly,"%s",pCxCStr_Var->pStr);
        dispMsg(sMsgOnly,DISP_USE_CUR_POS,1,1);
        if (pCxCStr_Var->pNextCxCStr != NULL)
        { dispMsg(",",DISP_USE_CUR_POS,1,1);
        } // IF
        else
        { dispMsg("]",DISP_USE_CUR_POS | DISP_POST_LF,1,1);
        } // ELSE
      } // IF
      jMessage(DISP_AND_EXIT,"BSDL attribute redefinition",1,pDP);
    } // IF m_CxCStr_InstrCaptureList != NULL

    checkAttrEntity(pDP);

    findStr(pDP,"\"",0);
    ++pDP->iBracketLevel;
    pCxCStr_Var = new CxCStr;
    pCxCStr_Var->pStr = new char [lInstrLength+1];
    pCxCStr_Var->pStr[lInstrLength] = '\0'; //==! NULL-terminate string
    for (lVar2 = 0;;)
    { xDataBsdl(pDP,XDATA_READ);
      if (!strcmp(pDP->sData,";") || !strcmp(pDP->sData,","))
      { if (lVar2 != lInstrLength)
        { sprintf(sMsgOnly,
                  "Only [%ld] of [%ld] INSTRUCTION_CAPTURE bits specified",
                  lVar2,lInstrLength);
          jMessage(DISP_AND_EXIT,sMsgOnly,1,pDP);
        } // IF
        pCxCStr_Var->pNextCxCStr = m_CxCStr_InstrCaptureList;
        m_CxCStr_InstrCaptureList = pCxCStr_Var;

        if (!strcmp(pDP->sData,";"))
        { break;
        } // IF

        lVar2 = 0;
        pCxCStr_Var = new CxCStr;
        pCxCStr_Var->pStr = new char [lInstrLength+1];
        pCxCStr_Var->pStr[lInstrLength] = '\0'; //==! NULL-terminate string
        continue;
      } // IF
      lVar0 = strlen(pDP->sData);
      validateData(pDP->sData,lVar0,"01xX",pDP);
      for (lVar1 = 0; lVar1 < lVar0; ++lVar1)
      { if (lVar2 >= lInstrLength)
        { sprintf(sMsgOnly,
"INSTRUCTION_CAPTURE bit vector exceeds INSTRUCTION_LENGTH of %ld bits",
                  lInstrLength);
          jMessage(DISP_AND_EXIT,sMsgOnly,1,pDP);
        } // IF
        pCxCStr_Var->pStr[lVar2] = pDP->sData[lVar1];
        ++lVar2;
      } // FOR lVar1 < lVar0
    } // FOR (;;)

    checkBsdlEntityEnd(pDP);
    return;
  } // IF attribute = instruction_capture

  if (!strcmp(pDP->sData,"boundary_length"))
  { if (lBscLength >= 0)
    { jMessage(DISP_AND_EXIT,"BSC length already defined",1,pDP);
    } // IF
    checkAttrEntity(pDP);
    lBscLength = lReadDecimal(pDP);
    findStrBsdl(pDP,";",0);
    return;
  } // IF attribute = boundary_length

//==! Read and store ISC_FLOWs
  if (!strcmp(pDP->sData,"isc_flow"))
  { if (m_pCIscFlowInfo != NULL)
    { jMessage(DISP_AND_EXIT,"ISC_FLOW already defined",1,pDP);
    } // IF pCIscFlowInfo != NULL
    checkAttrEntity(pDP);
    this->readIscFlows(pDP);
    checkBsdlEntityEnd(pDP);
    return;
  } // IF attribute = isc_flow

//==! Read and store ISC_PROCEDUREs
  if (!strcmp(pDP->sData,"isc_procedure"))
  { if (m_pCIscProcInfo != NULL)
    { jMessage(DISP_AND_EXIT,"ISC_PROCEDURE already defined",1,pDP);
    } // IF pCIscProcInfo != NULL
    checkAttrEntity(pDP);
    this->readIscProcedures(pDP);
    checkBsdlEntityEnd(pDP);
    return;
  } // IF attribute = isc_procedure

//==! Read and store ISC_ACTIONs
  if (!strcmp(pDP->sData,"isc_action"))
  { if (m_pCIscActInfo != NULL)
    { jMessage(DISP_AND_EXIT,"ISC_ACTION already defined",1,pDP);
    } // IF pCIscActInfo != NULL
    checkAttrEntity(pDP);
    this->readIscActions(pDP);
    checkBsdlEntityEnd(pDP);
    return;
  } // IF attribute = isc_action

//==! Read and store opcode names and encodings. Instruction register length
//==! should already have been defined.
  if (!strcmp(pDP->sData,"instruction_opcode"))
  { if (lInstrLength <= 0)
    { jMessage(DISP_AND_EXIT,"INSTRUCTION_LENGTH still undefined",1,pDP);
    } // IF
    if (pCBsdlOpInfo != NULL)
    { jMessage(DISP_AND_EXIT,"Instruction_opcode(s) already defined",1,pDP);
    } // IF pCBsdlOpInfo != NULL
    lVar0 = -1;
    checkAttrEntity(pDP);
    findStr(pDP,"\"",0);
    ++pDP->iBracketLevel;
    while (strcmp(pDP->sData,";"))
    { xGetData(pDP,1,eATTR_VAR,"Expecting instr name");

      for (pCBsdlOpVar = pCBsdlOpHead;
           pCBsdlOpVar != NULL;
           pCBsdlOpVar = pCBsdlOpVar->pNextCBsdlOp)
      { if (!strcmp(pCBsdlOpVar->csOpcodeName,pDP->sData))
        { sprintf(sMsgOnly,"Opcode [%s] already defined",pDP->sData);
          jMessage(DISP_AND_EXIT,sMsgOnly,1,pDP);
        } // IF
      } // FOR pCBsdlOpVar != NULL

      ++lVar0;
      pCBsdlOpVar = new CBsdlOp;
      pCBsdlOpVar->pNextCBsdlOp = pCBsdlOpHead;
      pCBsdlOpHead = pCBsdlOpVar;

      copyStr(&(pCBsdlOpVar->csOpcodeName),pDP->sData);
      if (!strcmp(pDP->sData,"sample"))
      { if (lOpcodeSamplePreload != -1)
        { jMessage(DISP_AND_EXIT,"SAMPLE opcode already defined",1,pDP);
        } // IF
        lOpcodeSamplePreload = lVar0;
      } // IF sData = sample
      if (!strcmp(pDP->sData,"extest"))
      { if (lOpcodeExtest != -1)
        { jMessage(DISP_AND_EXIT,"EXTEST opcode already defined",1,pDP);
        } // IF
        lOpcodeExtest = lVar0;
      } // IF sData = extest
      if (!strcmp(pDP->sData,"bypass"))
      { if (lOpcodeBypass != -1)
        { jMessage(DISP_AND_EXIT,"BYPASS opcode already defined",1,pDP);
        } // IF
        lOpcodeBypass = lVar0;
      } // IF sData = idcode
      if (!strcmp(pDP->sData,"idcode"))
      { if (lOpcodeIdcode != -1)
        { jMessage(DISP_AND_EXIT,"IDCODE opcode already defined",1,pDP);
        } // IF
        lOpcodeIdcode = lVar0;
      } // IF sData = idcode
      if (!strcmp(pDP->sData,"usercode"))
      { if (lOpcodeUsercode != -1)
        { jMessage(DISP_AND_EXIT,"USERCODE opcode already defined",1,pDP);
        } // IF
        lOpcodeUsercode = lVar0;
      } // IF sData = usercode
      if (!strcmp(pDP->sData,"user1"))
      { if (lOpcodeUser1 != -1)
        { jMessage(DISP_AND_EXIT,"USER1 opcode already defined",1,pDP);
        } // IF
        lOpcodeUser1 = lVar0;
      } // IF sData = user1
      if (!strcmp(pDP->sData,"user2"))
      { if (lOpcodeUser2 != -1)
        { jMessage(DISP_AND_EXIT,"USER2 opcode already defined",1,pDP);
        } // IF
        lOpcodeUser2 = lVar0;
      } // IF sData = user2

      findStrBsdl(pDP,"(",0);
      while (strcmp(pDP->sData,")"))
      { xDataBsdl(pDP,XDATA_READ);
        if ((long)strlen(pDP->sData) != lInstrLength)
        { sprintf(sMsgBuf00,"Opcode (%s) != specified IR length [%ld]",
                  pDP->sData,lInstrLength);
          jMessage(DISP_AND_EXIT,sMsgBuf00,1,pDP);
        } // IF
        validateData(pDP->sData,lInstrLength,"01xX",pDP);
        pCxCStr_Var = new CxCStr;
        copyStr(&(pCxCStr_Var->pStr),pDP->sData);
        pCxCStr_Var->pNextCxCStr = pCBsdlOpVar->m_CxCStr_OpcodeList;
        pCBsdlOpVar->m_CxCStr_OpcodeList = pCxCStr_Var;
        xDataBsdl(pDP,XDATA_READ);
        if (strcmp(pDP->sData,",") && strcmp(pDP->sData,")"))
        { jMessage(DISP_AND_EXIT,"Expecting ',' or ')'",1,pDP);
        } // IF
      } // WHILE not ')'
      xDataBsdl(pDP,XDATA_READ);
      if (strcmp(pDP->sData,",") && strcmp(pDP->sData,";"))
      { jMessage(DISP_AND_EXIT,"Expecting ',' or ';'",1,pDP);
      } // IF

    } // WHILE sData != ';'

//==! Convert opcode linked list into an array of opcode pointers.
    pCBsdlOpInfo = new CBsdlOp * [lVar0+1];
    for (lVar1 = lVar0; lVar1 >= 0; --lVar1)
    { if (pCBsdlOpHead == NULL)
      { jMessage(DISP_AND_EXIT,"OpcN!=lVar1",1,pDP);  // should never happen
      } // IF
      pCBsdlOpInfo[lVar1] = pCBsdlOpHead;
      pCBsdlOpHead = pCBsdlOpHead->pNextCBsdlOp;
      pCBsdlOpInfo[lVar1]->pNextCBsdlOp = NULL;
    } // FOR lVar1 >= 0
    if (lVar1 == -1 && pCBsdlOpHead != NULL)
    { jMessage(DISP_AND_EXIT,"Op-0/!=0",1,pDP);  // should never happen
    } // IF

    lNumOpcodes = lVar0+1;

    checkBsdlEntityEnd(pDP);
    return;
  } // IF attribute = instruction_opcode

//==! Read list of instructions that will force a device to exit ISC mode
  if (!strcmp(pDP->sData,"isc_illegal_exit"))
  { if (lNumOpcodes <= 0)
    { jMessage(DISP_AND_EXIT,"No instructions defined yet",1,pDP);
    } // IF lNumOpcodes <= 0
    checkAttrEntity(pDP);
    findStr(pDP,"\"",0);
    ++pDP->iBracketLevel;
    while (strcmp(pDP->sData,";"))
    { xGetData(pDP,1,eATTR_VAR,"Expecting instr name");
      if (this->getOpcodeIndex(1,pDP->sData,pDP) < 0)
      { sprintf(sMsgOnly,"Opcode [%s] undefined",pDP->sData);
        jMessage(DISP_AND_EXIT,sMsgOnly,1,pDP);
      } // IF

      xDataBsdl(pDP,XDATA_READ);
      if (strcmp(pDP->sData,",") && strcmp(pDP->sData,";"))
      { jMessage(DISP_AND_EXIT,"Expecting ',' or ';'",1,pDP);
      } // IF

    } // WHILE sData != ';'

    checkBsdlEntityEnd(pDP);
    return;
  } // IF attribute = isc_illegal_exit

//==! Check ISC_PIN_BEHAVIOR attribute (can only choose between HIGHZ and
//==! CLAMP instructions.
  if (!strcmp(pDP->sData,"isc_pin_behavior"))
  { checkAttrEntity(pDP);
    findStr(pDP,"\"",0);
    ++pDP->iBracketLevel;
    xGetData(pDP,1,eATTR_VAR,"Expecting instr name");
    if (strcmp(pDP->sData,"highz") && strcmp(pDP->sData,"clamp"))
    { jMessage(DISP_AND_EXIT,"Expected HIGHZ or CLAMP only",1,pDP);
    } // IF

    if (m_csPinBehavior == NULL)
    { copyStr(&m_csPinBehavior,pDP->sData);
      m_iPinBehaviorCheck = 1;

//==! ISC_PIN_BEHAVIOR can occur before the INSTRUCTION_OPCODE attribute. If
//==! so then set m_iPinBehaviorCheck = 1 to defer the check to the function
//==! CBsdlData->semanticCheckBsdl() after BSDL file is read.
      if (pCBsdlOpInfo != NULL)
      { m_iPinBehaviorCheck = 0;
        if (this->getOpcodeIndex(1,m_csPinBehavior,pDP) == -2)
        { sprintf(sMsgOnly,"Opcode [%s] undefined",m_csPinBehavior);
          jMessage(DISP_AND_EXIT,sMsgOnly,1,pDP);
        } // IF
      } // IF pCBsdlOpInfo != NULL
    } // IF m_csPinBehavior == NULL
    else
    { sprintf(sMsgOnly,"ISC_PIN_BEHAVIOR already defined as [%s]",
              m_csPinBehavior);
      jMessage(DISP_SHOW_MSG,sMsgOnly,1,pDP);
    } // ELSE

    xDataBsdl(pDP,XDATA_READ);
    if (strcmp(pDP->sData,";"))
    { jMessage(DISP_AND_EXIT,"Expecting ';'",1,pDP);
    } // IF

    checkBsdlEntityEnd(pDP);
    return;
  } // IF attribute = isc_pin_behavior

  if (!strcmp(pDP->sData,"idcode_register"    ) ||
      !strcmp(pDP->sData,"isc_blank_usercode" ) ||
      !strcmp(pDP->sData,"usercode_register"  ))
  { if (!strcmp(pDP->sData,"idcode_register"))
    { p_m_CxCStrPtrVar = &m_CxCStr_IdcodeList;
      if (lOpcodeIdcode < 0)
      { jMessage(DISP_AND_EXIT,
           "IDCODE_REGISTER defined but BSDL has no IDCODE opcode",1,pDP);
      } // IF
      sprintf(sMsgBuf00,"IDCODE");
    } // IF
    if (!strcmp(pDP->sData,"isc_blank_usercode"))
    { p_m_CxCStrPtrVar = &m_CxCStr_IscBlankUsercodeList;
      if (lOpcodeUsercode < 0)
      { jMessage(DISP_AND_EXIT,
           "BSDL has no USERCODE opcode",1,pDP);
      } // IF
      sprintf(sMsgBuf00,"ISC_BLANK_USERCODE");
    } // IF
    if (!strcmp(pDP->sData,"usercode_register"))
    { p_m_CxCStrPtrVar = &m_CxCStr_UsercodeList;
      if (lOpcodeUsercode < 0)
      { jMessage(DISP_AND_EXIT,
           "BSDL has no USERCODE opcode",1,pDP);
      } // IF
      sprintf(sMsgBuf00,"USERCODE");
    } // IF

    if (*p_m_CxCStrPtrVar != NULL)
    { sprintf(sMsgOnly,"%s already defined as:\n\t-> [",sMsgBuf00);
      dispMsg(sMsgOnly,DISP_PRE_LF,1,1);
      for (pCxCStr_Var = *p_m_CxCStrPtrVar; pCxCStr_Var != NULL;
           pCxCStr_Var = pCxCStr_Var->pNextCxCStr)
      { sprintf(sMsgOnly,"%s",pCxCStr_Var->pStr);
        dispMsg(sMsgOnly,DISP_USE_CUR_POS,1,1);
        if (pCxCStr_Var->pNextCxCStr != NULL)
        { dispMsg(",",DISP_USE_CUR_POS,1,1);
        } // IF
        else
        { dispMsg("]",DISP_USE_CUR_POS | DISP_POST_LF,1,1);
        } // ELSE
      } // IF
      jMessage(DISP_AND_EXIT,"BSDL attribute redefinition",1,pDP);
    } // IF *p_m_CxCStrPtrVar != NULL

    checkAttrEntity(pDP);

    findStr(pDP,"\"",0);
    ++pDP->iBracketLevel;
    pCxCStr_Var = new CxCStr;
    pCxCStr_Var->pStr = new char [33];
    pCxCStr_Var->pStr[32] = '\0';  //==! NULL-terminate the string
    for (lVar2 = 0;;)
    { xDataBsdl(pDP,XDATA_READ);
      if (!strcmp(pDP->sData,";") || !strcmp(pDP->sData,","))
      { if (lVar2 != 32)
        { sprintf(sMsgOnly,"Only [%ld] of 32 %s bits specified",
                  lVar2,sMsgBuf00);
          jMessage(DISP_AND_EXIT,sMsgOnly,1,pDP);
        } // IF
        pCxCStr_Var->pNextCxCStr = *p_m_CxCStrPtrVar;
        *p_m_CxCStrPtrVar = pCxCStr_Var;

        if (!strcmp(pDP->sData,";"))
        { break;
        } // IF

        lVar2 = 0;
        pCxCStr_Var = new CxCStr;
        pCxCStr_Var->pStr = new char [33];
        pCxCStr_Var->pStr[32] = '\0';  //==! NULL-terminate the string
        continue;
      } // IF
      lVar0 = strlen(pDP->sData);
      validateData(pDP->sData,lVar0,"01xX",pDP);
      for (lVar1 = 0; lVar1 < lVar0; ++lVar1)
      { if (lVar2 >= 32)
        { sprintf(sMsgOnly,"%s can only be 32 bits long",sMsgBuf00);
          jMessage(DISP_AND_EXIT,sMsgOnly,1,pDP);
        } // IF
        pCxCStr_Var->pStr[lVar2] = pDP->sData[lVar1];
        ++lVar2;
      } // FOR lVar1 < lVar0
    } // FOR (;;)

    checkBsdlEntityEnd(pDP);
    return;
  } // IF attribute = idcode_register | isc_blank_usercode | usercode_register

  if (!strcmp(pDP->sData,"boundary_register"))
  { if (pBscData != NULL)
    { jMessage(DISP_AND_EXIT,"BSC register already defined",1,pDP);
    } // IF
    checkAttrEntity(pDP);
    findStr(pDP,"\"",0);
    ++pDP->iBracketLevel;

    pBscData = new CBscData * [lBscLength];   //==! Search for NOTE_0008
    for (lVar0 = 0; lVar0 < lBscLength; ++lVar0)
    { pBscData[lVar0] = new CBscData;
    } // FOR lVar0 < lBscLength

    while (strcmp(pDP->sData,";"))
    { lVar0 = lReadDecimalBsdl(pDP,0);
      sprintf(sMsgBuf01,"BSC [%ld] out of range",lVar0);
      isInRange(1,lVar0,0,lBscLength-1,sMsgBuf01,DISP_AND_EXIT);
      getBscInfo(pDP,lVar0);
      xDataBsdl(pDP,XDATA_READ);
      if (strcmp(pDP->sData,";") && strcmp(pDP->sData,","))
      { jMessage(DISP_AND_EXIT,"Expecting ',' or ';'",1,pDP);
      } // IF
    } // WHILE sData != ';'
//==! Verify that boundary_register section has defined all boundary scan
//==! cells.
    for (lVar0 = 0; lVar0 < lBscLength; ++lVar0)
    { if (pBscData[lVar0]->lBscIndex != lVar0)
      { sprintf(sMsgBuf00,"BSC [%ld] not defined",lVar0);
        jMessage(DISP_AND_EXIT,sMsgBuf00,0,NULL);
      } // IF
    } // FOR lVar0 < lBscLength

    checkBsdlEntityEnd(pDP);
    return;
  } // IF attribute = boundary_register

  if (!strcmp(pDP->sData,"register_access"))
  { if (pCBsdlDataRegInfo != NULL)
    { jMessage(DISP_AND_EXIT,"Data reg access section already defined",1,pDP);
    } // IF pCBsdlDataRegInfo != NULL
    lVar0 = -1;
    checkAttrEntity(pDP);
    findStr(pDP,"\"",0);
    ++pDP->iBracketLevel;
    while (strcmp(pDP->sData,";"))
    { xGetData(pDP,1,eATTR_VAR,"Expecting data reg name");

//==! If reg name already exists, then this reg definition should have the
//==! same length.
      for (pCBsdlDataRegVar = m_pCBsdlDataRegListHead;
           pCBsdlDataRegVar != NULL;
           pCBsdlDataRegVar = pCBsdlDataRegVar->pNextCBsdlDataReg)
      { if (!strcmp(pCBsdlDataRegVar->csRegName,pDP->sData))
        { break;
        } // IF
      } // FOR pCBsdlDataRegVar != NULL
      if (pCBsdlDataRegVar == NULL)
      { ++lVar0;
        pCBsdlDataRegVar = new CBsdlDataReg;
        if (m_pCBsdlDataRegListHead == NULL)
        { m_pCBsdlDataRegListHead = pCBsdlDataRegVar;
        } // IF
        else
        { pCBsdlDataRegTail->pNextCBsdlDataReg = pCBsdlDataRegVar;
        } // ELSE
        pCBsdlDataRegTail = pCBsdlDataRegVar;
        copyStr(&(pCBsdlDataRegVar->csRegName),pDP->sData);
      } // IF

      xDataBsdl(pDP,XDATA_READ);
      if (!strcmp(pDP->sData,"["))
      { lVar1 = lReadDecimalBsdl(pDP,0);
        if (lVar1 <= 0)
        { jMessage(DISP_AND_EXIT,"register length must be > 0",1,pDP);
        } // IF
        findStrBsdl(pDP,"]",0);
        xDataBsdl(pDP,XDATA_READ);
      } // IF
      else
      {
//==! Lengths for BYPASS and DEVICE_ID are predefined. Use lBscLength if
//==! already defined (i.e., attribute BOUNDARY_LENGTH already specified).
        if (!strcmp(pCBsdlDataRegVar->csRegName,"boundary"))
        { lVar1 = -1;
          if (lBscLength >= 0)
          { lVar1 = lBscLength;
          } // IF
        } // IF
        else
        { if (!strcmp(pCBsdlDataRegVar->csRegName,"device_id"))
          { lVar1 = 32;
          } // IF
          else
          { lVar1 = 1;   //==! BYPASS reg length or default = 1
          } // ELSE
        } // ELSE
      } // ELSE
      pDP->iParserGetData = 0;

      if ((pCBsdlDataRegVar->lRegLength > 0) &&
          (pCBsdlDataRegVar->lRegLength != lVar1))
      { sprintf(sMsgOnly,"[%s] already defined with length [%ld]",
                pCBsdlDataRegVar->csRegName,pCBsdlDataRegVar->lRegLength);
        jMessage(DISP_AND_EXIT,sMsgOnly,1,pDP);
      } // IF
      else
      { if (lVar1 != -1)
        { pCBsdlDataRegVar->lRegLength = lVar1;
        } // IF
      } // ELSE

      findStrBsdl(pDP,"(",0);
      while (strcmp(pDP->sData,")"))
      { xGetData(pDP,1,eATTR_VAR,"Expecting instr opcode");
        lVar1 = lAssignDataRegToOpcode(lVar0,pDP->sData,pDP);
        pSListVar = new SList;
        pSListVar->lVal = lVar1;
        pSListVar->pNextSList = pCBsdlDataRegVar->pOpList;
        pCBsdlDataRegVar->pOpList = pSListVar;
        xDataBsdl(pDP,XDATA_READ);
        if (!strcmp(pDP->sData,"captures"))
        { xDataBsdl(pDP,XDATA_READ);
          if ((long)strlen(pDP->sData) != pCBsdlDataRegVar->lRegLength)
          { if (!strcmp(pCBsdlDataRegVar->csRegName,"boundary") &&
                pCBsdlDataRegVar->lRegLength == 0)
            { pCBsdlDataRegVar->lRegLength = (long)strlen(pDP->sData);
            } // IF
            else
            { sprintf(sMsgOnly,
                      "Reg capture val length [%ld] != reg length [%ld]",
                      strlen(pDP->sData),pCBsdlDataRegVar->lRegLength);
              jMessage(DISP_AND_EXIT,sMsgOnly,1,pDP);
            } // IF
          } // IF
          validateData(pDP->sData,pCBsdlDataRegVar->lRegLength,"01xX",pDP);
          xDataBsdl(pDP,XDATA_READ);
        } // IF
        if (!strcmp(pDP->sData,",") || !strcmp(pDP->sData,")"))
        { continue;
        } // IF
        jMessage(DISP_AND_EXIT,"Expected ',' or ')'",1,pDP);
      } // WHILE sData != ')'

      xDataBsdl(pDP,XDATA_READ);
      if (strcmp(pDP->sData,",") && strcmp(pDP->sData,";"))
      { jMessage(DISP_AND_EXIT,"Expecting ',' or ';'",1,pDP);
      } // IF

    } // WHILE sData != ';'

//==! See NOTE_0013
    checkRegOpcodeAssociation(1,"bypass","bypass",pDP,&lVar0,
                  &m_pCBsdlDataRegListHead,&pCBsdlDataRegTail);
    checkRegOpcodeAssociation(1,"extest","boundary",pDP,&lVar0,
                  &m_pCBsdlDataRegListHead,&pCBsdlDataRegTail);
    checkRegOpcodeAssociation(1,"sample","boundary",pDP,&lVar0,
                  &m_pCBsdlDataRegListHead,&pCBsdlDataRegTail);
    if (lJtagStdLevel >= 2001)
    { checkRegOpcodeAssociation(1,"preload","boundary",pDP,&lVar0,
                  &m_pCBsdlDataRegListHead,&pCBsdlDataRegTail);
    } // IF
    checkRegOpcodeAssociation(0,"idcode","device_id",pDP,&lVar0,
                  &m_pCBsdlDataRegListHead,&pCBsdlDataRegTail);
    checkRegOpcodeAssociation(0,"usercode","device_id",pDP,&lVar0,
                  &m_pCBsdlDataRegListHead,&pCBsdlDataRegTail);
    checkRegOpcodeAssociation(0,"clamp","bypass",pDP,&lVar0,
                  &m_pCBsdlDataRegListHead,&pCBsdlDataRegTail);
    checkRegOpcodeAssociation(0,"intest","boundary",pDP,&lVar0,
                  &m_pCBsdlDataRegListHead,&pCBsdlDataRegTail);
    checkRegOpcodeAssociation(0,"highz","bypass",pDP,&lVar0,
                  &m_pCBsdlDataRegListHead,&pCBsdlDataRegTail);

//==! Convert data reg linked list into an array of data reg pointers.
    pCBsdlDataRegInfo = new CBsdlDataReg * [lVar0+1];
    for (lVar1 = 0; lVar1 <= lVar0; ++lVar1)
    { if (m_pCBsdlDataRegListHead == NULL)
      { jMessage(DISP_AND_EXIT,"DRegN!=lVar1",1,pDP);  // should never happen
      } // IF

//==! Set the DEVICE_ID register (if defined) length after reading the
//==! register_access section. The BOUNDARY register (if defined) length will
//==! be set after reading the BOUNDARY_LENGTH attribute. This is done after
//==! reading the BSDL file.
      if (!strcmp(m_pCBsdlDataRegListHead->csRegName,"device_id"))
      { m_pCBsdlDataRegListHead->lRegLength = 32;
      } // IF

      pCBsdlDataRegInfo[lVar1] = m_pCBsdlDataRegListHead;
      m_pCBsdlDataRegListHead = m_pCBsdlDataRegListHead->pNextCBsdlDataReg;
      pCBsdlDataRegInfo[lVar1]->pNextCBsdlDataReg = NULL;
    } // FOR lVar1 <= lVar0
    if (lVar1 >= lVar0 && m_pCBsdlDataRegListHead != NULL)
    { jMessage(DISP_AND_EXIT,"DR-0/!=0",1,pDP);  // should never happen
    } // IF

    lNumDataRegs = lVar0+1;

    checkBsdlEntityEnd(pDP);
    return;

  } // IF attribute = register_access

  if (!strcmp(pDP->sData,"instruction_private"))
  { if (pCBsdlOpInfo == NULL)
    { jMessage(DISP_AND_EXIT,"Instruction opcodes not yet defined",1,pDP);
    } // IF pCBsdlOpInfo != NULL
//==! Check the iPrivate flag of any one of the opcodes (check 0th opcode)
//==! to determine if instruction_private attribute has already been read.
    isPtrNull((void *)(pCBsdlOpInfo[0]),"iPrv[0]");
    if (pCBsdlOpInfo[0]->iPrivate != -1)
    { jMessage(DISP_AND_EXIT,"Instruction_private attribute already read",1,pDP);
    } // IF

    checkAttrEntity(pDP);
    findStr(pDP,"\"",0);
    ++pDP->iBracketLevel;
    while (strcmp(pDP->sData,";"))
    { xGetData(pDP,1,eATTR_VAR,"Expecting private instruction");

      for (lVar0 = 0; lVar0 < lNumOpcodes; ++lVar0)
      { isPtrNull((void *)(pCBsdlOpInfo[lVar0]),"iPrv[x]");
        if (!strcmp(pCBsdlOpInfo[lVar0]->csOpcodeName,pDP->sData))
        { if (pCBsdlOpInfo[lVar0]->iPrivate != -1)
          { sprintf(sMsgOnly,"[%s] private attribute already specified",
                    pDP->sData);
            jMessage(DISP_AND_EXIT,sMsgOnly,1,pDP);
          } // IF
          pCBsdlOpInfo[lVar0]->iPrivate = 1;
          break;
        } // IF
      } // FOR lVar0 < lNumOpcodes
      if (lVar0 >= lNumOpcodes)
      { sprintf(sMsgOnly,"Opcode [%s] undefined",pDP->sData);
        jMessage(DISP_AND_EXIT,sMsgOnly,1,pDP);
      } // IF lVar0 >= lNumOpcodes

      xDataBsdl(pDP,XDATA_READ);
      if (strcmp(pDP->sData,",") && strcmp(pDP->sData,";"))
      { jMessage(DISP_AND_EXIT,"Expecting ',' or ';'",1,pDP);
      } // IF

    } // WHILE sData != ';'

//==! Set all other opcodes as public
    for (lVar0 = 0; lVar0 < lNumOpcodes; ++lVar0)
    { isPtrNull((void *)(pCBsdlOpInfo[lVar0]),"iPrv[x]");
      if (pCBsdlOpInfo[lVar0]->iPrivate == -1)
      { pCBsdlOpInfo[lVar0]->iPrivate = 0; // All other opcodes are public
      } // IF
    } // FOR lVar0 < lNumOpcodes

    checkBsdlEntityEnd(pDP);
    return;
  } // IF attribute = instruction_private

  if (!strcmp(pDP->sData,"design_warning") ||
      !strcmp(pDP->sData,"isc_design_warning"))
  { if (iGlobOptDisplayBsdlWarnings == 1)
    { sprintf(sMsgOnly,"    - ISC design warnings: --------------------");
      if (!strcmp(pDP->sData,"design_warning"))
      { sprintf(sMsgOnly,"    - IEEE 1149.1 design warnings: ------------");
      } // IF
      jMessage(DISP_SHOW_MSG,sMsgOnly,1,NULL);
      checkAttrEntity(pDP);
      findStr(pDP,"\"",0);
      ++pDP->iBracketLevel;
      for (lNumVar = 0;;++lNumVar)
      { getDpBsdlChar(pDP,0);
        if (pDP->iBsdlEntityEnd == 1)
        { break;
        } // IF
        sprintf(sMsgOnly,"%c",pDP->iCurChar);
        dispMsg(sMsgOnly,DISP_USE_CUR_POS,1,1);
        if (isspace(pDP->iCurChar) && (lNumVar > 70))
        { dispMsg(" ",DISP_POST_LF,1,1);
          lNumVar = 0;
        } // IF
      } // FOR (;;)
      checkBsdlEntityEnd(pDP);
      dispMsg("    -------------------------------------------",
              DISP_PRE_LF | DISP_POST_LF,1,1);
    } // IF iGlobOptDisplayBsdlWarnings == 1
    else
    { findStrBsdl(pDP,";",1);
    } // IF
    return;
  } // IF attribute = design_warning or isc_design_warning

//==! Skip this attribute if none of the above
  findStrBsdl(pDP,";",1);

} // CBsdlData::readBsdlAttributes()
//==!=====================================================================!==
//==! semanticCheckBsdl(): Perform additional semantic checks on BSDL file.
//==!
//==! Function Parameters: none
//==! Return Value       : none
//==!=====================================================================!==
void CBsdlData::semanticCheckBsdl()
{ long lVar0;

//==! If BOUNDARY register was defined in the REGISTER_ACCESS section, then:
//==! - compare defined BSC reg length against BOUNDARY_LENGTH
//==!   = length in REGISTER_ACCESS section can be defined by boundary [x],
//==!     or implicitly through the "captures" definition
//==! - if length not defined (i.e., default -1), then set to the boundary
//==!   length
  if (pCBsdlDataRegInfo != NULL)
  { for (lVar0 = 0; lVar0 < lNumDataRegs; ++lVar0)
    { if (!strcmp(pCBsdlDataRegInfo[lVar0]->csRegName,"boundary"))
      { if (pCBsdlDataRegInfo[lVar0]->lRegLength != 0 &&
            pCBsdlDataRegInfo[lVar0]->lRegLength != lBscLength)
        { sprintf(sMsgOnly,
             "BOUNDARY_LENGTH (%ld) != REGISTER_ACCESS BOUNDARY length (%ld)",
             lBscLength,pCBsdlDataRegInfo[lVar0]->lRegLength);
          jMessage(DISP_AND_EXIT,sMsgOnly,0,NULL);
        } // IF
        pCBsdlDataRegInfo[lVar0]->lRegLength = lBscLength;
        break;
      } // IF
    } // FOR lVar0 < lNumDataRegs
  } // IF pCBsdlDataRegInfo != NULL

  if (m_iPinBehaviorCheck == 1)
  { if (m_csPinBehavior == NULL)  // should never happen
    { jMessage(DISP_SHOW_MSG,"No PinBehavior opcode",0,NULL);
    } // IF
    else
    { if (pCBsdlOpInfo != NULL)
      { m_iPinBehaviorCheck = 0;
        if (this->getOpcodeIndex(1,m_csPinBehavior,NULL) == -2)
        { sprintf(sMsgOnly,"PIN BEHAVIOR opcode [%s] undefined",
                  m_csPinBehavior);
          jMessage(DISP_AND_EXIT,sMsgOnly,0,NULL);
        } // IF
      } // IF pCBsdlOpInfo != NULL
    } // ELSE
  } // IF m_iPinBehaviorCheck == 1

} // CBsdlData::semanticCheckBsdl()
//==!=====================================================================!==
//==! Debug function: display flows/procedures/actions.
//==!
//==! Function Parameters: none
//==! Return Value       : none
//==!=====================================================================!==
void CBsdlData::display1532Bsdl()
{ CIscAct *pCIA;
  CxProcObj *pCPO;
  CxFlowObj *pCFO;
  CIscProc *pCIP;
  CIscFlow *pCIF;

  if (pDP_Output == NULL)
  { return;
  } // IF
  if (pDP_Output->pFILE == NULL)
  { return;
  } // IF

  fprintf(pDP_Output->pFILE,"\n--- Flows/Procs/Actions from [%s] ---",
          csBsdlFileName);
  for (pCIA = m_pCIscActInfo; pCIA != NULL; pCIA = pCIA->m_pNextCIscAct)
  { fprintf(pDP_Output->pFILE,"\n  Action [%s(%s)]",
            pCIA->m_csActName,
            pCIA->m_csActDataName == NULL ? " " : pCIA->m_csActDataName);
    for (pCPO = pCIA->m_pActionProcs;
         pCPO != NULL;
         pCPO = pCPO->m_pNextCxProcObj)
    { fprintf(pDP_Output->pFILE,"\n  - Procedure: %s(%s|%s)",
              pCPO->m_pProcItem->m_csProcName,
              (pCPO->m_csProcParam == NULL
                 ? " " : pCPO->m_csProcParam),
              (pCPO->m_pProcItem->m_csProcDataName == NULL
                 ? " " : pCPO->m_pProcItem->m_csProcDataName)
             );
      for (pCFO = pCPO->m_pProcItem->m_pProcFlows;
           pCFO != NULL;
           pCFO = pCFO->m_pNextCxFlowObj)
      { fprintf(pDP_Output->pFILE,"\n\t= %s(%s|%s)",
              pCFO->m_pFlowItem->m_csFlowName,
              pCFO->m_csFlowParam == NULL
                 ? " " : pCFO->m_csFlowParam,
              pCFO->m_pFlowItem->m_csDataName == NULL
                 ? " " : pCFO->m_pFlowItem->m_csDataName);
      } // FOR pCFO != NULL
    } // FOR pCIPvar != NULL
  } // FOR pCIA != NULL

  fprintf(pDP_Output->pFILE,"\n--- Procedures from [%s] ---",
          csBsdlFileName);
  for (pCIP = m_pCIscProcInfo; pCIP != NULL; pCIP = pCIP->m_pNextCIscProc)
  { fprintf(pDP_Output->pFILE,"\n  %s(%s)",
            pCIP->m_csProcName,
            pCIP->m_csProcDataName == NULL
                ? " " : pCIP->m_csProcDataName);
  } // FOR pCIP != NULL

  fprintf(pDP_Output->pFILE,"\n--- Flows from [%s] --------",
          csBsdlFileName);
  for (pCIF = m_pCIscFlowInfo; pCIF != NULL; pCIF = pCIF->m_pNextCIscFlow)
  { fprintf(pDP_Output->pFILE,"\n  %s(%s)",
            pCIF->m_csFlowName,
            pCIF->m_csDataName == NULL
                ? " " : pCIF->m_csDataName);
  } // FOR pCIF != NULL

  fprintf(pDP_Output->pFILE,"\n--- end of Flows/Procs/Actions ------");

} // CBsdlData::display1532Bsdl()
//==!=====================================================================!==
//==! check1149StdLevel(): Check if 1149 standard level is defined. The std
//==! level is of the form "std_1149_*". For example:
//==! - std_1149_1_1990
//==! - std_1149_1_1994
//==! - std_1149_1_2001
//==!
//==! Function Parameters: pDP = BSDL data port
//==! Return Value       : none
//==!=====================================================================!==
void CBsdlData::check1149StdLevel(cDataPort *pDP)
{ char *pChr;

  xGetData(pDP,1,eATTR_VAR,"Expecting std_* string");
  if (!strncmp(pDP->sData,"std_1149_",9)) 
  { pChr = pDP->sData + 9;
    if (pChr == NULL)
    { return;
    } // IF
    pChr = strchr(pChr,'_');
    if (pChr == NULL)
    { return;
    } // IF
    if (strlen(pChr) > 4)
    { if (pChr[1] == '1')  // std level is [1990:1999]
      { lJtagStdLevel = 1994;
      } // IF
    } // IF strlen(pChr) > 1
  } // IF sData = std_1149_

} // CBsdlData::check1149StdLevel()
//==!=====================================================================!==
//==! readBsdl(): Read and store the necessary info from the BSDL file
//==!
//==! Function Parameters: pDP = BSDL data port
//==! Return Value       : 1 if BSDL file successfully read (e.g., end found)
//==!                      0 otherwise
//==!=====================================================================!==
int CBsdlData::readBsdl(cDataPort *pDP)
{ int iFuncState;
  long lVar0;

  copyStr(&csBsdlFileName,pDP->psPortName);

  for (xParseData(pDP,XDATA_READ),iFuncState = 0;
       pDP->m_iEndOfDataStream == 0;
       xParseData(pDP,XDATA_READ))
  { iGlobProgramState = dPROGSTATE_READBSDL;

    if (pDP->sData[0] == '"')
    { jMessage(DISP_AND_EXIT,"BSDL syntax error",1,pDP);
    } // IF
    pDP->iParserGetData = 0;
    xDataBsdl(pDP,XDATA_READ);

    if (pDP->iBracketLevel > 0)
    { pDP->iBracketLevel -= 2;
    } // IF
    if (pDP->iBracketLevel != 0)
    { jMessage(DISP_AND_EXIT,
        "Mismatched BSDL brackets (see previous BSDL section)",1,pDP);
    } // IF

    if (!strcmp(pDP->sData,"entity"))
    { xGetData(pDP,1,eATTR_VAR,"Expecting entity name");
      copyStr(&csEntityName,pDP->sData);
      findStrBsdl(pDP,"is",0);
      continue;
    } // IF data = entity

    if (!strcmp(pDP->sData,"port"))
    { lNumPorts = this->readBsdlPort(pDP);
      pCPortDataInfo = new CPortData * [lNumPorts];
      for (lVar0 = 0; lVar0 < lNumPorts; ++lVar0)
      { pCPortDataInfo[lVar0] = NULL;
      } // FOR
      continue;
    } // IF data = port

    if (!strcmp(pDP->sData,"use"))
    { check1149StdLevel(pDP);
      findStrBsdl(pDP,";",1);
      continue;
    } // IF data = use

    if (!strcmp(pDP->sData,"constant"))
    { xGetData(pDP,1,eATTR_VAR,"Expecting a constant");
      copyStr(&csPortMapName,pDP->sData);
      findStrBsdl(pDP,":",0);
      xGetData(pDP,1,eATTR_VAR,"Expecting a constant value");
      if (strcmp(pDP->sData,"pin_map_string"))
      { findStrBsdl(pDP,";",0);
        freeArray(csPortMapName);
        continue;
      } // IF
      findStrBsdl(pDP,":",0);
      findStrBsdl(pDP,"=",0);
      this->readPortMap(pDP);
      continue;
    } // IF data = constant

    if (!strcmp(pDP->sData,"attribute"))  //==! Look for certain attributes
    { this->readBsdlAttributes(pDP);
      continue;
    } // IF data = attribute

    if (!strcmp(pDP->sData,"end"))
    { findStrBsdl(pDP,csEntityName,0);
      findStrBsdl(pDP,";",0);
      iFuncState = 1;
      break;
    } // IF data = end

    findStrBsdl(pDP,";",1);   //==! Skip this BSDL section

  } // FOR (pDP->m_iEndOfDataStream == 0)

  this->semanticCheckBsdl();

  if (iFuncState == 0)
  { jMessage(DISP_SHOW_MSG,"end <entityname> not found",0,NULL);
  } // IF

  if (iOpt_PrintBsdl == 1)
  { bsdlDataOut(this);
    display1532Bsdl();
  } // IF

  return iFuncState;

} // CBsdlData::readBsdl()
//==!=====================================================================!==
//==! end of module XRD_BSDL.CPP
//==!=====================================================================!==

